@@ -1,7 +1,7 @@
 use std::{cell::RefCell, iter::once};
 
 use num_bigint::BigUint;
-use once_cell::unsync::Lazy;
+use once_cell::sync::Lazy;
 use rustc_hash::FxHashMap as HashMap;
 
 use crate::{CircuitNode, CircuitRc};
@@ -11,7 +11,7 @@ thread_local! {
         RefCell::new(HashMap::default());
 }
 
-const BIG_UINT_0: Lazy<BigUint> = Lazy::new(|| 0u8.into());
+static BIG_UINT_0: Lazy<BigUint> = Lazy::new(|| 0u8.into());
 pub fn max_non_leaf_size(circuit: CircuitRc) -> BigUint {
     if let Some(result) =
         MAX_NON_LEAF_SIZE.with(|cache| cache.borrow().get(&circuit.info().hash).cloned())
@@ -27,7 +27,7 @@ pub fn max_non_leaf_size(circuit: CircuitRc) -> BigUint {
             circuit.info().naive_mem_use(None)
         }))
         .max()
-        .unwrap_or(BIG_UINT_0.clone());
+        .unwrap();
     MAX_NON_LEAF_SIZE.with(|cache| {
         cache
             .borrow_mut()
