@@ -5,24 +5,25 @@ use std::{
     sync::Arc,
 };
 
-use anyhow::{anyhow, bail, Result};
+use anyhow::{anyhow, bail, ensure, Context, Result};
 use itertools::Itertools;
 use num_bigint::BigUint;
 use pyo3::{exceptions, prelude::*, PyTypeInfo};
 use rr_util::{
     name::Name,
-    py_types::{reduction_to_ints, PyOpAtAxes, Tensor},
-    rearrange_spec::{check_canon_axes_as, OpSize, RearrangeSpec},
+    py_types::{reduction_to_ints, PyOpAtAxes},
+    rearrange_spec::{check_canon_axes_as, RearrangeSpec},
+    shape::{Shape, Size},
     sv,
-    tensor_util::{Shape, TensorIndex},
+    tensor_util::TensorIndex,
     tu8v,
     util::{arc_ref_clone, arc_unwrap_or_clone, AsOp, HashBytes, NamedAxes},
 };
+use smallvec::ToSmallVec;
 
 use crate::{
     circuit_info::{CachedCircuitInfo, CircuitFlags},
-    print,
-    print::PrintOptions,
+    self_funcs::SELF_FUNCS,
     CircuitNodeInit, *,
 };
 
@@ -42,7 +43,7 @@ macro_rules! circuit_node_self_only_hash_default_impl {
     };
 }
 
-pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash + Clone {
+pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash + Clone + Debug {
     // ==== implementable section ===
     //
     // NOTE: ALL FNS IN THIS SECTION *MUST* BE COPIED TO THE MACRO AND CircuitRc impl
@@ -74,8 +75,6 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash + Clone {
         BigUint::from(0usize)
     }
 
-    fn eval_tensors(&self, tensors: &[Tensor]) -> Result<Tensor>;
-
     fn c(self) -> Circuit;
     fn rc(self) -> CircuitRc {
         CircuitRc(Arc::new(self.c()))
@@ -152,6 +151,10 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash + Clone {
         &self.info().children
     }
 
+    fn non_free_children_sl(&self) -> &[CircuitRc] {
+        &self.info().children[self.num_free_children()..]
+    }
+
     fn num_children(&self) -> usize {
         self.info().children.len()
     }
@@ -161,7 +164,15 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash + Clone {
             new_children.len() == self.num_children(),
             "replace_children: wrong number of children"
         );
-        self._replace_children(new_children)
+        // TODO disabled because reconstructing might need to trigger autoname
+        // if &new_children == self.children_sl() {
+        //     Ok(self.clone())
+        // } else {
+        self._replace_children(new_children.clone())
+            .with_context(|| {
+                format!("while replacing children of\n{self:?}\nnew_children={new_children:#?}")
+            })
+        // }
     }
 
     fn map_children_enumerate<F>(&self, mut f: F) -> Result<Self>
@@ -174,12 +185,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash + Clone {
             .enumerate()
             .map(|(i, x)| f(i, x))
             .collect::<Result<Vec<_>>>()?;
-        // TODO disabled because reconstructing might need to trigger autoname
-        // if &out == self.children_sl() {
-        //     Ok(self.clone())
-        // } else {
-        self._replace_children(out)
-        // }
+        self.replace_children(out)
     }
 
     fn map_non_free_children_enumerate<F>(&self, f: F) -> Result<Self>
@@ -193,11 +199,12 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash + Clone {
         self.map_children_enumerate(|ix, c| if ix < n { Ok(c) } else { f(ix - n, c) })
     }
 
-    fn map_children<F>(&self, mut f: F) -> Result<Self>
+    fn map_children<F>(&self, f: F) -> Result<Self>
     where
         F: FnMut(CircuitRc) -> Result<CircuitRc>,
     {
-        self.map_children_enumerate(|_i, x| f(x))
+        let out = self.children().map(f).collect::<Result<Vec<_>>>()?;
+        self.replace_children(out)
     }
     fn map_non_free_children<F>(&self, mut f: F) -> Result<Self>
     where
@@ -207,19 +214,23 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash + Clone {
         self.map_non_free_children_enumerate(|_i, x| f(x))
     }
 
-    fn map_children_unwrap<F>(&self, mut f: F) -> Self
+    fn map_children_unwrap<F>(&self, f: F) -> Self
     where
         Self: Sized,
         F: FnMut(CircuitRc) -> CircuitRc,
     {
-        self.map_children(|x| Ok(f(x))).unwrap()
+        let out = self.children().map(f).collect();
+        self.replace_children(out).unwrap()
     }
-    fn map_non_free_children_unwrap<F>(&self, mut f: F) -> Self
+    fn map_non_free_children_unwrap<F>(&self, f: F) -> Self
     where
         Self: Sized,
         F: FnMut(CircuitRc) -> CircuitRc,
     {
-        self.map_non_free_children(|x| Ok(f(x))).unwrap()
+        let mut f = f;
+        let n = self.num_free_children();
+
+        self.map_children_unwrap_enumerate(|ix, c| if ix < n { c } else { f(c) })
     }
 
     fn map_children_unwrap_enumerate<F>(&self, mut f: F) -> Self
@@ -227,7 +238,8 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash + Clone {
         Self: Sized,
         F: FnMut(usize, CircuitRc) -> CircuitRc,
     {
-        self.map_children_enumerate(|i, x| Ok(f(i, x))).unwrap()
+        let out = self.children().enumerate().map(|(i, x)| f(i, x)).collect();
+        self.replace_children(out).unwrap()
     }
 
     /// if any return Some, return child mapped, otherwise None
@@ -237,16 +249,20 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash + Clone {
         F: FnMut(CircuitRc) -> Option<CircuitRc>,
     {
         let mut any_modified = false;
-        let out = self.map_children_unwrap(|x| {
-            if let Some(new) = f(x.clone()) {
-                any_modified = true;
-                new
-            } else {
-                x
-            }
-        });
+        let out: Vec<CircuitRc> = self
+            .children_sl()
+            .iter()
+            .map(|x| {
+                if let Some(new) = f(x.clone()) {
+                    any_modified = true;
+                    new
+                } else {
+                    x.clone()
+                }
+            })
+            .collect();
         if any_modified {
-            Some(out)
+            Some(self.replace_children(out).unwrap())
         } else {
             None
         }
@@ -266,7 +282,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash + Clone {
             .unwrap()
     }
     fn repr(&self) -> Result<String> {
-        PrintOptions::repr(&Default::default(), self.crc())
+        (SELF_FUNCS.repr)(self.crc())
     }
 
     fn print(&self) -> Result<()> {
@@ -283,7 +299,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash + Clone {
     }
 
     fn debug_repr(&self) -> Result<String> {
-        print::debug_repr(self.crc())
+        (SELF_FUNCS.debug_repr)(self.crc())
     }
 
     fn debug_repru(&self) -> String {
@@ -307,21 +323,28 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash + Clone {
     }
 
     fn sum(&self, axes: &[i64], name: Option<Name>) -> Result<Einsum> {
-        let axes = check_canon_axes_as(self.info().rank(), axes)?;
+        let axes = check_canon_axes_as(self.rank(), axes)?;
         Ok(Einsum::new(
-            vec![(self.crc(), (0u8..self.info().rank() as u8).collect())],
-            (0u8..self.info().rank() as u8)
-                .filter(|i| !axes.contains(&i))
+            vec![(self.crc(), (0u8..self.rank() as u8).collect())],
+            (0u8..self.rank() as u8)
+                .filter(|i| !axes.contains(i))
                 .collect(),
             name,
         ))
     }
 
     fn mean(&self, axes: &[i64], name: Option<Name>, scalar_name: Option<Name>) -> Result<Einsum> {
-        let total_size: usize = check_canon_axes_as(self.info().rank(), axes)?
+        let total_size: Option<usize> = check_canon_axes_as(self.rank(), axes)?
             .into_iter()
-            .map(|i| self.info().shape[*i as usize])
+            .map(|i| self.info().shape[*i as usize].t())
             .product();
+        if total_size.is_none() {
+            bail!(
+                "mean: can't take mean over axis/axes {axes:?} with unknown size(s), shape={:?}",
+                self.shape()
+            );
+        }
+        let total_size = total_size.unwrap();
         let sum = self.sum(axes, None)?;
         Ok(Einsum::new(
             sum.args_cloned()
@@ -348,12 +371,12 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash + Clone {
             _ => (),
         }
 
-        let axes = check_canon_axes_as(self.info().rank(), axes)?;
+        let axes = check_canon_axes_as(self.rank(), axes)?;
 
         Ok(GeneralFunction::new_by_name(
             vec![Rearrange::nrc(
                 self.crc(),
-                RearrangeSpec::combine_axes_at_end(self.info().rank().try_into()?, axes).unwrap(),
+                RearrangeSpec::combine_axes_at_end(self.rank().try_into()?, axes).unwrap(),
                 None,
             )],
             op_name,
@@ -400,33 +423,30 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash + Clone {
     ) -> Result<Rearrange> {
         Ok(Rearrange::new(
             self.crc(),
-            RearrangeSpec::expand_at_axes(self.info().rank(), axes, counts)?,
+            RearrangeSpec::expand_at_axes(self.rank(), axes, counts)?,
             name,
         ))
     }
     fn unsqueeze(&self, axes: Vec<usize>, name: Option<Name>) -> Result<Rearrange> {
         Ok(Rearrange::new(
             self.crc(),
-            RearrangeSpec::unsqueeze(self.info().rank(), axes)?,
+            RearrangeSpec::unsqueeze(self.rank(), axes)?,
             name,
         ))
     }
     fn squeeze(&self, axes: PyOpAtAxes, name: Option<Name>) -> Result<Rearrange> {
-        let axes = check_canon_axes_as(
-            self.info().rank(),
-            &reduction_to_ints(Some(axes), self.info().rank()),
-        )?;
-        if axes.iter().any(|i| self.info().shape[*i as usize] != 1) {
+        let axes = check_canon_axes_as(self.rank(), &reduction_to_ints(Some(axes), self.rank()))?;
+        if axes.iter().any(|i| !self.info().shape[*i as usize].is(1)) {
             bail!(anyhow!(
                 "trying to squeeze non-1 axes, shape {:?} axes {:?}",
                 &self.info().shape,
                 &axes
             ))
         }
-        let num_ints = self.info().rank() - axes.len();
+        let num_ints = self.rank() - axes.len();
         let mut input_ints = sv![];
         let mut counter = 0;
-        for i in 0..self.info().rank() as u8 {
+        for i in 0..self.rank() as u8 {
             if axes.iter().contains(&i) {
                 input_ints.push(tu8v![])
             } else {
@@ -439,7 +459,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash + Clone {
             RearrangeSpec {
                 input_ints,
                 output_ints: (0..num_ints).map(|i| tu8v![i as u8]).collect(),
-                int_sizes: sv![OpSize::NONE;num_ints],
+                int_sizes: sv![Size::NONE;num_ints],
             },
             name,
         )
@@ -479,6 +499,17 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash + Clone {
     fn rearrange_str(&self, string: &str, name: Option<Name>) -> Result<Rearrange> {
         Rearrange::from_string(self.crc(), string, name)
     }
+    fn repeat_to(&self, new_shape: Shape, name: Option<Name>) -> Result<CircuitRc> {
+        ensure!(new_shape.len() >= self.ndim());
+        let n = new_shape.len() - self.ndim();
+        ensure!(self.shape()[..] == new_shape[n..]);
+        Ok(self
+            .rearrange(
+                RearrangeSpec::prepend_batch_shape(new_shape[..n].to_smallvec(), self.ndim())?,
+                name,
+            )?
+            .crc())
+    }
     fn add_suffix(self, suffix: Option<&str>) -> Self
     where
         Self: Sized,
@@ -554,9 +585,9 @@ macro_rules! circuit_node_eq_ord_debug {
 
         impl std::fmt::Display for $type_name {
             fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-                let mut printer = $crate::print::PrintOptions::default();
-                printer.bijection = false;
-                f.write_str(&printer.repr(self.crc()).unwrap())
+                f.write_str(&$crate::self_funcs::repr_circuit_default_no_bijection(
+                    self.crc(),
+                ))
             }
         }
     };
@@ -666,17 +697,14 @@ macro_rules! define_circuit_union_impl {
             }
 
             impl std::str::FromStr for [<$name Type>] {
-                type Err = anyhow::Error;
+                type Err = ();
                 fn from_str(s: &str) -> Result<Self, Self::Err> {
-                    let out = match s {
+                    match s {
                         $(
-                        stringify!($x) => Self::$x,
+                        stringify!($x) => Ok(Self::$x),
                         )*
-                        v => bail!($crate::parsing::ParseCircuitError::InvalidVariant {
-                            v: v.to_owned()
-                        }),
-                    };
-                    Ok(out)
+                        _ => Err(())
+                    }
                 }
             }
 
@@ -788,10 +816,6 @@ macro_rules! define_circuit_union_impl {
                 [<$name:snake _map>]!(self; node => node.self_flops())
             }
 
-            fn eval_tensors(&self, tensors: &[Tensor]) -> Result<Tensor> {
-                [<$name:snake _map>]!(self; node => node.eval_tensors(tensors))
-            }
-
             fn c(self) -> Circuit { ite_is_circuit!($name, self, [<$name:snake _map>]!(self; node => node.c())) }
             fn get_autoname(&self) -> Result<Option<Name>> { [<$name:snake _map>]!(self; node => node.get_autoname()) }
             fn child_axis_map(&self) -> Vec<Vec<Option<usize>>> { [<$name:snake _map>]!(self; node => node.child_axis_map()) }
@@ -889,7 +913,6 @@ define_circuit!(
     Scatter,
     Conv,
     Module,
-    SetSymbolicShape,
     Tag,
     DiscreteVar,
     StoredCumulantVar,
@@ -930,21 +953,6 @@ macro_rules! define_circuit_union {
                     [<$name:snake _map>]!(item; node => node.into())
                 }
             }
-
-            fn [<circuit_is_ $name:snake>](circuit: &$crate::Circuit) -> bool {
-                match circuit {
-                    $(
-                        $crate::Circuit::$x(_) => true,
-                    )*
-                    _ => false
-                }
-            }
-
-            #[pyfunction]
-            #[pyo3(name="circuit_is_" $name:snake)]
-            pub fn [<circuit_is_ $name:snake _py>](circuit: $crate::CircuitRc) -> bool {
-                [<circuit_is_ $name:snake>](&circuit)
-            }
         }
     }
 }
@@ -955,7 +963,12 @@ macro_rules! define_circuit_union_special {
         paste::paste! {
             impl Circuit {
                 pub fn [<is_ $name:snake>](&self) -> bool {
-                    [<circuit_is_ $name:snake>](self)
+                    match self {
+                        $(
+                            $crate::Circuit::$x(_) => true,
+                        )*
+                        _ => false
+                    }
                 }
                 pub fn [<into_ $name:snake>](self) -> Option<$name> {
                     self.into()
@@ -1095,16 +1108,15 @@ impl CircuitNode for CircuitRc {
         self.0.self_flops()
     }
 
-    fn eval_tensors(&self, tensors: &[Tensor]) -> Result<Tensor> {
-        self.0.eval_tensors(tensors)
-    }
-
     fn c(self) -> Circuit {
         arc_unwrap_or_clone(self.0)
     }
     fn rc(self) -> CircuitRc {
         self
     }
+    fn crc(&self) -> CircuitRc {
+        self.clone()
+    }
 
     // fn static_node_type_uuid() -> [u8; 16] { self.0.static_node_type_uuid() }
 
