@@ -1,20 +1,11 @@
 // ! CircuitNodeInit and various helpers for nicely implementing it
-use std::{collections::BTreeMap, iter::zip, sync::Arc};
+use std::{collections::BTreeMap, iter::zip};
 
-use anyhow::Context;
-use itertools::Itertools;
-use rr_util::{
-    name::Name,
-    symbolic_size::{SymbolicSizeConstraint, SymbolicSizeConstraints},
-    tensor_util::TorchDeviceDtypeOp,
-    util::HashBytes,
-    IndexSet,
-};
+use rr_util::{name::Name, util::HashBytes};
 use uuid::uuid;
 
 use crate::{
     CachedCircuitInfo, CircuitFlags, CircuitNode, CircuitRc, ConstructError, NamedAxes, Result,
-    Symbol,
 };
 
 pub trait CircuitNodeInit {
@@ -47,57 +38,6 @@ pub trait CircuitNodeComputeInfoImpl: CircuitNode {
     fn compute_flags(&self) -> CircuitFlags {
         self.compute_flags_default()
     }
-
-    fn device_dtype_extra(&self) -> Option<Vec<TorchDeviceDtypeOp>> {
-        None
-    }
-    fn device_dtype_override(&self) -> Option<Result<TorchDeviceDtypeOp>> {
-        None
-    }
-    fn symbolic_size_constraints_extra(&self) -> Result<Vec<SymbolicSizeConstraint>> {
-        Ok(Vec::new())
-    }
-}
-
-fn compute_device_dtype(
-    children: &[CircuitRc],
-    extra: Option<Vec<TorchDeviceDtypeOp>>,
-) -> Result<TorchDeviceDtypeOp> {
-    children
-        .iter()
-        .map(|c| c.info().device_dtype.clone())
-        .chain(extra.clone().into_iter().flatten())
-        .try_fold(TorchDeviceDtypeOp::NONE, |acc, new| {
-            TorchDeviceDtypeOp::combine(acc, new)
-        })
-        .with_context(|| {
-            format!(
-                "Could not combine device_dtype for node with children {:?} and extra {:?}",
-                children
-                    .iter()
-                    .map(|c| c.info().clone())
-                    .collect::<Vec<_>>(),
-                extra
-            )
-        })
-}
-
-fn compute_symbolic_size_constraints(
-    children: &[CircuitRc],
-    extra: Vec<SymbolicSizeConstraint>,
-) -> Result<SymbolicSizeConstraints> {
-    SymbolicSizeConstraints::new(
-        children
-            .iter()
-            .flat_map(|c| {
-                c.info()
-                    .symbolic_size_constraints
-                    .clone()
-                    .into_constraints()
-            })
-            .chain(extra)
-            .collect(),
-    )
 }
 
 fn compute_named_axes(
@@ -108,34 +48,13 @@ fn compute_named_axes(
     for (mp, child) in zip(child_axis_map, children) {
         for (ax, name) in &child.info().named_axes {
             if let Some(top_ax) = mp[(*ax) as usize] {
-                result.insert(top_ax as u8, name.clone());
+                result.insert(top_ax as u8, *name);
             }
         }
     }
     result
 }
 
-fn compute_free_symbols(children: &[CircuitRc]) -> Option<Arc<IndexSet<Symbol>>> {
-    // ideally, symbol would just return itself rather than doing this check here
-    // but, then the symbol would have a circular reference to itself
-    // So we can only include children and then we special case syms in various places. : /
-    children
-        .iter()
-        .filter_map(|x| {
-            if let Some(sym) = x.as_symbol() {
-                Some(Arc::new(IndexSet::from_iter([sym.clone()])))
-            } else {
-                x.info().free_symbols.clone()
-            }
-        })
-        .at_most_one()
-        .unwrap_or_else(|mut it| {
-            let mut a = (*it.next().unwrap()).clone();
-            it.for_each(|x| a.extend(x.iter().cloned()));
-            Some(Arc::new(a))
-        })
-}
-
 pub trait CircuitNodeSetNonHashInfo: CircuitNodePrivate {
     fn set_non_hash_info(&mut self) -> Result<()>;
 }
@@ -143,15 +62,6 @@ pub trait CircuitNodeSetNonHashInfo: CircuitNodePrivate {
 impl<T: CircuitNodeComputeInfoImpl + CircuitNodePrivate> CircuitNodeSetNonHashInfo for T {
     fn set_non_hash_info(&mut self) -> Result<()> {
         self.info_mut().flags = self.compute_flags();
-        self.info_mut().device_dtype = if let Some(z) = self.device_dtype_override() {
-            z
-        } else {
-            compute_device_dtype(self.children_sl(), self.device_dtype_extra())
-        }?;
-        self.info_mut().symbolic_size_constraints = compute_symbolic_size_constraints(
-            self.children_sl(),
-            self.symbolic_size_constraints_extra()?,
-        )?;
         self.info_mut().named_axes = if let Some(out) = self.already_set_named_axes() {
             out
         } else if self
@@ -163,7 +73,6 @@ impl<T: CircuitNodeComputeInfoImpl + CircuitNodePrivate> CircuitNodeSetNonHashIn
         } else {
             compute_named_axes(self.children_sl(), &self.child_axis_map())
         };
-        self.info_mut().free_symbols = compute_free_symbols(self.children_sl());
         Ok(())
     }
 }
@@ -222,7 +131,7 @@ where
             })
         }
 
-        for (axis, _) in &self.info().named_axes {
+        for axis in self.info().named_axes.keys() {
             if *axis as usize >= self.info().shape.len() {
                 return Err(ConstructError::NamedAxisAboveRank {}.into());
             }
