@@ -9,16 +9,14 @@ use rr_util::{
     compact_data::U8Set,
     name::Name,
     opt_einsum::EinsumSpec,
-    py_types::{
-        einops_repeat, einsum_py, make_diagonal_py, scalar_to_tensor, scalar_to_tensor_py,
-        ExtraPySelfOps, PyEinsumAxes, Tensor, PY_UTILS,
-    },
+    py_types::{PyEinsumAxes, PY_UTILS},
     pycall,
-    rearrange_spec::{OpSize, RInnerInts, RearrangeSpec},
+    rearrange_spec::{RInnerInts, RearrangeSpec},
+    shape::{broadcast_shapes, shape_eq_and_known, shape_eq_if_known, shape_is_known, Size},
     sv,
     tensor_util::{
-        broadcast_shapes, check_canon_axes, upcast_tensor_device_dtypes, IndexError, IndexTensor,
-        Slice, TensorAxisIndex, TensorIndex, TensorIndexSync, TorchDeviceDtypeOp,
+        check_canon_axes, IndexError, IndexTensor, Slice, TensorAxisIndex, TensorIndex,
+        TensorIndexSync, TorchDeviceDtypeOp,
     },
     tu8v,
     util::{cumsum, dict_to_list, hashmap_collect_except_duplicates, AxisInt, EinsumAxes},
@@ -57,9 +55,8 @@ impl CircuitNodeHashWithChildren for Einsum {
         for (c, axes) in self.args() {
             // no need to delimit because we have hash each loop
             hasher.update(&c.info().hash);
-            hasher.update(&axes);
+            hasher.update(axes);
         }
-        ()
     }
 }
 impl CircuitNodeSelfOnlyHash for Einsum {
@@ -109,32 +106,6 @@ impl CircuitNode for Einsum {
     fn self_flops(&self) -> BigUint {
         self.get_spec().flops()
     }
-
-    fn eval_tensors(&self, tensors: &[Tensor]) -> Result<Tensor> {
-        let tensors = upcast_tensor_device_dtypes(tensors);
-        Python::with_gil(|py| {
-            if self.in_axes.is_empty() {
-                return Ok(scalar_to_tensor_py(py, 1., sv![], Default::default())?);
-            }
-            let out_axes_deduped: EinsumAxes = self.out_axes.unique();
-            let result_non_diag: Tensor = einsum_py(
-                py,
-                zip(tensors.iter().cloned(), self.in_axes.iter().cloned()).collect(),
-                out_axes_deduped.clone(),
-            )?;
-
-            if out_axes_deduped.len() != self.out_axes.len() {
-                Ok(make_diagonal_py(
-                    py,
-                    &result_non_diag,
-                    out_axes_deduped,
-                    self.out_axes.clone(),
-                )?)
-            } else {
-                Ok(result_non_diag)
-            }
-        })
-    }
 }
 
 pub type EinsumArgs = Vec<(CircuitRc, EinsumAxes)>;
@@ -143,35 +114,35 @@ impl Einsum {
         children: &Vec<CircuitRc>,
         in_axes: &Vec<EinsumAxes>,
         name: Option<Name>,
-    ) -> Result<HashMap<AxisInt, usize>> {
+    ) -> Result<HashMap<AxisInt, Size>> {
         let mut out = HashMap::default();
         for (circ, axes) in zip(children, in_axes) {
-            if circ.info().rank() != axes.len() {
+            if circ.rank() != axes.len() {
                 bail!(ConstructError::EinsumLenShapeDifferentFromAxes {
                     len_axes: axes.len(),
-                    child_len_shape: circ.info().rank(),
+                    child_len_shape: circ.rank(),
                     einsum_name: name,
                     child_circuit: circ.clone(),
                 });
             }
             for (&circuit_size, axis) in circ.info().shape.iter().zip(axes) {
-                let existing_size = *out.entry(*axis).or_insert(circuit_size);
-                if existing_size != circuit_size {
-                    bail!(ConstructError::EinsumAxisSizeDifferent {
-                        new_size: circuit_size.into(),
-                        existing_size: existing_size.into(),
+                let existing_size = out.entry(*axis).or_insert(circuit_size);
+                *existing_size = existing_size.join_eq(&circuit_size, || {
+                    anyhow!(ConstructError::EinsumAxisSizeDifferent {
+                        new_size: circuit_size,
+                        existing_size: *existing_size,
                         axis: *axis as usize,
                         einsum_name: name,
                         child_circuit: circ.clone(),
-                    });
-                }
+                    })
+                })?;
             }
         }
 
         Ok(out)
     }
 
-    pub fn shape_map(&self) -> Result<HashMap<AxisInt, usize>> {
+    pub fn shape_map(&self) -> Result<HashMap<AxisInt, Size>> {
         Self::shape_map_(&self.info.children, &self.in_axes, self.info().name)
     }
 
@@ -195,7 +166,7 @@ impl Einsum {
         // TODO: reduce axes nums!
         let out = Self {
             out_axes,
-            info: CachedCircuitInfo::incomplete(name, shape, children),
+            info: CachedCircuitInfo::incomplete(name, shape, children)?,
             in_axes,
         };
 
@@ -321,7 +292,7 @@ impl Einsum {
         name: Option<Name>,
     ) -> Result<Self> {
         let (mut input_ints, output_ints) = EinsumSpec::string_to_ints(string.to_owned())?;
-        if input_ints.len() == 1 && input_ints[0].is_empty() && nodes.len() == 0 {
+        if input_ints.len() == 1 && input_ints[0].is_empty() && nodes.is_empty() {
             input_ints.clear()
         }
         Self::from_nodes_ints(nodes, input_ints, output_ints, name)
@@ -335,7 +306,7 @@ impl Einsum {
         name: Option<Name>,
     ) -> Result<Self> {
         let (mut input_ints, output_ints) = EinsumSpec::fancy_string_to_ints(string.to_owned())?;
-        if input_ints.len() == 1 && input_ints[0].is_empty() && nodes.len() == 0 {
+        if input_ints.len() == 1 && input_ints[0].is_empty() && nodes.is_empty() {
             input_ints.clear()
         }
         Self::from_nodes_ints(nodes, input_ints, output_ints, name)
@@ -369,7 +340,7 @@ impl Einsum {
         name: Option<Name>,
         scalar_name: Option<Name>,
     ) -> Self {
-        let axes: EinsumAxes = (0..node.info().rank()).map(|x| x as u8).collect();
+        let axes: EinsumAxes = (0..node.rank()).map(|x| x as u8).collect();
         Einsum::try_new(
             vec![
                 (node, axes.clone()),
@@ -380,22 +351,18 @@ impl Einsum {
         )
         .unwrap()
     }
+
     #[staticmethod]
     #[pyo3(signature=(*nodes, name = None))]
     pub fn elementwise_broadcasted(nodes: Vec<CircuitRc>, name: Option<Name>) -> Result<Einsum> {
-        let out_shape = broadcast_shapes(
-            &nodes
-                .iter()
-                .map(|x| &x.info().shape[..])
-                .collect::<Vec<_>>(),
-        )
-        .context("failed to broadcast for einsum mul")?;
+        let out_shape = broadcast_shapes(&nodes.iter().map(|x| &x.shape()[..]).collect::<Vec<_>>())
+            .context("failed to broadcast for einsum mul")?;
         let mut prev_one_shape = out_shape.len().saturating_sub(1) as u8;
         Einsum::try_new(
             nodes
                 .iter()
                 .map(|node| {
-                    let rank_difference = out_shape.len() - node.info().rank();
+                    let rank_difference = out_shape.len() - node.rank();
                     (
                         node.clone(),
                         node.info()
@@ -403,7 +370,7 @@ impl Einsum {
                             .iter()
                             .enumerate()
                             .map(|(i, l)| {
-                                if *l != out_shape[i + rank_difference] {
+                                if !l.eq_if_known(out_shape[i + rank_difference]) {
                                     prev_one_shape += 1;
                                     prev_one_shape
                                 } else {
@@ -418,6 +385,7 @@ impl Einsum {
             name,
         )
     }
+
     #[staticmethod]
     pub fn empty(name: Option<Name>) -> Self {
         Einsum::try_new(vec![], tu8v![], name).unwrap()
@@ -425,7 +393,7 @@ impl Einsum {
 
     #[staticmethod]
     pub fn identity(node: CircuitRc, name: Option<Name>) -> Self {
-        let axes: EinsumAxes = (0..node.info().rank()).map(|x| x as u8).collect();
+        let axes: EinsumAxes = (0..node.rank()).map(|x| x as u8).collect();
         Einsum::try_new(vec![(node, axes.clone())], axes, name).unwrap()
     }
 
@@ -436,7 +404,7 @@ impl Einsum {
         name: Option<Name>,
         out_axes_permutation: Option<Vec<usize>>,
     ) -> Self {
-        let sections: Vec<_> = nodes.iter().map(|n| n.info().rank()).collect();
+        let sections: Vec<_> = nodes.iter().map(|n| n.rank()).collect();
         let starts = cumsum(&sections);
         let mut out_axes: EinsumAxes = nodes
             .iter()
@@ -504,6 +472,7 @@ impl Einsum {
                     .map(|(key, val)| (*key as usize, *val))
                     .collect(),
                 None,
+                Size::known(0),
             ),
         }
     }
@@ -534,7 +503,7 @@ impl Add {
         .with_context(|| format!("Sum isn't broadcastable name={:?}", name))?;
         // TODO: reduce axes nums!
         let out = Self {
-            info: CachedCircuitInfo::incomplete(name, shape, nodes),
+            info: CachedCircuitInfo::incomplete(name, shape, nodes)?,
         };
         out.initial_init_info()
     }
@@ -561,14 +530,12 @@ impl Add {
     }
 
     pub fn has_broadcast(&self) -> bool {
-        !self
-            .children()
-            .all(|node| node.info().shape == self.info().shape)
+        !self.children().all(|node| node.shape() == self.shape())
     }
 
     pub fn nodes_and_rank_differences(&self) -> Vec<(CircuitRc, usize)> {
         self.children()
-            .map(|node| (node.clone(), self.info().rank() - node.info().rank()))
+            .map(|node| (node.clone(), self.rank() - node.info().rank()))
             .collect()
     }
 
@@ -626,7 +593,8 @@ impl CircuitNode for Add {
                     .iter()
                     .enumerate()
                     .map(|(i, _x)| {
-                        if self.info().shape[i + rank_difference] == node.info().shape[i] {
+                        if self.info().shape[i + rank_difference].eq_and_known(node.info().shape[i])
+                        {
                             Some(i + rank_difference)
                         } else {
                             None
@@ -644,20 +612,6 @@ impl CircuitNode for Add {
     fn self_flops(&self) -> BigUint {
         self.info.numel() * self.num_children()
     }
-
-    fn eval_tensors(&self, tensors: &[Tensor]) -> Result<Tensor> {
-        if tensors.is_empty() {
-            return Ok(scalar_to_tensor(0.0, sv!(), Default::default())?);
-        }
-        let tensors = upcast_tensor_device_dtypes(tensors);
-        let mut out = tensors[0].clone();
-        Python::with_gil(|py| {
-            for tensor in tensors[1..].iter() {
-                out = tensor.clone().py_add(py, out.clone()).unwrap();
-            }
-        });
-        Ok(out)
-    }
 }
 
 impl CircuitNodeAutoName for Add {
@@ -704,7 +658,7 @@ impl Rearrange {
         // TODO: reduce axes nums!
         let out = Self {
             spec: Box::new(spec),
-            info: CachedCircuitInfo::incomplete(name, shape, vec![node]),
+            info: CachedCircuitInfo::incomplete(name, shape, vec![node])?,
         };
 
         out.initial_init_info()
@@ -712,7 +666,7 @@ impl Rearrange {
 
     pub fn evolve(&self, node: Option<CircuitRc>, spec: Option<RearrangeSpec>) -> Rearrange {
         Rearrange::try_new(
-            node.unwrap_or_else(|| self.node().clone()),
+            node.unwrap_or_else(|| self.node()),
             spec.unwrap_or_else(|| (*self.spec).clone()),
             self.info().name,
         )
@@ -769,21 +723,13 @@ impl CircuitNode for Rearrange {
     fn _replace_children(&self, children: Vec<CircuitRc>) -> Result<Self> {
         Self::try_new(children[0].clone(), (*self.spec).clone(), self.info().name)
     }
-
-    fn eval_tensors(&self, tensors: &[Tensor]) -> Result<Tensor> {
-        // avoid using self.spec.apply so we get shape errors correctly
-        let (string, letter_sizes) = self
-            .conform_to_input_shape_spec()
-            .to_einops_string_and_letter_sizes();
-        einops_repeat(&tensors[0], string, letter_sizes)
-    }
 }
 
 impl CircuitNodeAutoName for Rearrange {
     const PRIORITY: OperatorPriority = OperatorPriority::PostFix {};
 
     fn auto_name(&self) -> Option<Name> {
-        child_name_with_maybe_paren(&Self::PRIORITY, self.node().clone())
+        child_name_with_maybe_paren(&Self::PRIORITY, self.node())
             .map(|n| (n.string() + " rearrange").into())
     }
 }
@@ -814,7 +760,7 @@ impl Rearrange {
 
     #[staticmethod]
     fn reshape(node: CircuitRc, shape: Shape) -> Result<Self> {
-        let rank: u8 = node.info().rank().try_into().unwrap();
+        let rank: u8 = node.rank().try_into().unwrap();
         let flat_rs = RearrangeSpec::flatten(rank);
         let unflat_rs = RearrangeSpec::unflatten(shape)?;
 
@@ -837,12 +783,7 @@ impl Rearrange {
     }
 
     pub fn conform_to_input_shape(&self) -> Self {
-        Rearrange::try_new(
-            self.node().clone(),
-            self.conform_to_input_shape_spec(),
-            None,
-        )
-        .unwrap()
+        Rearrange::try_new(self.node(), self.conform_to_input_shape_spec(), None).unwrap()
     }
 }
 
@@ -863,7 +804,7 @@ impl Index {
     #[apply(new_rc_unwrap)]
     pub fn try_new(node: CircuitRc, index: TensorIndex, name: Option<Name>) -> Result<Self> {
         let mut index = index;
-        let node_rank = node.info().rank();
+        let node_rank = node.rank();
         match index.0.len().cmp(&node_rank) {
             Ordering::Greater => {
                 bail!(IndexError::IndexRankTooHigh {
@@ -897,11 +838,23 @@ impl Index {
         index.validate(&node.info().shape)?;
 
         let out = Self {
-            info: CachedCircuitInfo::incomplete(
+            info: CachedCircuitInfo::with_device_dtype_extra(
                 name,
                 index.apply_to_shape(&node.info().shape),
                 vec![node],
-            ),
+                &index
+                    .0
+                    .iter()
+                    .filter_map(|x| match x {
+                        TensorAxisIndex::Tensor(tensor) if !tensor.shape().is_empty() => {
+                            let mut out: TorchDeviceDtypeOp = tensor.device_dtype().into();
+                            out.dtype = None;
+                            Some(out)
+                        }
+                        _ => None,
+                    })
+                    .collect::<Vec<_>>(),
+            )?,
             index,
         };
 
@@ -911,24 +864,7 @@ impl Index {
 
 circuit_node_extra_impl!(Index, self_hash_default);
 
-impl CircuitNodeComputeInfoImpl for Index {
-    fn device_dtype_extra(&self) -> Option<Vec<TorchDeviceDtypeOp>> {
-        Some(
-            self.index
-                .0
-                .iter()
-                .filter_map(|x| match x {
-                    TensorAxisIndex::Tensor(tensor) if !tensor.shape().is_empty() => {
-                        let mut out: TorchDeviceDtypeOp = tensor.device_dtype().into();
-                        out.dtype = None;
-                        Some(out)
-                    }
-                    _ => None,
-                })
-                .collect(),
-        )
-    }
-}
+impl CircuitNodeComputeInfoImpl for Index {}
 
 impl CircuitNodeHashItems for Index {
     fn compute_hash_non_name_non_children(&self, hasher: &mut blake3::Hasher) {
@@ -941,7 +877,7 @@ impl CircuitNode for Index {
 
     fn child_axis_map(&self) -> Vec<Vec<Option<usize>>> {
         let mut cur: i32 = -1;
-        let result = vec![zip(&self.node().info().shape, &self.index.0)
+        vec![zip(&self.node().info().shape, &self.index.0)
             .map(|(l, idx)| {
                 if !matches!(idx, TensorAxisIndex::Single(_)) {
                     cur += 1;
@@ -952,32 +888,19 @@ impl CircuitNode for Index {
                     None
                 }
             })
-            .collect()];
-        result
+            .collect()]
     }
 
     fn _replace_children(&self, children: Vec<CircuitRc>) -> Result<Self> {
         Self::try_new(children[0].clone(), self.index.clone(), self.info().name)
     }
-
-    fn eval_tensors(&self, tensors: &[Tensor]) -> Result<Tensor> {
-        assert_eq!(tensors.len(), 1);
-        let tensors = upcast_tensor_device_dtypes(tensors);
-        Python::with_gil(|py| {
-            PY_UTILS
-                .index_apply_dimwise
-                .call(py, (tensors[0].clone(), self.index.clone()), None)
-                .map_err(|err| err.into())
-                .map(|x| x.extract(py).unwrap())
-        })
-    }
 }
 
 impl CircuitNodeAutoName for Index {
     const PRIORITY: OperatorPriority = OperatorPriority::PostFix {};
 
     fn auto_name(&self) -> Option<Name> {
-        child_name_with_maybe_paren(&Self::PRIORITY, self.node().clone())
+        child_name_with_maybe_paren(&Self::PRIORITY, self.node())
             .map(|n| (n.string() + " idx").into())
     }
 }
@@ -1037,7 +960,7 @@ impl Index {
                 output_ints: std::iter::once(tensor_part_is.clone())
                     .chain(nontensor_part_is.iter().map(|x| tu8v![*x]))
                     .collect(),
-                int_sizes: sv![OpSize::NONE;index.0.len()],
+                int_sizes: sv![Size::NONE;index.0.len()],
             },
             name.as_ref().map(|x| format!("{}_pre_idx_flat", x).into()),
         )?
@@ -1106,8 +1029,8 @@ impl Index {
                     .collect(),
                 int_sizes: broadcasted_shape
                     .iter()
-                    .map(|z| OpSize::from(Some(*z)))
-                    .chain(nontensor_part_is.iter().map(|_| OpSize::NONE))
+                    .copied()
+                    .chain(nontensor_part_is.iter().map(|_| Size::NONE))
                     .collect(),
             },
             name.as_ref().map(|x| format!("{}_post_t_shape", x).into()),
@@ -1172,8 +1095,9 @@ impl Concat {
         if nodes.is_empty() {
             return Err(ConstructError::ConcatZeroNodes {}.into());
         }
-        let node_shape = &nodes[0].info().shape;
-        for node in nodes.iter() {
+        let mut shape = nodes[0].info().shape.clone();
+
+        for node in &nodes[1..] {
             if node.ndim() != nodes[0].ndim() {
                 return Err(ConstructError::ConcatShapeDifferent {
                     shapes: nodes.iter().map(|x| x.info().shape.clone()).collect(),
@@ -1181,21 +1105,21 @@ impl Concat {
                 .into());
             }
             for (i, s) in node.info().shape.iter().enumerate() {
-                if i != axis && *s != node_shape[i] {
-                    return Err(ConstructError::ConcatShapeDifferent {
-                        shapes: nodes.iter().map(|x| x.info().shape.clone()).collect(),
-                    }
-                    .into());
+                if i != axis {
+                    shape[i] = shape[i].join_eq(s, || {
+                        anyhow!(ConstructError::ConcatShapeDifferent {
+                            shapes: nodes.iter().map(|x| x.info().shape.clone()).collect()
+                        })
+                    })?;
                 }
             }
         }
 
-        let mut shape = nodes[0].info().shape.clone();
         shape[axis] = nodes.iter().map(|n| n.info().shape[axis]).sum();
 
         let out = Self {
             axis,
-            info: CachedCircuitInfo::incomplete(name, shape, nodes),
+            info: CachedCircuitInfo::incomplete(name, shape, nodes)?,
         };
         out.initial_init_info()
     }
@@ -1205,7 +1129,7 @@ impl Concat {
             bail!(ConstructError::ConcatZeroNodes {})
         }
         let extra = if is_stack { 1 } else { 0 };
-        Ok(check_canon_axes(nodes[0].info().rank() + extra, &[axis])?[0])
+        Ok(check_canon_axes(nodes[0].rank() + extra, &[axis])?[0])
     }
 
     pub fn new_signed_axis(nodes: Vec<CircuitRc>, axis: i64, name: Option<Name>) -> Result<Concat> {
@@ -1234,7 +1158,7 @@ impl CircuitNode for Concat {
 
     fn child_axis_map(&self) -> Vec<Vec<Option<usize>>> {
         vec![
-            (0..self.info().rank())
+            (0..self.rank())
                 .map(|i| match i == self.axis {
                     true => None,
                     false => Some(i),
@@ -1243,19 +1167,6 @@ impl CircuitNode for Concat {
             self.num_children()
         ]
     }
-
-    fn eval_tensors(&self, tensors: &[Tensor]) -> Result<Tensor> {
-        let tensors = upcast_tensor_device_dtypes(tensors);
-        Python::with_gil(|py| {
-            PY_UTILS
-                .torch
-                .getattr(py, "cat")
-                .unwrap()
-                .call(py, (tensors.to_vec(), self.axis), None)
-                .map_err(|err| err.into())
-                .map(|x| x.extract(py).unwrap())
-        })
-    }
 }
 
 impl CircuitNodeAutoName for Concat {
@@ -1285,7 +1196,7 @@ impl Concat {
         Self::new_signed_axis(nodes, axis, name).map(|x| x.into_init())
     }
 
-    pub fn get_sizes_at_axis(&self) -> Vec<usize> {
+    pub fn get_sizes_at_axis(&self) -> Vec<Size> {
         self.children().map(|x| x.info().shape[self.axis]).collect()
     }
 
@@ -1334,10 +1245,10 @@ impl Scatter {
     ) -> Result<Self> {
         let index = index.canonicalize(&shape);
         let index_shape = index.apply_to_shape(&shape);
-        if index.all_uslices().is_none() {
+        if index.all_uslices().is_none() || !shape_is_known(&shape) {
             return Err(ConstructError::ScatterIndexTypeUnimplemented { index }.into());
         }
-        if index_shape[..] != node.info().shape[..] {
+        if !shape_eq_if_known(&index_shape, node.shape()) {
             return Err(ConstructError::ScatterShapeWrong {
                 shape,
                 index,
@@ -1348,7 +1259,7 @@ impl Scatter {
 
         let out = Self {
             index,
-            info: CachedCircuitInfo::incomplete(name, shape, vec![node]),
+            info: CachedCircuitInfo::incomplete(name, shape, vec![node])?,
         };
         // todo check not too many axes / oob
         out.initial_init_info()
@@ -1380,7 +1291,7 @@ impl CircuitNode for Scatter {
                     cur += 1;
                 }
                 if idx.is_identity(*l) {
-                    Some(i as usize)
+                    Some(i)
                 } else {
                     None
                 }
@@ -1396,20 +1307,6 @@ impl CircuitNode for Scatter {
             self.info().name,
         )
     }
-
-    fn eval_tensors(&self, tensors: &[Tensor]) -> Result<Tensor> {
-        Python::with_gil(|py| {
-            PY_UTILS
-                .scatter
-                .call(
-                    py,
-                    (self.index.clone(), self.shape().clone(), tensors[0].clone()),
-                    None,
-                )
-                .map_err(|err| err.into())
-                .map(|x| x.extract(py).unwrap())
-        })
-    }
 }
 
 impl CircuitNodeAutoName for Scatter {
@@ -1441,7 +1338,7 @@ impl Scatter {
     }
 
     pub fn is_identity(&self) -> bool {
-        self.info().shape[..] == self.node().info().shape[..] && self.index.all_uslices().is_some()
+        shape_eq_and_known(self.shape(), self.node().shape()) && self.index.all_uslices().is_some()
     }
 }
 
@@ -1452,7 +1349,7 @@ pub fn flat_concat(circuits: Vec<CircuitRc>) -> Result<Concat> {
         .map(|x| -> Result<_> {
             Ok(Rearrange::try_new(
                 x.clone(),
-                RearrangeSpec::flatten_usize(x.info().rank())?,
+                RearrangeSpec::flatten_usize(x.rank())?,
                 Some("flatten".into()),
             )
             .unwrap()
@@ -1466,6 +1363,10 @@ pub fn flat_concat(circuits: Vec<CircuitRc>) -> Result<Concat> {
 pub fn flat_concat_back(circuits: Vec<CircuitRc>) -> Result<(Concat, Vec<CircuitRc>)> {
     let flat = flat_concat(circuits.clone())?;
     let sections = flat.get_sizes_at_axis();
+    if !sections.iter().all(|x| x.is_known()) {
+        bail!("flat_concat_back: some concattands of unknown size {sections:?}")
+    }
+    let sections: Vec<usize> = sections.iter().map(|x| x.unwrap()).collect();
     let starts = cumsum(&sections);
     let out = (
         flat.clone(),
@@ -1517,7 +1418,7 @@ impl Conv {
         let out = Self {
             stride,
             padding,
-            info: CachedCircuitInfo::incomplete(name, shape, vec![input, filter]),
+            info: CachedCircuitInfo::incomplete(name, shape, vec![input, filter])?,
         };
         // todo check not too many axes / oob
         out.initial_init_info()
@@ -1548,11 +1449,10 @@ impl CircuitNode for Conv {
                 .map(Some)
                 .chain(vec![
                     None;
-                    self.input().info().rank()
-                        - self.input_batch_shape().len()
+                    self.input().rank() - self.input_batch_shape().len()
                 ])
                 .collect(),
-            vec![None; self.filter().info().rank()],
+            vec![None; self.filter().rank()],
         ]
     }
 
@@ -1565,21 +1465,6 @@ impl CircuitNode for Conv {
             self.info().name,
         )
     }
-
-    fn eval_tensors(&self, tensors: &[Tensor]) -> Result<Tensor> {
-        let tensors = upcast_tensor_device_dtypes(tensors);
-        pycall!(
-            PY_UTILS.conv,
-            (
-                self.dims(),
-                tensors[0].clone(),
-                tensors[1].clone(),
-                self.stride.clone(),
-                self.padding.clone(),
-            ),
-            anyhow
-        )
-    }
 }
 
 impl CircuitNodeAutoName for Conv {
@@ -1629,7 +1514,7 @@ impl Conv {
         self.stride.len()
     }
     pub fn input_batch_shape(&self) -> Shape {
-        self.input().info().shape[..self.info().rank() - 1 - self.dims()]
+        self.input().info().shape[..self.rank() - 1 - self.dims()]
             .iter()
             .cloned()
             .collect()
