@@ -8,9 +8,9 @@ use pyo3::{
 use rr_util::{
     lru_cache::TensorCacheRrfs,
     name::Name,
-    py_types::{scalar_to_tensor, Tensor, PY_UTILS},
+    py_types::{Tensor, PY_UTILS},
+    shape::{shape_as_known_result, shape_is_known},
     sv,
-    symbolic_size::{shape_first_symbolic_dim, SymbolicSizeProduct},
     tensor_db::{get_tensor_prefix, save_tensor},
     tensor_util::TorchDeviceDtypeOp,
 };
@@ -22,7 +22,7 @@ use crate::{
     circuit_node_private::{CircuitNodeComputeInfoImpl, CircuitNodeHashItems},
     new_rc, new_rc_unwrap,
     prelude::*,
-    CachedCircuitInfo, CircuitFlags, PyCircuitBase, Shape, TensorEvalError,
+    CachedCircuitInfo, CircuitFlags, PyCircuitBase, Shape,
 };
 
 macro_rules! circuit_node_auto_leaf_impl {
@@ -56,11 +56,7 @@ pub struct Array {
 
 circuit_node_extra_impl!(Array, self_hash_default);
 
-impl CircuitNodeComputeInfoImpl for Array {
-    fn device_dtype_extra(&self) -> Option<Vec<TorchDeviceDtypeOp>> {
-        Some(vec![self.value.device_dtype().into()])
-    }
-}
+impl CircuitNodeComputeInfoImpl for Array {}
 
 impl CircuitNodeHashItems for Array {
     fn compute_hash_non_name_non_children(&self, hasher: &mut blake3::Hasher) {
@@ -70,24 +66,19 @@ impl CircuitNodeHashItems for Array {
 
 impl CircuitNode for Array {
     circuit_node_auto_leaf_impl!("b2aac9d5-1bfa-4c2a-9684-e3f9ecbc1b94");
-
-    fn eval_tensors(&self, _tensors: &[Tensor]) -> Result<Tensor> {
-        Ok(self.value.clone())
-    }
 }
 
 impl Array {
     #[apply(new_rc_unwrap)]
     pub fn try_new(value: Tensor, name: Option<Name>) -> Result<Self> {
         let value = value.hashed()?;
-        if let Some(i) = shape_first_symbolic_dim(value.shape()) {
-            bail!(ConstructError::ArrayHasReservedSymbolicShape {
-                number: value.shape()[i],
-                sym: SymbolicSizeProduct::from(value.shape()[i]).to_string(),
-            })
-        }
         let out = Self {
-            info: CachedCircuitInfo::incomplete(name, value.shape().clone(), vec![]),
+            info: CachedCircuitInfo::with_device_dtype(
+                name,
+                value.shape().clone(),
+                vec![],
+                value.device_dtype().into(),
+            ),
             value,
         };
         out.initial_init_info()
@@ -105,11 +96,15 @@ impl Array {
         get_tensor_prefix(hash_base16).map(|value| Array::new(value, name))
     }
 
-    pub fn randn(shape: Shape) -> Self {
+    pub fn randn(shape: Shape) -> Result<Self> {
         Self::randn_full(shape, None, TorchDeviceDtypeOp::TENSOR_DEFAULT, None)
     }
 
-    pub fn randn_named(shape: Shape, name: Option<Name>, device_dtype: TorchDeviceDtypeOp) -> Self {
+    pub fn randn_named(
+        shape: Shape,
+        name: Option<Name>,
+        device_dtype: TorchDeviceDtypeOp,
+    ) -> Result<Self> {
         Self::randn_full(shape, name, device_dtype, None)
     }
 }
@@ -134,7 +129,7 @@ impl Array {
         name: Option<Name>,
         device_dtype: TorchDeviceDtypeOp,
         seed: Option<usize>,
-    ) -> Self {
+    ) -> Result<Self> {
         Python::with_gil(|py| {
             if let Some(seed) = seed {
                 PY_UTILS
@@ -153,14 +148,14 @@ impl Array {
             if let Some(device) = device_dtype.device {
                 kwargs.insert("device", device.into_py(py));
             }
-            Array::new(
+            Array::try_new(
                 PY_UTILS
                     .torch
                     .getattr(py, "randn")
                     .unwrap()
                     .call(
                         py,
-                        (PyTuple::new(py, shape),),
+                        (PyTuple::new(py, shape_as_known_result(&shape)?),),
                         Some(kwargs.into_py_dict(py)),
                     )
                     .unwrap()
@@ -223,13 +218,6 @@ impl CircuitNodeHashItems for Symbol {
 
 impl CircuitNode for Symbol {
     circuit_node_auto_leaf_impl!("13c3ee63-76e9-4afb-8057-40309d17b458");
-
-    fn eval_tensors(&self, _tensors: &[Tensor]) -> Result<Tensor> {
-        Err(TensorEvalError::NotExplicitlyComputableInternal {
-            circuit: self.crc(),
-        }
-        .into())
-    }
 }
 
 impl Symbol {
@@ -237,10 +225,24 @@ impl Symbol {
     pub fn new(shape: Shape, uuid: Uuid, name: Option<Name>) -> (Self) {
         let out = Self {
             uuid,
-            info: CachedCircuitInfo::incomplete(name, shape, vec![]),
+            info: CachedCircuitInfo::incomplete(name, shape, vec![]).unwrap(),
         };
         out.initial_init_info().unwrap()
     }
+
+    // HACK: currently only used for Schedule & not exported to python
+    pub fn new_device_dtype_random_uuid(
+        shape: Shape,
+        name: Option<Name>,
+        device_dtype: TorchDeviceDtypeOp,
+    ) -> Self {
+        Self {
+            uuid: Uuid::new_v4(),
+            info: CachedCircuitInfo::with_device_dtype(name, shape, vec![], device_dtype),
+        }
+        .initial_init_info()
+        .unwrap()
+    }
 }
 
 #[pymethods]
@@ -289,25 +291,21 @@ impl CircuitNodeHashItems for Scalar {
 
 impl CircuitNode for Scalar {
     circuit_node_auto_leaf_impl!("78a77905-8b3f-4471-bb77-255673941fef");
-
-    fn eval_tensors(&self, _tensors: &[Tensor]) -> Result<Tensor> {
-        Ok(scalar_to_tensor(
-            self.value,
-            self.info().shape.clone(),
-            Default::default(),
-        )?)
-    }
 }
 
 impl Scalar {
-    #[apply(new_rc)]
-    pub fn new(value: f64, shape: Shape, name: Option<Name>) -> (Self) {
+    #[apply(new_rc_unwrap)]
+    pub fn try_new(value: f64, shape: Shape, name: Option<Name>) -> Result<Self> {
+        if !shape_is_known(&shape) {
+            bail!(ConstructError::ScalarUnknownSizes { shape, name, value })
+        }
+
         let out = Self {
             value,
-            info: CachedCircuitInfo::incomplete(name, shape, vec![]),
+            info: CachedCircuitInfo::incomplete(name, shape, vec![])?,
         };
 
-        out.initial_init_info().unwrap()
+        out.initial_init_info()
     }
 }
 
@@ -315,8 +313,8 @@ impl Scalar {
 impl Scalar {
     #[new]
     #[pyo3(signature=(value, shape = sv![], name = None))]
-    fn py_new(value: f64, shape: Shape, name: Option<Name>) -> PyClassInitializer<Self> {
-        Self::new(value, shape, name).into_init()
+    fn py_new(value: f64, shape: Shape, name: Option<Name>) -> Result<PyClassInitializer<Self>> {
+        Ok(Self::try_new(value, shape, name)?.into_init())
     }
 
     pub fn is_zero(&self) -> bool {
@@ -327,14 +325,7 @@ impl Scalar {
         self.value == 1.
     }
 
-    pub fn evolve_shape(&self, shape: Shape) -> Self {
-        Self::new(self.value, shape, self.info().name)
+    pub fn evolve_shape(&self, shape: Shape) -> Result<Self> {
+        Self::try_new(self.value, shape, self.info().name)
     }
 }
-
-#[test]
-fn test_nrc() {
-    pyo3::prepare_freethreaded_python();
-    let ex = Scalar::nrc(0.0, sv![1, 2], None);
-    ex.print().unwrap();
-}
