@@ -12,14 +12,17 @@ use pyo3::{
 use regex::Regex;
 use rr_util::{
     name::Name,
-    py_types::{PyShape, Tensor, PY_UTILS, SELF_MODULE},
-    pycall, python_error_exception, simple_from,
+    py_types::{scalar_to_tensor, PySymShape, Tensor, PY_UTILS, SELF_MODULE},
+    pycall, python_error_exception,
+    shape::{broadcast_shapes_impl, shape_eq_if_known, shape_join_eq, shape_to_py, Shape, Size},
+    simple_from, sv,
     tensor_util::{
-        broadcast_shapes_impl, check_canon_idxs, eq_or_symbolic, upcast_tensor_device_dtypes,
-        upcast_tensor_devices, MiscInputError, Shape, TorchDevice, TorchDeviceDtypeOp, TorchDtype,
+        check_canon_idxs, upcast_tensor_device_dtypes, upcast_tensor_devices, MiscInputError,
+        TorchDevice, TorchDeviceDtype, TorchDeviceDtypeOp, TorchDtype,
     },
 };
 use rustc_hash::FxHashMap as HashMap;
+use smallvec::ToSmallVec;
 use thiserror::Error;
 use uuid::uuid;
 
@@ -123,13 +126,16 @@ pub trait SpecTrait: fmt::Debug + ToPyObject {
     fn is_official(&self) -> bool {
         false
     }
-    fn serialize(&self) -> Result<Option<String>>;
+    fn serialize(&self) -> Result<String>;
     fn get_device_dtype_override(
         &self,
         _device_dtypes: &[TorchDeviceDtypeOp],
-    ) -> Option<Result<TorchDeviceDtypeOp>> {
-        None
+    ) -> Result<Option<TorchDeviceDtypeOp>> {
+        Ok(None)
     } // returning None means use normal device_dtype inheritance. if Some, then error if dtypes are incorrect and dtype if correct
+    fn should_const_fold(&self) -> bool {
+        true
+    }
 }
 
 #[pyclass]
@@ -146,8 +152,8 @@ pub struct GeneralFunctionShapeInfo {
 #[pymethods]
 impl GeneralFunctionShapeInfo {
     #[getter]
-    fn shape(&self) -> PyShape {
-        PyShape(self.shape.clone())
+    fn shape(&self) -> PySymShape {
+        PySymShape(self.shape.clone())
     }
 
     /// no checking done here, validation done in validate
@@ -176,18 +182,20 @@ impl GeneralFunctionShapeInfo {
 
         if self.input_batchability.iter().any(|b| *b) {
             let batch_dims = self.shape.len() - self.num_non_batchable_output_dims as usize;
-            let out_batch = &self.shape[..batch_dims];
-            let batches: Vec<&[usize]> = shapes
+            let mut batch_shape = self.shape[..batch_dims].to_smallvec();
+            let batches: Vec<&[Size]> = shapes
                 .iter()
                 .zip(&self.input_batchability)
                 .filter(|(_, b)| **b)
                 .map(|(s, _)| &s[..batch_dims])
                 .collect();
-            if !batches.iter().all(|x| eq_or_symbolic(x, &out_batch)) {
-                bail!(GeneralFunctionShapeError::BatchDimsNotEqual {
-                    batches: batches.into_iter().map(|x| x.into()).collect(),
-                    output_batch: out_batch.into()
-                })
+            for x in &batches[..] {
+                batch_shape = shape_join_eq(&batch_shape, x, || {
+                    anyhow!(GeneralFunctionShapeError::BatchDimsNotEqual {
+                        batches: batches.iter().cloned().map(|x| x.into()).collect(),
+                        output_batch: self.shape[..batch_dims].into()
+                    })
+                })?
             }
         }
 
@@ -298,8 +306,8 @@ impl SpecTrait for GeneralFunctionSimpleSpec {
     fn is_official(&self) -> bool {
         true
     }
-    fn serialize(&self) -> Result<Option<String>> {
-        Ok(Some(self.name().into()))
+    fn serialize(&self) -> Result<String> {
+        Ok(self.name().into())
     }
 }
 
@@ -328,14 +336,14 @@ impl ToPyObject for GeneralFunctionIndexSpec {
 
 fn index_get_device_dtype_override(
     device_dtypes: &[TorchDeviceDtypeOp],
-) -> Option<Result<TorchDeviceDtypeOp>> {
+) -> Result<Option<TorchDeviceDtypeOp>> {
     if let Some(dev1) = device_dtypes[0].device && let Some(dev2) = device_dtypes[1].device && dev1!=dev2{
-            return Some(Err(MiscInputError::ChildrenMultipleDevices { a: device_dtypes[0].device, b: device_dtypes[1].device}.into()))
+            return Err(MiscInputError::ChildrenMultipleDevices { a: device_dtypes[0].device, b: device_dtypes[1].device}.into())
         }
     if let Some(dtype) = device_dtypes[1].dtype && dtype != TorchDtype::int64 {
-            return Some(Err(MiscInputError::IndexDtypeNotI64 {}.into()));
+            return Err(MiscInputError::IndexDtypeNotI64 {}.into());
         }
-    Some(Ok(TorchDeviceDtypeOp {
+    Ok(Some(TorchDeviceDtypeOp {
         device: device_dtypes[0].device.or(device_dtypes[1].device),
         dtype: device_dtypes[0].dtype,
     }))
@@ -406,7 +414,7 @@ impl SpecTrait for GeneralFunctionIndexSpec {
                 bail!(get_err())
             }
 
-            if &x_shape[..prefix_len] != &index_shape[..] {
+            if !shape_eq_if_known(&x_shape[..prefix_len], &index_shape[..]) {
                 bail!(get_err())
             }
 
@@ -473,7 +481,7 @@ impl SpecTrait for GeneralFunctionIndexSpec {
     fn get_device_dtype_override(
         &self,
         device_dtypes: &[TorchDeviceDtypeOp],
-    ) -> Option<Result<TorchDeviceDtypeOp>> {
+    ) -> Result<Option<TorchDeviceDtypeOp>> {
         index_get_device_dtype_override(device_dtypes)
     }
     fn name(&self) -> &'static str {
@@ -493,8 +501,8 @@ impl SpecTrait for GeneralFunctionIndexSpec {
     fn is_official(&self) -> bool {
         true
     }
-    fn serialize(&self) -> Result<Option<String>> {
-        Ok(Some(self.name().into()))
+    fn serialize(&self) -> Result<String> {
+        Ok(self.name().into())
     }
 }
 
@@ -612,7 +620,7 @@ impl SpecTrait for GeneralFunctionExplicitIndexSpec {
             bail!(get_err())
         }
 
-        if x_shape[..batch_len] != index_shape[..batch_len] {
+        if !shape_eq_if_known(&x_shape[..batch_len], &index_shape[..batch_len]) {
             bail!(get_err())
         }
 
@@ -639,7 +647,7 @@ impl SpecTrait for GeneralFunctionExplicitIndexSpec {
     fn get_device_dtype_override(
         &self,
         device_dtypes: &[TorchDeviceDtypeOp],
-    ) -> Option<Result<TorchDeviceDtypeOp>> {
+    ) -> Result<Option<TorchDeviceDtypeOp>> {
         index_get_device_dtype_override(device_dtypes)
     }
     fn name(&self) -> &'static str {
@@ -654,8 +662,8 @@ impl SpecTrait for GeneralFunctionExplicitIndexSpec {
     fn is_official(&self) -> bool {
         true
     }
-    fn serialize(&self) -> Result<Option<String>> {
-        Ok(Some(self.name().into()))
+    fn serialize(&self) -> Result<String> {
+        Ok(self.name().into())
     }
 }
 
@@ -704,25 +712,24 @@ impl SpecTrait for GeneralFunctionSetDDSpec {
         Ok(GeneralFunctionShapeInfo {
             shape: shapes[0].clone(),
             num_non_batchable_output_dims: 0,
-            input_batchability: vec![true].try_into().unwrap(),
+            input_batchability: vec![true],
         })
     }
     fn get_device_dtype_override(
         &self,
         device_dtypes: &[TorchDeviceDtypeOp],
-    ) -> Option<Result<TorchDeviceDtypeOp>> {
-        Some(
+    ) -> Result<Option<TorchDeviceDtypeOp>> {
+        Ok(Some(
             self.input_required_compatibility
                 .combine(device_dtypes[0])
                 .map_err(|_e| {
-                    MiscInputError::CastIncompatibleDeviceDtype {
+                    anyhow!(MiscInputError::CastIncompatibleDeviceDtype {
                         required: self.input_required_compatibility,
                         actual: device_dtypes[0],
-                    }
-                    .into()
+                    })
                 })
-                .map(|_| self.output.override_other(device_dtypes[0])),
-        )
+                .map(|_| self.output.override_other(device_dtypes[0]))?,
+        ))
     }
 
     // fn has_jacobian(&self) -> Result<bool> {
@@ -737,8 +744,8 @@ impl SpecTrait for GeneralFunctionSetDDSpec {
         ))
         .into()
     }
-    fn serialize(&self) -> Result<Option<String>> {
-        Ok(Some(self.name().to_owned()))
+    fn serialize(&self) -> Result<String> {
+        Ok(self.name().to_owned())
     }
 }
 
@@ -822,8 +829,8 @@ impl SpecTrait for GeneralFunctionPowSpec {
         true
     }
 
-    fn serialize(&self) -> Result<Option<String>> {
-        Ok(Some(self.name().to_owned()))
+    fn serialize(&self) -> Result<String> {
+        Ok(self.name().to_owned())
     }
 }
 
@@ -848,8 +855,8 @@ pub struct GeneralFunctionMultinomialSpec {
 impl GeneralFunctionMultinomialSpec {
     #[getter]
     #[pyo3(name = "shape")]
-    fn py_shape(&self) -> PyShape {
-        PyShape(self.shape.clone())
+    fn py_shape(&self) -> PySymShape {
+        PySymShape(self.shape.clone())
     }
 }
 
@@ -937,44 +944,42 @@ impl SpecTrait for GeneralFunctionMultinomialSpec {
     fn get_device_dtype_override(
         &self,
         device_dtypes: &[TorchDeviceDtypeOp],
-    ) -> Option<Result<TorchDeviceDtypeOp>> {
+    ) -> Result<Option<TorchDeviceDtypeOp>> {
         let (probs_dd, seed_dd) = (device_dtypes[0], device_dtypes[1]);
         if probs_dd
             .dtype
             .map(|dtype| !dtype.is_floating_point())
             .unwrap_or(false)
         {
-            return Some(Err(anyhow!("probs dtype not floating point")));
+            bail!("probs dtype not floating point")
         }
         if seed_dd
             .dtype
             .map(|dtype| dtype != TorchDtype::int64)
             .unwrap_or(false)
         {
-            return Some(Err(anyhow!(
-                "seed should be int64 (we could support other integers if we wanted)"
-            )));
+            bail!("seed should be int64 (we could support other integers if we wanted)")
         }
         if seed_dd
             .device
             .map(|device| device != TorchDevice::Cpu)
             .unwrap_or(false)
         {
-            return Some(Err(anyhow!("seed should be on cpu (for now at least)")));
+            bail!("seed should be on cpu (for now at least)")
         }
 
-        Some(Ok(TorchDeviceDtypeOp {
+        Ok(Some(TorchDeviceDtypeOp {
             device: probs_dd.device,
             dtype: Some(TorchDtype::int64),
         }))
     }
 
-    fn serialize(&self) -> Result<Option<String>> {
-        Ok(Some(format!(
+    fn serialize(&self) -> Result<String> {
+        Ok(format!(
             "multinomial{}_{:?}",
             if !self.replacement { "_no_replace" } else { "" },
             self.shape
-        )))
+        ))
     }
 }
 
@@ -994,6 +999,49 @@ pub fn multinomial(
     )
 }
 
+#[pyclass]
+#[derive(Debug, Clone)]
+/// used internally to optimize & evaluate multiple circuits at once
+pub struct GeneralFunctionOutputSpec();
+
+impl SpecTrait for GeneralFunctionOutputSpec {
+    fn compute_hash(&self) -> HashBytes {
+        let mut hasher = blake3::Hasher::new();
+        hasher.update(uuid!("b79f926e-431d-47dc-a571-182c01eda8c9").as_bytes());
+        *hasher.finalize().as_bytes()
+    }
+
+    fn function(&self, _tensors: &[Tensor]) -> Result<Tensor> {
+        scalar_to_tensor(0., sv![], TorchDeviceDtype::default())
+    }
+
+    fn get_shape_info(&self, shapes: &[Shape]) -> Result<GeneralFunctionShapeInfo> {
+        Ok(GeneralFunctionShapeInfo {
+            shape: sv![],
+            num_non_batchable_output_dims: 0,
+            input_batchability: shapes.iter().map(|_| false).collect(),
+        })
+    }
+
+    fn name(&self) -> &'static str {
+        "Output"
+    }
+
+    fn serialize(&self) -> Result<String> {
+        Ok("Output".to_owned())
+    }
+
+    fn should_const_fold(&self) -> bool {
+        false
+    }
+}
+
+impl ToPyObject for GeneralFunctionOutputSpec {
+    fn to_object(&self, py: Python<'_>) -> PyObject {
+        self.clone().into_py(py)
+    }
+}
+
 enum PyModuleLocator {
     Path(String),
     Module(String),
@@ -1001,7 +1049,7 @@ enum PyModuleLocator {
 
 impl PyModuleLocator {
     fn parse(s: String) -> Self {
-        if s.starts_with("/") && s.ends_with(".py") {
+        if s.starts_with('/') && s.ends_with(".py") {
             PyModuleLocator::Path(s)
         } else {
             PyModuleLocator::Module(s)
@@ -1017,7 +1065,7 @@ impl PyModuleLocator {
 
                 let importlib_util = py.import("importlib.util")?;
                 let spec = importlib_util
-                    .call_method1("spec_from_file_location", ("user_funcs", rrfs_dir + &path))?;
+                    .call_method1("spec_from_file_location", ("user_funcs", rrfs_dir + path))?;
                 let function_spec_module =
                     importlib_util.call_method1("module_from_spec", (spec,))?;
                 spec.getattr("loader")?
@@ -1036,17 +1084,17 @@ struct PyClassLocator {
 
 impl PyClassLocator {
     fn parse(s: &String) -> Result<Self> {
-        return (|| -> Result<Self> {
-            let mut split = s.splitn(2, ":");
+        (|| -> Result<Self> {
+            let mut split = s.splitn(2, ':');
             let file_path_or_module = convert_op_str_to_format_err(split.next(), s)?;
             let spec_name = convert_op_str_to_format_err(split.next(), s)?;
             Ok(PyClassLocator {
                 module: PyModuleLocator::parse(file_path_or_module),
-                name: spec_name.to_string(),
+                name: spec_name,
             })
         })().context(
             "Function locator must either be of the form module.to.import:MyCustomSpec or /dir/sub_dir/.../file.py:MyCustomSpec (path relative to rrfs)"
-        );
+        )
     }
 
     fn get_class<'py>(&self, py: Python<'py>) -> Result<&'py PyAny> {
@@ -1154,8 +1202,7 @@ impl SpecTrait for Box<PyWrap> {
                 .call_method1(
                     py,
                     "get_shape_info",
-                    // PyShape here is important!
-                    PyTuple::new(py, shapes.iter().map(|x| PyShape(x.clone()).into_py(py))),
+                    PyTuple::new(py, shapes.iter().map(|x| shape_to_py(x))),
                 )?
                 .extract(py)?)
         })
@@ -1163,8 +1210,8 @@ impl SpecTrait for Box<PyWrap> {
     fn get_device_dtype_override(
         &self,
         device_dtypes: &[TorchDeviceDtypeOp],
-    ) -> Option<Result<TorchDeviceDtypeOp>> {
-        let resulty: Result<Option<TorchDeviceDtypeOp>> = Python::with_gil(|py| {
+    ) -> Result<Option<TorchDeviceDtypeOp>> {
+        Python::with_gil(|py| {
             Ok(self
                 .ob
                 .call_method1(
@@ -1179,20 +1226,16 @@ impl SpecTrait for Box<PyWrap> {
                     ),
                 )?
                 .extract(py)?)
-        });
-        if let Err(e) = resulty {
-            return Some(Err(e));
-        }
-        return resulty.unwrap().map(|x| Ok(x));
+        })
     }
     fn name(&self) -> &'static str {
         self.name.into()
     }
-    fn serialize(&self) -> Result<Option<String>> {
+    fn serialize(&self) -> Result<String> {
         match &self.path {
-            Some(p) => Ok(Some(format!("{} at {}", &self.name, p))),
+            Some(p) => Ok(format!("{} at {}", &self.name, p)),
             None => bail!(anyhow!(format!(
-                "Can't serialize function {}: no path given",
+                "Can't serialize function {}: no path given, override GeneralFunctionSpecBase.path to specify path",
                 &self.name
             ))),
         }
@@ -1207,6 +1250,7 @@ pub enum GeneralFunctionSpec {
     SetDD(GeneralFunctionSetDDSpec),
     Pow(GeneralFunctionPowSpec),
     Multinomial(GeneralFunctionMultinomialSpec),
+    Output(GeneralFunctionOutputSpec),
     Py(Box<PyWrap>), // boxed bc it was long pole for Circuit size
 }
 
@@ -1220,6 +1264,7 @@ impl GeneralFunctionSpec {
             Self::SetDD(x) => x,
             Self::Pow(x) => x,
             Self::Multinomial(x) => x,
+            Self::Output(x) => x,
             Self::Py(x) => x,
         }
     }
@@ -1249,7 +1294,7 @@ impl SpecTrait for GeneralFunctionSpec {
     fn get_device_dtype_override(
         &self,
         device_dtypes: &[TorchDeviceDtypeOp],
-    ) -> Option<Result<TorchDeviceDtypeOp>> {
+    ) -> Result<Option<TorchDeviceDtypeOp>> {
         self.as_trait_obj().get_device_dtype_override(device_dtypes)
     }
     // fn has_jacobian(&self) -> Result<bool>;
@@ -1260,9 +1305,12 @@ impl SpecTrait for GeneralFunctionSpec {
     fn is_official(&self) -> bool {
         self.as_trait_obj().is_official()
     }
-    fn serialize(&self) -> Result<Option<String>> {
+    fn serialize(&self) -> Result<String> {
         self.as_trait_obj().serialize()
     }
+    fn should_const_fold(&self) -> bool {
+        self.as_trait_obj().should_const_fold()
+    }
 }
 
 #[pyclass(extends=PyCircuitBase)]
@@ -1288,7 +1336,7 @@ impl GeneralFunction {
             .map(|x| x.shape().clone())
             .collect::<Vec<Shape>>();
 
-        let si = spec.get_shape_info(&shapes).with_context(|| {
+        let si = spec.get_shape_info(&shapes[..]).with_context(|| {
             format!(
                 "failed to compute shape info in new with spec={:?} input_shapes={:?}",
                 spec,
@@ -1301,7 +1349,7 @@ impl GeneralFunction {
 
         si.validate(&shapes).with_context(|| {
             format!(
-                "invalid shape info in new with spec={:?} input_shapes={:?}",
+                "invalid shape info in new with spec={:?} input_shapes={:?} (likely input shape error)",
                 spec,
                 nodes
                     .iter()
@@ -1316,9 +1364,19 @@ impl GeneralFunction {
             input_batchability,
         } = si;
 
+        let ddto = spec.get_device_dtype_override(
+            &nodes
+                .iter()
+                .map(|x| x.info().device_dtype)
+                .collect::<Vec<_>>(),
+        )?;
+
         let out = Self {
             spec,
-            info: CachedCircuitInfo::incomplete(name, shape, nodes),
+            info: match ddto {
+                Some(ddt) => CachedCircuitInfo::with_device_dtype(name, shape, nodes, ddt),
+                None => CachedCircuitInfo::incomplete(name, shape, nodes)?,
+            },
             num_non_batchable_output_dims,
             input_batchability,
         };
@@ -1332,16 +1390,7 @@ impl GeneralFunction {
 
 circuit_node_extra_impl!(GeneralFunction, self_hash_default);
 
-impl CircuitNodeComputeInfoImpl for GeneralFunction {
-    fn device_dtype_override(&self) -> Option<Result<TorchDeviceDtypeOp>> {
-        self.spec.get_device_dtype_override(
-            &self
-                .children()
-                .map(|x| x.info().device_dtype)
-                .collect::<Vec<_>>(),
-        )
-    }
-}
+impl CircuitNodeComputeInfoImpl for GeneralFunction {}
 
 impl CircuitNodeHashItems for GeneralFunction {
     fn compute_hash_non_name_non_children(&self, hasher: &mut blake3::Hasher) {
@@ -1357,13 +1406,13 @@ impl CircuitNode for GeneralFunction {
     }
 
     fn child_axis_map(&self) -> Vec<Vec<Option<usize>>> {
-        let num_batchable_axes = self.info().rank() as u8 - self.num_non_batchable_output_dims;
+        let num_batchable_axes = self.rank() as u8 - self.num_non_batchable_output_dims;
         zip(self.children(), &self.input_batchability)
             .map(|(child, batchable)| {
                 if !batchable {
-                    vec![None; child.info().rank()]
+                    vec![None; child.rank()]
                 } else {
-                    (0..child.info().rank())
+                    (0..child.rank())
                         .map(|i| match i < num_batchable_axes as usize {
                             true => Some(i),
                             false => None,
@@ -1373,10 +1422,6 @@ impl CircuitNode for GeneralFunction {
             })
             .collect()
     }
-
-    fn eval_tensors(&self, tensors: &[Tensor]) -> Result<Tensor> {
-        self.spec.function(tensors)
-    }
 }
 
 impl CircuitNodeAutoName for GeneralFunction {
@@ -1469,7 +1514,7 @@ impl GeneralFunction {
                 .parse()
                 .context("failed to parse index_dim for gen_index_at")?;
             let batch_x = re_captures.get(2).is_some();
-            let batch_index = !re_captures.get(3).is_some();
+            let batch_index = re_captures.get(3).is_none();
             let check_index_ints = re_captures.get(4).is_some();
             GeneralFunctionIndexSpec {
                 index_dim,
@@ -1493,7 +1538,7 @@ impl GeneralFunction {
             }
             let shape = axis_strs.into_iter().map(|x| x.parse().unwrap()).collect();
             GeneralFunctionMultinomialSpec {
-                replacement: !re_captures.get(1).is_some(),
+                replacement: re_captures.get(1).is_none(),
                 shape,
             }
             .into()
@@ -1551,6 +1596,8 @@ impl GeneralFunction {
             .into()
         } else if spec_name.str() == "pow" {
             GeneralFunctionPowSpec {}.into()
+        } else if spec_name.str() == "Output" {
+            GeneralFunctionSpec::Output(GeneralFunctionOutputSpec {})
         } else if let Some(spec) = SPECS.get(&spec_name) {
             spec.clone()
         } else {
