@@ -7,6 +7,7 @@ use std::{
 };
 
 use anyhow::{bail, Context, Result};
+use itertools::izip;
 use macro_rules_attribute::apply;
 use pyo3::{exceptions::PyValueError, prelude::*, pyclass::CompareOp};
 use rr_util::{
@@ -17,12 +18,12 @@ use rr_util::{
     py_types::{use_rust_comp, PyCallable},
     pycall, python_error_exception,
     rearrange_spec::RearrangeSpec,
-    symbolic_size::SymbolicSizeProduct,
-    tensor_util::{right_align_shapes, Shape},
+    shape::{right_align_shapes, shape_is_known, Shape, Size},
     util::{arc_unwrap_or_clone, counts_g_1, is_unique, HashBytes},
     IndexSet,
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet, FxHasher};
+use smallvec::ToSmallVec;
 use thiserror::Error;
 use uuid::{uuid, Uuid};
 
@@ -36,11 +37,11 @@ use crate::{
         deep_map_preorder_unwrap, visit_circuit_unwrap,
     },
     constant::Symbol,
-    expand_node::{replace_expand_map, ExpandError},
     named_axes::{deep_strip_axis_names, get_axis_leaves, propagate_named_axes},
     new_rc_unwrap,
     prelude::*,
-    PyCircuitBase, Rearrange, TensorEvalError,
+    self_funcs::replace_expand_map,
+    PyCircuitBase, Rearrange,
 };
 
 /// can also be thought of as lambda from lambda calculus (expression with free variables + list of these variables)
@@ -67,7 +68,7 @@ where
         .into_iter()
         .rev()
         .scan(HashSet::default(), |prior_syms, spec| {
-            let out = (!prior_syms.contains(&spec.symbol)) && has_free_sym(&circuit, &spec.symbol);
+            let out = (!prior_syms.contains(&spec.symbol)) && has_free_sym(circuit, &spec.symbol);
             prior_syms.insert(&spec.symbol);
             Some(out)
         })
@@ -142,7 +143,7 @@ impl ModuleSpec {
         Uuid::from_bytes(x)
     }
 
-    pub fn batch_shapes<'a>(&self, nodes: &'a [CircuitRc]) -> Vec<&'a [usize]> {
+    pub fn batch_shapes<'a>(&self, nodes: &'a [CircuitRc]) -> Vec<&'a [Size]> {
         zip(self.arg_specs.iter(), nodes)
             .map(|(arg_spec, node)| &node.shape()[..node.ndim() - arg_spec.symbol.ndim()])
             .collect()
@@ -152,18 +153,38 @@ impl ModuleSpec {
         right_align_shapes(&self.batch_shapes(nodes))
     }
 
-    pub fn expand_raw(&self, nodes: &Vec<CircuitRc>) -> Result<CircuitRc> {
+    pub fn out_shape(&self, nodes: &[CircuitRc]) -> Result<Shape> {
+        let bs = self.aligned_batch_shape(nodes)?;
+        return Ok([&bs[..], self.circuit.shape()].concat().to_smallvec());
+    }
+
+    pub fn check_expand_valid_shapes(
+        &self,
+        nodes: &[CircuitRc],
+        check_inherited_restrictions: bool, /* currently only checked on module construction, not on resize */
+    ) -> Result<()> {
         if self.arg_specs.len() != nodes.len() {
             bail!(ConstructError::ModuleWrongNumberChildren {
                 expected: self.arg_specs.len(),
                 got: nodes.len(),
                 arg_specs: self.arg_specs.clone(),
-                nodes: nodes.clone(),
+                nodes: nodes.to_vec(),
             });
         }
-        for (arg_spec, node) in zip(self.arg_specs.iter(), nodes) {
-            if node.info().rank() < arg_spec.symbol.info().rank() {
-                bail!(ExpandError::ModuleRankReduced {
+
+        let spec_restrictions =
+            check_inherited_restrictions.then(|| get_symbol_restrictions(&self.circuit));
+        for (arg_spec, node) in izip!(self.arg_specs.iter(), nodes) {
+            let (a, b, c) = spec_restrictions
+                .as_ref()
+                .and_then(|m| m.get(&arg_spec.symbol).cloned())
+                .unwrap_or((true, true, false));
+            let batchable = a && arg_spec.batchable;
+            let expandable = b && arg_spec.expandable;
+            let ban_non_symbolic_size_expand = c || arg_spec.ban_non_symbolic_size_expand;
+
+            if node.rank() < arg_spec.symbol.info().rank() {
+                bail!(ConstructError::ModuleRankReduced {
                     node_rank: node.rank(),
                     symbol_rank: arg_spec.symbol.rank(),
                     arg_spec: arg_spec.clone(),
@@ -171,47 +192,53 @@ impl ModuleSpec {
                     spec_circuit: self.circuit.clone()
                 });
             }
-            if !arg_spec.batchable && node.info().rank() > arg_spec.symbol.info().rank() {
-                bail!(ExpandError::ModuleTriedToBatchUnbatchableInput {
+            if !batchable && node.rank() > arg_spec.symbol.info().rank() {
+                bail!(ConstructError::ModuleTriedToBatchUnbatchableInput {
                     node_rank: node.rank(),
                     symbol_rank: arg_spec.symbol.rank(),
                     arg_spec: arg_spec.clone(),
                     spec_circuit: self.circuit.clone()
                 });
             }
-            if !arg_spec.expandable
-                && node.info().shape[node.info().rank() - arg_spec.symbol.info().rank()..]
-                    != arg_spec.symbol.info().shape[..]
+            if !expandable
+                && node.shape()[node.rank() - arg_spec.symbol.rank()..]
+                    != arg_spec.symbol.shape()[..]
             {
-                bail!(ExpandError::ModuleTriedToExpandUnexpandableInput {
+                bail!(ConstructError::ModuleTriedToExpandUnexpandableInput {
                     node_shape: node.shape().clone(),
                     symbol_shape: arg_spec.symbol.shape().clone(),
                     arg_spec: arg_spec.clone(),
                     spec_circuit: self.circuit.clone()
                 });
             }
-            if arg_spec.ban_non_symbolic_size_expand {
+            if ban_non_symbolic_size_expand {
                 for (dim, (&new_size, &old_size)) in node.shape()
                     [node.rank() - arg_spec.symbol.rank()..]
                     .iter()
                     .zip(arg_spec.symbol.shape())
                     .enumerate()
                 {
-                    if new_size != old_size && !SymbolicSizeProduct::has_symbolic(old_size) {
-                        bail!(ExpandError::ModuleTriedToExpandOnNonSymbolicSizeAndBanned {
-                            old_size,
-                            new_size,
-                            dim,
-                            node_shape: node.shape().clone(),
-                            arg_spec: arg_spec.clone(),
-                            spec_circuit: self.circuit.clone()
-                        });
+                    if new_size != old_size && old_size.is_known() {
+                        bail!(
+                            ConstructError::ModuleTriedToExpandOnNonSymbolicSizeAndBanned {
+                                old_size,
+                                new_size,
+                                dim,
+                                node_shape: node.shape().clone(),
+                                arg_spec: arg_spec.clone(),
+                                spec_circuit: self.circuit.clone()
+                            }
+                        );
                     }
                 }
             }
         }
+        Ok(())
+    }
+
+    pub fn expand_raw(&self, nodes: &Vec<CircuitRc>) -> Result<CircuitRc> {
+        self.check_expand_valid_shapes(nodes, false)?;
 
-        // TODO: maybe we should allow for inconsistent symbolic batch shapes?
         let aligned_batch_shape = self
             .aligned_batch_shape(nodes)
             .context("batch shapes didn't match for module")?;
@@ -279,7 +306,12 @@ impl ModuleSpec {
                 )
             })
             .collect();
-        let result = self.expand_raw(&symbols)?;
+        let result = self.expand_raw(&symbols).with_context(|| {
+            format!(
+                "module spec {:?} expansion failed for argument shapes {shapes:?}",
+                self.circuit.info().name
+            )
+        })?;
         let out_bytes: Vec<_> = symbols.into_iter().map(|x| x.info().hash).collect();
         MODULE_EXPANSIONS_SHAPE.with(|cache| {
             cache.borrow_mut().insert(
@@ -297,21 +329,23 @@ impl ModuleSpec {
             .iter()
             .map(|arg_spec| (arg_spec.symbol.ident(), &arg_spec.symbol))
             .collect();
-        let sym_single_map = self
-            .circuit
-            .as_symbol()
-            .map(|x| [x.clone()].into_iter().collect());
+        let sym_single_map = self.circuit.is_symbol().then(|| {
+            [self.circuit.clone()]
+                .into_iter()
+                .collect::<IndexSet<CircuitRc>>()
+        });
         let syms = sym_single_map
             .as_ref()
-            .unwrap_or(self.circuit.info().get_raw_free_symbols());
+            .unwrap_or_else(|| self.circuit.info().get_raw_free_symbols());
 
         for sym_here in syms {
-            if let Some(&my_sym) = my_idents.get(&sym_here.ident()) {
-                if my_sym != sym_here {
+            let sym_here_ = sym_here.as_symbol_unwrap();
+            if let Some(&my_sym) = my_idents.get(&sym_here_.ident()) {
+                if my_sym != sym_here_ {
                     bail!(
                         SubstitutionError::FoundNEQFreeSymbolWithSameIdentification {
                             sym: my_sym.clone(),
-                            matching_sym: sym_here.clone()
+                            matching_sym: sym_here_.clone()
                         }
                     )
                 }
@@ -505,7 +539,7 @@ impl ModuleSpec {
             let new = if let Some(m) = circ.as_module() {
                 let mut removed: BTreeSet<_> = symbols
                     .iter()
-                    .filter(|x| m.info().get_raw_free_symbols().contains(*x))
+                    .filter(|x| m.info().get_raw_free_symbols().contains(&x.crc()))
                     .cloned()
                     .collect();
                 for bound in &m.spec.arg_specs {
@@ -541,12 +575,12 @@ impl ModuleSpec {
 }
 
 pub enum FreeSymbolsWrapper<'a> {
-    Sym(IndexSet<Symbol>),
-    Other(&'a IndexSet<Symbol>),
+    Sym(IndexSet<CircuitRc>),
+    Other(&'a IndexSet<CircuitRc>),
 }
 
 impl<'a> Deref for FreeSymbolsWrapper<'a> {
-    type Target = IndexSet<Symbol>;
+    type Target = IndexSet<CircuitRc>;
     fn deref(&self) -> &Self::Target {
         match self {
             Self::Sym(x) => x,
@@ -555,19 +589,19 @@ impl<'a> Deref for FreeSymbolsWrapper<'a> {
     }
 }
 
-pub fn get_free_symbols(circuit: &Circuit) -> FreeSymbolsWrapper<'_> {
+pub fn get_free_symbols(circuit: &CircuitRc) -> FreeSymbolsWrapper<'_> {
     // returns syms in circuit visit order (order maybe matters here).
     // The fact the ordering matters here is why we use indexmap in free_symbols
-    if let Some(sym) = circuit.as_symbol() {
-        FreeSymbolsWrapper::Sym([sym.clone()].into_iter().collect())
+    if circuit.is_symbol() {
+        FreeSymbolsWrapper::Sym([circuit.clone()].into_iter().collect())
     } else {
-        FreeSymbolsWrapper::Other(&circuit.info().get_raw_free_symbols())
+        FreeSymbolsWrapper::Other(circuit.info().get_raw_free_symbols())
     }
 }
 
 #[pyfunction]
 #[pyo3(name = "get_free_symbols")]
-pub fn py_get_free_symbols(circuit: CircuitRc) -> Vec<Symbol> {
+pub fn py_get_free_symbols(circuit: CircuitRc) -> Vec<CircuitRc> {
     get_free_symbols(&circuit).iter().cloned().collect()
 }
 
@@ -646,7 +680,7 @@ impl ModuleSpec {
     pub fn new_free_symbols(circuit: CircuitRc, check_unique_arg_names: bool) -> Result<Self> {
         let arg_specs = get_free_symbols(&circuit)
             .iter()
-            .map(|x| ModuleArgSpec::new(x.clone(), true, true, true))
+            .map(|x| ModuleArgSpec::new(x.as_symbol_unwrap().clone(), true, true, true))
             .collect();
         let out =
             Self::new(circuit, arg_specs, true, false).expect("method should guarantee valid");
@@ -904,7 +938,7 @@ impl Module {
             format!("module construction failed because substitution isn't valid {MSG} (name={name:?})")
         })?;
         let mut substitute_cached_info = ModuleSpec {
-            // substituing all
+            // substituing all to avoid exponential worst case (and also to avoid exponential due to expand_node rerun)
             circuit: substitute_all_modules(spec.circuit.clone()),
             arg_specs: spec.arg_specs.clone(),
         }
@@ -912,19 +946,26 @@ impl Module {
         .with_context(|| format!("module construction failed on substitute name={name:?}"))?
         .info()
         .clone();
+        // need to check because we substituted all
+        spec.check_expand_valid_shapes(&nodes, true).with_context(|| format!("module construction name={name:?} failed due to nested module with stronger restrictions"))?;
 
         // the substituted free symbols is *nearly* right, but has a few issues:
         // - nodes might be unused which would result in missing their free symbols (and we want to include them)
-        // - If the spec.circuit is a single free symbol, it isn't counter
+        // - If the spec.circuit is a single free symbol, it isn't counted
         // - the ordering isn't ideal
         // Due to these issues we recompute the free symbols 'by hand'
         let mut free_syms: IndexSet<_> = get_free_symbols(&spec.circuit)
             .iter()
-            .filter(|x| !spec.arg_specs.iter().any(|arg_spec| &arg_spec.symbol == *x))
+            .filter(|x| {
+                !spec
+                    .arg_specs
+                    .iter()
+                    .any(|arg_spec| arg_spec.symbol == *x.as_symbol_unwrap())
+            })
             .cloned()
             .collect();
         for n in &nodes {
-            free_syms.extend(get_free_symbols(&**n).iter().cloned());
+            free_syms.extend(get_free_symbols(n).iter().cloned());
         }
         substitute_cached_info.free_symbols = Some(Arc::new(free_syms));
 
@@ -1004,12 +1045,12 @@ impl Module {
         let spec_resized = self
             .spec
             .resize(self.args().map(|x| x.info().shape.clone()).collect())?;
-        let spec_circuit_resized = substitute_all_modules(spec_resized.circuit.clone());
+        let spec_circuit_resized = substitute_all_modules(spec_resized.circuit);
         assert!(spec_circuit_resized.info().shape[..] == self.info().shape[..]);
         let stripped = deep_strip_axis_names(spec_circuit_resized, &None);
         let out_named = propagate_named_axes(
             stripped,
-            (0..self.info().rank())
+            (0..self.rank())
                 .map(|i| (i as u8, i.to_string().into()))
                 .collect(),
             true,
@@ -1018,7 +1059,7 @@ impl Module {
         let mut seen: Vec<bool> = vec![false; self.num_children() - 1];
         let mut axes_dead: Vec<bool> = vec![false; self.rank()];
         let mut result: Vec<Vec<Option<usize>>> =
-            self.args().map(|c| vec![None; c.info().rank()]).collect();
+            self.args().map(|c| vec![None; c.rank()]).collect();
 
         visit_circuit_unwrap(out_named, |x| {
             if let Some(sym) = x.as_symbol() {
@@ -1036,9 +1077,9 @@ impl Module {
                         for (k, v) in vec.iter().enumerate() {
                             if result[i][k] != *v {
                                 result[i][k] = None;
-                                (*v).or(result[i][k]).map(|ax| {
+                                if let Some(ax) = (*v).or(result[i][k]) {
                                     axes_dead[ax] = true;
-                                });
+                                }
                             }
                         }
                     }
@@ -1048,8 +1089,8 @@ impl Module {
                 for m in x.children() {
                     children_named_axes.extend(m.info().named_axes.values())
                 }
-                for (_, v) in &x.info().named_axes {
-                    if !children_named_axes.contains(&v) {
+                for v in x.info().named_axes.values() {
+                    if !children_named_axes.contains(v) {
                         axes_dead[v.parse::<usize>().unwrap()] = true;
                     }
                 }
@@ -1087,7 +1128,7 @@ impl CircuitNode for Module {
 
     fn child_axis_map(&self) -> Vec<Vec<Option<usize>>> {
         let map_inputs = self.child_axis_map_inputs().unwrap();
-        once(vec![None; self.spec.circuit.info().rank()])
+        once(vec![None; self.spec.circuit.rank()])
             .chain(map_inputs)
             .collect()
     }
@@ -1114,15 +1155,6 @@ impl CircuitNode for Module {
     fn num_free_children(&self) -> usize {
         1
     }
-
-    fn eval_tensors(
-        &self,
-        _tensors: &[rr_util::py_types::Tensor],
-    ) -> Result<rr_util::py_types::Tensor> {
-        bail!(TensorEvalError::ModulesCantBeDirectlyEvalutedInternal {
-            module: self.clone(),
-        })
-    }
 }
 
 impl CircuitNodeAutoName for Module {
@@ -1156,7 +1188,7 @@ impl Module {
         name: Option<Name>,
         kwargs: Option<HashMap<Name, CircuitRc>>,
     ) -> PyResult<PyClassInitializer<Module>> {
-        Ok(Module::new_kwargs(&kwargs.unwrap_or_else(HashMap::default), spec, name)?.into_init())
+        Ok(Module::new_kwargs(&kwargs.unwrap_or_default(), spec, name)?.into_init())
     }
 
     #[staticmethod]
@@ -1423,10 +1455,7 @@ pub fn inline_single_callsite_modules(circuit: CircuitRc) -> CircuitRc {
 pub fn get_children_with_symbolic_sizes(circuit: CircuitRc) -> HashSet<CircuitRc> {
     let mut circuits_with_symbolic_sizes = HashSet::default();
     visit_circuit_unwrap(circuit, |x| {
-        if x.shape()
-            .iter()
-            .any(|s| SymbolicSizeProduct::has_symbolic(*s))
-        {
+        if !shape_is_known(x.shape()) {
             circuits_with_symbolic_sizes.insert(x);
         }
     });
@@ -1438,29 +1467,27 @@ pub fn any_children_with_symbolic_sizes(circuit: CircuitRc) -> bool {
     !get_children_with_symbolic_sizes(circuit).is_empty()
 }
 
+pub fn any_modules(circuit: CircuitRc) -> bool {
+    let mut b = false;
+    visit_circuit_unwrap(circuit, |x| {
+        if x.is_module() {
+            b = true;
+        }
+    });
+    b
+}
+
 pub fn has_free_sym(circuit: &CircuitRc, sym: &Symbol) -> bool {
     if let Some(other) = circuit.as_symbol() {
         other == sym
     } else {
-        circuit.info().get_raw_free_symbols().contains(sym)
+        circuit.info().get_raw_free_symbols().contains(&sym.crc())
     }
 }
 
-pub fn is_intersecting_free_syms_iter<'a>(
-    circuit: &CircuitRc,
-    syms: impl IntoIterator<Item = &'a Symbol>,
-) -> bool {
-    if let Some(other) = circuit.as_symbol() {
-        syms.into_iter().any(|x| x == other)
-    } else {
-        syms.into_iter()
-            .any(|x| circuit.info().get_raw_free_symbols().contains(x))
-    }
-}
-
-pub fn is_intersecting_free_syms(circuit: &CircuitRc, syms: &IndexSet<Symbol>) -> bool {
-    if let Some(other) = circuit.as_symbol() {
-        syms.contains(other)
+pub fn is_intersecting_free_syms(circuit: &CircuitRc, syms: &IndexSet<CircuitRc>) -> bool {
+    if circuit.is_symbol() {
+        syms.contains(circuit)
     } else {
         !circuit.info().get_raw_free_symbols().is_disjoint(syms)
     }
@@ -1513,7 +1540,7 @@ fn recur_check_no_free_bound_by_nested(
                     .arg_specs
                     .iter()
                     .all(|arg_spec| &arg_spec.symbol != **s)
-                    && has_free_sym(&m.spec.circuit, *s)
+                    && has_free_sym(&m.spec.circuit, s)
             })
             .unzip();
 
@@ -1555,6 +1582,53 @@ fn recur_check_no_free_bound_by_nested(
     Ok(())
 }
 
+fn get_symbol_restrictions(circ: &Circuit) -> HashMap<Symbol, (bool, bool, bool)> {
+    if circ.num_children() == 0 {
+        return HashMap::default();
+    }
+    if let Some(r) = SYMBOL_RESTRICTIONS.with(|c| c.borrow().get(&circ.info().hash).cloned()) {
+        return r;
+    }
+
+    let merge = |out: &mut HashMap<_, _>, s: Symbol, bs| {
+        let (a, b, c) = bs;
+        if let Some(&(d, e, f)) = out.get(&s) {
+            out.insert(s, (a && d, b && e, c || f));
+        } else {
+            out.insert(s, bs);
+        }
+    };
+
+    let mut out = HashMap::default();
+    for c in circ.children_sl() {
+        for (s, bs) in get_symbol_restrictions(c) {
+            merge(&mut out, s, bs);
+        }
+    }
+    if let Some(m) = circ.as_module() {
+        for (c, arg_spec) in zip(m.args(), &m.spec.arg_specs) {
+            if !arg_spec.batchable || !arg_spec.expandable || arg_spec.ban_non_symbolic_size_expand
+            {
+                for sym in (*get_free_symbols(c)).iter() {
+                    merge(
+                        &mut out,
+                        sym.as_symbol_unwrap().clone(),
+                        (
+                            arg_spec.batchable,
+                            arg_spec.expandable,
+                            arg_spec.ban_non_symbolic_size_expand,
+                        ),
+                    )
+                }
+            }
+        }
+    }
+
+    SYMBOL_RESTRICTIONS.with(|c| c.borrow_mut().insert(circ.info().hash, out.clone()));
+
+    out
+}
+
 const SAME_IDENT_P1: &str = concat!(
     "Substituting for a symbol when there is a not-equal free symbol",
     " with the same name and uuid is not allowed (aka, a near miss)."
@@ -1602,6 +1676,7 @@ thread_local! {
         HashSet<(ModuleSpecHashable, Vec<HashBytes>)>,
     > = RefCell::new(HashSet::default());
     static ALL_MODULE_SUBSTITUTOR: RefCell<AllModuleSubstituter> = RefCell::new(Default::default());
+    static SYMBOL_RESTRICTIONS: RefCell<HashMap<HashBytes, HashMap<Symbol, (bool, bool, bool)>>> = RefCell::new(Default::default());
 }
 
 #[pyfunction]
