@@ -59,7 +59,7 @@ pub fn named_axes_backward<T: CircuitNode + Clone>(
                     ax.and_then(|i| {
                         named_axes
                             .get(&(i as u8))
-                            .map(|name| (child_i as u8, name.clone()))
+                            .map(|name| (child_i as u8, *name))
                     })
                 })
                 .collect()
@@ -79,7 +79,7 @@ pub fn propagate_named_axes(
             let child_axis_names: Vec<NamedAxes> = if abort_on_branch
                 && circuit
                     .as_einsum()
-                    .map(|x| x.args().any(|(_, i)| !is_unique(&i)))
+                    .map(|x| x.args().any(|(_, i)| !is_unique(i)))
                     .unwrap_or(false)
             {
                 circuit
@@ -198,7 +198,7 @@ pub fn get_axis_leaves(
         &mut result_no_axis,
         &mut seen,
         name,
-        circ_named_axes.clone(),
+        circ_named_axes,
     )?;
     Ok((name, circ_new, result_axis, result_no_axis))
 }
