@@ -141,21 +141,20 @@ pub trait HasTerm {
 impl<Matcher: Clone + HasTerm> UpdatedIterativeMatcher<Matcher> {
     // make this pyfunction as needed
     pub fn per_child(self, num_children: usize) -> Vec<Option<Matcher>> {
-        let out = match self {
+        match self {
             Self::Single(Update(item)) => vec![item; num_children],
             Self::Many(items) => {
                 // should already have been checked!
                 assert_eq!(items.len(), num_children);
                 items.into_iter().map(|x| x.0).collect::<Vec<_>>()
             }
-        };
-        out
+        }
     }
 
     pub fn per_child_with_term(self, num_children: usize) -> Vec<Matcher> {
         self.per_child(num_children)
             .into_iter()
-            .map(|x| x.unwrap_or_else(|| Matcher::term().into()))
+            .map(|x| x.unwrap_or_else(|| Matcher::term()))
             .collect()
     }
 
