@@ -3,7 +3,11 @@ use std::collections::BTreeMap;
 use anyhow::{bail, Result};
 use macro_rules_attribute::apply;
 use pyo3::prelude::*;
-use rr_util::{name::Name, py_types::Tensor, sv, tensor_util::Shape};
+use rr_util::{
+    name::Name,
+    shape::{shape_eq_if_known, Shape, Size},
+    sv,
+};
 use uuid::Uuid;
 
 use crate::{
@@ -11,7 +15,7 @@ use crate::{
     circuit_node_auto_impl, circuit_node_extra_impl,
     circuit_node_private::{CircuitNodeComputeInfoImpl, CircuitNodeHashItems},
     new_rc, new_rc_unwrap, CachedCircuitInfo, CircuitFlags, CircuitNode, CircuitNodeAutoName,
-    CircuitRc, ConstructError, PyCircuitBase, Scalar, TensorEvalError,
+    CircuitRc, ConstructError, PyCircuitBase, Scalar,
 };
 
 /// Tags a Circuit with a UUID. Use this to make two otherwise equal nodes distinct.
@@ -34,11 +38,12 @@ pub struct Tag {
 impl Tag {
     #[apply(new_rc)]
     pub fn new(node: CircuitRc, uuid: Uuid, name: Option<Name>) -> (Self) {
-        let out = Self {
+        Self {
             uuid,
-            info: CachedCircuitInfo::incomplete(name, node.shape().clone(), vec![node]),
-        };
-        out.initial_init_info().unwrap()
+            info: CachedCircuitInfo::incomplete(name, node.shape().clone(), vec![node]).unwrap(),
+        }
+        .initial_init_info()
+        .unwrap()
     }
 }
 
@@ -46,7 +51,7 @@ impl CircuitNodeAutoName for Tag {
     const PRIORITY: OperatorPriority = OperatorPriority::PostFix {};
 
     fn auto_name(&self) -> Option<Name> {
-        child_name_with_maybe_paren(&Self::PRIORITY, self.node().clone())
+        child_name_with_maybe_paren(&Self::PRIORITY, self.node())
             .map(|n| (n.string() + " tag").into())
     }
 }
@@ -65,17 +70,12 @@ impl CircuitNode for Tag {
     circuit_node_auto_impl!("63fdc4ce-2f1b-40b3-b8b6-13991b54cbd7");
 
     fn child_axis_map(&self) -> Vec<Vec<Option<usize>>> {
-        vec![(0..self.node().info().rank()).map(Some).collect()]
+        vec![(0..self.node().rank()).map(Some).collect()]
     }
 
     fn _replace_children(&self, children: Vec<CircuitRc>) -> Result<Self> {
         Ok(Self::new(children[0].clone(), self.uuid, self.info().name))
     }
-
-    fn eval_tensors(&self, tensors: &[Tensor]) -> Result<Tensor> {
-        assert_eq!(tensors.len(), 1);
-        Ok(tensors[0].clone())
-    }
 }
 
 #[pymethods]
@@ -115,29 +115,29 @@ impl DiscreteVar {
         probs_and_group: CircuitRc,
         name: Option<Name>,
     ) -> Result<Self> {
-        if probs_and_group.info().rank() != 1 {
+        if probs_and_group.rank() != 1 {
             bail!(ConstructError::DiscreteVarProbsMustBe1d {
                 shape: probs_and_group.info().shape.clone(),
             });
         }
-        if values.info().rank() < 1 {
+        if values.rank() < 1 {
             bail!(ConstructError::DiscreteVarNoSamplesDim {});
         }
-        if values.info().shape[0] != probs_and_group.info().shape[0] {
+        if !values.info().shape[0].eq_if_known(probs_and_group.info().shape[0]) {
             bail!(ConstructError::DiscreteVarWrongSamplesDim {
                 node: values.info().shape[0],
                 probs: probs_and_group.info().shape[0],
             });
         }
 
-        let out = Self {
+        Self {
             info: CachedCircuitInfo::incomplete(
                 name,
                 values.info().shape[1..].iter().cloned().collect(),
                 vec![values, probs_and_group],
-            ),
-        };
-        out.initial_init_info()
+            )?,
+        }
+        .initial_init_info()
     }
 
     pub fn new_for_py(
@@ -145,11 +145,15 @@ impl DiscreteVar {
         probs_and_group: Option<CircuitRc>,
         name: Option<Name>,
     ) -> Result<Self> {
-        if values.info().rank() < 1 {
+        if values.rank() < 1 {
             bail!(ConstructError::DiscreteVarNoSamplesDim {});
         }
+        let size = values.shape()[0];
+        if !size.is_known() {
+            bail!(ConstructError::DiscreteVarSamplesDimUnknownSize {});
+        }
         let probs_and_group = probs_and_group
-            .unwrap_or_else(|| Self::uniform_probs_and_group(values.info().shape[0], None).rc()); // TODO: name
+            .unwrap_or_else(|| Self::uniform_probs_and_group(size.unwrap(), None).rc()); // TODO: name
         Self::try_new(values, probs_and_group, name)
     }
 }
@@ -182,7 +186,7 @@ impl CircuitNode for DiscreteVar {
     fn child_axis_map(&self) -> Vec<Vec<Option<usize>>> {
         vec![
             std::iter::once(None)
-                .chain((0..self.info().rank()).map(Some))
+                .chain((0..self.rank()).map(Some))
                 .collect(),
             vec![None],
         ]
@@ -191,13 +195,6 @@ impl CircuitNode for DiscreteVar {
     fn _replace_children(&self, children: Vec<CircuitRc>) -> Result<Self> {
         Self::try_new(children[0].clone(), children[1].clone(), self.info().name)
     }
-
-    fn eval_tensors(&self, _tensors: &[Tensor]) -> Result<Tensor> {
-        Err(TensorEvalError::NotExplicitlyComputableInternal {
-            circuit: self.crc(),
-        }
-        .into())
-    }
 }
 
 #[pymethods]
@@ -218,7 +215,10 @@ impl DiscreteVar {
     #[staticmethod]
     pub fn uniform_probs_and_group(size: usize, name: Option<Name>) -> Tag {
         // TODO: naming
-        Tag::new_with_random_uuid(Scalar::nrc(1.0 / (size as f64), sv![size], None), name)
+        Tag::new_with_random_uuid(
+            Scalar::nrc(1.0 / (size as f64), sv![Size::known(size)], None),
+            name,
+        )
     }
     #[getter]
     pub fn values(&self) -> &CircuitRc {
@@ -260,7 +260,7 @@ impl StoredCumulantVar {
                 .take(k * shape.len())
                 .copied()
                 .collect();
-            if shape_here != v.info().shape {
+            if !shape_eq_if_known(&shape_here, v.shape()) {
                 bail!(ConstructError::StoredCumulantVarCumulantWrongShape {
                     cumulant_shape: v.info().shape.clone(),
                     cumulant_number: *k,
@@ -269,13 +269,13 @@ impl StoredCumulantVar {
             }
         }
         let out = Self {
-            cumulant_ixs: cumulants.iter().map(|(a, _)| *a).collect(),
+            cumulant_ixs: cumulants.keys().copied().collect(),
             uuid,
             info: CachedCircuitInfo::incomplete(
                 name,
                 cumulants[&1].info().shape.clone(),
-                cumulants.iter().map(|(_, b)| b).cloned().collect(),
-            ),
+                cumulants.values().cloned().collect(),
+            )?,
         };
         out.initial_init_info()
     }
@@ -290,11 +290,7 @@ impl StoredCumulantVar {
         let mut higher_cumulants = higher_cumulants;
         higher_cumulants.insert(1, mean);
         higher_cumulants.insert(2, variance);
-        Self::try_new(
-            higher_cumulants,
-            uuid.unwrap_or_else(|| Uuid::new_v4()),
-            name,
-        )
+        Self::try_new(higher_cumulants, uuid.unwrap_or_else(Uuid::new_v4), name)
     }
 }
 
@@ -351,9 +347,7 @@ impl CircuitNode for StoredCumulantVar {
     circuit_node_auto_impl!("f36da959-d160-484d-b6b8-7685ef7521c0");
 
     fn child_axis_map(&self) -> Vec<Vec<Option<usize>>> {
-        self.children()
-            .map(|x| vec![None; x.info().rank()])
-            .collect()
+        self.children().map(|x| vec![None; x.rank()]).collect()
     }
 
     fn _replace_children(&self, children: Vec<CircuitRc>) -> Result<Self> {
@@ -363,13 +357,6 @@ impl CircuitNode for StoredCumulantVar {
             self.info().name,
         )
     }
-
-    fn eval_tensors(&self, _tensors: &[Tensor]) -> Result<Tensor> {
-        Err(TensorEvalError::NotExplicitlyComputableInternal {
-            circuit: self.crc(),
-        }
-        .into())
-    }
 }
 
 #[pymethods]
@@ -380,7 +367,7 @@ impl StoredCumulantVar {
         uuid: Option<Uuid>,
         name: Option<Name>,
     ) -> PyResult<PyClassInitializer<StoredCumulantVar>> {
-        let uuid = uuid.unwrap_or_else(|| Uuid::new_v4());
+        let uuid = uuid.unwrap_or_else(Uuid::new_v4);
         let out = Self::try_new(cumulants, uuid, name)?;
         Ok(out.into_init())
     }
