@@ -19,11 +19,11 @@ use rr_util::{
     name::Name,
     opt_einsum::optimize_einsum_spec_cached,
     python_error_exception,
-    rearrange_spec::{shape_to_op_shape, OpSize, RInts, RearrangeSpec},
+    rearrange_spec::{RInts, RearrangeSpec},
+    shape::{shape_as_known, shape_is_known, Shape, Size},
     sv,
-    symbolic_size::SymbolicSizeProduct,
     tensor_util::{
-        check_canon_idxs, compose, IndexTensor, Shape, Slice, TensorAxisIndex, TensorIndex, USlice,
+        check_canon_idxs, compose, IndexTensor, Slice, TensorAxisIndex, TensorIndex, USlice,
     },
     tu8v,
     union_find::UnionFind,
@@ -74,7 +74,9 @@ pub fn add_collapse_scalar_inputs(add: &Add) -> Option<Add> {
         let new_scalar: f64 = scalar_inputs.iter().map(|node| node.value).sum();
         let mut new_inputs: Vec<CircuitRc> =
             non_scalar_inputs.iter().map(|x| x.clone().into()).collect();
-        let new_scalar_input = Scalar::new(new_scalar, add.info().shape.clone(), None).rc();
+        let new_scalar_input = Scalar::try_new(new_scalar, add.info().shape.clone(), None)
+            .ok()?
+            .rc();
         new_inputs.push(new_scalar_input);
         Some(Add::new(new_inputs, add.info().name))
     }
@@ -442,14 +444,14 @@ pub fn make_broadcast_py(node: CircuitRc, out_shape: Shape) -> Option<CircuitRc>
 }
 
 pub fn make_broadcast(node: &CircuitRc, out_shape: &Shape) -> Result<CircuitRc> {
-    let rank_dif = out_shape.len() - node.info().rank();
+    let rank_dif = out_shape.len() - node.rank();
     let input_ints = node
         .info()
         .shape
         .iter()
         .enumerate()
         .map(|(i, x)| {
-            if *x == 1 {
+            if x.is(1) {
                 tu8v![]
             } else {
                 tu8v![(i + rank_dif) as u8]
@@ -460,7 +462,7 @@ pub fn make_broadcast(node: &CircuitRc, out_shape: &Shape) -> Result<CircuitRc>
     let int_sizes = out_shape.clone();
     Ok(Rearrange::nrc_elim_identity(
         node.clone(),
-        RearrangeSpec::new_canon(input_ints, output_ints, shape_to_op_shape(&int_sizes)),
+        RearrangeSpec::new_canon(input_ints, output_ints, int_sizes),
         None,
     ))
 }
@@ -490,7 +492,7 @@ pub fn index_merge_scalar(index: &Index) -> Option<CircuitRc> {
 #[pyfunction]
 pub fn index_elim_identity(index: &Index) -> Option<CircuitRc> {
     if index.index.is_identity(&index.node().info().shape) {
-        Some(index.node().clone())
+        Some(index.node())
     } else {
         None
     }
@@ -498,9 +500,9 @@ pub fn index_elim_identity(index: &Index) -> Option<CircuitRc> {
 
 #[pyfunction]
 pub fn index_fuse(index: &Index) -> Option<Index> {
-    index.node().as_index().and_then(|inner| {
-        Some(Index::new(
-            inner.node().clone(),
+    index.node().as_index().map(|inner| {
+        Index::new(
+            inner.node(),
             compose(
                 &index.index,
                 &inner.index.canonicalize(inner.node().shape()),
@@ -513,7 +515,7 @@ pub fn index_fuse(index: &Index) -> Option<Index> {
                     n
                 }
             }),
-        ))
+        )
     })
 }
 
@@ -532,7 +534,7 @@ pub fn rearrange_merge_scalar(rearrange: &Rearrange) -> Option<CircuitRc> {
 #[pyfunction]
 pub fn rearrange_elim_identity(rearrange: &Rearrange) -> Option<CircuitRc> {
     if rearrange.spec.is_identity() {
-        Some(rearrange.node().clone())
+        Some(rearrange.node())
     } else {
         None
     }
@@ -573,7 +575,7 @@ pub fn concat_elim_identity(concat: &Concat) -> Option<CircuitRc> {
 pub fn get_removable_axes(circuit: &CircuitRc) -> HashSet<usize> {
     // plausibly this should have a flag for conforming to input shape
     match &***circuit {
-        Circuit::Scalar(_scalar) => (0..circuit.info().rank()).collect(),
+        Circuit::Scalar(_scalar) => (0..circuit.rank()).collect(),
         Circuit::Rearrange(rearrange) => rearrange
             .spec
             .out_broadcast_axes()
@@ -622,7 +624,7 @@ pub fn remove_axes_gen(
                 (*rearrange.spec).clone()
             };
             Some(Rearrange::nrc_elim_identity(
-                rearrange.node().clone(),
+                rearrange.node(),
                 spec.filter_out_axes(axes).unwrap(),
                 get_name(rearrange),
             ))
@@ -633,12 +635,12 @@ pub fn remove_axes_gen(
                 .shape
                 .iter()
                 .enumerate()
-                .all(|(i, l)| !axes.contains(&i) || *l == 1)
+                .all(|(i, l)| !axes.contains(&i) || l.is(1))
             {
                 Some(Rearrange::nrc_elim_identity(
                     circuit.clone(),
                     // wow this chained construction is long, should change at some point
-                    RearrangeSpec::ident_usize(circuit.info().rank())
+                    RearrangeSpec::ident_usize(circuit.rank())
                         .unwrap()
                         .conform_to_input_shape(&circuit.info().shape)
                         .unwrap()
@@ -662,7 +664,7 @@ pub fn generalfunction_pull_removable_axes(node: &GeneralFunction) -> Option<Cir
     let removable_axes_inp = get_removable_axes(&node.children_sl()[0]);
     let removable_batchable_inp: HashSet<usize> = removable_axes_inp
         .iter()
-        .filter(|x| (**x + node.num_non_batchable_output_dims as usize) < node.info().rank())
+        .filter(|x| (**x + node.num_non_batchable_output_dims as usize) < node.rank())
         .cloned()
         .collect();
     if removable_batchable_inp.is_empty() {
@@ -740,7 +742,12 @@ pub fn einsum_pull_removable_axes(einsum: &Einsum) -> Option<CircuitRc> {
         .collect();
 
     // multiply by product of removed reduced axes
-    let shape_map = einsum.shape_map().unwrap();
+    let shape_map = einsum
+        .shape_map()
+        .unwrap()
+        .into_iter()
+        .map(|(k, x)| Some((k, x.t()?)))
+        .collect::<Option<HashMap<_, _>>>()?;
     let reduced_axes = einsum.reduced_axes();
     let scalar_mul: f64 = reduced_axes
         .into_iter()
@@ -804,7 +811,7 @@ pub fn add_pull_removable_axes(add: &Add, remove_non_common_axes: bool) -> Optio
                 && !removed_any_non_one_axes
                 && pre_broadcast_with_ones
                     .iter()
-                    .any(|i| x.info().shape[*i] != 1)
+                    .any(|i| !x.info().shape[*i].is(1))
             {
                 removed_any_non_one_axes = true;
             }
@@ -823,16 +830,16 @@ pub fn add_pull_removable_axes(add: &Add, remove_non_common_axes: bool) -> Optio
         return None;
     }
     let post_rearrange_spec = RearrangeSpec::new_canon(
-        (0..add.info().rank())
+        (0..add.rank())
             .filter(|x| !intersection.contains(x))
             .map(|x| tu8v![x as u8])
             .collect(),
-        (0..add.info().rank()).map(|x| tu8v![x as u8]).collect(),
-        shape_to_op_shape(&add.info().shape),
+        (0..add.rank()).map(|x| tu8v![x as u8]).collect(),
+        add.shape().clone(),
     );
     let new_operands = zip(add.nodes_and_rank_differences(), removable_axes)
         .map(|((node, _rank_dif), removable_here_base)| {
-            let rank_difference = add.info().rank() - node.info().rank();
+            let rank_difference = add.rank() - node.info().rank();
             if !remove_non_common_axes {
                 let removable_common_in_rank: HashSet<usize> = intersection
                     .iter()
@@ -851,7 +858,7 @@ pub fn add_pull_removable_axes(add: &Add, remove_non_common_axes: bool) -> Optio
                 let mut output_ints: RInts = sv![];
                 let mut count = 0;
                 let mut started = false;
-                for i in 0..node.info().rank() {
+                for i in 0..node.rank() {
                     if removable_here.contains(&i) {
                         if started && !intersection.contains(&(i + rank_difference)) {
                             output_ints.push(tu8v![]);
@@ -864,11 +871,9 @@ pub fn add_pull_removable_axes(add: &Add, remove_non_common_axes: bool) -> Optio
                 }
 
                 let broadcast_rspec = RearrangeSpec::new_canon(
-                    (0..raw_removed.info().rank())
-                        .map(|i| tu8v![i as u8])
-                        .collect(),
+                    (0..raw_removed.rank()).map(|i| tu8v![i as u8]).collect(),
                     output_ints,
-                    shape_to_op_shape(&raw_removed.info().shape),
+                    raw_removed.shape().clone(),
                 );
                 Rearrange::nrc_elim_identity(raw_removed, broadcast_rspec, None)
             }
@@ -892,7 +897,7 @@ pub fn add_make_broadcasts_explicit(add: &Add) -> Option<Add> {
                 did_anything = true;
                 make_broadcast(&node, &add.info().shape).unwrap()
             } else {
-                node.clone()
+                node
             }
         })
         .collect();
@@ -948,7 +953,7 @@ pub fn distribute_once_raw(
                     new_args[operand_idx].0 = node.clone();
                     on_sub(
                         i,
-                        node.clone(),
+                        node,
                         Einsum::new(new_args, einsum.out_axes.clone(), None),
                     )
                 })
@@ -1029,7 +1034,7 @@ pub fn einsum_of_permute_merge(einsum: &Einsum) -> Option<Einsum> {
                     rearrange.spec.get_fwd_permutation().map(|permutation| {
                         did_anything = true;
                         (
-                            rearrange.node().clone(),
+                            rearrange.node(),
                             permutation.iter().map(|i| ints[*i]).collect(),
                         )
                     })
@@ -1069,11 +1074,7 @@ pub fn einsum_elim_zero(einsum: &Einsum) -> Option<Scalar> {
         Circuit::Scalar(scalar) => scalar.value == 0.0,
         _ => false,
     }) {
-        Some(Scalar::new(
-            0.0,
-            einsum.info().shape.clone(),
-            einsum.info().name,
-        ))
+        Some(Scalar::try_new(0.0, einsum.info().shape.clone(), einsum.info().name).ok()?)
     } else {
         None
     }
@@ -1152,8 +1153,8 @@ pub fn einsum_concat_to_add(einsum: &Einsum) -> Option<Add> {
                     return None;
                 }
                 // This transformation only works if we're summing across the concat dimension.
-                if !einsum.out_axes.contains(&(concat_dim as u8)) {
-                    return Some((concat_idx.clone(), circ));
+                if !einsum.out_axes.contains(&{ concat_dim }) {
+                    return Some((concat_idx, circ));
                 }
             }
             None
@@ -1172,7 +1173,7 @@ pub fn einsum_concat_to_add(einsum: &Einsum) -> Option<Add> {
             let mut new_args = einsum.args_cloned();
             // Replace the concatted `a` dimension with the partial `a` we're mapping over, then
             // leave everything else in the output alone.
-            new_args[*concat_idx] = (node.clone(), einsum.in_axes[*concat_idx].clone());
+            new_args[*concat_idx] = (node, einsum.in_axes[*concat_idx].clone());
             Einsum::nrc(new_args, einsum.out_axes.clone(), None)
         })
         .collect();
@@ -1183,7 +1184,7 @@ pub fn einsum_concat_to_add(einsum: &Einsum) -> Option<Add> {
 pub fn index_split_axes(node: &Index, top_axes: HashSet<usize>) -> Option<Index> {
     let mut bottom: Vec<TensorAxisIndex> = vec![];
     let mut top: Vec<TensorAxisIndex> = vec![];
-    if top_axes.iter().any(|i| *i >= node.node().info().rank()) {
+    if top_axes.iter().any(|i| *i >= node.node().rank()) {
         return None;
     }
     for (i, idx) in node.index.0.iter().enumerate() {
@@ -1225,7 +1226,7 @@ pub fn rearrange_fuse(node: &Rearrange) -> Option<Rearrange> {
 
 pub fn axis_index_split_sections_for_concat(
     axis_index: &TensorAxisIndex,
-    sections: &Vec<usize>,
+    sections: &[usize],
 ) -> Result<Vec<Option<TensorAxisIndex>>> {
     let starts = cumsum(sections);
     let len = sections.iter().sum();
@@ -1266,7 +1267,7 @@ pub fn axis_index_split_sections_for_concat(
                     bail!(PushDownIndexError::ZeroLengthIndexAmbiguous {})
                 }
             }
-            if sections.len() == 1 && result[0] == None {
+            if sections.len() == 1 && result[0].is_none() {
                 result[0] = Some(TensorAxisIndex::Slice(uslice.into()))
             }
             result
@@ -1319,7 +1320,7 @@ pub fn push_down_index_raw(
             - (0..rank_to_pass)
                 .filter(|i| matches!(node.index.0[*i], TensorAxisIndex::Single(_)))
                 .count();
-        let top_axis_indices: Vec<_> = node.index.0[rank_to_pass..].iter().cloned().collect();
+        let top_axis_indices: Vec<_> = node.index.0[rank_to_pass..].to_vec();
         if top_axis_indices
             .iter()
             .zip(&new_circ.shape()[passed_rank_now..])
@@ -1350,7 +1351,7 @@ pub fn push_down_index_raw(
                 .iter()
                 .enumerate()
                 .map(|(i, (operand, rank_difference))| {
-                    let index_here = (0..operand.info().rank())
+                    let index_here = (0..operand.rank())
                         .map(|i| {
                             let idx_here = node.index.0[i + rank_difference].clone();
                             if operand.info().shape[i] == inner.info().shape[i + rank_difference] {
@@ -1383,10 +1384,18 @@ pub fn push_down_index_raw(
                     })
                     .collect(),
             );
-            let sections = &inner.get_sizes_at_axis();
+            let sections = inner.get_sizes_at_axis();
+            if !shape_is_known(&sections) {
+                bail!(PushDownIndexError::ConcatUnknownSizeAtAxis {
+                    sizes_at_axis: inner.get_sizes_at_axis().into(),
+                    index: node.clone()
+                });
+            }
             let old_index_on_concat_axis = node.index.0[inner.axis].clone();
-            let concat_axis_indices_maybe =
-                axis_index_split_sections_for_concat(&old_index_on_concat_axis, sections);
+            let concat_axis_indices_maybe = axis_index_split_sections_for_concat(
+                &old_index_on_concat_axis,
+                shape_as_known(&sections),
+            );
             let concat_axis_indices_maybe = if allow_partial_pushdown {
                 if let Err(Some(PushDownIndexError::ConcatSplitSectionsTensorUnsupported {
                     ..
@@ -1461,14 +1470,13 @@ pub fn push_down_index_raw(
                 let final_index = TensorIndex::new_single(
                     node.index.0[inner.axis].clone(),
                     new_axis,
-                    new_concat.info().rank(),
+                    new_concat.rank(),
                 );
                 Index::nrc(new_concat, final_index, idx_partial_name())
             }
         }
         Circuit::GeneralFunction(inner) => {
-            if !inner.is_batchable()
-                || inner.num_non_batchable_output_dims as usize == inner.info().rank()
+            if !inner.is_batchable() || inner.num_non_batchable_output_dims as usize == inner.rank()
             {
                 bail!(PushDownIndexError::NoopOnGeneralFunction {
                     index_node: node.clone(),
@@ -1477,13 +1485,12 @@ pub fn push_down_index_raw(
                     inner_rank: inner.rank()
                 })
             } else {
-                let rank_to_pass =
-                    inner.info().rank() - inner.num_non_batchable_output_dims as usize;
+                let rank_to_pass = inner.rank() - inner.num_non_batchable_output_dims as usize;
                 let new_operands = zip(inner.children(), &inner.input_batchability)
                     .enumerate()
                     .map(|(i, (child, &batchable))| {
                         if batchable {
-                            let ident_rank = child.info().rank() - rank_to_pass;
+                            let ident_rank = child.rank() - rank_to_pass;
                             let index_passed = TensorIndex(
                                 node.index.0[..rank_to_pass]
                                     .iter()
@@ -1497,8 +1504,7 @@ pub fn push_down_index_raw(
                         }
                     })
                     .collect::<Result<_>>()?;
-                let new_gf =
-                    GeneralFunction::nrc(new_operands, inner.spec.clone(), out_name.clone());
+                let new_gf = GeneralFunction::nrc(new_operands, inner.spec.clone(), out_name);
 
                 handle_non_batchable_remaining(rank_to_pass, new_gf).map_err(
                     |(_, top_axis_indices)| {
@@ -1562,7 +1568,7 @@ pub fn push_down_index_raw(
                     }
                     Ok((
                         on_sub(i, node, index_here, None)?,
-                        ints_filter_index(&ints), // todo make function work with sv
+                        ints_filter_index(ints), // todo make function work with sv
                     ))
                 })
                 .collect::<Result<_>>()?;
@@ -1571,11 +1577,8 @@ pub fn push_down_index_raw(
                     index_node: node.clone()
                 });
             }
-            let new_einsum = Einsum::nrc(
-                new_operands,
-                ints_filter_index(&inner.out_axes),
-                out_name.clone(),
-            );
+            let new_einsum =
+                Einsum::nrc(new_operands, ints_filter_index(&inner.out_axes), out_name);
             let out_index = TensorIndex(
                 zip(&inner.out_axes, &node.index.0)
                     .filter_map(|(i, idx)| match int_indices.get(i).unwrap() {
@@ -1605,8 +1608,8 @@ pub fn push_down_index_raw(
                 .conform_to_input_shape_spec()
                 .canonicalize(true)
                 .index_before(
-                    node.index.canonicalize(&node.node().shape()),
-                    node.info().device_dtype.clone().unwrap_or_defaults().device,
+                    node.index.canonicalize(node.node().shape()),
+                    node.info().device_dtype.unwrap_or_defaults().device,
                 ) {
                 x
             } else {
@@ -1615,7 +1618,7 @@ pub fn push_down_index_raw(
                     spec: (*inner.spec).clone()
                 })
             };
-            let new_index = on_sub(0, inner.node().clone(), index, None)?;
+            let new_index = on_sub(0, inner.node(), index, None)?;
             Rearrange::nrc_elim_identity(new_index, spec, out_name)
         }
         Circuit::Scatter(inner) => {
@@ -1646,7 +1649,7 @@ pub fn push_down_index_raw(
                     index_node: node.clone()
                 });
             }
-            let inner_node = on_sub(0, inner.node().clone(), TensorIndex(passed_index), None)?;
+            let inner_node = on_sub(0, inner.node(), TensorIndex(passed_index), None)?;
 
             let new_scatter = Scatter::nrc(
                 inner_node,
@@ -1669,7 +1672,7 @@ pub fn push_down_index_raw(
             }
         }
         Circuit::Tag(tag) => Tag::nrc(
-            on_sub(0, tag.node().clone(), node.index.clone(), None)?,
+            on_sub(0, tag.node(), node.index.clone(), None)?,
             tag.uuid,
             out_name,
         ),
@@ -1709,10 +1712,10 @@ pub fn push_down_index_raw(
                             let mut ix = vec![TensorAxisIndex::IDENT; arg_spec.symbol.rank()];
                             let this_num_batch = input_node.rank() - arg_spec.symbol.rank();
                             m2.iter().enumerate().for_each(|(child_axis, p)| {
-                                if let Some(parent_axis) = p.clone() &&
+                                if let Some(parent_axis) = *p &&
                                         parent_axis >= rank_to_pass &&
                                         arg_spec.expandable &&
-                                        (SymbolicSizeProduct::has_symbolic(arg_spec.symbol.shape()[child_axis - this_num_batch])
+                                        (!arg_spec.symbol.shape()[child_axis - this_num_batch].is_known()
                                          || !arg_spec.ban_non_symbolic_size_expand)
                                         {
                                         let parent_axis_c = parent_axis - rank_to_pass;
@@ -1756,41 +1759,39 @@ pub fn push_down_index_raw(
 
             if idxnb.is_identity(&m.shape()[rank_to_pass..]) {
                 Module::nrc(new_nodes, m.spec.clone(), out_name)
+            } else if into_module_spec {
+                Module::nrc(
+                    new_nodes,
+                    ModuleSpec::new(
+                        on_sub(0, m.spec.circuit.clone(), idxnb, None)?,
+                        m.spec.arg_specs.clone(),
+                        false,
+                        false,
+                    )?,
+                    out_name,
+                )
             } else {
-                if into_module_spec {
-                    Module::nrc(
-                        new_nodes,
-                        ModuleSpec::new(
-                            on_sub(0, m.spec.circuit.clone(), idxnb, None)?,
-                            m.spec.arg_specs.clone(),
-                            false,
-                            false,
-                        )?,
-                        out_name,
-                    )
+                let rank_done_now = rank_to_pass
+                    - (0..rank_to_pass)
+                        .filter(|i| matches!(node.index.0[*i], TensorAxisIndex::Single(_)))
+                        .count();
+                let remaining_index = TensorIndex(
+                    [&vec![TensorAxisIndex::IDENT; rank_done_now], &idxnb.0[..]].concat(),
+                );
+                let index = Index::new(
+                    Module::nrc(new_nodes, m.spec.clone(), out_name),
+                    remaining_index,
+                    idx_partial_name(),
+                );
+                if allow_partial_pushdown {
+                    index.rc()
                 } else {
-                    let rank_done_now = rank_to_pass
-                        - (0..rank_to_pass)
-                            .filter(|i| matches!(node.index.0[*i], TensorAxisIndex::Single(_)))
-                            .count();
-                    let remaining_index = TensorIndex(
-                        [&vec![TensorAxisIndex::IDENT; rank_done_now], &idxnb.0[..]].concat(),
-                    );
-                    let index = Index::new(
-                        Module::nrc(new_nodes, m.spec.clone(), out_name),
-                        remaining_index,
-                        idx_partial_name(),
-                    );
-                    if allow_partial_pushdown {
-                        index.rc()
-                    } else {
-                        bail!(PushDownIndexError::ModuleSomeAxesNotPossible {
-                            index_node: node.clone(),
-                            num_non_batch_out: m.rank() - rank_to_pass,
-                            inner_rank: m.rank(),
-                            remaining_index: idxnb,
-                        })
-                    }
+                    bail!(PushDownIndexError::ModuleSomeAxesNotPossible {
+                        index_node: node.clone(),
+                        num_non_batch_out: m.rank() - rank_to_pass,
+                        inner_rank: m.rank(),
+                        remaining_index: idxnb,
+                    })
                 }
             }
         }
@@ -1853,25 +1854,23 @@ pub fn concat_repeat_to_rearrange(concat: &Concat) -> Option<Concat> {
             if same_seen > 1 {
                 did_anything = true;
                 let mut int_sizes = prev_node.info().shape.clone();
-                int_sizes.push(same_seen);
+                int_sizes.push(Size::known(same_seen));
                 new_nodes.push(Rearrange::nrc_elim_identity(
                     prev_node.clone(),
                     RearrangeSpec::new(
-                        (0..prev_node.info().rank())
-                            .map(|x| tu8v![x as u8])
-                            .collect(),
-                        (0..prev_node.info().rank())
+                        (0..prev_node.rank()).map(|x| tu8v![x as u8]).collect(),
+                        (0..prev_node.rank())
                             .map(|x| {
                                 if x == concat.axis {
                                     // create group which is (count, axis_size) for total dim size of
                                     // count * axis_size()
-                                    tu8v![prev_node.info().rank() as u8, x as u8]
+                                    tu8v![prev_node.rank() as u8, x as u8]
                                 } else {
                                     tu8v![x as u8]
                                 }
                             })
                             .collect(),
-                        shape_to_op_shape(&int_sizes),
+                        int_sizes,
                     )
                     .unwrap(),
                     None,
@@ -1893,10 +1892,14 @@ pub fn concat_repeat_to_rearrange(concat: &Concat) -> Option<Concat> {
 
 #[test]
 fn basic_check_same_concat_shape() {
-    let a = Symbol::new_with_random_uuid(sv![2, 3, 4, 5, 6], None).rc();
+    let a = Symbol::new_with_random_uuid(
+        vec![2, 3, 4, 5, 6].into_iter().map(Size::known).collect(),
+        None,
+    )
+    .rc();
     let out =
         concat_repeat_to_rearrange(&Concat::new(vec![a.clone(), a.clone(), a], 3, None)).unwrap();
-    assert_eq!(out.info().shape[3], 15)
+    assert_eq!(out.info().shape[3].unwrap(), 15)
 }
 
 #[inline]
@@ -1949,7 +1952,7 @@ pub fn add_outer_product_broadcasts_on_top(add: &Add) -> Option<Add> {
         .map(|(child, rank_dif)| {
             let mut bitmask: BitMask128 = 0;
             for (i, l) in child.info().shape.iter().enumerate() {
-                if *l != 1 {
+                if !l.is(1) {
                     bitmask |= 1 << (rank_dif + i)
                 }
             }
@@ -2015,7 +2018,7 @@ pub fn add_fuse_scalar_multiples(add: &Add) -> Option<Add> {
                 let no_scalar_children: EinsumArgs = e
                     .args_cloned().into_iter()
                     .filter(|(child, _)| {
-                        if let Some(x) = child.as_scalar() && x.info().shape.len() == 0  {
+                        if let Some(x) = child.as_scalar() && x.info().shape.is_empty()  {
                             scalar *= x.value;
                             reduced_any += 1;
                             if x.info().name.is_some() {
@@ -2041,7 +2044,7 @@ pub fn add_fuse_scalar_multiples(add: &Add) -> Option<Add> {
                 )
             }
             Circuit::Scalar(s) => {
-                if s.info().shape.len() == 0 {
+                if s.info().shape.is_empty() {
                     (Einsum::empty(None), s.value, s.info().name)
                 } else {
                     (Einsum::identity(c, None), 1.0, None)
@@ -2099,7 +2102,7 @@ pub fn einsum_permute_to_rearrange(einsum: &Einsum) -> Option<Rearrange> {
             RearrangeSpec::new(
                 einsum.in_axes[0].iter().map(|z| tu8v![*z]).collect(),
                 einsum.out_axes.iter().map(|z| tu8v![*z]).collect(),
-                sv![OpSize::NONE;(*einsum.out_axes.iter().max().unwrap()+1) as usize],
+                sv![Size::NONE;(*einsum.out_axes.iter().max().unwrap()+1) as usize],
             )
             .unwrap()
             .canonicalize(false),
@@ -2119,6 +2122,8 @@ const MODULE_AXIS_NO_MSG: &str = concat!(
 #[base_exception(PyValueError)]
 #[derive(Error, Debug, Clone)]
 pub enum PushDownIndexError {
+    #[error("sizes_at_axis={sizes_at_axis:?} index={index} ({e_name})")]
+    ConcatUnknownSizeAtAxis { sizes_at_axis: Shape, index: Index },
     #[error("index_node={index_node:?} index_non_axis={index_non_axis} index={index} ({e_name})")]
     NoopOnConcat {
         index_node: Index,
