@@ -13,6 +13,7 @@ use pyo3::{exceptions::PyValueError, prelude::*};
 use rr_util::{
     python_error_exception,
     set_cover::set_cover,
+    shape::Size,
     tensor_util::{TensorAxisIndex, TensorIndex},
     util::EinsumAxes,
 };
@@ -35,7 +36,7 @@ pub fn batch_to_concat(
     if axis >= circuit.info().shape.len() {
         bail!("batch_to_concat on nonexistent axis");
     }
-    let l = circuit.info().shape[axis];
+    let l = circuit.info().shape[axis].unwrap_err()?;
     if batch_size == 0 {
         bail!(BatchError::BatchSizeIsZero {});
     }
@@ -44,7 +45,7 @@ pub fn batch_to_concat(
     }
     let circuit = deep_map_unwrap(circuit, |x| {
         match &**x {
-            Circuit::Index(index) => Index::slice_edges_to_none(&index).rc(),
+            Circuit::Index(index) => Index::slice_edges_to_none(index).rc(),
             // Circuit::Scatter(scatter) => Scatter::nrc(
             //     scatter.node(),
             //     scatter.index.edges_to_none(&scatter.info().shape),
@@ -70,7 +71,7 @@ pub fn batch_to_concat(
                 ModuleArgSpec {
                     symbol: set_named_axes(
                         Symbol::new_with_random_uuid(sub.info().shape.clone(), None),
-                        BTreeMap::from([(*i as u8, name.clone())]),
+                        BTreeMap::from([(*i as u8, name)]),
                     )
                     .context("Failed extracting named axis from old name.")?,
                     ban_non_symbolic_size_expand: false, // I think this needs to be false?
@@ -116,7 +117,7 @@ pub fn batch_to_concat(
                         .find(|(_c, spec)| spec.symbol.uuid == inp_spec.symbol.uuid)
                     {
                         let mut result = c.info().shape.clone();
-                        result[leaves_axis[c]] = axis_size;
+                        result[leaves_axis[c]] = Size::known(axis_size);
                         return result;
                     }
                     inp_spec.symbol.info().shape.clone()
@@ -135,7 +136,7 @@ pub fn batch_to_concat(
                 {
                     return Index::nrc(
                         c.clone(),
-                        TensorIndex::new_single(index.clone(), leaves_axis[c], c.info().rank()),
+                        TensorIndex::new_single(index.clone(), leaves_axis[c], c.rank()),
                         None,
                     );
                 }
@@ -259,7 +260,7 @@ pub fn batch_einsum(einsum: &Einsum, context: &mut OptimizationContext) -> Resul
     // get the set cover of axes to batch
     let mut covers: Vec<u128> = vec![0; 64];
     for (i, intermediate) in intermediates_above_mem.iter().enumerate() {
-        for (_, v) in &intermediate.info().named_axes {
+        for v in intermediate.info().named_axes.values() {
             let int = v.parse::<usize>().unwrap();
             let v = covers[int] | 1 << i;
             covers[int] = v
@@ -273,7 +274,7 @@ pub fn batch_einsum(einsum: &Einsum, context: &mut OptimizationContext) -> Resul
     let batch_int = ints_needed[0];
     let batch_str = batch_int.to_string().into();
 
-    let batch_int_len = einsum.shape_map().unwrap()[&(batch_int as u8)];
+    let batch_int_len = einsum.shape_map().unwrap()[&(batch_int as u8)].unwrap_err()?;
     let max_interm_mem = intermediates_above_mem
         .iter()
         .map(|x| x.info().numel_usize())
