@@ -6,7 +6,7 @@ use crate::circuit_optimizer::OptimizationContext;
 
 /// takes circuitrc bc convenient
 pub fn numel_sort_key(node: &CircuitRc) -> Vec<u8> {
-    (usize::MAX - *node.info().numel().to_u64_digits().get(0).unwrap_or(&0) as usize)
+    (usize::MAX - *node.info().numel().to_u64_digits().first().unwrap_or(&0) as usize)
         .to_be_bytes()
         .iter()
         .copied()
@@ -24,7 +24,7 @@ pub fn canonicalize_node_py(circuit: CircuitRc) -> CircuitRc {
 pub fn canonicalize_node_op(circuit: CircuitRc) -> Option<CircuitRc> {
     match &**circuit {
         Circuit::Rearrange(rearrange) => Some(Rearrange::nrc(
-            rearrange.node().clone(),
+            rearrange.node(),
             rearrange
                 .spec
                 .conform_to_input_shape(&rearrange.node().info().shape)
@@ -33,7 +33,7 @@ pub fn canonicalize_node_op(circuit: CircuitRc) -> Option<CircuitRc> {
             circuit.info().name,
         )),
         Circuit::Index(index) => Some(Index::nrc(
-            index.node().clone(),
+            index.node(),
             index.index.canonicalize(&index.node().info().shape),
             index.info().name,
         )),
@@ -81,7 +81,7 @@ pub fn normalize_node_py(circuit: CircuitRc) -> CircuitRc {
 pub fn normalize_node_op(circuit: CircuitRc) -> Option<CircuitRc> {
     match &**circuit {
         Circuit::Rearrange(rearrange) => Some(Rearrange::nrc(
-            rearrange.node().clone(),
+            rearrange.node(),
             rearrange.spec.canonicalize(false),
             circuit.info().name,
         )),
