@@ -1,22 +1,23 @@
 use std::io::prelude::*;
 
-use anyhow::{bail, Context, Result};
+use anyhow::Result;
 use circuit_base::{
-    flat_concat,
     module::{
-        any_children_with_symbolic_sizes, conform_all_modules, inline_single_callsite_modules,
+        any_children_with_symbolic_sizes, any_modules, conform_all_modules,
+        inline_single_callsite_modules,
     },
-    print::{print_circuit_stats, PrintOptions},
+    self_funcs::{compiler_default_print, print_circuit_stats},
     CircuitNode, CircuitRc, ModuleSpec,
 };
 use num_bigint::BigUint;
+use once_cell::sync::Lazy;
 use pyo3::prelude::*;
 use rr_util::{
     opt_einsum::EinsumSpec,
     py_types::{is_python_running, Tensor, PY_UTILS},
     python_println,
     repr::RustRepr,
-    tensor_util::Shape,
+    shape::Shape,
     timed,
     util::{apply_fn_until_same, HashBytes},
 };
@@ -32,7 +33,7 @@ use crate::{
     },
     module_rewrite::deep_module_remove_unused_inputs,
     scheduled_execution::{
-        circuit_to_schedule, circuit_to_schedule_naive_toposort, Schedule, SchedulingError,
+        circuit_to_schedule, circuit_to_schedule_naive_toposort, schedule_out_many, Schedule,
     },
 };
 
@@ -51,10 +52,6 @@ pub struct OptimizationSettings {
     #[pyo3(get, set)]
     pub max_single_tensor_memory: usize,
     #[pyo3(get, set)]
-    pub max_memory_fallback: Option<usize>,
-    #[pyo3(get, set)]
-    pub scheduling_num_mem_chunks: usize,
-    #[pyo3(get, set)]
     pub distribute_min_size: Option<usize>,
     #[pyo3(get, set)]
     pub scheduling_naive: bool,
@@ -69,10 +66,6 @@ pub struct OptimizationSettings {
     #[pyo3(get, set)]
     pub numerical_scale_max: f64,
     #[pyo3(get, set)]
-    pub capture_and_print: bool,
-    #[pyo3(get, set)]
-    pub create_tests: bool,
-    #[pyo3(get, set)]
     pub log_simplifications: bool,
     #[pyo3(get, set)]
     pub log_slow_einsums: bool,
@@ -93,25 +86,27 @@ pub struct OptimizationSettings {
 
 impl Default for OptimizationSettings {
     fn default() -> Self {
+        const NUM_CPUS: Lazy<usize> = Lazy::new(|| {
+            std::thread::available_parallelism()
+                .map_or(1, |x| x.get() - 2)
+                .max(1)
+                .min(8)
+        });
         Self {
             verbose: 0,
             max_memory: 9_000_000_000,
             max_single_tensor_memory: 9_000_000_000,
-            max_memory_fallback: None,
-            scheduling_num_mem_chunks: 200,
             distribute_min_size: None,
             scheduling_naive: false,
-            scheduling_timeout: 5_000,
+            scheduling_timeout: 500,
             scheduling_simplify: true,
             adjust_numerical_scale: false,
             numerical_scale_min: 1e-8,
             numerical_scale_max: 1e8,
-            capture_and_print: false,
-            create_tests: false,
             log_simplifications: false,
             log_slow_einsums: false,
             save_circ: false,
-            optimization_parallelism: Some(8),
+            optimization_parallelism: Some(*NUM_CPUS),
             keep_all_names: false,
             push_down_index: true,
             deep_maybe_distribute: true,
@@ -127,8 +122,6 @@ impl OptimizationSettings {
         verbose = OptimizationSettings::default().verbose,
         max_memory = OptimizationSettings::default().max_memory,
         max_single_tensor_memory = OptimizationSettings::default().max_single_tensor_memory,
-        max_memory_fallback = None,
-        scheduling_num_mem_chunks = OptimizationSettings::default().scheduling_num_mem_chunks,
         distribute_min_size = None,
         scheduling_naive = OptimizationSettings::default().scheduling_naive,
         scheduling_timeout = OptimizationSettings::default().scheduling_timeout,
@@ -136,12 +129,10 @@ impl OptimizationSettings {
         adjust_numerical_scale = OptimizationSettings::default().adjust_numerical_scale,
         numerical_scale_min = OptimizationSettings::default().numerical_scale_min,
         numerical_scale_max = OptimizationSettings::default().numerical_scale_max,
-        capture_and_print = OptimizationSettings::default().capture_and_print,
-        create_tests = OptimizationSettings::default().create_tests,
         log_simplifications = OptimizationSettings::default().log_simplifications,
         log_slow_einsums = OptimizationSettings::default().log_slow_einsums,
         save_circ = OptimizationSettings::default().save_circ,
-        optimization_parallelism = None,
+        optimization_parallelism = OptimizationSettings::default().optimization_parallelism.unwrap(),
         keep_all_names = OptimizationSettings::default().keep_all_names,
         push_down_index = OptimizationSettings::default().push_down_index,
         deep_maybe_distribute = OptimizationSettings::default().deep_maybe_distribute,
@@ -151,8 +142,6 @@ impl OptimizationSettings {
         verbose: usize,
         max_memory: usize,
         max_single_tensor_memory: usize,
-        max_memory_fallback: Option<usize>,
-        scheduling_num_mem_chunks: usize,
         distribute_min_size: Option<usize>,
         scheduling_naive: bool,
         scheduling_timeout: usize,
@@ -160,8 +149,6 @@ impl OptimizationSettings {
         adjust_numerical_scale: bool,
         numerical_scale_min: f64,
         numerical_scale_max: f64,
-        capture_and_print: bool,
-        create_tests: bool,
         log_simplifications: bool,
         log_slow_einsums: bool,
         save_circ: bool,
@@ -175,8 +162,6 @@ impl OptimizationSettings {
             verbose,
             max_memory,
             max_single_tensor_memory,
-            max_memory_fallback,
-            scheduling_num_mem_chunks,
             distribute_min_size,
             scheduling_naive,
             scheduling_timeout,
@@ -184,8 +169,6 @@ impl OptimizationSettings {
             adjust_numerical_scale,
             numerical_scale_min,
             numerical_scale_max,
-            capture_and_print,
-            create_tests,
             log_simplifications,
             log_slow_einsums,
             save_circ,
@@ -200,7 +183,7 @@ impl OptimizationSettings {
 
 #[derive(Debug, Clone, Default)]
 pub struct OptimizationCache {
-    pub simplified: HashMap<HashBytes, CircuitRc>,
+    pub simplified: HashMap<HashBytes, Option<CircuitRc>>,
     pub distributed: HashMap<HashBytes, Option<CircuitRc>>,
     pub module_removed_unused: HashMap<HashBytes, Option<CircuitRc>>,
     pub flops: HashMap<HashBytes, BigUint>,
@@ -248,10 +231,7 @@ pub fn optimize_circuit_py(
     circuit: CircuitRc,
     settings: OptimizationSettings,
 ) -> Result<CircuitRc> {
-    optimize_circuit(
-        circuit.clone(),
-        &mut OptimizationContext::new_settings(settings),
-    )
+    optimize_circuit(circuit, &mut OptimizationContext::new_settings(settings))
 }
 
 pub fn optimize_circuit(
@@ -276,11 +256,21 @@ pub fn optimize_circuit(
             )
         })?;
     }
+    if is_python_running() && any_modules(circuit.clone()) {
+        Python::with_gil(|py| {
+            pyo3::PyErr::warn(
+                py,
+                PY_UTILS.optimizing_modules_warning.as_ref(py),
+                "you're optimizing a circuit with modules. this is currently buggy, you might want to substitute_all_modules first",
+                1,
+            )
+        })?;
+    }
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
     if context.settings.verbose >= 4 {
         python_println!("Original Circuit");
         circuit.printu();
-        print_circuit_stats(&*circuit);
+        print_circuit_stats(&circuit);
     }
     if context.settings.save_circ {
         let s = circuit.repru();
@@ -320,10 +310,8 @@ pub fn optimize_circuit(
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
     if context.settings.verbose >= 3 {
         python_println!("Simplified Circuit");
-        PrintOptions::compiler_default()
-            .print(circuit.clone())
-            .unwrap();
-        print_circuit_stats(&*circuit);
+        compiler_default_print(&circuit);
+        print_circuit_stats(&circuit);
     }
     let circuit = if context.settings.deep_maybe_distribute {
         timed!(
@@ -341,10 +329,8 @@ pub fn optimize_circuit(
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
     if context.settings.verbose >= 3 {
         python_println!("Distributed Circuit");
-        PrintOptions::compiler_default()
-            .print(circuit.clone())
-            .unwrap();
-        print_circuit_stats(&*circuit);
+        compiler_default_print(&circuit);
+        print_circuit_stats(&circuit);
     }
     let circuit = timed!(deep_heuristic_nest_adds(circuit), 10, print_timings);
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
@@ -354,10 +340,8 @@ pub fn optimize_circuit(
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
     if context.settings.verbose >= 3 {
         python_println!("Final Circuit");
-        PrintOptions::compiler_default()
-            .print(circuit.clone())
-            .unwrap();
-        print_circuit_stats(&*circuit);
+        compiler_default_print(&circuit);
+        print_circuit_stats(&circuit);
     }
     Ok(circuit)
 }
@@ -368,18 +352,22 @@ pub fn optimize_to_schedule(
     circuit: CircuitRc,
     settings: OptimizationSettings,
 ) -> Result<Schedule> {
+    optimize_to_schedule_many(vec![circuit], settings)
+}
+
+#[pyfunction]
+#[pyo3(signature=(circuits, settings = Default::default()))]
+pub fn optimize_to_schedule_many(
+    circuits: Vec<CircuitRc>,
+    settings: OptimizationSettings,
+) -> Result<Schedule> {
+    let circuit = schedule_out_many(circuits);
     let context = &mut OptimizationContext::new_settings(settings.clone());
     let verbose = settings.verbose;
     if verbose > 0 {
         python_println!("Optimizing verbose {}", verbose)
     }
-    if settings.capture_and_print {
-        PrintOptions::compiler_default()
-            .print(circuit.clone())
-            .context("capture print fail!")?;
-    }
     let optimized_circuit = timed!(optimize_circuit(circuit, context), 10, verbose >= 1)?;
-    // return evaluate_fn(&*optimized_circuit).unwrap();
     let schedule = if settings.scheduling_naive {
         circuit_to_schedule_naive_toposort(optimized_circuit)?
     } else {
@@ -407,38 +395,14 @@ pub fn optimize_and_evaluate(circuit: CircuitRc, settings: OptimizationSettings)
 /// in python, lots of functions take in collections of circuits and operate on them at once
 /// with node caching for just that batch
 /// because functions that take one circuit already cache nodes, it's convenient to compute multiple nodes
-/// by flat-concatting and then passing around as one circuit
-/// flat concatting then splitting is an extra copy on all return data,
-/// which we could get rid of by removing flat-concat after rewrites (which would only work with a black box flat_concat node bc simplification isn't guaranteed to preserve concat at at end)
+/// wrapping in generalfunction Output and then passing around as one circuit
+/// generalfunction Output is removed in circuit_to_schedule before scheduling
 #[pyfunction]
 #[pyo3(signature=(circuits, settings = Default::default()))]
 pub fn optimize_and_evaluate_many(
     circuits: Vec<CircuitRc>,
     settings: OptimizationSettings,
 ) -> Result<Vec<Tensor>> {
-    for c in &circuits {
-        if !c.info().is_explicitly_computable() {
-            bail!(SchedulingError::NotExplicitlyComputable { circuit: c.clone() });
-        }
-    }
-
     let schedule = optimize_to_schedule_many(circuits, settings.clone())?;
     schedule.evaluate_many(settings)
 }
-
-#[pyfunction]
-#[pyo3(signature=(circuits, settings = Default::default()))]
-pub fn optimize_to_schedule_many(
-    circuits: Vec<CircuitRc>,
-    settings: OptimizationSettings,
-) -> Result<Schedule> {
-    let flat_concatted = flat_concat(circuits.clone())?.rc();
-    let mut schedule = optimize_to_schedule(flat_concatted, settings)?;
-    schedule.split_shapes = Some(
-        circuits
-            .iter()
-            .map(|x| x.info().shape.clone())
-            .collect::<Vec<Shape>>(),
-    );
-    Ok(schedule)
-}
