@@ -38,7 +38,7 @@ pub fn maybe_distribute_uncached(
         return None;
     }
     for (i, operand) in node.children().enumerate() {
-        if let Circuit::Add(add) = &**operand && !(add.num_children() == 0) && ( operand.info().numel() >= BigUint::from(context.settings.distribute_min_size.unwrap_or(context.settings.max_single_tensor_memory)))
+        if let Circuit::Add(add) = &**operand && add.num_children() != 0 && ( operand.info().numel() >= BigUint::from(context.settings.distribute_min_size.unwrap_or(context.settings.max_single_tensor_memory)))
 
         {
             let result = timed!(deep_canonicalize( compiler_simp(distribute_once_op(node, i, true).unwrap().rc(),context),context),1,context.settings.verbose>=3);
