@@ -16,13 +16,13 @@ pub fn strip_names_and_tags(circuit: CircuitRc, context: &mut OptimizationContex
         circuit.clone(),
         &|circuit, _| {
             if let Some(tag) = circuit.as_tag() {
-                return Some(tag.node().clone());
+                return Some(tag.node());
             }
             if circuit.info().name.is_some() {
                 if circuit.is_irreducible_node() {
                     return Some(Tag::nrc(circuit, Uuid::nil(), None).rename(None));
                 } else {
-                    return Some(circuit.clone().rename(None));
+                    return Some(circuit.rename(None));
                 }
             }
             None
