@@ -5,6 +5,7 @@ use itertools::Itertools;
 use num_bigint::BigUint;
 use pyo3::prelude::*;
 use rr_util::{
+    shape::{shape_try_as_known, shape_try_into_known},
     tensor_util::{Slice, TensorAxisIndex, TensorIndex, USlice},
     util::{cumsum, outer_product, AsOp},
 };
@@ -77,7 +78,7 @@ fn einsum_pull_concat(einsum: &Einsum, on_ix: &dyn Fn(Index) -> CircuitRc) -> Op
             && let int_here = ints[concat.axis]
             && einsum.out_axes.iter().filter(|z| **z == int_here).count() <= 1 {
             did_anything = true;
-            let sections_here = concat.get_sizes_at_axis();
+            let sections_here = shape_try_into_known(concat.get_sizes_at_axis().into())?;
             match int_sections.get(&int_here) {
                 None => {
                     int_sections.insert(int_here, sections_here);
@@ -148,7 +149,7 @@ fn einsum_pull_concat(einsum: &Einsum, on_ix: &dyn Fn(Index) -> CircuitRc) -> Op
                                 })
                                 .collect(),
                         );
-                        (on_ix(Index::new(node.clone(), index, None)), ints.clone())
+                        (on_ix(Index::new(node, index, None)), ints.clone())
                     })
                     .collect();
 
@@ -197,8 +198,7 @@ fn einsum_pull_concat(einsum: &Einsum, on_ix: &dyn Fn(Index) -> CircuitRc) -> Op
 fn add_pull_concat(add: &Add, on_ix: &dyn Fn(Index) -> CircuitRc) -> Option<CircuitRc> {
     for (child, rank_difference) in add.nodes_and_rank_differences() {
         if let Circuit::Concat(concat) = &**child {
-            if concat.info().shape[concat.axis] != 1 {
-                let sections_here = concat.get_sizes_at_axis();
+            if !concat.info().shape[concat.axis].is(1) && let Some(sections_here) = shape_try_as_known(&concat.get_sizes_at_axis()) {
                 let concat_axis = rank_difference + concat.axis;
                 let add_explicit = add_make_broadcasts_explicit(add)
                     .unwrap_or(add.clone())
@@ -227,21 +227,21 @@ fn generalfunction_pull_concat(
     on_ix: &dyn Fn(Index) -> CircuitRc,
 ) -> Option<Concat> {
     let batch_rank =
-        generalfunction.info().rank() - generalfunction.num_non_batchable_output_dims as usize;
+        generalfunction.rank() - generalfunction.num_non_batchable_output_dims as usize;
     for operand in generalfunction.children() {
         if let Circuit::Concat(concat) = &**operand && concat.axis < batch_rank {
             let mut cur = 0;
             let new_operands = concat
                 .children()
                 .map(|cnode| {
-                    let span_here = cnode.info().shape[concat.axis];
+                    let span_here = cnode.info().shape[concat.axis].t()?;
                     let new_operands = zip(
                         generalfunction.children(),
                         &generalfunction.input_batchability,
                     )
                     .map(|(node, &batchability)| {
                         if !batchability {
-                            node.clone()
+                            node
                         } else if node == operand {
                             cnode.clone()
                         } else {
@@ -250,7 +250,7 @@ fn generalfunction_pull_concat(
                                 TensorIndex::new_single(
                                     TensorAxisIndex::new_plain_slice(cur, cur + span_here),
                                     concat.axis,
-                                    node.info().rank(),
+                                    node.rank(),
                                 ),
                                 None,
                             ))
@@ -258,9 +258,9 @@ fn generalfunction_pull_concat(
                     })
                     .collect();
                     cur += span_here;
-                    GeneralFunction::nrc(new_operands, generalfunction.spec.clone(), None)
+                    Some(GeneralFunction::nrc(new_operands, generalfunction.spec.clone(), None))
                 })
-                .collect();
+                .collect::<Option<Vec<_>>>()?;
             return Some(Concat::new(new_operands, concat.axis, None));
         }
     }
@@ -293,7 +293,7 @@ pub fn concat_drop_size_zero(concat: &Concat) -> Option<Concat> {
     let new_nodes = concat
         .children()
         .filter(|x| {
-            if x.info().shape[concat.axis] == 0 {
+            if x.info().shape[concat.axis].is(0) {
                 did_anything = true;
                 false
             } else {
@@ -312,7 +312,7 @@ pub fn index_concat_drop_unreached(index: &Index) -> Option<CircuitRc> {
     // canon so we only have to work with positives
     let index_canon = index.index.canonicalize(&index.node().info().shape).0;
     let child = index.node();
-    let concat: &Concat = (&**child).as_op()?;
+    let concat: &Concat = (**child).as_op()?;
     if concat
         .children()
         .any(|x| x.info().numel() == BigUint::from(0usize))
@@ -321,12 +321,12 @@ pub fn index_concat_drop_unreached(index: &Index) -> Option<CircuitRc> {
         return None;
     }
     let axis_index = &index_canon[concat.axis];
-    let operand_starts: Vec<usize> = cumsum(
+    let operand_starts: Vec<usize> = cumsum(shape_try_as_known(
         &concat
             .children()
-            .map(|x| x.info().shape[concat.axis])
+            .map(|x| x.shape()[concat.axis])
             .collect::<Vec<_>>(),
-    );
+    )?);
     let (new_operands, new_index) = match axis_index {
         TensorAxisIndex::Single(single) => {
             let operand_idx = operand_starts
@@ -376,11 +376,11 @@ pub fn index_concat_drop_unreached(index: &Index) -> Option<CircuitRc> {
 pub fn split_to_concat_cb(
     circuit: CircuitRc,
     axis: usize,
-    sections: Vec<usize>,
+    sections: &[usize],
     on_ix: &dyn Fn(Index) -> CircuitRc,
 ) -> Concat {
-    assert!(axis < circuit.info().rank());
-    let starts = cumsum(&sections);
+    assert!(axis < circuit.rank());
+    let starts = cumsum(sections);
     let operands: Vec<CircuitRc> = (0..sections.len())
         .map(|i| {
             on_ix(Index::new(
@@ -388,7 +388,7 @@ pub fn split_to_concat_cb(
                 TensorIndex::new_single(
                     TensorAxisIndex::new_plain_slice(starts[i], starts[i + 1]),
                     axis,
-                    circuit.info().rank(),
+                    circuit.rank(),
                 ),
                 None,
             ))
@@ -399,7 +399,7 @@ pub fn split_to_concat_cb(
 
 #[pyfunction]
 pub fn split_to_concat(circuit: CircuitRc, axis: usize, sections: Vec<usize>) -> Concat {
-    split_to_concat_cb(circuit, axis, sections, &mut |x| x.rc())
+    split_to_concat_cb(circuit, axis, &sections, &|x| x.rc())
 }
 
 #[pyfunction]
@@ -408,7 +408,7 @@ pub fn concat_elim_split(circuit: &Concat) -> Option<Index> {
         .children()
         .map(|x| x.as_index().cloned())
         .collect::<Option<Vec<Index>>>()?;
-    let child = indexes[0].node().clone();
+    let child = indexes[0].node();
     let child_axis_map = indexes[0].child_axis_map()[0].clone();
     if !indexes
         .iter()
@@ -428,8 +428,8 @@ pub fn concat_elim_split(circuit: &Concat) -> Option<Index> {
         .collect::<Option<Vec<Slice>>>()?;
     let sections: Vec<USlice> = sections
         .iter()
-        .map(|x| x.to_uslice(child.info().shape[axis_lower]))
-        .collect();
+        .map(|x| Some(x.to_uslice(child.info().shape[axis_lower].t()?)))
+        .collect::<Option<_>>()?;
     let mut place = 0;
     for s in &sections {
         if s.start != place {
@@ -442,7 +442,7 @@ pub fn concat_elim_split(circuit: &Concat) -> Option<Index> {
         stop: Some(place as i64),
     });
     let mut new_index = vec![];
-    for i in 0..child.info().rank() {
+    for i in 0..child.rank() {
         if i == axis_lower {
             new_index.push(new_axis_index.clone());
         } else {
