@@ -1,6 +1,5 @@
-use circuit_base::{
-    circuit_utils::toposort_circuit, evaluate, Array, Circuit, CircuitNode, CircuitRc,
-};
+use anyhow::Result;
+use circuit_base::{circuit_utils::toposort_circuit, evaluate, Array, CircuitNode, CircuitRc};
 use circuit_utils::replace_nodes;
 use num_bigint::BigUint;
 use pyo3::prelude::*;
@@ -56,8 +55,8 @@ pub fn is_close(a: Tensor, b: Tensor) -> Result<PyObject, PyErr> {
 }
 
 pub fn assert_is_close(a: CircuitRc, b: CircuitRc) {
-    let a_ten = evaluate(a.clone()).unwrap();
-    let b_ten = evaluate(b.clone()).unwrap();
+    let a_ten = evaluate(a).unwrap();
+    let b_ten = evaluate(b).unwrap();
     let close_result = is_close(a_ten.clone(), b_ten.clone());
     if close_result.is_err() {
         println!("{:?}", close_result);
@@ -70,7 +69,7 @@ pub fn assert_is_close(a: CircuitRc, b: CircuitRc) {
 #[pyfunction]
 pub fn opt_eval_each_subcircuit_until_fail(circuit: CircuitRc, settings: OptimizationSettings) {
     let max_numel = BigUint::from(2_000_000_000usize);
-    let topo = toposort_circuit(circuit.clone());
+    let topo = toposort_circuit(circuit);
     for circ in topo {
         if circ.info().numel() < max_numel {
             let reference = evaluate(circ.clone()).unwrap();
@@ -98,21 +97,21 @@ pub fn opt_eval_each_subcircuit_until_fail(circuit: CircuitRc, settings: Optimiz
 /// `ablated == rewrite(ablated)`. However, if you serialize both sides to Rust via
 /// rust_expression_notation_circuit the results may differ. For a concrete example,
 /// take Index(Index(X, tensor1), tensor2), which can be simplified to Index(X, tensor3)
-pub fn randn_ablate_great_grandchildren(circ: CircuitRc) -> CircuitRc {
+pub fn randn_ablate_great_grandchildren(circ: CircuitRc) -> Result<CircuitRc> {
     let great_grandchildren = circ
         .children()
         .flat_map(|child| child.children().collect::<Vec<_>>())
         .flat_map(|grandchild| grandchild.children().collect::<Vec<_>>());
     let mut replacements: HashMap<HashBytes, CircuitRc> = HashMap::default();
     for great_grandchild in great_grandchildren {
-        let ablated = Circuit::Array(Array::randn_full(
+        let ablated = Array::randn_full(
             great_grandchild.info().shape.clone(),
             great_grandchild.info().name,
-            great_grandchild.info().device_dtype.clone(),
+            great_grandchild.info().device_dtype,
             Some(great_grandchild.info().hash_usize()),
-        ))
-        .rc();
+        )?
+        .crc();
         replacements.insert(great_grandchild.info().hash, ablated);
     }
-    replace_nodes(circ, &replacements)
+    Ok(replace_nodes(circ, &replacements))
 }
