@@ -15,7 +15,9 @@ use rr_util::{
     python_error_exception,
     tensor_util::{TensorAxisIndex, TensorIndex},
     timed, timed_value,
+    union_find::UnionFind,
     util::{apply_fn_until_same, mapping_until_end, transpose, AsOp, HashBytes},
+    IndexSet,
 };
 use rustc_hash::FxHashMap as HashMap;
 use thiserror::Error;
@@ -145,7 +147,7 @@ macro_rules! mk_simp_fns {
     )
 }
 fn all_simp_fns() -> &'static SimpFns {
-    &*ALL_SIMP_FNS_CACHE
+    &ALL_SIMP_FNS_CACHE
 }
 on_circuit_names!(mk_simp_fns);
 static ALL_SIMP_FNS_CACHE: Lazy<SimpFns> = Lazy::new(all_simp_fns_raw);
@@ -395,7 +397,7 @@ impl SimpFnSubset {
 
     #[pyo3(name = "simp_step")]
     pub fn simp_step_py(&self, circuit: CircuitRc) -> Option<CircuitRc> {
-        self.simp_step(&*circuit)
+        self.simp_step(&circuit)
     }
 
     pub fn simp(&self, circuit: CircuitRc) -> CircuitRc {
@@ -410,7 +412,7 @@ impl SimpFnSubset {
     pub fn simp_step(&self, circuit: &Circuit) -> Option<CircuitRc> {
         match_circ_with_simp_fns!(
             (circuit, &self.1);
-            (node, fns_sub) => run_simp(node, &fns_sub, |_| ())
+            (node, fns_sub) => run_simp(node, fns_sub, |_| ())
         )
     }
 
@@ -430,7 +432,7 @@ impl SimpFnSubset {
 
         match_circ_with_simp_fns!(
             (circuit, &self.1);
-            (node, fns_sub) => run_simp(node, &fns_sub, on_name)
+            (node, fns_sub) => run_simp(node, fns_sub, on_name)
         )
     }
 
@@ -460,7 +462,7 @@ impl SimpFnSubset {
             }
         })
         .into_iter()
-        .filter_map(|x| x)
+        .flatten()
         .join("\n")
     }
 }
@@ -506,39 +508,35 @@ pub fn deep_simp<F>(circ: CircuitRc, opt_context: &mut OptimizationContext, f: F
 where
     F: Fn(&Circuit, &mut OptimizationContext) -> Option<CircuitRc>,
 {
-    /// check if any new children have not been simplified yet, and simplify them if so
-    fn simplify_changed_descendants<F>(
+    /// fully simplify a circuit and all its descendants recursively until we hit a fixed point
+    fn fully_simplify<F>(
         circ: CircuitRc,
         context: &mut OptimizationContext,
         f: &F,
     ) -> Option<CircuitRc>
-    where
-        F: Fn(&Circuit, &mut OptimizationContext) -> Option<CircuitRc>,
-    {
-        circ.map_children_op(&mut |x: CircuitRc| {
-            if context.cache.simplified.contains_key(&x.info().hash) {
-                None
-            } else {
-                Some(fully_simplify(x, context, f))
-            }
-        })
-        .map(|c| c.rc())
-    }
-    /// fully simplify a circuit and all its descendants recursively until we hit a fixed point
-    fn fully_simplify<F>(circ: CircuitRc, context: &mut OptimizationContext, f: &F) -> CircuitRc
     where
         F: Fn(&Circuit, &mut OptimizationContext) -> Option<CircuitRc>,
     {
         if let Some(result) = context.cache.simplified.get(&circ.info().hash) {
             return result.clone();
         }
+        let mut did_anything = false;
         let mut result: CircuitRc = circ
-            .map_children_unwrap(&mut |x: CircuitRc| fully_simplify(x, context, f))
-            .rc();
+            .map_children_op(&mut |x: CircuitRc| fully_simplify(x, context, f))
+            .map_or_else(
+                || circ.clone(),
+                |x| {
+                    did_anything = true;
+                    x
+                },
+            );
         for iter_count in 0.. {
             match f(&result, context) {
                 Some(r) => {
-                    result = simplify_changed_descendants(r.clone(), context, f).unwrap_or(r)
+                    did_anything = true;
+                    result = r
+                        .map_children_op(&mut |x: CircuitRc| fully_simplify(x, context, f))
+                        .unwrap_or(r);
                 }
                 None => break,
             }
@@ -548,22 +546,23 @@ where
                 panic!();
             }
         }
+        let result = if did_anything { Some(result) } else { None };
         context
             .cache
             .simplified
             .insert(circ.info().hash, result.clone());
         result
     }
-    fully_simplify(circ, opt_context, &f)
+    fully_simplify(circ.clone(), opt_context, &f).unwrap_or(circ)
 }
 
 #[pyfunction]
 pub fn deep_push_down_index_raw(circ: CircuitRc, min_size: Option<usize>) -> CircuitRc {
-    deep_replace_parents_fix(circ.clone(), |c, parents| {
+    deep_replace_parents_fix(circ, |c, parents| {
         if min_size.is_some() && c.info().numel() < BigUint::from(min_size.unwrap()) {
             return None;
         }
-        if parents.len() < 1 || !parents.iter().all(|n| n.is_index()) {
+        if parents.is_empty() || !parents.iter().all(|n| n.is_index()) {
             return None;
         }
         let idxs: Vec<_> = parents
@@ -584,19 +583,19 @@ pub fn deep_push_down_index_raw(circ: CircuitRc, min_size: Option<usize>) -> Cir
             if c.is_index() {
                 return Some(vec![index_fuse(ix).unwrap().crc()]);
             } else {
-                push_down_index_op(ix).map_or(None, |x| Some(vec![x]))
+                push_down_index_op(ix).map(|x| vec![x])
             };
         }
 
         let ddt = get_compatible_dtype(&c);
         let (base_ix, new_ixs_t): (Vec<TensorAxisIndex>, Vec<Vec<TensorAxisIndex>>) = (0..c.rank())
             .map(|i| {
-                let l = c.shape()[i];
-                TensorAxisIndex::indices_union_and_rebased(
-                    &idxs.iter().map(|ix| ix.0[i].clone()).collect(),
-                    l,
-                    ddt,
-                )
+                let ixs = idxs.iter().map(|ix| ix.0[i].clone()).collect();
+                if let Some(l) = c.shape()[i].t() {
+                    TensorAxisIndex::indices_union_and_rebased(&ixs, l, ddt)
+                } else {
+                    (TensorAxisIndex::IDENT, ixs)
+                }
             })
             .unzip();
         let base_ix = TensorIndex(base_ix);
@@ -606,7 +605,7 @@ pub fn deep_push_down_index_raw(circ: CircuitRc, min_size: Option<usize>) -> Cir
             return None;
         }
 
-        let base = Index::new(c.clone(), base_ix, None);
+        let base = Index::new(c, base_ix, None);
         let base = push_down_index_op(&base).unwrap_or_else(|| base.crc());
 
         let out: Vec<_> = izip!(parents, new_ixs)
@@ -666,111 +665,155 @@ impl CircBloomFilter {
     }
 }
 
-/// we want adds to be nested rather than flat so arguments can be dropped if they're only needed
-/// in future adds
-/// this is suboptimal in many ways. one is broadcasts allow outer products which should be avoided but aren't
-/// for each add, greedily nest into preexisting adds
 #[inline]
 fn child_bloom_filter(add: &Add) -> CircBloomFilter {
     let mut result: CircBloomFilter = Default::default();
-    for node in add.children() {
+    for node in add.children_sl() {
         result.insert(node.info().hash_usize() as u64);
     }
     result
 }
+/// we want adds to be nested rather than flat so arguments can be dropped if they're only needed
+/// in future adds
+/// this is suboptimal in many ways. one is broadcasts allow outer products which should be avoided but aren't
+/// for each add, greedily nest into preexisting adds
 #[pyfunction]
 pub fn deep_heuristic_nest_adds(circ: CircuitRc) -> CircuitRc {
     let circ = deep_canonicalize(circ, &mut Default::default());
-    let mut seen_adds: Vec<(Add, CircBloomFilter)> = vec![];
+    let mut add_unions: Vec<(CircuitRc, usize)> = vec![];
+    let mut add_ids: IndexSet<HashBytes> = IndexSet::default();
+    let mut uf = UnionFind::new(0);
     visit_circuit_unwrap(circ.clone(), |c: CircuitRc| {
-        if let Some(add) = c.as_add() {
-            seen_adds.push((add.clone(), child_bloom_filter(add)));
+        if c.is_add() {
+            if let Some(gr) = c
+                .children_sl()
+                .iter()
+                .chain(std::iter::once(&c))
+                .map(|x| {
+                    let (ix, b) = add_ids.insert_full(x.info().hash);
+                    if b {
+                        uf.grow(1);
+                    }
+                    ix
+                })
+                .collect::<Vec<usize>>()
+                .iter()
+                .reduce(|x, y| {
+                    uf.union(*x, *y);
+                    x
+                })
+            {
+                add_unions.push((c.clone(), *gr))
+            }
         }
     });
-    // let mut stats = (0, 0, 0);
-    // let mut add_width_tracker: Vec<usize> = vec![0; 300];
-    // let mut intersections: HashMap<Add, CircBloomFilter> = HashMap::default();
-    // skipping intersections bc we're too slow as it is
-    // for circ in &seen_adds {
-    //     for circ2 in &seen_adds {
-    //         if circ.1.intersection_count(&circ2.1) >= 2 && circ.0 != circ2.0 {
-    //             let intersection = Add::new(
-    //                 circ.0
-    //                     .nodes
-    //                     .iter()
-    //                     .filter(|x| circ2.0.nodes.contains(x))
-    //                     .cloned()
-    //                     .collect(),
-    //                 None,
-    //             );
-    //             if intersection.nodes.len() >= 2
-    //                 && &intersection != &circ.0
-    //                 && &intersection != &circ2.0
-    //             {
-    //                 intersections.insert(intersection.clone(), child_bloom_filter(&intersection));
-    //             }
-    //         }
-    //     }
-    // }
-    // seen_adds.extend(intersections);
-    let mut seen_adds_index: HashMap<HashBytes, usize> = seen_adds
-        .iter()
-        .enumerate()
-        .map(|(i, (c, _b))| (c.info().hash, i))
-        .collect();
-    // for ad in &seen_adds {
-    //     add_width_tracker[ad.0.nodes.len()] += 1;
-    // }
+
+    let add_groups = add_unions.iter().into_group_map_by(|x| uf.find_(x.1));
+
     let mut mapping: HashMap<Add, Add> = HashMap::default();
-    for _ in 0..2 {
-        // should allow this forever but limiting to 2 bc slow
-        let timeout_timer = std::time::Instant::now();
-        let mut to_do: Vec<((Add, CircBloomFilter), (Add, CircBloomFilter))> = vec![];
-        for cand_sup in &seen_adds {
-            if cand_sup.0.num_children() >= 3 {
-                for cand_sub in &seen_adds {
-                    // stats.0 += 1;
-                    if cand_sup.1.is_superset(&cand_sub.1)
-                        && cand_sub.0.num_children() < cand_sup.0.num_children()
-                        && cand_sub.0 != cand_sup.0
-                    {
-                        // stats.1 += 1;
-                        if let Some(new) = extract_add(&cand_sup.0, &cand_sub.0) {
-                            // stats.2 += 1;
-                            let new_entry = (new.clone(), child_bloom_filter(&new));
-                            to_do.push(((cand_sup.0.clone(), cand_sup.1.clone()), new_entry));
-                            break;
-                        }
-                    }
-                }
-                if timeout_timer.elapsed().as_millis() > 1000 {
-                    println!("nest adds timed out");
-                    break;
-                }
-            }
-        }
-        // println!("stats {:?} {:?}", stats, add_width_tracker);
-        if to_do.is_empty() {
+    for add_group in add_groups.values() {
+        if add_group.len() == 1 {
             break;
         }
+        let mut seen_adds: Vec<(Add, CircBloomFilter)> = add_group
+            .iter()
+            .map(|(x, _)| {
+                let y = x.as_add_unwrap().clone();
+                let bl = child_bloom_filter(&y);
+                (y, bl)
+            })
+            .collect();
 
-        for (sup, new_sup) in to_do {
-            let old_sup_index = seen_adds_index
-                .insert(sup.0.info().hash, usize::MAX)
-                .unwrap();
-            if old_sup_index != usize::MAX {
-                if !seen_adds_index.contains_key(&new_sup.0.info().hash) {
-                    seen_adds_index.insert(new_sup.0.info().hash, old_sup_index);
-                    seen_adds[old_sup_index] = new_sup.clone();
-                } else if let Some(popped) = seen_adds.pop() {
-                    // if the new superset is already seen, pop last thing ad put it at old sup index and fix seen_adds_index
-                    seen_adds_index.insert(popped.0.info().hash, old_sup_index);
-                    seen_adds[old_sup_index] = popped;
+        // let mut stats = (0, 0, 0);
+        // let mut add_width_tracker: Vec<usize> = vec![0; 300];
+        let mut intersections: HashMap<Add, CircBloomFilter> = HashMap::default();
+        // skipping intersections bc we're too slow as it is
+        for circ in &seen_adds {
+            for circ2 in &seen_adds {
+                if circ.1.intersection_count(&circ2.1) >= 2 && circ.0 != circ2.0 {
+                    let intersection = Add::new(
+                        circ.0
+                            .children_sl()
+                            .iter()
+                            .filter(|x| circ2.0.children_sl().contains(x))
+                            .cloned()
+                            .collect(),
+                        None,
+                    );
+                    if intersection.num_children() >= 2
+                        && intersection != circ.0
+                        && intersection != circ2.0
+                    {
+                        intersections
+                            .insert(intersection.clone(), child_bloom_filter(&intersection));
+                    }
                 }
             }
-            mapping.insert(sup.0.clone(), new_sup.0.clone());
         }
-        // break
+        // println!(
+        //     "add group {} elements + {} intersections",
+        //     seen_adds.len(),
+        //     intersections.len()
+        // );
+        seen_adds.extend(intersections);
+        let mut seen_adds_index: HashMap<HashBytes, usize> = seen_adds
+            .iter()
+            .enumerate()
+            .map(|(i, x)| (x.0.info().hash, i))
+            .collect();
+        // for ad in &seen_adds {
+        //     add_width_tracker[ad.0.nodes.len()] += 1;
+        // }
+        let timeout_timer = std::time::Instant::now();
+        for _ in 0..2 {
+            // should allow this forever but limiting to 2 bc slow
+            let mut to_do: Vec<((Add, CircBloomFilter), (Add, CircBloomFilter))> = vec![];
+            for cand_sup in &seen_adds {
+                if cand_sup.0.num_children() >= 3 {
+                    for cand_sub in &seen_adds {
+                        // stats.0 += 1;
+                        if cand_sup.1.is_superset(&cand_sub.1)
+                            && cand_sub.0.num_children() < cand_sup.0.num_children()
+                            && cand_sub.0 != cand_sup.0
+                        {
+                            // stats.1 += 1;
+                            if let Some(new) = extract_add(&cand_sup.0, &cand_sub.0) {
+                                // stats.2 += 1;
+                                let new_entry = (new.clone(), child_bloom_filter(&new));
+                                to_do.push(((cand_sup.0.clone(), cand_sup.1.clone()), new_entry));
+                                break;
+                            }
+                        }
+                    }
+                    if timeout_timer.elapsed().as_millis() > 20 {
+                        // println!("nest adds group timed out, continuing with current nesting");
+                        break;
+                    }
+                }
+            }
+            // println!("stats {:?} {:?}", stats, add_width_tracker);
+            if to_do.is_empty() {
+                break;
+            }
+
+            for (sup, new_sup) in to_do {
+                let old_sup_index = seen_adds_index
+                    .insert(sup.0.info().hash, usize::MAX)
+                    .unwrap();
+                if old_sup_index != usize::MAX {
+                    if !seen_adds_index.contains_key(&new_sup.0.info().hash) {
+                        seen_adds_index.insert(new_sup.0.info().hash, old_sup_index);
+                        seen_adds[old_sup_index] = new_sup.clone();
+                    } else if let Some(popped) = seen_adds.pop() {
+                        // if the new superset is already seen, pop last thing ad put it at old sup index and fix seen_adds_index
+                        seen_adds_index.insert(popped.0.info().hash, old_sup_index);
+                        seen_adds[old_sup_index] = popped;
+                    }
+                }
+                mapping.insert(sup.0.clone(), new_sup.0.clone());
+            }
+            // break
+        }
     }
 
     deep_map_preorder_unwrap(circ, |c| {
@@ -801,7 +844,7 @@ pub fn add_nest_ltr_py(add: Add) -> Add {
 
 #[pyfunction]
 pub fn deep_pull_concat_messy(circuit: CircuitRc, min_size: Option<usize>) -> CircuitRc {
-    deep_map_unwrap(circuit, &|x: CircuitRc| {
+    deep_map_unwrap(circuit, |x: CircuitRc| {
         if min_size.is_none()
             || x.children()
                 .chain(std::iter::once(x.clone()))
@@ -819,7 +862,7 @@ pub fn deep_pull_concat_messy(circuit: CircuitRc, min_size: Option<usize>) -> Ci
 
 #[pyfunction]
 pub fn deep_pull_concat_new(circuit: CircuitRc, min_size: Option<usize>) -> CircuitRc {
-    deep_map_unwrap(circuit, &|x: CircuitRc| {
+    deep_map_unwrap(circuit, |x: CircuitRc| {
         if min_size.is_none()
             || x.children()
                 .chain(std::iter::once(x.clone()))
@@ -857,7 +900,7 @@ pub fn deep_optimize_einsums(circ: CircuitRc, context: &mut OptimizationContext)
             Circuit::Einsum(ein) => {
                 let (result, took) = timed_value!(einsum_nest_optimize(ein, context));
                 if result.is_err() {
-                    Some(timed!(batch_einsum(ein, context).unwrap()))
+                    Some(timed!(batch_einsum(ein, context).unwrap_or(x)))
                 } else {
                     let result = result.ok()?;
                     if context.settings.log_slow_einsums && took.as_millis() > 10 {
