@@ -35,7 +35,7 @@ pub fn diags_union(diags: &Vec<EinsumAxes>) -> EinsumAxes {
             for (i2, int2) in d.iter().enumerate() {
                 if i1 < i2 && int1 == int2 {
                     println!("{} {}", i1, i2);
-                    uf.union(i1 as usize, i2 as usize);
+                    uf.union(i1, i2);
                 }
             }
         }
@@ -50,18 +50,16 @@ pub fn einsum_push_down_trace(einsum: &Einsum) -> Option<Einsum> {
     let new_args = einsum
         .args()
         .map(|(node, ints)| {
-            if !is_unique(&ints) {
+            if !is_unique(ints) {
                 match &**node {
                     Circuit::Add(child) => {
                         // have to rearrange
                         let explicit_add =
-                            add_make_broadcasts_explicit(&child).unwrap_or(child.clone());
+                            add_make_broadcasts_explicit(child).unwrap_or(child.clone());
                         let new_add = Add::nrc(
                             explicit_add
                                 .children()
-                                .map(|node| {
-                                    Einsum::new_trace(node.clone(), ints.clone(), None).rc()
-                                })
+                                .map(|node| Einsum::new_trace(node, ints.clone(), None).rc())
                                 .collect(),
                             None,
                         );
@@ -70,7 +68,7 @@ pub fn einsum_push_down_trace(einsum: &Einsum) -> Option<Einsum> {
                     }
                     Circuit::Rearrange(child) => {
                         let axis_map = &child.child_axis_map()[0];
-                        let ints_for_intersection = (0..child.info().rank())
+                        let ints_for_intersection = (0..child.rank())
                             .map(|i| {
                                 if axis_map.iter().any(|z| *z == Some(i)) {
                                     0
@@ -106,15 +104,15 @@ pub fn einsum_push_down_trace(einsum: &Einsum) -> Option<Einsum> {
                                     if let Some(z) = x {
                                         intersected_ints[*z]
                                     } else {
-                                        i as u8 + child.info().rank() as u8
+                                        i as u8 + child.rank() as u8
                                     }
                                 })
                                 .collect();
                             // can't just use new_trace bc which one gets eliminated is determined by output, not input
                             // so for instance `aba` might go to `ba` instead of normal `ab`
                             let new_trace = Einsum::nrc(
-                                vec![(child.node().clone(), thingy.clone())],
-                                (0..child.node().info().rank())
+                                vec![(child.node(), thingy.clone())],
+                                (0..child.node().rank())
                                     .filter(|i| {
                                         !tuples_to_remove.contains(&child.spec.input_ints[*i][..])
                                     })
@@ -164,16 +162,16 @@ pub fn einsum_push_down_trace(einsum: &Einsum) -> Option<Einsum> {
                                 .enumerate()
                                 .map(|(i, x)| {
                                     if let Some(z) = x {
-                                        intersected_ints[*z] as u8
+                                        intersected_ints[*z]
                                     } else {
-                                        (i + child.node().info().rank()) as u8
+                                        (i + child.node().rank()) as u8
                                     }
                                 })
                                 .collect();
                             let intersected_ints_input_deduped: EinsumAxes =
                                 intersected_ints_input.iter().unique().cloned().collect();
                             let new_trace = Einsum::new_trace(
-                                child.node().clone(),
+                                child.node(),
                                 intersected_ints_input.clone(),
                                 None,
                             )
@@ -219,10 +217,7 @@ pub fn einsum_push_down_trace(einsum: &Einsum) -> Option<Einsum> {
                             let new_concat = Concat::nrc(
                                 child
                                     .children()
-                                    .map(|x| {
-                                        Einsum::new_trace(x.clone(), new_raw_ints.clone(), None)
-                                            .rc()
-                                    })
+                                    .map(|x| Einsum::new_trace(x, new_raw_ints.clone(), None).rc())
                                     .collect(),
                                 new_concat_axis,
                                 None,
@@ -275,8 +270,8 @@ pub fn add_pull_diags(add: &Add) -> Option<Einsum> {
             (0..rank_difference)
                 .chain(zip(&einsum.info().shape, &einsum.out_axes).enumerate().map(
                     |(i, (l, x))| {
-                        if *l > 1 {
-                            x + rank_difference as u8
+                        if !l.is(1) {
+                            x + rank_difference
                         } else {
                             i as u8 + rank_difference + *einsum.out_axes.iter().max().unwrap() + 1
                         }
