@@ -6,7 +6,6 @@ use circuit_base::{
 use pyo3::prelude::*;
 use rr_util::{
     sv,
-    symbolic_size::SymbolicSizeProduct,
     tensor_util::{TensorAxisIndex, TensorIndex},
 };
 #[pyfunction]
@@ -23,14 +22,13 @@ pub fn generalfunction_merge_inverses(circuit: &GeneralFunction) -> Option<Circu
 #[pyfunction]
 pub fn generalfunction_special_case_simplification(circuit: &GeneralFunction) -> Option<CircuitRc> {
     if circuit.spec.is_official() {
-        let name_str: &str = &circuit.spec.name();
+        let name_str: &str = circuit.spec.name();
         match name_str {
             "softmax" => {
                 if let Circuit::Scalar(_sc) = &**circuit.children_sl()[0] {
                     let scalar: f64 = 1.0
-                        / circuit.children_sl()[0].info().shape
-                            [circuit.children_sl()[0].info().rank() - 1]
-                            as f64;
+                        / circuit.children_sl()[0].info().shape[circuit.children_sl()[0].rank() - 1]
+                            .t()? as f64;
                     return Some(Scalar::nrc(
                         scalar,
                         circuit.info().shape.clone(),
@@ -41,9 +39,8 @@ pub fn generalfunction_special_case_simplification(circuit: &GeneralFunction) ->
             "log_softmax" => {
                 if let Circuit::Scalar(_sc) = &**circuit.children_sl()[0] {
                     let scalar: f64 = (1.0
-                        / circuit.children_sl()[0].info().shape
-                            [circuit.children_sl()[0].info().rank() - 1]
-                            as f64)
+                        / circuit.children_sl()[0].info().shape[circuit.children_sl()[0].rank() - 1]
+                            .t()? as f64)
                         .ln();
                     return Some(Scalar::nrc(
                         scalar,
@@ -53,19 +50,15 @@ pub fn generalfunction_special_case_simplification(circuit: &GeneralFunction) ->
                 }
             }
             "last_dim_size" => {
-                let last_dim_maybe_symbolic = circuit.children_sl()[0].info().shape
-                    [circuit.children_sl()[0].info().rank() - 1];
-                if !SymbolicSizeProduct::has_symbolic(last_dim_maybe_symbolic) {
-                    return Some(Scalar::nrc(
-                        last_dim_maybe_symbolic as f64,
-                        circuit.children_sl()[0].info().shape
-                            [..circuit.children_sl()[0].info().rank() - 1]
-                            .iter()
-                            .cloned()
-                            .collect(),
-                        circuit.info().name,
-                    ));
-                }
+                return Some(Scalar::nrc(
+                    circuit.children_sl()[0].info().shape[circuit.children_sl()[0].rank() - 1]
+                        .t()? as f64,
+                    circuit.children_sl()[0].info().shape[..circuit.children_sl()[0].rank() - 1]
+                        .iter()
+                        .cloned()
+                        .collect(),
+                    circuit.info().name,
+                ));
             }
             _ => {}
         }
@@ -77,9 +70,10 @@ pub fn generalfunction_special_case_simplification(circuit: &GeneralFunction) ->
 #[pyfunction]
 pub fn generalfunction_evaluate_simple(node: &GeneralFunction) -> Option<CircuitRc> {
     if node.num_non_batchable_output_dims == 0
-        && node.info().rank() == 0
+        && node.rank() == 0
         && node.num_children() == 1
-        && node.children_sl()[0].info().rank() == 0
+        && node.children_sl()[0].rank() == 0
+        && node.spec.should_const_fold()
     {
         return node.children_sl()[0].as_scalar().map(|_inner| {
             Python::with_gil(|py| {
