@@ -8,8 +8,8 @@ use circuit_base::{
 use circuit_utils::replace_nodes_py;
 use pyo3::prelude::*;
 use rr_util::{
-    name::Name, py_types::PyOpAtAxes, rearrange_spec::RearrangeSpec,
-    tensor_util::right_align_shapes, util::HashBytes,
+    name::Name, py_types::PyOpAtAxes, rearrange_spec::RearrangeSpec, shape::right_align_shapes,
+    util::HashBytes,
 };
 use rustc_hash::FxHashMap as HashMap;
 
@@ -112,7 +112,7 @@ pub fn module_remove_unused_inputs_op(
             if add_suffix_on_remove_unused {
                 format!("{} rem_unused", s).into()
             } else {
-                s.into()
+                s
             }
         }),
     );
@@ -236,11 +236,11 @@ pub fn fuse_concat_modules(
     if !modules.iter().all(|x| x.spec == spec) {
         bail!(anyhow!("modules_one_batch not all same"))
     }
-    let pure_rank = spec.circuit.info().rank();
+    let pure_rank = spec.circuit.rank();
     let modules_one_batch = modules
         .iter()
         .map(|x| {
-            let dif = x.info().rank() - spec.circuit.info().rank();
+            let dif = x.rank() - spec.circuit.info().rank();
             match dif {
                 0 => Ok(Module::new(
                     zip(x.args(), &x.spec.arg_specs)
@@ -293,7 +293,7 @@ pub fn fuse_concat_modules(
                 }
             })
             .collect(),
-        spec.clone(),
+        spec,
         name,
     );
     let splitted_modules: Vec<CircuitRc> = split_to_concat(
@@ -301,14 +301,14 @@ pub fn fuse_concat_modules(
         0,
         modules_one_batch
             .iter()
-            .map(|x| x.info().shape[0])
-            .collect(),
+            .map(|x| x.info().shape[0].unwrap_err())
+            .collect::<Result<Vec<_>>>()?,
     )
     .children()
     .collect();
     let splitted_unsqueezed: Vec<CircuitRc> = (0..modules.len())
         .map(|i| {
-            if modules[i].info().rank() == pure_rank {
+            if modules[i].rank() == pure_rank {
                 splitted_modules[i]
                     .squeeze(PyOpAtAxes::Single(0), None)
                     .unwrap()
