@@ -45,7 +45,7 @@ pub fn einsum_elim_removable_axes_weak(einsum: &Einsum) -> Option<Einsum> {
             }
         }
     }
-    if int_to_input_removable.len() == 0 {
+    if int_to_input_removable.is_empty() {
         return None;
     }
 
@@ -82,12 +82,12 @@ pub fn einsum_elim_removable_axes_weak(einsum: &Einsum) -> Option<Einsum> {
                     .cloned()
                     .collect();
 
-                if removable_axes.len() == 0 {
-                    (node.clone(), ints.clone())
+                if removable_axes.is_empty() {
+                    (node, ints.clone())
                 } else {
                     (
                         remove_axes_weak(&node, &removable_axes).unwrap(),
-                        filter_out_idx(&ints, &removable_axes).into_iter().collect(),
+                        filter_out_idx(ints, &removable_axes).into_iter().collect(),
                     )
                 }
             })
@@ -104,7 +104,7 @@ pub fn add_elim_removable_axes_weak(add: &Add) -> Option<Add> {
     let mut axis_to_inputs_removable: HashMap<usize, _> = HashMap::default();
     for (i, v) in axis_map.iter().enumerate() {
         let node = &add.children_sl()[i];
-        let removable = get_removable_axes(&node);
+        let removable = get_removable_axes(node);
         for (oj, j) in v
             .iter()
             .enumerate()
@@ -137,7 +137,7 @@ pub fn add_elim_removable_axes_weak(add: &Add) -> Option<Add> {
             }
         }
     }
-    if axis_to_inputs_removable.len() == 0 {
+    if axis_to_inputs_removable.is_empty() {
         return None;
     }
 
@@ -151,7 +151,7 @@ pub fn add_elim_removable_axes_weak(add: &Add) -> Option<Add> {
             .iter()
             .enumerate()
             .map(|(i, (node, rank_dif))| {
-                let removable_axes: Vec<usize> = (*rank_dif..add.info().rank())
+                let removable_axes: Vec<usize> = (*rank_dif..add.rank())
                     .enumerate()
                     .filter_map(|(j, int)| {
                         if axis_to_inputs_removable
@@ -173,7 +173,7 @@ pub fn add_elim_removable_axes_weak(add: &Add) -> Option<Add> {
                     .filter_map(|(i, ax)| (i == ax).then_some(ax))
                     .collect();
 
-                if removable_axes.len() == 0 {
+                if removable_axes.is_empty() {
                     node.clone()
                 } else {
                     remove_axes_weak(node, &removable_axes).unwrap()
@@ -185,5 +185,5 @@ pub fn add_elim_removable_axes_weak(add: &Add) -> Option<Add> {
     if &result == add {
         return None;
     }
-    return Some(result);
+    Some(result)
 }
