@@ -1,12 +1,14 @@
 use std::iter::zip;
 
+use anyhow::Result;
 use circuit_base::{prelude::*, Add, Concat, Einsum, Index, Rearrange, Scalar, Scatter};
 use pyo3::prelude::*;
 use rr_util::{
     compact_data::TinyVecU8,
     rearrange_spec::RearrangeSpec,
+    shape::{shape_try_as_known, Shape, Size},
     tensor_util::{
-        compose, uslices_shrink_base, uslices_to_index, Shape, TensorAxisIndex, TensorIndex, USlice,
+        compose, uslices_shrink_base, uslices_to_index, TensorAxisIndex, TensorIndex, USlice,
     },
     util::{cumsum, filter_out_idx},
 };
@@ -21,7 +23,7 @@ pub fn scatter_fuse(scatter: &Scatter) -> Option<Scatter> {
     if let Circuit::Scatter(inner) = &**scatter.node() {
         let index_composed = compose(&inner.index, &scatter.index)?;
         Some(Scatter::new(
-            inner.node().clone(),
+            inner.node(),
             index_composed,
             scatter.info().shape.clone(),
             scatter.info().name,
@@ -50,23 +52,20 @@ pub fn einsum_pull_scatter(einsum: &Einsum) -> Option<CircuitRc> {
     let int_sizes = einsum.shape_map().unwrap();
     let mut int_slices: HashMap<u8, USlice> = HashMap::default();
     for (node, ints) in einsum.args() {
-        match &**node {
-            Circuit::Scatter(scatter) => {
-                did_anything = true;
-                for (slice, i) in zip(scatter.index.all_uslices().unwrap(), ints.iter()) {
-                    int_slices.insert(
-                        *i,
-                        int_slices
-                            .get(i)
-                            .unwrap_or(&USlice {
-                                start: 0,
-                                stop: int_sizes[i],
-                            })
-                            .intersection(&slice),
-                    );
-                }
+        if let Circuit::Scatter(scatter) = &**node {
+            did_anything = true;
+            for (slice, i) in zip(scatter.index.all_uslices().unwrap(), ints.iter()) {
+                int_slices.insert(
+                    *i,
+                    int_slices
+                        .get(i)
+                        .unwrap_or(&USlice {
+                            start: 0,
+                            stop: int_sizes[i].t()?,
+                        })
+                        .intersection(&slice),
+                );
             }
-            _ => {}
         }
     }
     if !did_anything {
@@ -75,7 +74,11 @@ pub fn einsum_pull_scatter(einsum: &Einsum) -> Option<CircuitRc> {
         .iter()
         .any(|(_i, slice)| slice.start == slice.stop)
     {
-        Some(Scalar::new(0.0, einsum.info().shape.clone(), None).rc())
+        Some(
+            Scalar::try_new(0.0, einsum.info().shape.clone(), None)
+                .ok()?
+                .rc(),
+        )
     } else {
         let new_args = einsum
             .args()
@@ -85,12 +88,12 @@ pub fn einsum_pull_scatter(einsum: &Einsum) -> Option<CircuitRc> {
                     let index =
                         uslices_shrink_base(&index_orig_base, &inner.index.all_uslices().unwrap());
                     (
-                        Index::nrc(inner.node().clone(), uslices_to_index(&index), None),
+                        Index::nrc(inner.node(), uslices_to_index(&index), None),
                         ints.clone(),
                     )
                 }
                 _ => (
-                    Index::nrc(node.clone(), uslice_map_get_index(&ints, &int_slices), None),
+                    Index::nrc(node.clone(), uslice_map_get_index(ints, &int_slices), None),
                     ints.clone(),
                 ),
             })
@@ -126,18 +129,18 @@ pub fn add_pull_scatter(add: &Add) -> Option<Scatter> {
             } else {
                 slices[i + rank_difference] = USlice {
                     start: 0,
-                    stop: add.info().shape[i + rank_difference],
+                    stop: add.info().shape[i + rank_difference].t()?,
                 };
             }
         }
         for i in 0..rank_difference {
             slices[i] = USlice {
                 start: 0,
-                stop: add.info().shape[i],
+                stop: add.info().shape[i].t()?,
             };
         }
     }
-    if zip(&slices, &add.info().shape).all(|(s, l)| s.start == 0 && s.stop == *l) {
+    if zip(&slices, &add.info().shape).all(|(s, l)| s.start == 0 && l.is(s.stop)) {
         None
     } else {
         let new_operands = add
@@ -164,7 +167,7 @@ pub fn add_pull_scatter(add: &Add) -> Option<Scatter> {
         let new_add = Add::nrc(new_operands, add.info().name);
         Some(Scatter::new(
             new_add,
-            uslices_to_index(&(0..add.info().rank()).map(|i| slices[i]).collect()),
+            uslices_to_index(&(0..add.rank()).map(|i| slices[i]).collect()),
             add.info().shape.clone(),
             None,
         ))
@@ -174,7 +177,7 @@ pub fn add_pull_scatter(add: &Add) -> Option<Scatter> {
 #[pyfunction]
 pub fn scatter_elim_identity(scatter: &Scatter) -> Option<CircuitRc> {
     if scatter.is_identity() {
-        Some(scatter.node().clone())
+        Some(scatter.node())
     } else {
         None
     }
@@ -199,7 +202,7 @@ pub fn index_einsum_to_scatter(node: &Index) -> Option<CircuitRc> {
                 Some(uslice_here) => {
                     if int_slices.contains_key(int) {
                         let new_here = uslice_here.intersection(&int_slices[int]);
-                        if &new_here != &int_slices[int] || &new_here != uslice_here {
+                        if new_here != int_slices[int] || &new_here != uslice_here {
                             did_anything = true;
                             int_slices.insert(*int, new_here);
                         }
@@ -212,13 +215,17 @@ pub fn index_einsum_to_scatter(node: &Index) -> Option<CircuitRc> {
         if !did_anything {
             None
         } else if int_slices.iter().any(|(_i, slice)| slice.length() == 0) {
-            Some(Scalar::new(0.0, node.info().shape.clone(), None).rc())
+            Some(
+                Scalar::try_new(0.0, node.info().shape.clone(), None)
+                    .ok()?
+                    .rc(),
+            )
         } else {
             let new_args = inner
                 .args()
                 .map(|(node, ints)| {
                     (
-                        Index::nrc(node, uslice_map_get_index(&ints, &int_slices), None),
+                        Index::nrc(node, uslice_map_get_index(ints, &int_slices), None),
                         ints.clone(),
                     )
                 })
@@ -226,7 +233,7 @@ pub fn index_einsum_to_scatter(node: &Index) -> Option<CircuitRc> {
             let new_einsum = Einsum::nrc(new_args, inner.out_axes.clone(), inner.info().name);
             let new_shape = containing_uslices
                 .iter()
-                .map(|x| x.unwrap().stop - x.unwrap().start)
+                .map(|x| Size::known(x.unwrap().stop - x.unwrap().start))
                 .collect();
             let new_scatter = Scatter::nrc(
                 new_einsum,
@@ -274,11 +281,9 @@ pub fn scatter_pull_removable_axes(scatter: &Scatter) -> Option<Rearrange> {
     let removable_axes: HashSet<usize> = removable_axes
         .iter()
         .filter(|i| {
-            scatter_uslices[**i]
-                == USlice {
-                    start: 0,
-                    stop: scatter.info().shape[**i],
-                }
+            scatter.info().shape[**i]
+                .t()
+                .is_some_and(|x| scatter_uslices[**i] == USlice { start: 0, stop: x })
         })
         .copied()
         .collect();
@@ -305,26 +310,26 @@ pub fn scatter_pull_removable_axes(scatter: &Scatter) -> Option<Rearrange> {
 }
 
 #[pyfunction]
-pub fn scatter_to_concat(scatter: &Scatter) -> CircuitRc {
-    let mut result = scatter.node().clone();
+pub fn scatter_to_concat(scatter: &Scatter) -> Result<CircuitRc> {
+    let mut result = scatter.node();
     for (i, (idx, l)) in
         zip(scatter.index.all_uslices().unwrap(), &scatter.info().shape).enumerate()
     {
         let mut lower_pad_shape: Shape = result.info().shape.clone();
-        lower_pad_shape[i] = idx.start;
+        lower_pad_shape[i] = Size::known(idx.start);
         let mut upper_pad_shape: Shape = result.info().shape.clone();
         upper_pad_shape[i] = l - idx.stop;
         result = Concat::nrc(
             vec![
-                Scalar::new(0.0, lower_pad_shape, None).rc(),
+                Scalar::try_new(0.0, lower_pad_shape, None)?.rc(),
                 result,
-                Scalar::new(0.0, upper_pad_shape, None).rc(),
+                Scalar::try_new(0.0, upper_pad_shape, None)?.rc(),
             ],
             i,
             scatter.info().name,
         );
     }
-    result
+    Ok(result)
 }
 
 #[pyfunction]
@@ -354,9 +359,9 @@ pub fn concat_to_scatter(concat: &Concat) -> Option<Scatter> {
     if pre_zeros == 0 && post_zeros == 0 || (pre_zeros + post_zeros >= concat.num_children()) {
         return None;
     }
-    let starts = cumsum(&concat.get_sizes_at_axis());
+    let starts = cumsum(shape_try_as_known(&concat.get_sizes_at_axis())?);
     let cslice = TensorAxisIndex::new_plain_slice(starts[pre_zeros], starts[end]);
-    let scatter_index = TensorIndex::new_single(cslice, concat.axis, concat.info().rank())
+    let scatter_index = TensorIndex::new_single(cslice, concat.axis, concat.rank())
         .canonicalize(&concat.info().shape);
     let new_concat = Concat::nrc(
         concat.children_sl()[pre_zeros..end].to_vec(),
