@@ -1,13 +1,18 @@
 // temporary schedule sending setup
 use anyhow::{Context, Result};
-use circuit_base::{parsing::Parser, print::PrintOptions, CircuitNode, IrreducibleNode, Scalar};
+use circuit_base::{
+    self_funcs::{parse_circuit, repr_shape_always},
+    CircuitNode, IrreducibleNode, Scalar,
+};
 use miniserde::{json, Deserialize, Serialize};
 use pyo3::{prelude::*, types::PyByteArray};
 use rr_util::{
     lru_cache::TensorCacheRrfs,
-    py_types::{un_flat_concat, Tensor, PY_UTILS},
-    pycall, sv,
-    tensor_util::{Shape, TorchDevice, TorchDeviceDtype, TorchDtype},
+    py_types::{Tensor, PY_UTILS},
+    pycall,
+    shape::{shape_into_known, Shape},
+    sv,
+    tensor_util::{TorchDevice, TorchDeviceDtype, TorchDtype},
     unwrap,
 };
 use rustc_hash::FxHashMap as HashMap;
@@ -29,51 +34,55 @@ pub struct ScheduleToSend {
     pub instructions: Vec<InstructionToSend>,
     pub constants: ::std::collections::HashMap<usize, String>,
     pub scalars: ::std::collections::HashMap<usize, String>,
-    pub output_circuit: (usize, Vec<usize>),
-    pub split_shapes: Option<Vec<Vec<usize>>>,
+    pub output_circuits: Vec<(usize, Vec<usize>)>,
     pub old_constant_hashes: Vec<(Vec<u8>, usize)>,
 }
 
 #[pymethods]
 impl ScheduleToSend {
-    pub fn evaluate_remote(&self, remote_url: String, device: TorchDevice) -> Option<Tensor> {
-        let self_string = json::to_string(&self);
-        let response = ureq::post(&remote_url).send_string(&self_string).unwrap();
-        let mut body: Vec<u8> = vec![];
-        // dbg!(&response);
-        response.into_reader().read_to_end(&mut body).unwrap();
-        let dtype = unsafe { std::mem::transmute::<u8, TorchDtype>(body[0]) };
-        let body_pybytes: PyObject =
-            Python::with_gil(|py| PyByteArray::new(py, &body[1..]).clone().into());
-        let out_shape = self.output_circuit.1.clone();
-        let count: usize = out_shape.iter().product();
-        pycall!(
-            PY_UTILS.tensor_from_bytes,
-            (
-                TorchDeviceDtype { dtype, device },
-                out_shape,
-                body_pybytes,
-                count,
-            )
-        )
-    }
     pub fn evaluate_remote_many(
         &self,
         remote_url: String,
         device: TorchDevice,
-    ) -> Option<Vec<Tensor>> {
-        self.evaluate_remote(remote_url, device).map(|tensor| {
-            un_flat_concat(
-                &tensor,
-                self.split_shapes
-                    .clone()
-                    .unwrap()
-                    .iter()
-                    .map(|y| y.iter().cloned().collect())
-                    .collect(),
-            )
-            .unwrap()
-        })
+    ) -> Result<Vec<Tensor>> {
+        let self_string = json::to_string(&self);
+        let response = ureq::post(&remote_url)
+            .send_string(&self_string)
+            .context("during evaluate_remote http request")?;
+        let mut body: Vec<u8> = vec![];
+        response.into_reader().read_to_end(&mut body).unwrap();
+        let mut offset = 0usize;
+        self.output_circuits
+            .iter()
+            .map(|circ| {
+                let len = usize::from_le_bytes(body[offset..offset + 8].try_into().unwrap());
+                let dtype = unsafe { std::mem::transmute::<u8, TorchDtype>(body[offset + 8]) };
+                offset += 9;
+                let body_pybytes: PyObject =
+                    Python::with_gil(|py| PyByteArray::new(py, &body[offset..offset + len]).into());
+                offset += len;
+                let out_shape = circ.1.clone();
+                let count: usize = out_shape.iter().product();
+                pycall!(
+                    PY_UTILS.tensor_from_bytes,
+                    (
+                        TorchDeviceDtype { dtype, device },
+                        out_shape,
+                        body_pybytes,
+                        count,
+                    ),
+                    anyhow
+                )
+            })
+            .collect()
+    }
+    pub fn evaluate_remote(&self, remote_url: String, device: TorchDevice) -> Result<Tensor> {
+        let r = self.evaluate_remote_many(remote_url, device)?;
+        anyhow::ensure!(
+            r.len() == 1,
+            "evaluate_remote: got multiple tensors (maybe you want evaluate_remote_many?)"
+        );
+        Ok(r[0].clone())
     }
 }
 
@@ -83,29 +92,25 @@ impl ScheduleToSend {
             instructions: vec![],
             constants: Default::default(),
             scalars: Default::default(),
-            output_circuit: Some((self.output_circuit.0, Scalar::nrc(0.0, sv![], None))),
-            split_shapes: self
-                .split_shapes
-                .map(|z| z.iter().map(|z| z.iter().cloned().collect()).collect()),
+            output_circuits: Some(
+                self.output_circuits
+                    .iter()
+                    .map(|x| (x.0, Scalar::nrc(0.0, sv![], None)))
+                    .collect(),
+            ),
             old_constant_hashes: self
                 .old_constant_hashes
                 .iter()
-                .map(|(b, i)| {
-                    (
-                        b.iter().cloned().collect::<Vec<_>>().try_into().unwrap(),
-                        *i,
-                    )
-                })
+                .map(|(b, i)| (b.to_vec().try_into().unwrap(), *i))
                 .collect(),
         };
 
-        let parser = Parser::default();
         let mut shapes: HashMap<usize, Shape> = <HashMap<usize, Shape> as Default>::default();
         result.constants = self
             .constants
             .iter()
             .map(|(k, v)| {
-                let parsed = parser.parse_circuit(&v, cache)?;
+                let parsed = parse_circuit(v, cache)?;
                 shapes.insert(*k, parsed.info().shape.clone());
                 let irreducible: Option<IrreducibleNode> = (**parsed).clone().into();
                 Ok((*k, ScheduleConstant::Circ(irreducible.unwrap())))
@@ -116,11 +121,7 @@ impl ScheduleToSend {
             .iter()
             .map(|(k, v)| {
                 Ok((*k, {
-                    let resulty = parser
-                        .parse_circuit(&v, cache)?
-                        .as_scalar()
-                        .unwrap()
-                        .clone();
+                    let resulty = parse_circuit(v, cache)?.as_scalar().unwrap().clone();
                     shapes.insert(*k, resulty.info().shape.clone());
                     resulty
                 }))
@@ -134,7 +135,7 @@ impl ScheduleToSend {
                 match v {
                     "Drop" => Ok(Instruction::Drop(ins.key)),
                     "Compute" => Ok(Instruction::Compute(ins.key, {
-                        let result = parser.parse_circuit(&ins.info, cache)?;
+                        let result = parse_circuit(&ins.info, cache)?;
                         shapes.insert(ins.key, result.info().shape.clone());
                         result
                     })),
@@ -159,26 +160,15 @@ impl TryFrom<&Schedule> for ScheduleToSend {
         ))?;
         x.check_no_raw_tensor_constants()
             .context("schedules with already replaced raw tensors can't be sent")?;
-        fn options() -> PrintOptions {
-            PrintOptions {
-                shape_only_when_necessary: false,
-                ..PrintOptions::default()
-            }
-            .validate_ret()
-            .unwrap()
-        }
 
         fn map_to_str<'a, N: CircuitNode + Clone + 'a>(
             map: impl IntoIterator<Item = (&'a usize, &'a N)>,
         ) -> Result<std::collections::HashMap<usize, String>> {
-            let options = options();
             map.into_iter()
-                .map(|(k, v)| Ok((*k, options.repr(v.crc())?)))
+                .map(|(k, v)| Ok((*k, repr_shape_always(v.crc())?)))
                 .collect::<Result<_>>()
         }
 
-        let options = options();
-
         let out = Self {
             instructions: x
                 .instructions
@@ -188,7 +178,7 @@ impl TryFrom<&Schedule> for ScheduleToSend {
                         Instruction::Compute(key, circ) => InstructionToSend {
                             key: *key,
                             variant: "Compute".to_owned(),
-                            info: options.repr(circ.clone())?,
+                            info: repr_shape_always(circ.clone())?,
                             children: get_children_keys(circ.clone()),
                         },
                         Instruction::Drop(key) => InstructionToSend {
@@ -207,26 +197,17 @@ impl TryFrom<&Schedule> for ScheduleToSend {
                     .map(|(h, x)| (h, unwrap!(x, ScheduleConstant::Circ))), /* we checked no raw tensors above */
             )?,
             scalars: map_to_str(&x.scalars)?,
-            output_circuit: (
-                x.output_circuit.clone().unwrap().0,
-                x.output_circuit
-                    .clone()
-                    .unwrap()
-                    .1
-                    .info()
-                    .shape
-                    .iter()
-                    .cloned()
-                    .collect(),
-            ),
-            split_shapes: x
-                .split_shapes
-                .clone()
-                .map(|z| z.iter().map(|y| y.iter().cloned().collect()).collect()),
+            output_circuits: x
+                .output_circuits
+                .as_ref()
+                .unwrap()
+                .iter()
+                .map(|c| (c.0, shape_into_known(c.1.info().shape.clone())))
+                .collect(),
             old_constant_hashes: x
                 .old_constant_hashes
                 .iter()
-                .map(|(b, i)| (b.iter().cloned().collect::<Vec<_>>(), *i))
+                .map(|(b, i)| (b.to_vec(), *i))
                 .collect(),
         };
         Ok(out)
