@@ -1,19 +1,22 @@
 use std::{
     collections::hash_map::Entry,
     fmt::{self, Display},
+    iter::zip,
 };
 
 use anyhow::{anyhow, bail, Context, Result};
 use circuit_base::{
     cached_circuit_properties::max_non_leaf_size,
     circuit_utils::toposort_circuit,
-    flat_concat,
-    module::{any_children_with_symbolic_sizes, substitute_all_modules},
+    generalfunction::GeneralFunctionOutputSpec,
+    module::{
+        any_children_with_symbolic_sizes, any_modules, conform_all_modules, substitute_all_modules,
+    },
     prelude::*,
-    print::{oom_fmt, PrintOptions},
-    visit_circuit_non_free, Array, IrreducibleNode, Leaf, ModuleSpec, Scalar, Symbol,
+    visit_circuit_non_free, Array, GeneralFunction, GeneralFunctionSpec, IrreducibleNode, Leaf,
+    ModuleSpec, Scalar, Symbol,
 };
-use circuit_utils::{hash_to_node_non_free, is_definitely_view_on_child};
+use circuit_utils::{biggest_non_leaf, hash_to_node_non_free, is_definitely_view_on_child};
 use itertools::Itertools;
 use macro_rules_attribute::apply;
 use miniserde::json;
@@ -21,12 +24,14 @@ use num_bigint::BigUint;
 use pyo3::{exceptions::PyValueError, prelude::*, types::PyBytes};
 use rr_util::{
     lru_cache::TensorCacheRrfs,
+    print::oom_fmt,
     py_types::{
-        is_python_running, scalar_to_tensor, tensor_scale, un_flat_concat, ExtraPySelfOps, Tensor,
-        PY_UTILS,
+        is_python_running, scalar_to_tensor, tensor_scale, ExtraPySelfOps, Tensor, PY_UTILS,
     },
-    pycall, python_error_exception, sv,
-    tensor_util::{Shape, TorchDevice},
+    pycall, python_error_exception,
+    shape::Shape,
+    sv,
+    tensor_util::{TorchDevice, TorchDeviceDtype},
     timed, unwrap,
     util::{arc_ref_clone, hash_to_hex, with_context_failable, HashBytes},
 };
@@ -34,6 +39,7 @@ use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use thiserror::Error;
 
 use crate::{
+    circuit_evaluate::eval_tensors,
     circuit_optimizer::{OptimizationContext, OptimizationSettings},
     module_rewrite::deep_module_remove_unused_inputs,
     schedule_send::ScheduleToSend,
@@ -106,6 +112,12 @@ impl ScheduleConstant {
             Self::Tensor(tensor) => tensor.shape(),
         }
     }
+    pub fn device_dtype(&self) -> TorchDeviceDtype {
+        match self {
+            ScheduleConstant::Circ(c) => c.info().device_dtype.unwrap_or_defaults(),
+            ScheduleConstant::Tensor(t) => t.device_dtype,
+        }
+    }
 }
 
 /// Schedule is an optimized sequence of operations to perform to compute a circuit
@@ -125,8 +137,7 @@ pub struct Schedule {
     // adjustment needed
     #[pyo3(get)]
     pub scalars: HashMap<usize, Scalar>,
-    pub output_circuit: Option<(usize, CircuitRc)>,
-    pub split_shapes: Option<Vec<Shape>>,
+    pub output_circuits: Option<Vec<(usize, CircuitRc)>>,
     pub old_constant_hashes: HashMap<HashBytes, usize>,
 }
 
@@ -164,9 +175,15 @@ impl Schedule {
         }
 
         if validate_output {
-            let output_id = self.output_circuit.as_ref().ok_or(anyhow!("no output!"))?.0;
+            let output_ids: HashSet<usize> = self
+                .output_circuits
+                .as_ref()
+                .ok_or(anyhow!("no output!"))?
+                .iter()
+                .map(|x| x.0)
+                .collect();
             for computed in prior_compute_numbers {
-                if !prior_drop_numbers.contains(computed) && *computed != output_id {
+                if !prior_drop_numbers.contains(computed) && !output_ids.contains(computed) {
                     bail!("computed={computed} is never dropped (or returned)");
                 }
             }
@@ -226,17 +243,21 @@ impl Schedule {
         Ok(result)
     }
     #[pyo3(signature=(settings = Default::default()))]
-    pub fn evaluate(&self, settings: OptimizationSettings) -> Result<Tensor> {
+    pub fn evaluate_many(&self, settings: OptimizationSettings) -> Result<Vec<Tensor>> {
         let timed = settings.verbose >= 1;
         let eval = || -> Result<_> {
-            let result = if settings.adjust_numerical_scale {
-                evaluate_schedule_adjust_numerical_scale(self, settings)?
-                    [&self.output_circuit.as_ref().unwrap().0]
-                    .clone()
+            let v = if settings.adjust_numerical_scale {
+                evaluate_schedule_adjust_numerical_scale(self, settings)
             } else {
-                evaluate_schedule(self)?[&self.output_circuit.as_ref().unwrap().0].clone()
-            };
-            Ok(result)
+                evaluate_schedule(self)
+            }?;
+            Ok(self
+                .output_circuits
+                .as_ref()
+                .unwrap()
+                .iter()
+                .map(|(id, _)| v[id].clone())
+                .collect())
         };
         Ok(timed!(eval()?, 10, timed))
     }
@@ -268,13 +289,12 @@ impl Schedule {
     }
 
     #[pyo3(signature=(settings = Default::default()))]
-    pub fn evaluate_many(&self, settings: OptimizationSettings) -> Result<Vec<Tensor>> {
-        let single = self.evaluate(settings)?;
-        Ok(self.split(single))
-    }
-
-    pub fn split(&self, tensor: Tensor) -> Vec<Tensor> {
-        un_flat_concat(&tensor, self.split_shapes.clone().unwrap()).unwrap()
+    pub fn evaluate(&self, settings: OptimizationSettings) -> Result<Tensor> {
+        let many = self.evaluate_many(settings)?;
+        if many.len() != 1 {
+            bail!("Schedule.evaluate but schedule returned more than one tensor (did you mean to use evaluate_many?)")
+        }
+        Ok(many[0].clone())
     }
 
     pub fn get_stats(&self) -> ScheduleStats {
@@ -286,11 +306,11 @@ impl Schedule {
             match instruction {
                 Instruction::Drop(drop) => {
                     let dropped = current.remove(&drop).unwrap();
-                    mem -= dropped.info().numel();
+                    mem -= dropped.info().naive_mem_use(None);
                 }
                 Instruction::Compute(key, circuit) => {
                     current.insert(key, circuit.clone());
-                    mem += circuit.info().numel();
+                    mem += circuit.info().naive_mem_use(None);
                     if mem > max_mem {
                         max_mem = mem.clone();
                         biggest = current.clone();
@@ -302,8 +322,11 @@ impl Schedule {
             max_mem,
             constant_mem: self
                 .constants
-                .iter()
-                .map(|(_h, t)| t.shape().iter().product::<usize>())
+                .values()
+                .map(|t| {
+                    t.shape().iter().map(|x| x.unwrap()).product::<usize>()
+                        * t.device_dtype().size()
+                })
                 .sum(),
             max_circuit_set: biggest.values().cloned().collect(),
         }
@@ -339,15 +362,12 @@ impl Schedule {
         self.try_into()
     }
 
-    pub fn evaluate_remote(&self, remote_url: String) -> Result<Option<Tensor>> {
-        let out = self.tosend()?.evaluate_remote(remote_url, TorchDevice::Cpu);
-        Ok(out)
+    pub fn evaluate_remote(&self, remote_url: String) -> Result<Tensor> {
+        self.tosend()?.evaluate_remote(remote_url, TorchDevice::Cpu)
     }
-    pub fn evaluate_remote_many(&self, remote_url: String) -> Result<Option<Vec<Tensor>>> {
-        let out = self
-            .tosend()?
-            .evaluate_remote_many(remote_url, TorchDevice::Cpu);
-        Ok(out)
+    pub fn evaluate_remote_many(&self, remote_url: String) -> Result<Vec<Tensor>> {
+        self.tosend()?
+            .evaluate_remote_many(remote_url, TorchDevice::Cpu)
     }
     fn __repr__(&self) -> String {
         format!("{}", &self)
@@ -403,8 +423,8 @@ impl Display for Schedule {
                 .collect::<Vec<String>>()
                 .join(" "),
             self.scalars
-                .iter()
-                .map(|(k, _c)| k.to_string())
+                .keys()
+                .map(|k| k.to_string())
                 .collect::<Vec<String>>()
                 .join(" "),
         )
@@ -422,31 +442,29 @@ pub struct ScheduleStats {
     max_circuit_set: HashSet<CircuitRc>,
 }
 
+static SCHEDULING_ERROR_NOTE: &'static str = "(see circuit_to_schedule docstring for tips)";
+
 #[apply(python_error_exception)]
 #[base_error_name(SchedulingOOM)]
 #[base_exception(PyValueError)]
 #[derive(Error, Clone, Debug)]
 pub enum SchedulingOOMError {
-    #[error("{string} ({e_name})")]
-    Many {
+    #[error("Schedule doesn't fit into max_memory={max_mem}, {stats} ({e_name}) {SCHEDULING_ERROR_NOTE}", max_mem=oom_fmt(*max_memory))]
+    OverMemoryLimit {
         max_memory: usize,
-        memory_chunks: usize,
-        node_memories: Vec<usize>,
-        string: String,
+        stats: ScheduleStats,
+    },
+    #[error("Single {} element doesn't fit max_single_tensor_memory={} max_memory={} {circuit:?} ({e_name})", oom_fmt(mem_usage.clone()), oom_fmt(*max_single_tensor_memory), oom_fmt(*max_memory))]
+    Single {
+        max_single_tensor_memory: usize,
+        max_memory: usize,
+        mem_usage: BigUint,
+        circuit: CircuitRc,
     },
-    #[error("Single element doesn't fit {formatted} ({e_name})")]
-    Single { numel: BigUint, formatted: String },
-    #[error("Simplification can't find anything within bounds, likely each isolated subdag fits but together they don't. maybe try scheduling_simplify=False opt setting graph {cyto} ({e_name})")]
-    Simp { cyto: String },
     #[error("Threading failed! bug! ({e_name})")]
     ThreadsLost {},
-
-    #[error("Schedule exhaustive timed out, at size {size} iters {iters} graph (click to see circuit structure) {cyto} ({e_name})")]
-    ExhaustiveTimeout {
-        cyto: String,
-        iters: usize,
-        size: usize,
-    },
+    #[error("BUG: Schedule exhaustive timed out, at size {size} iters {iters}. maybe your circuit is gigantic (exabytes mem usage)? ({e_name}) {SCHEDULING_ERROR_NOTE}")]
+    ExhaustiveTimeout { iters: usize, size: usize },
 }
 
 #[apply(python_error_exception)]
@@ -463,21 +481,19 @@ pub enum SchedulingError {
 
 impl Display for ScheduleStats {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        let mut shapes = self
+        let shapes_and_percents = self
             .max_circuit_set
             .iter()
-            .map(|x| x.info().shape.clone())
-            .collect::<Vec<Shape>>();
-        shapes.sort_by_key(|x| std::cmp::Reverse(x.iter().product::<usize>()));
-        let shapes_and_percents: String = shapes
-            .iter()
-            .map(|x| {
+            .map(|x| (x.info().shape.clone(), x.info().naive_mem_use_usize(None)))
+            .counts()
+            .into_iter()
+            .sorted_by_key(|((_, mem), count)| std::cmp::Reverse(count * mem))
+            .map(|(x, c)| {
                 format!(
-                    "{:?} {}%",
-                    x,
+                    "{c} {:?} {}%",
+                    x.0,
                     // biguint doesn't have cast f64? that would be less lossy than truncate
-                    (x.iter().product::<usize>() as f64 / self.max_mem.to_u64_digits()[0] as f64
-                        * 100.0) as i64
+                    ((x.1 * c) as f64 / self.max_mem.to_u64_digits()[0] as f64 * 100.0) as i64
                 )
             })
             .collect::<Vec<String>>()
@@ -503,8 +519,12 @@ fn get_child_key(circuit: CircuitRc) -> usize {
         .unwrap()
 }
 
-fn child_from_key(key: usize, shape: Shape) -> Symbol {
-    Symbol::new_with_random_uuid(shape, Some(key.to_string().into()))
+fn child_from_key(key: usize, child: CircuitRc) -> Symbol {
+    Symbol::new_device_dtype_random_uuid(
+        child.info().shape.clone(),
+        Some(key.to_string().into()),
+        child.info().device_dtype,
+    )
 }
 
 pub fn get_children_keys(circuit: CircuitRc) -> Vec<usize> {
@@ -535,7 +555,7 @@ pub fn evaluate_schedule(schedule: &Schedule) -> Result<HashMap<usize, Tensor>>
         schedule
             .scalars
             .iter()
-            .map(|x| (*x.0, x.1.eval_tensors(&[]).unwrap())),
+            .map(|x| (*x.0, eval_tensors(&x.1.crc(), &[]).unwrap())),
     );
     for s in &schedule.instructions {
         let run = |_py: Python| match s {
@@ -550,7 +570,7 @@ pub fn evaluate_schedule(schedule: &Schedule) -> Result<HashMap<usize, Tensor>>
                     .iter()
                     .map(|child_key| live[child_key].clone())
                     .collect();
-                let result_err = circ.eval_tensors(&tensors);
+                let result_err = eval_tensors(circ, &tensors);
                 if result_err.is_err() {
                     println!("errored evaluate");
                     circ.print().unwrap()
@@ -657,12 +677,13 @@ pub fn evaluate_schedule_adjust_numerical_scale(
                 let result = match &***circ {
                     Circuit::Einsum(_) => {
                         let new_scale = tensors_and_scales.iter().map(|x| x.1).product();
-                        clamp(&(circ.eval_tensors(&tensors).unwrap(), new_scale))
+                        clamp(&(eval_tensors(circ, &tensors).unwrap(), new_scale))
                     }
                     Circuit::Add(_) | Circuit::Concat(_) => {
                         let new_ts = uniformize(&tensors_and_scales);
                         (
-                            circ.eval_tensors(
+                            eval_tensors(
+                                circ,
                                 &new_ts.iter().map(|x| x.0.clone()).collect::<Vec<_>>(),
                             )
                             .unwrap(),
@@ -675,7 +696,8 @@ pub fn evaluate_schedule_adjust_numerical_scale(
                             .map(|x| set_scale(x, 1.0))
                             .collect();
                         clamp(&(
-                            circ.eval_tensors(
+                            eval_tensors(
+                                circ,
                                 &new_ts.iter().map(|x| x.0.clone()).collect::<Vec<_>>(),
                             )
                             .unwrap(),
@@ -683,7 +705,8 @@ pub fn evaluate_schedule_adjust_numerical_scale(
                         ))
                     }
                     Circuit::Index(_) | Circuit::Rearrange(_) | Circuit::Scatter(_) => (
-                        circ.eval_tensors(
+                        eval_tensors(
+                            circ,
                             &tensors_and_scales
                                 .iter()
                                 .map(|x| x.0.clone())
@@ -722,8 +745,9 @@ pub fn order_to_schedule(
     order: &Vec<CircuitRc>,
     constants: &Vec<IrreducibleNode>,
     scalars: &Vec<Scalar>,
-    to_keep: HashSet<HashBytes>,
+    to_keep: &[HashBytes],
 ) -> (Schedule, Vec<usize>) {
+    let to_keep_set: HashSet<HashBytes> = to_keep.iter().cloned().collect();
     let mut circ_to_id: HashMap<HashBytes, usize> = Default::default();
     let constants: HashMap<usize, _> = constants
         .iter()
@@ -741,8 +765,7 @@ pub fn order_to_schedule(
                 (circ_to_id.len() - 1, x.clone())
             })
             .collect(),
-        split_shapes: None,
-        output_circuit: None,
+        output_circuits: None,
         old_constant_hashes: constants
             .iter()
             .map(|(id, node)| (node.info().hash, *id))
@@ -759,7 +782,7 @@ pub fn order_to_schedule(
             let dep = *circ_to_id.entry(child.info().hash).or_insert(next_id);
             if !Leaf::matches(&child)
                 && seen_dependencies.insert(dep)
-                && !to_keep.contains(&child.info().hash)
+                && !to_keep_set.contains(&child.info().hash)
             {
                 result.instructions.push(Instruction::Drop(dep));
             }
@@ -767,7 +790,7 @@ pub fn order_to_schedule(
         let next_id = circ_to_id.len();
         let our_id = *circ_to_id.entry(ex.info().hash).or_insert(next_id);
         let node_here_symbol_children = (**ex.map_non_free_children_unwrap(|child| {
-            child_from_key(circ_to_id[&child.info().hash], child.info().shape.clone()).rc()
+            child_from_key(circ_to_id[&child.info().hash], child).rc()
         }))
         .clone()
         .rc();
@@ -780,7 +803,7 @@ pub fn order_to_schedule(
     (result, to_keep.iter().map(|x| circ_to_id[x]).collect())
 }
 
-pub fn circuits_to_dag(circuits: &[CircuitRc]) -> Dag {
+pub fn circuit_to_dag(circuit: CircuitRc) -> Dag {
     let mut result: Dag = Default::default();
     let mut to_merge_with_child: Vec<u32> = vec![];
     // append the circuit to the dag if it's not already there, and return its index
@@ -797,53 +820,47 @@ pub fn circuits_to_dag(circuits: &[CircuitRc]) -> Dag {
         result.parents.push(sv![]);
         let output = (result.node_hashes.len() - 1) as u32;
 
-        if is_definitely_view_on_child(c.clone()) {
+        if is_definitely_view_on_child(c) {
             to_merge_with_child.push(output)
         }
         output
     };
-    for circuit in circuits {
-        // non free so we avoid recuring into module spec.circuit
-        visit_circuit_non_free(
-            circuit.clone(),
-            |c: CircuitRc| {
-                // Arrays are never added to the dag to begin with because they're always 0 cost
+    // non free so we avoid recuring into module spec.circuit
+    visit_circuit_non_free(
+        circuit,
+        |c: CircuitRc| {
+            // Arrays are never added to the dag to begin with because they're always 0 cost
 
-                if !c.is_leaf() {
-                    let my_number: u32 =
-                        number_node(c.clone(), &mut result, &mut to_merge_with_child);
+            if !c.is_leaf() {
+                let my_number: u32 = number_node(c.clone(), &mut result, &mut to_merge_with_child);
 
-                    let children_to_consider: Vec<CircuitRc> = c
-                        .non_free_children()
-                        .filter(|child| !Leaf::matches(child))
-                        .collect();
-                    result.children[my_number as usize] = children_to_consider
+                let children_to_consider: Vec<CircuitRc> = c
+                    .non_free_children()
+                    .filter(|child| !Leaf::matches(child))
+                    .collect();
+                result.children[my_number as usize] = children_to_consider
+                    .iter()
+                    .map(|child| number_node(child.clone(), &mut result, &mut to_merge_with_child))
+                    .unique()
+                    .collect();
+                for child in children_to_consider {
+                    let new_num = number_node(child.clone(), &mut result, &mut to_merge_with_child);
+                    if !result.parents[new_num as usize]
                         .iter()
-                        .map(|child| {
-                            number_node(child.clone(), &mut result, &mut to_merge_with_child)
-                        })
-                        .unique()
-                        .collect();
-                    for child in children_to_consider {
-                        let new_num =
-                            number_node(child.clone(), &mut result, &mut to_merge_with_child);
-                        if !result.parents[new_num as usize]
-                            .iter()
-                            .contains(&(my_number as u32))
-                        {
-                            result.parents[new_num as usize].push(my_number as u32);
-                        }
+                        .contains(&(my_number as u32))
+                    {
+                        result.parents[new_num as usize].push(my_number as u32);
                     }
                 }
+            }
 
-                Ok(())
-            },
-            false,
-        )
-        .unwrap();
-    }
+            Ok(())
+        },
+        false,
+    )
+    .unwrap();
     result.node_to_orig = (0..result.node_costs.len() as u32)
-        .map(|x| (x, sv![x].clone()))
+        .map(|x| (x, sv![x]))
         .collect();
     for tmp in to_merge_with_child {
         assert!(result.children[tmp as usize].len() < 2);
@@ -867,96 +884,97 @@ pub fn circuit_to_schedule(
         true,
         true,
     );
+    let circuit = conform_all_modules(circuit);
     if is_python_running() && any_children_with_symbolic_sizes(circuit.clone()) {
         Python::with_gil(|py| {
             pyo3::PyErr::warn(
                 py,
                 PY_UTILS.optimizing_symbolic_size_warning.as_ref(py),
-                concat!(
-                    "you're scheduling with symbolic sizes.",
-                    " Perhaps you need to conform_all_modules?",
-                    " (this is typically done in optimize_circuit)."
-                ),
+                concat!("you're scheduling with symbolic sizes. This is not recommended!",),
+                1,
+            )
+        })?;
+    }
+    if is_python_running() && any_modules(circuit.clone()) {
+        Python::with_gil(|py| {
+            pyo3::PyErr::warn(
+                py,
+                PY_UTILS.optimizing_modules_warning.as_ref(py),
+                "you're scheduling a circuit with modules. this is currently buggy, you might want to substitute_all_modules first",
                 1,
             )
         })?;
     }
 
-    let mut dag = circuits_to_dag(&[circuit.clone()]);
-    if max_non_leaf_size(circuit.clone()) > BigUint::from(context.settings.max_single_tensor_memory)
+    if let Some((circ, mem)) = biggest_non_leaf(circuit.clone()) && mem > BigUint::from(context.settings.max_single_tensor_memory.min(context.settings.max_memory))
     {
         bail!(SchedulingOOMError::Single {
-            numel: max_non_leaf_size(circuit.clone()),
-            formatted: oom_fmt(max_non_leaf_size(circuit.clone()))
+            max_single_tensor_memory: context.settings.max_single_tensor_memory,
+            max_memory: context.settings.max_memory,
+            mem_usage: max_non_leaf_size(circuit.clone()),
+            circuit: circ
         });
     }
+    let mut dag = circuit_to_dag(circuit.clone());
     let order_result = {
-        let mut result: Result<Vec<u32>, SchedulingOOMError> = Err(SchedulingOOMError::Many {
-            max_memory: 0,
-            memory_chunks: 0,
-            node_memories: vec![],
-            string: "".to_owned(),
-        });
-        let mut mem = context.settings.max_memory;
-        while mem
-            <= context
-                .settings
-                .max_memory_fallback
-                .unwrap_or(context.settings.max_memory)
-            && result.is_err()
-        {
-            if context.settings.scheduling_simplify {
-                let mut dag_simp_settings: DagSimpSettings = Default::default();
-                dag_simp_settings.acceptable_subdag_max = context.settings.max_memory;
-                dag_simp_settings.mem_limit = mem;
-                dag_simp_settings.num_mem_chunks = context.settings.scheduling_num_mem_chunks;
-                dag_simp_settings.timeout = context.settings.scheduling_timeout;
-                dag_simp_settings.verbose = context.settings.verbose;
-                dag_simp_settings.parallelism = context.settings.optimization_parallelism;
-                timed!(
-                    dag.simplify(&dag_simp_settings)?,
-                    10,
-                    context.settings.verbose >= 2
-                );
-            }
-            result = dag.compute_schedule(
-                context.settings.verbose,
-                mem,
-                context.settings.scheduling_num_mem_chunks,
-                context.settings.scheduling_timeout,
+        let mut dag_simp_settings: DagSimpSettings = Default::default();
+        dag_simp_settings.verbose = context.settings.verbose;
+        dag_simp_settings.parallelism = context.settings.optimization_parallelism;
+        dag_simp_settings.exhaustive_give_up_ns = context.settings.scheduling_timeout * 1_000_000;
+        dag_simp_settings.mem_limit = context.settings.max_memory;
+        if context.settings.scheduling_simplify {
+            timed!(
+                dag.simplify(&dag_simp_settings)?,
+                10,
+                context.settings.verbose >= 2
             );
-            mem *= 2;
         }
-        result
+        timed!(
+            dag.compute_schedule(&dag_simp_settings),
+            10,
+            context.settings.verbose >= 2
+        )
     };
     let order = with_context_failable(order_result, || {
-        Ok(format!(
-            "getting schedule failed for circuit:\n{}",
-            PrintOptions::compiler_default().repr(circuit.clone())?
-        ))
+        Ok(format!("getting schedule failed for circuit:\n{circuit:?}",))
     })?;
 
     let to_node = hash_to_node_non_free(circuit.clone(), false);
-    let circuit_order: Vec<CircuitRc> = order
+    let mut circuit_order: Vec<CircuitRc> = order
         .iter()
         .map(|x| to_node[&dag.node_hashes[*x as usize]].clone())
         .collect();
 
+    let circuits = if let Circuit::GeneralFunction(g) = &**circuit
+                   && let GeneralFunctionSpec::Output(_) = g.spec {
+        circuit_order.pop();
+        g.children_sl().to_vec()
+    } else {
+        vec![circuit.clone()]
+    };
     let (mut out, kept_keys) = order_to_schedule(
         &circuit_order,
         &to_node
             .iter()
-            .filter_map(|x| Option::<IrreducibleNode>::from(arc_ref_clone(&x.1)))
+            .filter_map(|x| Option::<IrreducibleNode>::from(arc_ref_clone(x.1)))
             .collect(),
         &to_node
             .iter()
             .filter_map(|x| x.1.as_scalar().cloned())
             .collect(),
-        HashSet::from_iter([circuit.info().hash]),
+        &circuits.iter().map(|x| x.info().hash).collect::<Vec<_>>(),
     );
-    out.output_circuit = Some((kept_keys[0], circuit));
-    let out = replace_module_schedules(&out, context);
-    out.validate(true).expect("should be valid");
+    out.output_circuits = Some(zip(kept_keys, circuits.clone()).collect());
+    let out = replace_module_schedules(&out, context)?;
+    out.validate(true)?;
+    let stats = out.get_stats();
+    // scheduling simplify & exhaustive currently overestimates mem usage by a lot (todo debug?), so we currently don't error during scheduling & just recheck & error here. also this is nice since it tells user schedule memory usage & takes modules into account!
+    if stats.max_mem > context.settings.max_memory.into() {
+        bail!(SchedulingOOMError::OverMemoryLimit {
+            max_memory: context.settings.max_memory,
+            stats: stats
+        })
+    }
     Ok(out)
 }
 
@@ -977,17 +995,34 @@ pub fn circuit_to_schedule_naive_toposort(circuit: CircuitRc) -> Result<Schedule
         }
     }
 
+    let circuits = if let Circuit::GeneralFunction(g) = &**circuit
+                   && let GeneralFunctionSpec::Output(_) = g.spec {
+        order.pop();
+        g.children_sl().to_vec()
+    } else {
+        vec![circuit]
+    };
     let (mut result, kept_keys) = order_to_schedule(
         &order,
         &constants,
         &scalars,
-        HashSet::from_iter([circuit.info().hash]),
+        &circuits.iter().map(|x| x.info().hash).collect::<Vec<_>>(),
     );
-    result.output_circuit = Some((kept_keys[0], circuit));
+    result.output_circuits = Some(zip(kept_keys, circuits).collect());
     result.validate(true).expect("should be valid");
     Ok(result)
 }
 
+pub fn schedule_out_many(circuits: Vec<CircuitRc>) -> CircuitRc {
+    GeneralFunction::try_new(
+        circuits,
+        GeneralFunctionSpec::Output(GeneralFunctionOutputSpec()),
+        None,
+    )
+    .unwrap()
+    .crc()
+}
+
 #[pyfunction]
 pub fn scheduled_evaluate(circuit: CircuitRc, settings: OptimizationSettings) -> Result<Tensor> {
     let schedule = if settings.scheduling_naive {
@@ -995,10 +1030,16 @@ pub fn scheduled_evaluate(circuit: CircuitRc, settings: OptimizationSettings) ->
     } else {
         circuit_to_schedule(
             circuit.clone(),
-            &mut OptimizationContext::new_settings(settings),
+            &mut OptimizationContext::new_settings(settings.clone()),
         )
     }?;
-    Ok(evaluate_schedule(&schedule)?[&schedule.output_circuit.unwrap().0].clone())
+    schedule.evaluate(settings)
+}
+
+#[pyfunction]
+pub fn scheduled_evaluate_naive(circuit: CircuitRc) -> Result<Tensor> {
+    let schedule = circuit_to_schedule_naive_toposort(circuit)?;
+    schedule.evaluate(Default::default())
 }
 
 #[pyfunction]
@@ -1007,10 +1048,7 @@ pub fn py_circuit_to_schedule(
     circuit: CircuitRc,
     settings: OptimizationSettings,
 ) -> Result<Schedule> {
-    circuit_to_schedule(
-        circuit.clone(),
-        &mut OptimizationContext::new_settings(settings),
-    )
+    circuit_to_schedule(circuit, &mut OptimizationContext::new_settings(settings))
 }
 
 #[pyfunction]
@@ -1019,35 +1057,33 @@ pub fn py_circuit_to_schedule_many(
     circuits: Vec<CircuitRc>,
     settings: OptimizationSettings,
 ) -> Result<Schedule> {
-    let flat_concatted = flat_concat(circuits.clone())?.rc();
-    let mut schedule = py_circuit_to_schedule(flat_concatted, settings)?;
-    schedule.split_shapes = Some(
-        circuits
-            .iter()
-            .map(|x| x.info().shape.clone())
-            .collect::<Vec<Shape>>(),
-    );
-    Ok(schedule)
+    py_circuit_to_schedule(schedule_out_many(circuits), settings)
 }
 
-fn schedule_module_spec(spec: &ModuleSpec, context: &mut OptimizationContext) -> Schedule {
+fn schedule_module_spec(spec: &ModuleSpec, context: &mut OptimizationContext) -> Result<Schedule> {
     context
         .cache
         .module_specs_scheduled_same_settings
         .get(spec)
         .cloned()
-        .unwrap_or_else(|| {
-            let result = circuit_to_schedule(spec.circuit.clone(), context).unwrap();
-            context
-                .cache
-                .module_specs_scheduled_same_settings
-                .insert(spec.clone(), result.clone());
-            result
-        })
+        .map_or_else(
+            || {
+                let result = circuit_to_schedule(spec.circuit.clone(), context)?;
+                context
+                    .cache
+                    .module_specs_scheduled_same_settings
+                    .insert(spec.clone(), result.clone());
+                Ok(result)
+            },
+            |x| Ok(x),
+        )
 }
 
 /// note: this doesn't allow for any tensors in constants
-fn replace_module_schedules(schedule: &Schedule, context: &mut OptimizationContext) -> Schedule {
+fn replace_module_schedules(
+    schedule: &Schedule,
+    context: &mut OptimizationContext,
+) -> Result<Schedule> {
     let mut result = schedule.clone();
     let mut scalar_to_id: HashMap<Scalar, usize> = schedule
         .scalars
@@ -1073,7 +1109,8 @@ fn replace_module_schedules(schedule: &Schedule, context: &mut OptimizationConte
                     let original_children: Vec<usize> =
                         mn.args().cloned().map(get_child_key).collect();
 
-                    let inner_schedule = schedule_module_spec(&mn.spec, context);
+                    let inner_schedule = schedule_module_spec(&mn.spec, context)
+                        .with_context(|| format!("while scheduling Module spec {mn:?}"))?;
                     let mut inner_to_outer_key = HashMap::default();
                     for (inner_k, sc) in &inner_schedule.scalars {
                         let outer = scalar_to_id.get(sc).cloned().unwrap_or_else(|| {
@@ -1105,6 +1142,10 @@ fn replace_module_schedules(schedule: &Schedule, context: &mut OptimizationConte
                                             this_key,
                                             ScheduleConstant::Circ(irreducible.clone()),
                                         );
+                                        assert!(result
+                                            .old_constant_hashes
+                                            .insert(irreducible.info().hash, this_key)
+                                            .is_none());
                                         next_key += 1;
                                         this_key
                                     }
@@ -1119,7 +1160,13 @@ fn replace_module_schedules(schedule: &Schedule, context: &mut OptimizationConte
 
                     inner_to_outer_key
                         .try_insert(
-                            inner_schedule.output_circuit.clone().unwrap().0,
+                            inner_schedule
+                                .output_circuits
+                                .unwrap()
+                                .iter()
+                                .exactly_one()
+                                .unwrap()
+                                .0,
                             *outermost_k,
                         )
                         .unwrap();
@@ -1138,8 +1185,7 @@ fn replace_module_schedules(schedule: &Schedule, context: &mut OptimizationConte
                                         let child_inner_k = get_child_key(child.clone());
                                         let child_outer_k =
                                             inner_to_outer_key.get(&child_inner_k).unwrap();
-                                        child_from_key(*child_outer_k, child.info().shape.clone())
-                                            .rc()
+                                        child_from_key(*child_outer_k, child).rc()
                                     })
                                     .rc();
                                 let key =
@@ -1162,7 +1208,7 @@ fn replace_module_schedules(schedule: &Schedule, context: &mut OptimizationConte
             },
         }
     }
-    result.validate(result.output_circuit.is_some()).unwrap();
+    result.validate(result.output_circuits.is_some())?;
 
-    result
+    Ok(result)
 }
