@@ -1,26 +1,29 @@
 use std::{
     cmp::max,
-    sync::{atomic::AtomicBool, mpsc::channel, Arc},
+    sync::{
+        atomic::{AtomicBool, AtomicUsize, Ordering},
+        mpsc::channel,
+        Arc,
+    },
+    time::Instant,
 };
 
 use anyhow::{bail, Result};
 use itertools::Itertools;
 use rand::{rngs::SmallRng, Rng, SeedableRng};
-use rr_util::{python_println, sv, timed, timed_value, util::HashBytes};
+use rr_util::{print::oom_fmt, python_println, sv, timed, timed_value, util::HashBytes};
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use smallvec::SmallVec as Sv;
 
-use crate::{scheduled_execution::SchedulingOOMError, scheduling_z3::schedule_dag_strategy_ints};
+use crate::scheduled_execution::SchedulingOOMError;
+
 #[derive(Clone, Debug)]
 pub struct DagSimpSettings {
     pub same_ratio: f64,
     pub tiny_size: usize,
-    pub acceptable_subdag_max: usize,
     pub exhaustive_settle_ns: usize,
     pub exhaustive_give_up_ns: usize,
     pub mem_limit: usize,
-    pub num_mem_chunks: usize,
-    pub timeout: usize,
     pub verbose: usize,
     pub parallelism: Option<usize>,
 }
@@ -30,13 +33,10 @@ impl Default for DagSimpSettings {
         Self {
             same_ratio: 0.25,
             tiny_size: 200_000,
-            acceptable_subdag_max: 100_000_000,
-            exhaustive_settle_ns: 2_000_000,
-            exhaustive_give_up_ns: 400_000_000,
+            mem_limit: 100_000_000, // overridden by OptimizationSettings.max_memory
+            exhaustive_settle_ns: 8_000_000,
+            exhaustive_give_up_ns: 400_000_000, /* overridden by OptimizationSettings.scheduling_timeout */
             verbose: 0,
-            timeout: 5_000, // useless defaults, todo fix
-            num_mem_chunks: 200,
-            mem_limit: 10,
             parallelism: Some(8),
         }
     }
@@ -59,7 +59,7 @@ impl Dag {
         result
     }
 
-    pub fn expand_order(&self, order: &Vec<u32>) -> Vec<u32> {
+    pub fn expand_order(&self, order: Vec<u32>) -> Vec<u32> {
         self.pre_root_nodes
             .iter()
             .chain(order.iter().flat_map(|x| self.node_to_orig[x].iter()))
@@ -75,40 +75,20 @@ impl Dag {
             .unwrap_or(0)
     }
 
-    pub fn compute_schedule_top_ints(
-        &self,
-        verbose: usize,
-        mem_limit: usize,
-        num_mem_chunks: usize,
-        timeout: usize,
-        required_first: Option<usize>,
-    ) -> Result<Vec<u32>, SchedulingOOMError> {
-        if self.node_to_orig.len() == 1 {
-            return Ok(self.node_to_orig.keys().cloned().collect::<Vec<_>>());
-        }
-        schedule_dag_strategy_ints(
-            &self,
-            verbose,
-            mem_limit,
-            num_mem_chunks,
-            timeout,
-            required_first,
-        )
-        .map(|x| x.into_iter().map(|x| x as u32).collect())
-    }
-
-    pub fn compute_schedule(
-        &self,
-        verbose: usize,
-        mem_limit: usize,
-        num_mem_chunks: usize,
-        timeout: usize,
-    ) -> Result<Vec<u32>, SchedulingOOMError> {
+    pub fn compute_schedule(&self, settings: &DagSimpSettings) -> Result<Vec<u32>> {
         let mut self_useds = self.clone();
         self_useds.used_to_beginning();
-        self_useds
-            .compute_schedule_top_ints(verbose, mem_limit, num_mem_chunks, timeout, None)
-            .map(|x| self_useds.expand_order(&x))
+        Ok(
+            self_useds.expand_order(if self_useds.node_to_orig.len() <= 1 {
+                self_useds.node_to_orig.keys().cloned().collect::<Vec<_>>()
+            } else {
+                ExhaustiveScheduler::new(self_useds.clone(), None, settings.clone(), None)
+                    .run_portfolio_maybe()?
+                    .into_iter()
+                    .map(|(a, _b)| a)
+                    .collect()
+            }),
+        )
     }
 
     // Renumbers everything! breaks all node references!
@@ -168,11 +148,6 @@ impl Dag {
             }
             result = true;
         }
-        if self.cost() > simp_settings.mem_limit {
-            bail!(SchedulingOOMError::Simp {
-                cyto: self.repr_cytoscape_link(),
-            })
-        }
         Ok(result)
     }
 
@@ -232,10 +207,9 @@ impl Dag {
                     if self.children[*parent as usize]
                         .iter()
                         .all(|x| nodes_visited.contains(x))
+                        && !nodes_visited.contains(parent)
                     {
-                        if !nodes_visited.contains(parent) {
-                            bottom_frontier.push(*parent);
-                        }
+                        bottom_frontier.push(*parent);
                     }
                 }
             } else {
@@ -243,10 +217,9 @@ impl Dag {
                     if self.children[*parent as usize]
                         .iter()
                         .all(|x| nodes_visited.contains(x))
+                        && !nodes_visited.contains(parent)
                     {
-                        if !nodes_visited.contains(parent) {
-                            bottom_frontier.push(*parent);
-                        }
+                        bottom_frontier.push(*parent);
                     }
                 }
             }
@@ -297,19 +270,15 @@ impl Dag {
         node: u32,
         _simp_settings: &DagSimpSettings,
     ) -> Option<Vec<u32>> {
-        if let [a, b] = self.parents[node as usize][..]  && a!=b{
-            if self.parents[a as usize].len() == 1
-                && self.parents[b as usize].len() == 1
-                && self.parents[a as usize][0] == self.parents[b as usize][0]
-            {
-                let top = self.parents[a as usize][0];
-                let new_cost = self.node_costs[a as usize] + self.node_costs[b as usize];
-                let intermediate = self.split_node_extra_below(top);
-                self.merge_add(intermediate, a, true);
-                self.merge_add(intermediate, b, true);
-                self.node_costs[intermediate as usize] = new_cost;
-                return Some(Default::default());
-            }
+        if let [a, b] = self.parents[node as usize][..] && a!=b && self.parents[a as usize].len() == 1
+                && self.parents[b as usize].len() == 1 && self.parents[a as usize][0] == self.parents[b as usize][0] {
+            let top = self.parents[a as usize][0];
+            let new_cost = self.node_costs[a as usize] + self.node_costs[b as usize];
+            let intermediate = self.split_node_extra_below(top);
+            self.merge_add(intermediate, a, true);
+            self.merge_add(intermediate, b, true);
+            self.node_costs[intermediate as usize] = new_cost;
+            return Some(Default::default());
         }
         None
     }
@@ -616,22 +585,14 @@ impl Dag {
                     .unwrap()
                     .0,
             );
-            let exhaustive_scheduler = ExhaustiveScheduler::new(
+            let resy = ExhaustiveScheduler::new(
                 direct_only_sub.clone(),
                 required_first,
                 settings.clone(),
                 None,
-            );
-            let resy = if direct_only_sub.node_to_orig.len() > 120 && settings.parallelism.is_some()
-            {
-                timed!(
-                    exhaustive_scheduler.run_portfolio_release_gil(),
-                    10,
-                    settings.verbose >= 2
-                )
-            } else {
-                timed!(exhaustive_scheduler.run(None), 10, settings.verbose >= 2)
-            };
+            )
+            .run_portfolio_maybe();
+
             if resy.is_err() && did_anything {
                 return Ok(true);
             }
@@ -703,10 +664,10 @@ impl Dag {
         // we don't need to fix up any backrefs
         for (i, (node, cost)) in order.iter().enumerate() {
             if i != 0 {
-                self.children[*node as usize] = sv![order[i - 1 as usize].0];
+                self.children[*node as usize] = sv![order[i - 1_usize].0];
             }
             if i != order.len() - 1 {
-                self.parents[*node as usize] = sv![order[i + 1 as usize].0];
+                self.parents[*node as usize] = sv![order[i + 1_usize].0];
                 self.node_costs[*node as usize] = *cost;
             }
         }
@@ -720,12 +681,10 @@ impl Dag {
         b: u32,
         simp_settings: &DagSimpSettings,
     ) -> bool {
-        let result = (self.node_costs[a as usize] as f64 - self.node_costs[b as usize] as f64)
-            .abs()
+        (self.node_costs[a as usize] as f64 - self.node_costs[b as usize] as f64).abs()
             < simp_settings.same_ratio
                 * (self.node_costs[a as usize] as f64 + self.node_costs[b as usize] as f64)
-                / 2.0;
-        result
+                / 2.0
     }
 
     pub fn get_sole_parent(&self, node: u32) -> Option<u32> {
@@ -734,7 +693,7 @@ impl Dag {
         {
             return Some(self.parents[node as usize][0]);
         }
-        return None;
+        None
     }
 
     pub fn is_tiny(&self, node: u32, settings: &DagSimpSettings) -> bool {
@@ -744,7 +703,7 @@ impl Dag {
     pub fn get_bottom_frontier(&self) -> Vec<u32> {
         self.node_to_orig
             .iter()
-            .filter(|(k, _n)| self.children[**k as usize].len() == 0)
+            .filter(|(k, _n)| self.children[**k as usize].is_empty())
             .map(|(k, _n)| *k)
             .collect()
     }
@@ -835,7 +794,10 @@ impl Dag {
                 .keys()
                 .map(|i| format!(
                     "{{\"id\":\"{}\",\"name\":{:?},\"label\": \"{}: {}\"}}",
-                    i, "", *i, self.node_costs[*i as usize]
+                    i,
+                    "",
+                    *i,
+                    oom_fmt(self.node_costs[*i as usize])
                 ))
                 .collect::<Vec<_>>()
                 .join(","),
@@ -867,19 +829,102 @@ where
         .filter(move |(_i, x)| f(x))
         .map(|(i, _x)| i as u32)
 }
+
+#[derive(Clone, Debug)]
+struct VecStack<T> {
+    storage: Vec<T>,
+    levels: Vec<usize>,
+    start: usize,
+}
+impl<T> VecStack<T> {
+    fn new() -> Self {
+        Self {
+            storage: vec![],
+            levels: vec![],
+            start: 0,
+        }
+    }
+    #[inline]
+    fn pop_level(&mut self) {
+        let n = self.levels.pop().unwrap();
+        self.storage.truncate(n);
+        self.start = self.levels.last().copied().unwrap_or(0);
+    }
+    #[inline]
+    fn push_level(&mut self) {
+        self.levels.push(self.storage.len());
+        self.start = self.storage.len();
+    }
+
+    fn clear(&mut self) {
+        self.storage.clear();
+        self.levels.clear();
+        self.start = 0;
+    }
+
+    fn push_copy_last(&mut self)
+    where
+        T: Copy,
+    {
+        let l = self.storage.len();
+        self.storage.extend_from_within(self.start..);
+        self.levels.push(l);
+        self.start = l;
+    }
+
+    #[inline]
+    fn extend(&mut self, x: impl IntoIterator<Item = T>) {
+        self.storage.extend(x)
+    }
+    #[inline]
+    fn last_swap_remove(&mut self, ix: usize) -> T {
+        self.storage.swap_remove(self.start + ix)
+    }
+    #[inline]
+    fn last(&self) -> &[T] {
+        &self.storage[self.start..]
+    }
+    #[inline]
+    fn last_mut(&mut self) -> &mut [T] {
+        &mut self.storage[self.start..]
+    }
+    #[inline]
+    fn last_len(&self) -> usize {
+        self.storage.len() - self.start
+    }
+    #[inline]
+    fn push(&mut self, x: T) {
+        self.storage.push(x)
+    }
+}
+
 #[derive(Clone, Debug)]
 struct ExhaustiveScheduler {
     dag: Dag,
-    parents_left: Vec<i32>,
-    children_left: Vec<i32>,
+    parents_left: Vec<u16>,
+    num_parents: Vec<u16>,
+    entry_point: Option<u32>, // alive at start of schedule
     best: (Vec<(u32, usize)>, usize),
-    order: Vec<(u32, u32, usize)>,
-    options_stack: Vec<Sv<[u32; 16]>>,
+    order_nodes: VecStack<(u32, usize)>, /* (node, max cost) for chosen node & propagations. max cost always <= max_cost_so_far. stack has a level per entry in order */
+    order: Vec<(u32, usize, usize)>, /* (choice ix in options_stack entry, alive cost before choice, max cost so far before choice) */
+    alive_cost: usize,
+    max_cost_so_far: usize,
+    options_stack: VecStack<u32>,
     settings: DagSimpSettings,
-    found_anything: bool,
     rng: Option<SmallRng>,
+    found_anything: bool,
+    numiters: usize,
 }
 
+// todo:
+// nogoods: (S, x): S alive => no solution of cost < x
+// learn a nogood when backtracking due to checked all alternatives, x = min of achieveable cost returned from failed alternatives (when backtrack keep track of actual cost achieved), S = set alive before backtracking
+// do one watch literal scheme on nogoods
+// might be interesting to figure out proof rules and/or clause minimization:
+// - log nogood size diff assuming we can delete from both sides? to see potential for minimization
+// - maybe can delete leaves from S (if nothing else in S depends on them) & subtract their cost from x I think? not totally sure this is valid, and probably won't implement it for now anyways
+// - machanical way to dervie proof rules: look at MIP/PB proofs of learned clauses for MIP/pesudo boolean encoding
+
 impl ExhaustiveScheduler {
     pub fn new(
         dag: Dag,
@@ -888,183 +933,405 @@ impl ExhaustiveScheduler {
         random_seed: Option<u64>,
     ) -> Self {
         let mut result = Self {
-            parents_left: dag.parents.iter().map(|x| x.len() as i32).collect(),
-            children_left: dag.children.iter().map(|x| x.len() as i32).collect(),
-            best: (vec![], settings.acceptable_subdag_max + 1),
-            order: Vec::with_capacity(dag.node_costs.len()),
-            options_stack: Vec::with_capacity(dag.node_costs.len() + 2),
+            parents_left: vec![],
+            num_parents: dag
+                .parents
+                .iter()
+                .map(|x| {
+                    x.len()
+                        .try_into()
+                        .expect("node with more than 2^16-1 parents!")
+                })
+                .collect(),
+            entry_point,
+            best: (vec![], usize::MAX),
+            order: vec![],
+            order_nodes: VecStack::new(),
+            options_stack: VecStack::new(),
+            alive_cost: 0,
+            max_cost_so_far: 0,
             dag,
             settings,
             found_anything: false,
             rng: random_seed.map(|rs| SmallRng::seed_from_u64(rs)),
+            numiters: 0,
         };
-        result.options_stack.push(
-            entry_point
-                .map(|z| sv![z])
-                .unwrap_or(filter_to_idx(&result.children_left, |x| *x == 0).collect()),
-        );
+        entry_point.map(|x| {
+            // ensure x is alive at start, special cased
+            result.num_parents[x as usize] += 1;
+        });
+        result.reset();
         result
     }
+    fn reset(&mut self) {
+        self.parents_left = self.num_parents.clone();
+        self.order.clear();
+        self.order_nodes.clear();
+        self.options_stack.clear();
+        self.alive_cost = 0;
+        self.max_cost_so_far = 0;
+
+        self.options_stack.push_level();
+        self.options_stack
+            .extend(filter_to_idx(&self.parents_left, |x| *x == 0).map(|x| {
+                self.alive_cost += self.dag.node_costs[x as usize];
+                x
+            }));
+        self.max_cost_so_far = self.alive_cost;
+    }
 
     #[inline]
-    fn choose(&mut self, i: u32) {
-        let options = self.options_stack.last().unwrap();
-        let choice = options[i as usize] as u32;
-        for child in &self.dag.children[choice as usize] {
-            self.parents_left[*child as usize] -= 1;
+    fn upd_alive_cost(&mut self, x: u32) {
+        for &child in &self.dag.children[x as usize] {
+            if self.parents_left[child as usize] == self.num_parents[child as usize] {
+                self.alive_cost += self.dag.node_costs[child as usize];
+            }
         }
-        for parent in &self.dag.parents[choice as usize] {
-            self.children_left[*parent as usize] -= 1;
+    }
+    #[inline]
+    fn upd_rest(&mut self, x: u32) {
+        for &child in &self.dag.children[x as usize] {
+            self.parents_left[child as usize] -= 1;
+            if self.parents_left[child as usize] == 0 {
+                self.options_stack.push(child)
+            }
         }
-        self.children_left[choice as usize] -= 1;
-        let cost: usize = (0..self.parents_left.len() as u32)
-            .filter(|i| self.children_left[*i as usize] < 0 && self.parents_left[*i as usize] >= 1)
-            .map(|x| self.dag.node_costs[x as usize])
-            .sum();
-        self.order.push((choice, i, cost));
-        self.options_stack
-            .push(filter_to_idx(&self.children_left, |x| *x == 0).collect());
-        let endel1 = self.options_stack.len() - 1;
+    }
+
+    fn choose(&mut self, i: u32) -> bool {
+        let choice = self.options_stack.last()[i as usize] as u32;
+        let cost_before = self.alive_cost;
+        let max_cost_so_far_before = self.max_cost_so_far;
+
+        self.upd_alive_cost(choice);
+
+        if self.alive_cost >= self.best.1 {
+            self.alive_cost = cost_before;
+            return false;
+        }
+        self.max_cost_so_far = self.max_cost_so_far.max(self.alive_cost);
+
+        self.order_nodes.push_level();
+        self.order_nodes.push((choice, self.alive_cost));
+
+        self.alive_cost -= self.dag.node_costs[choice as usize];
+
+        self.options_stack.push_copy_last();
+        self.options_stack.last_swap_remove(i as usize);
+
+        self.upd_rest(choice);
+
+        loop {
+            let mut did_anything = false;
+            for ix in (0..self.options_stack.last_len()).into_iter().rev() {
+                let x = self.options_stack.last()[ix];
+                let cost: usize = self.dag.children[x as usize]
+                    .iter()
+                    .map(|&ch| {
+                        if self.parents_left[ch as usize] == self.num_parents[ch as usize] {
+                            self.dag.node_costs[ch as usize]
+                        } else {
+                            0
+                        }
+                    })
+                    .sum();
+                if (cost <= self.dag.node_costs[x as usize])
+                    && (cost + self.alive_cost <= self.max_cost_so_far)
+                {
+                    did_anything = true;
+                    self.options_stack.last_swap_remove(ix);
+                    self.upd_alive_cost(x);
+                    self.order_nodes.push((x, self.alive_cost));
+                    assert!(self.alive_cost <= self.max_cost_so_far);
+                    self.alive_cost -= self.dag.node_costs[x as usize];
+                    self.upd_rest(x);
+                }
+            }
+            if !did_anything {
+                break;
+            }
+        }
+
+        self.order.push((i, cost_before, max_cost_so_far_before));
+
+        // todo maybe take choice here when options_stack.last_len() == 1?
+
         if let Some(rng) = &mut self.rng {
             let rand_byte: u8 = rng.gen_range(0..5);
             if rand_byte == 0 {
-                self.options_stack[endel1].reverse();
-                return;
+                self.options_stack
+                    .last_mut()
+                    .sort_unstable_by_key(|&x| -(x as i32));
+                return true;
             } else if rand_byte == 1 {
-                return;
+                self.options_stack.last_mut().sort_unstable();
+                return true;
             } else if rand_byte == 2 {
-                self.options_stack[endel1]
-                    .sort_by_cached_key(|&x| self.dag.children[x as usize].len());
-                return;
+                self.options_stack
+                    .last_mut()
+                    .sort_by_cached_key(|&x| self.dag.parents[x as usize].len());
+                return true;
             } else if rand_byte == 3 {
-                self.options_stack[endel1].sort_by_cached_key(|&x| {
-                    self.dag.parents[x as usize]
+                // alive_cost diff
+                self.options_stack.last_mut().sort_by_cached_key(|&x| {
+                    self.dag.children[x as usize]
                         .iter()
-                        .map(|&parent| self.children_left[parent as usize])
-                        .max()
-                        .unwrap_or(0)
+                        .map(|&ch| {
+                            if self.parents_left[ch as usize] == self.num_parents[ch as usize] {
+                                self.dag.node_costs[ch as usize]
+                            } else {
+                                0
+                            }
+                        })
+                        .sum::<usize>() as i64
+                        - self.dag.node_costs[x as usize] as i64
                 });
-                return;
+                return true;
             }
         }
-        // trying out sorting options, must do this before stack is used
-        self.options_stack[endel1].sort_by_cached_key(|&x| {
+
+        self.options_stack.last_mut().sort_by_cached_key(|&x| {
             // lower means tried sooner
-            self.dag.parents[x as usize]
+            self.dag.children[x as usize]
                 .iter()
-                .map(|&parent| self.children_left[parent as usize])
+                .map(|&child| self.parents_left[child as usize])
                 .max()
                 .unwrap_or(0)
-                + (self.dag.children[x as usize].is_empty() as i32) * 10
-        })
+        });
+        return true;
     }
     #[inline]
     fn pop(&mut self) -> u32 {
         let popped = self.order.pop().unwrap();
-        self.options_stack.pop();
-        let choice = popped.0;
-        for child in &self.dag.children[choice as usize] {
-            self.parents_left[*child as usize] += 1;
-        }
-        for parent in &self.dag.parents[choice as usize] {
-            self.children_left[*parent as usize] += 1;
-        }
-        self.children_left[choice as usize] += 1;
-
-        popped.1
-    }
-    #[inline]
-    fn backtrack(&mut self) {
-        while !self.order.is_empty()
-            && self.order.last().unwrap().1 + 1
-                >= self.options_stack[self.order.len() - 1].len() as u32
-        {
-            self.pop();
-        }
-        if !self.order.is_empty() {
-            let popped_idx = self.pop();
-            if !self.order.is_empty() {
-                self.choose(popped_idx + 1);
+        for (x, _) in self.order_nodes.last().iter().rev() {
+            for child in &self.dag.children[*x as usize] {
+                self.parents_left[*child as usize] += 1;
             }
         }
+        self.order_nodes.pop_level();
+        self.options_stack.pop_level();
+        self.alive_cost = popped.1;
+        self.max_cost_so_far = popped.2;
+        popped.0
     }
-    fn run(mut self, halt_signal: Option<Arc<AtomicBool>>) -> Result<Vec<(u32, usize)>> {
+    fn search_loop(
+        &mut self,
+        halt_signal: Option<Arc<AtomicBool>>,
+        best_cost: Option<Arc<AtomicUsize>>,
+        max_cost: usize,
+    ) {
         let start_instant = std::time::Instant::now();
-        let mut numiters: usize = 0;
         loop {
-            // println!("{}", order.iter().map(|x| x.0.to_string()).join(","));
-            if self.order.len() == self.dag.node_costs.len() {
-                self.best = (
-                    self.order.iter().map(|(a, _b, c)| (*a, *c)).collect(),
-                    *self.order.iter().map(|(_n, _i, c)| c).max().unwrap(),
-                );
+            // println!("{} {:?}", self.alive_cost, self.order);
+            // if self.
+            // println!(
+            //     "{} {} {}",
+            //     self.alive_cost,
+            //     self.order.len(),
+            //     self.options_stack.last().unwrap().len()
+            // );
+            if self.options_stack.last_len() == 0 {
+                let max_cost = self
+                    .order_nodes
+                    .storage
+                    .iter()
+                    .map(|(_, a)| *a)
+                    .max()
+                    .unwrap();
+                assert!(max_cost < self.best.1);
+                self.best = (self.order_nodes.storage.clone(), max_cost);
                 self.found_anything = true;
-                if start_instant.elapsed().as_nanos() > self.settings.exhaustive_settle_ns as u128 {
+                best_cost
+                    .as_ref()
+                    .map(|x| x.fetch_min(max_cost, std::sync::atomic::Ordering::SeqCst));
+                // println!(
+                //     "backjump {} {}",
+                //     oom_fmt(max_cost),
+                //     self.order.len(),
+                //     // self.options_stack.levels
+                // );
+            }
+            let best_cost = best_cost
+                .as_ref()
+                .map_or(usize::MAX, |x| x.load(std::sync::atomic::Ordering::Relaxed))
+                .min(self.best.1)
+                .min(max_cost);
+            // loop until we make a successful choice
+            let mut i = 0;
+            loop {
+                while !self.order.is_empty()
+                    && (i >= self.options_stack.last_len() as u32
+                        || self.max_cost_so_far >= best_cost)
+                {
+                    i = self.pop() + 1;
+                }
+                if self.order.len() == 0 && i != 0 {
                     break;
                 }
-                self.backtrack();
-            } else {
-                self.choose(0);
-                if self.order.last().unwrap().2 >= self.best.1 {
-                    self.backtrack();
+                if !self.choose(i) {
+                    i += 1;
+                } else {
+                    break;
                 }
             }
-            if self.order.len() == 0
-                || numiters.trailing_zeros() > 8
-                    && (start_instant.elapsed().as_nanos()
-                        > self.settings.exhaustive_give_up_ns as u128
-                        || halt_signal
-                            .as_ref()
-                            .map(|h| h.load(std::sync::atomic::Ordering::Relaxed))
-                            .unwrap_or(false))
-            {
+            if self.done(start_instant, &halt_signal) {
                 break;
             }
-            numiters += 1;
+            self.numiters += 1;
         }
+        // println!(
+        //     "dag size {} numiters {} cur order len {} best cost {}{}",
+        //     self.dag.node_costs.len(),
+        //     self.numiters,
+        //     self.order.len(),
+        //     if self.best.0.len() == 0 {
+        //         "no solution".to_owned()
+        //     } else {
+        //         oom_fmt(self.best.1)
+        //     },
+        //     if self.order.len() == 0 {
+        //         " exhaustive"
+        //     } else {
+        //         ""
+        //     }
+        // );
+    }
+    fn done(&mut self, start_instant: Instant, halt_signal: &Option<Arc<AtomicBool>>) -> bool {
+        (self.numiters % 100 == 0
+            && ((self.best.1 <= self.settings.mem_limit
+                && start_instant.elapsed().as_nanos()
+                    > self.settings.exhaustive_settle_ns as u128)
+                || (start_instant.elapsed().as_nanos()
+                    > self.settings.exhaustive_give_up_ns as u128)
+                || (halt_signal
+                    .as_ref()
+                    .map(|h| h.load(std::sync::atomic::Ordering::Relaxed))
+                    .unwrap_or(false))))
+            || self.order.len() == 0
+    }
+    fn get_out(self, halt_signal: Option<Arc<AtomicBool>>) -> Result<(Vec<(u32, usize)>, usize)> {
+        halt_signal.map(|x| x.swap(true, std::sync::atomic::Ordering::Relaxed));
         if self.found_anything {
-            return Ok(self.best.0);
+            let mut o = self.best.0;
+            self.entry_point.map(|x| {
+                let last = o.last().unwrap();
+                let last_cost = self.dag.node_costs[last.0 as usize];
+                o.push((x, last.1 - last_cost))
+            });
+            o.reverse();
+            assert_eq!(o.len(), self.parents_left.len());
+            return Ok((o, self.best.1));
         }
         bail!(SchedulingOOMError::ExhaustiveTimeout {
-            cyto: self.dag.repr_cytoscape_link(),
-            iters: numiters,
+            iters: self.numiters,
             size: self.dag.node_to_orig.len(),
         })
-        // self.print_cytoscape_link();
     }
 
-    pub fn run_portfolio_release_gil(self) -> Result<Vec<(u32, usize)>> {
-        pyo3::Python::with_gil(|py| py.allow_threads(move || self.run_portfolio_raw()))
+    fn run(
+        mut self,
+        halt_signal: Option<Arc<AtomicBool>>,
+        best_cost: Option<Arc<AtomicUsize>>,
+    ) -> Result<(Vec<(u32, usize)>, usize)> {
+        self.search_loop(halt_signal.clone(), best_cost, usize::MAX);
+        self.get_out(halt_signal)
+        // println!("{}", self.dag.repr_cytoscape_link());
+    }
+
+    pub fn run_binary_search(
+        mut self,
+        halt_signal: Option<Arc<AtomicBool>>,
+        best_cost: Option<Arc<AtomicUsize>>,
+    ) -> Result<(Vec<(u32, usize)>, usize)> {
+        let mut failed_cost = 0;
+        let mut best_so_far = usize::MAX;
+        let start_instant = std::time::Instant::now();
+        let real_timeout = self.settings.exhaustive_settle_ns;
+        let mut prev_mid = usize::MAX;
+        loop {
+            let top = best_cost
+                .as_ref()
+                .map_or(self.best.1, |x| x.load(Ordering::Relaxed));
+            let mid = failed_cost + (top - failed_cost) / 2;
+            let remaining_steps = (prev_mid as f64 - mid as f64).log2() - 16.0; // 2^16 = 64kb
+            if mid > prev_mid {
+                self.reset();
+            }
+            prev_mid = mid;
+
+            self.settings.exhaustive_settle_ns = real_timeout / (remaining_steps as usize).max(1);
+            self.search_loop(halt_signal.clone(), best_cost.clone(), mid);
+            self.settings.exhaustive_settle_ns = real_timeout;
+
+            if self.best.1 == best_so_far {
+                failed_cost = mid;
+            } else {
+                // println!("binary search {} -> {}", oom_fmt(mid), oom_fmt(self.best.1));
+                best_so_far = self.best.1;
+            }
+            if self.done(start_instant, &halt_signal) {
+                break;
+            }
+        }
+        self.get_out(halt_signal)
     }
 
     pub fn run_portfolio_raw(self) -> Result<Vec<(u32, usize)>> {
         // println!("portfolio!");
         let halt_signal: Arc<AtomicBool> = Arc::new(AtomicBool::from(false));
+        let best_cost: Arc<AtomicUsize> = Arc::new(AtomicUsize::from(usize::MAX));
         let (tx, rx) = channel();
         for i in 0..self.settings.parallelism.unwrap() {
             let tx = tx.clone();
             let mut self_here = self.clone();
-            self_here.rng = if i != 0 {
+            self_here.rng = if i == 0 {
                 None
             } else {
                 Some(SmallRng::seed_from_u64(i as u64))
             };
             let halt_signal_clone = halt_signal.clone();
+            let best_cost = best_cost.clone();
             std::thread::spawn(move || {
                 // let randy = self_here.use_rand;
-                let (result, _time) = timed_value!(self_here.run(Some(halt_signal_clone)));
-                // println!("finished {} {} took {:.2?}", result.is_ok(), randy, time);
+                let (result, _time) = timed_value!(if i < 2 {
+                    self_here.run_binary_search(Some(halt_signal_clone), Some(best_cost))
+                } else {
+                    self_here.run(Some(halt_signal_clone), Some(best_cost))
+                });
+                // println!("finished {} took {:.2?}", result.is_ok(), _time);
                 tx.send(result).unwrap_or(());
             });
         }
-        let mut resulty: Result<Vec<(u32, usize)>> = Err(SchedulingOOMError::ThreadsLost {}.into());
-        for recv in rx.iter().take(self.settings.parallelism.unwrap()) {
-            resulty = recv;
-            if resulty.is_ok() {
-                break;
-            }
+        let results = rx
+            .iter()
+            .take(self.settings.parallelism.unwrap())
+            .collect::<Vec<_>>();
+        if results.iter().any(|x| x.is_ok()) {
+            let x = results
+                .into_iter()
+                .filter_map(|x| x.ok())
+                .min_by_key(|t| t.1)
+                .unwrap();
+            // println!("best {}", oom_fmt(x.1));
+            Ok(x.0)
+        } else {
+            Ok(results.into_iter().next().unwrap()?.0)
         }
-        halt_signal.swap(true, std::sync::atomic::Ordering::Relaxed);
-        return resulty;
+    }
+
+    pub fn run_portfolio_maybe(self) -> Result<Vec<(u32, usize)>> {
+        let v = self.settings.verbose >= 2;
+        pyo3::Python::with_gil(|py| {
+            py.allow_threads(move || {
+                if self.dag.node_to_orig.len() > 30
+                    && self.settings.parallelism.is_some()
+                    && self.settings.parallelism.unwrap() > 1
+                {
+                    timed!(self.run_portfolio_raw(), 10, v)
+                } else {
+                    timed!(self.run_binary_search(None, None).map(|x| x.0), 10, v)
+                }
+            })
+        })
     }
 }
