@@ -68,7 +68,7 @@ simple_from!(|x: Matcher| -> IterativeMatcher { IterativeMatcherData::Match(x).i
 simple_from!(|x: MatcherData| -> IterativeMatcher { Matcher::from(x).into() });
 simple_from!(|x: MatcherFromPyBase| -> IterativeMatcher { Matcher::from(x).into() });
 simple_from!(|x: MatcherFromPyBase| -> IterativeMatcherFromPy {
-    IterativeMatcherFromPy::BaseMatch(x.into())
+    IterativeMatcherFromPy::BaseMatch(x)
 });
 simple_from!(|x: Matcher| -> IterativeMatcherRc { IterativeMatcher::from(x).into() });
 simple_from!(|x: MatcherData| -> IterativeMatcherRc { IterativeMatcher::from(x).into() });
@@ -90,7 +90,7 @@ impl From<IterativeMatcherFromPy> for IterativeMatcher {
 
             // TODO: maybe Finished should be convertible to Term
             // IterativeMatcherFromPy::Finished(_) => IterativeMatcherData::Term.into(),
-            IterativeMatcherFromPy::IterativeMatcher(x) => x.into(),
+            IterativeMatcherFromPy::IterativeMatcher(x) => x,
             // we intentionally do a matcher here as the default - if users want
             // a IterativeMatcher pyfunc, they can explicitly use IterativeMatcher
             // factory.
@@ -561,7 +561,6 @@ pub fn function_per_child_op(
     mut func: impl FnMut(CircuitRc, IterativeMatcherRc) -> Result<CircuitRc>,
 ) -> Result<Option<CircuitRc>> {
     if all_finished(&updated) {
-        // special casing this isn't strictly required
         return Ok(None);
     }
     let mut new_matchers = per_child(updated, matcher, circuit.num_children());
@@ -840,7 +839,7 @@ impl IterativeMatcher {
                     }
 
                     fn with_first(&'a self, first: IterativeMatcherRc) -> ChainItem {
-                        ChainItem::new(first, self.rest().to_vec()).into()
+                        ChainItem::new(first, self.rest().to_vec())
                     }
                 }
 
@@ -963,7 +962,7 @@ impl IterativeMatcher {
                         found,
                         updated: None,
                     }
-                } else if all_updated.iter().any(|x| all_finished(&x)) {
+                } else if all_updated.iter().any(|x| all_finished(x)) {
                     IterateMatchResults {
                         found,
                         updated: Some(None.into()),
@@ -1042,7 +1041,7 @@ impl IterativeMatcher {
         IterativeMatcherData::Chains(
             matchers
                 .into_iter()
-                .map(|x| ChainItem::new(x.into(), Vec::new()))
+                .map(|x| ChainItem::new(x, Vec::new()))
                 .collect(),
         )
         .into()
@@ -1410,7 +1409,7 @@ pub fn restrict(
 ) -> IterativeMatcher {
     // TODO: flatten
     IterativeMatcherData::Restrict(RestrictIterativeMatcher::new(
-        matcher.clone().into(),
+        matcher,
         term_if_matches,
         start_depth,
         end_depth,
