@@ -5,15 +5,14 @@ use std::{
 };
 
 use anyhow::Result;
-use circuit_base::{
-    clicolor,
-    print::{color, last_child_arrows, CliColor, PrintOptions},
-    CircuitNode, CircuitNodeUnion, CircuitRc,
-};
+use circuit_base::{CircuitNode, CircuitNodeUnion, CircuitRc};
+use circuit_print::print::PrintOptions;
 use itertools::Itertools;
-use once_cell::unsync::Lazy;
+use once_cell::sync::Lazy;
 use pyo3::prelude::*;
 use rr_util::{
+    clicolor,
+    print::{color, last_child_arrows, CliColor},
     python_println,
     util::{indent, indent_butfirst},
 };
@@ -25,7 +24,7 @@ use crate::{
     IterateMatchResults, IterativeMatcherData, IterativeMatcherRc,
 };
 
-const MAX_SELF_SINGLE_LINE_WIDTH: Lazy<usize> = Lazy::new(|| {
+static MAX_SELF_SINGLE_LINE_WIDTH: Lazy<usize> = Lazy::new(|| {
     std::env::var("MATCHER_PRINT_SELF_WIDTH")
         .ok()
         .and_then(|x| x.parse::<usize>().ok())
@@ -34,7 +33,7 @@ const MAX_SELF_SINGLE_LINE_WIDTH: Lazy<usize> = Lazy::new(|| {
 const MATCHER_PRINT_INDENT_WIDTH: usize = 4;
 
 fn indented_if_multiline(x: &str) -> String {
-    if x.contains("\n") {
+    if x.contains('\n') {
         return format!("\n{}\n", indent(x.to_owned(), MATCHER_PRINT_INDENT_WIDTH));
     }
     x.to_owned()
@@ -46,7 +45,7 @@ fn maybe_indented_list(
     surrounding_newlines: bool,
 ) -> String {
     let total_length: usize = strings.iter().map(|s| s.len() + 3).sum();
-    if total_length < *MAX_SELF_SINGLE_LINE_WIDTH && !strings.iter().any(|s| s.contains("\n")) {
+    if total_length < *MAX_SELF_SINGLE_LINE_WIDTH && !strings.iter().any(|s| s.contains('\n')) {
         return strings.iter().join(sep);
     }
     let sep: String = if sep_after_line {
@@ -94,10 +93,10 @@ impl Display for IterativeMatcher {
                         op_debug(&filter.end_depth)
                     ));
                 }
-                if &**filter.iterative_matcher != &Matcher::true_matcher().into() {
+                if **filter.iterative_matcher != Matcher::true_matcher().into() {
                     entries.push(format!("matcher: {}", filter.iterative_matcher.0));
                 }
-                if &**filter.term_early_at != &Matcher::false_matcher() {
+                if **filter.term_early_at != Matcher::false_matcher() {
                     entries.push(format!("term_at: {}", filter.term_early_at.0));
                 }
                 if filter.term_if_matches {
@@ -233,7 +232,10 @@ impl Display for Matcher {
                     } else {
                         "Never".to_owned()
                     },
-                MatcherData::Name(names) => repr_single_many(names, |x| format!("\"{}\"", x)),
+                MatcherData::Name(names) => repr_single_many(
+                    &names.iter().map(|x| x.0).collect(),
+                    |x| format!("\"{}\"", x)
+                ),
                 MatcherData::Type(types) => repr_single_many(types, |x| format!("{}", x)),
                 MatcherData::Regex(regex) => format!(
                     "re{}'{}'",
@@ -368,9 +370,9 @@ pub fn repr_matcher_debug(
     ) -> Result<()> {
         let key = (circuit.clone(), matcher.clone());
         if !seen.insert(key) {
-            write!(
+            writeln!(
                 result,
-                "{}'{}'...\n",
+                "{}'{}'...",
                 last_child_arrows(&last_child_stack, true, print_options.arrows),
                 circuit.info().name.unwrap_or("".into())
             )
@@ -383,7 +385,7 @@ pub fn repr_matcher_debug(
             last_child_arrows(&last_child_stack, true, print_options.arrows),
             print_options.repr_line_info(circuit.clone())?,
             if matched.found {
-                color(&" # Found", clicolor!(Green))
+                color(" # Found", clicolor!(Green))
             } else {
                 "".to_owned()
             },
@@ -394,7 +396,7 @@ pub fn repr_matcher_debug(
         );
         result.push_str(&line_repr);
 
-        let child_matchers = per_child(matched.updated, matcher.clone(), circuit.num_children());
+        let child_matchers = per_child(matched.updated, matcher, circuit.num_children());
         for (i, (child, child_matcher)) in zip(circuit.children(), child_matchers).enumerate() {
             if let Some(child_matcher) = child_matcher {
                 let new_last_child_stack = last_child_stack
