@@ -5,10 +5,12 @@ use std::{
 
 use anyhow::{anyhow, bail, Result};
 use circuit_base::{
-    expand_node::{expand_node, ReplaceMapRc},
-    CircuitNode, CircuitRc,
+    // expand_node::{expand_node, ReplaceMapRc},
+    CircuitNode,
+    CircuitRc,
 };
 use circuit_utils::SetOfCircuitIdentities;
+use expand_node::{expand_node, ReplaceMapRc};
 use macro_rules_attribute::apply;
 use pyo3::prelude::*;
 use rr_util::{
@@ -220,7 +222,7 @@ impl BoundUpdater {
 }
 
 #[pyclass]
-#[derive(Debug, Clone)]
+#[derive(Debug, Clone, Default)]
 pub struct Getter {
     #[pyo3(get, set)]
     pub(super) default_fancy_validate: bool,
@@ -231,17 +233,6 @@ pub struct Getter {
         FastUnboundedCache<(HashBytes, IterativeMatcherRc), Vec<CircuitLinkedList>>,
 }
 
-impl Default for Getter {
-    fn default() -> Self {
-        Self {
-            default_fancy_validate: false,
-            cache: FastUnboundedCache::default(),
-            paths_cache: FastUnboundedCache::default(),
-            all_paths_cache: FastUnboundedCache::default(),
-        }
-    }
-}
-
 #[derive(Clone, Debug)]
 pub enum CircuitLinkedList {
     None,
@@ -255,13 +246,9 @@ impl CircuitLinkedList {
     fn list(&self) -> Vec<CircuitRc> {
         let mut out = vec![];
         let mut head = self;
-        loop {
-            if let CircuitLinkedList::Some(c, n) = &head {
-                out.push(c.clone());
-                head = &*n;
-            } else {
-                break;
-            }
+        while let CircuitLinkedList::Some(c, n) = &head {
+            out.push(c.clone());
+            head = n;
         }
         out
     }
@@ -498,10 +485,7 @@ pub struct BoundGetter {
 impl BoundGetter {
     #[new]
     pub fn new(getter: Getter, matcher: IterativeMatcherRc) -> Self {
-        Self {
-            getter,
-            matcher: matcher.into(),
-        }
+        Self { getter, matcher }
     }
 
     fn __call__(
@@ -582,7 +566,7 @@ impl AnyFound {
         circuit: CircuitRc,
         matcher: IterativeMatcherRc,
     ) -> Result<bool> {
-        let out = self.any_impl(circuit, matcher.clone())?;
+        let out = self.any_impl(circuit, matcher)?;
         Ok(out)
     }
 
@@ -633,10 +617,7 @@ pub struct BoundAnyFound {
 impl BoundAnyFound {
     #[new]
     pub fn new(any_found: AnyFound, matcher: IterativeMatcherRc) -> Self {
-        Self {
-            any_found,
-            matcher: matcher.into(),
-        }
+        Self { any_found, matcher }
     }
 
     fn __call__(&mut self, _py: Python<'_>, circuit: CircuitRc) -> Result<bool> {
@@ -649,7 +630,7 @@ impl BoundAnyFound {
 }
 
 #[pyclass]
-#[derive(Debug, Clone)]
+#[derive(Debug, Clone, Default)]
 pub struct Expander {
     /// Note: we don't currently cache these transforms individually. We could
     /// do this.
@@ -675,21 +656,6 @@ pub struct Expander {
     pub(super) validation_getter: Getter,
 }
 
-impl Default for Expander {
-    fn default() -> Self {
-        Self {
-            replacements: Vec::new(),
-            matchers: Vec::new(),
-            ban_multiple_matches_on_node: false,
-            default_fancy_validate: false,
-            default_assert_any_found: false,
-            suffix: None,
-            batch_cache: FastUnboundedCache::default(),
-            validation_getter: Default::default(),
-        }
-    }
-}
-
 #[pymethods]
 impl Expander {
     #[new]
@@ -707,10 +673,7 @@ impl Expander {
         default_assert_any_found: bool,
         suffix: Option<String>,
     ) -> Self {
-        let (matchers, replacements) = expanders
-            .into_iter()
-            .map(|(a, b)| (a.into(), b.into()))
-            .unzip();
+        let (matchers, replacements) = expanders.into_iter().map(|(a, b)| (a, b)).unzip();
         Self {
             replacements,
             matchers,
@@ -837,7 +800,7 @@ impl Expander {
             Ok(out.updated)
         })?;
         let expanded = if expanded != circuit {
-            expanded.add_suffix(self_.suffix.as_ref().map(|x| &**x))
+            expanded.add_suffix(self_.suffix.as_deref())
         } else {
             expanded
         };
@@ -849,6 +812,6 @@ impl Expander {
     }
 
     pub fn suffix(&self) -> Option<&str> {
-        self.suffix.as_ref().map(|x| &**x)
+        self.suffix.as_deref()
     }
 }
