@@ -16,11 +16,13 @@ use rr_util::{
     caching::FastUnboundedCache,
     eq_by_big_hash::EqByBigHash,
     fn_struct,
-    py_types::{i64_to_tensor, random_torch_i64, PyShape},
+    py_types::{i64_to_tensor, random_torch_i64, PySymShape},
     python_error_exception,
+    shape::{shape_into_known, Shape},
+    sv,
     tensor_util::{
-        Shape, Slice, TensorAxisIndex, TensorIndex, TorchDevice, TorchDeviceDtype,
-        TorchDeviceDtypeOp, TorchDtype,
+        Slice, TensorAxisIndex, TensorIndex, TorchDevice, TorchDeviceDtype, TorchDeviceDtypeOp,
+        TorchDtype,
     },
     tu8v,
     util::HashBytes,
@@ -45,7 +47,7 @@ pub fn default_hash_seeder(base_seed: Option<i64>) -> Seeder {
 }
 
 fn count_elements(s: &Shape) -> usize {
-    s.iter().product()
+    s.iter().map(|x| x.unwrap()).product()
 }
 
 #[pyfunction]
@@ -132,8 +134,8 @@ macro_rules! py_forward_sample_spec {
                 self.get_overall_weight()
             }
             #[pyo3(name = "get_sample_shape")]
-            fn get_sample_shape_py(&self) -> PyShape {
-                PyShape(self.get_sample_shape())
+            fn get_sample_shape_py(&self) -> PySymShape {
+                PySymShape(self.get_sample_shape())
             }
             #[pyo3(name = "get_transform")]
             fn get_transform_py(&self) -> TransformRc {
@@ -197,7 +199,6 @@ impl RandomSampleSpec {
             replace,
             simplify,
             seeder,
-            ..Default::default()
         }
         .into_init()
     }
@@ -209,7 +210,7 @@ impl RandomSampleSpec {
         let seed_array = Array::nrc(
             i64_to_tensor(
                 seed,
-                Shape::new(),
+                sv![],
                 TorchDeviceDtype {
                     device: TorchDevice::Cpu,
                     dtype: TorchDtype::int64,
@@ -300,7 +301,7 @@ impl SampleSpecTrait for RandomSampleSpec {
 }
 
 #[pyclass(extends=PySampleSpecBase)]
-#[derive(Debug, Clone)]
+#[derive(Debug, Clone, Default)]
 pub struct RunDiscreteVarAllSpec {
     #[pyo3(get)]
     groups: Vec<CircuitRc>,
@@ -310,7 +311,7 @@ pub struct RunDiscreteVarAllSpec {
 
 impl RunDiscreteVarAllSpec {
     fn try_new(groups: Vec<CircuitRc>, subsets: Option<Vec<Slice>>) -> Result<Self> {
-        if groups.iter().any(|g| g.info().rank() != 1) {
+        if groups.iter().any(|g| g.rank() != 1) {
             bail!(SampleError::GroupWithIncorrectNdim { groups })
         }
         if let Some(sub) = &subsets && sub.len() != groups.len() {
@@ -319,7 +320,7 @@ impl RunDiscreteVarAllSpec {
 
         Ok(Self {
             groups: groups.clone(),
-            subsets: subsets.clone().unwrap_or_else(|| {
+            subsets: subsets.unwrap_or_else(|| {
                 groups
                     .iter()
                     .map(|_| Slice {
@@ -348,15 +349,6 @@ impl RunDiscreteVarAllSpec {
 
 py_forward_sample_spec!(RunDiscreteVarAllSpec);
 
-impl Default for RunDiscreteVarAllSpec {
-    fn default() -> Self {
-        RunDiscreteVarAllSpec {
-            groups: vec![],
-            subsets: vec![],
-        }
-    }
-}
-
 #[pymethods]
 impl RunDiscreteVarAllSpec {
     #[new]
@@ -412,13 +404,13 @@ impl SampleSpecTrait for RunDiscreteVarAllSpec {
             None,
         );
 
-        if axes_to_repeat.len() != 0 {
+        if !axes_to_repeat.is_empty() {
             let counts = self
                 .groups
                 .iter()
                 .filter(|&g| g != var.probs_and_group())
-                .map(|g| g.info().shape[0])
-                .collect();
+                .map(|g| g.info().shape[0].unwrap_err())
+                .collect::<Result<Vec<_>>>()?;
             sampled_values = sampled_values
                 .expand_at_axes(axes_to_repeat, counts, None)
                 .unwrap()
@@ -431,13 +423,12 @@ impl SampleSpecTrait for RunDiscreteVarAllSpec {
         let all_weights_sum = Einsum::nrc(
             vec![(
                 all_weights.clone(),
-                (0..(all_weights.info().rank() as u8)).collect(),
+                (0..(all_weights.rank() as u8)).collect(),
             )],
             tu8v![],
             None,
         );
         all_weights
-            .clone()
             .mul(
                 GeneralFunction::new_by_name(vec![all_weights_sum], "reciprocal".into(), None)
                     .unwrap()
@@ -576,7 +567,7 @@ impl Sampler {
         let out = if is_orig_shape {
             out.expand_at_axes(
                 (0..self.sample_spec.get_sample_shape().len()).collect(),
-                self.sample_spec.get_sample_shape().into_vec(),
+                shape_into_known(self.sample_spec.get_sample_shape()),
                 None,
             )?
             .rc()
@@ -665,7 +656,7 @@ impl Sampler {
     }
 
     fn get_expectation(&self, circuit: CircuitRc) -> CircResult {
-        let circuit_rank = circuit.info().rank() as u8;
+        let circuit_rank = circuit.rank() as u8;
         let spec_nb_axes = self.spec_nb_axes() as u8;
         let name = self.suffix().and_then(|s| {
             circuit
@@ -719,16 +710,12 @@ pub fn factored_cumulant_expectation_rewrite(cumulant: &Cumulant) -> CircuitRc {
 
     let centered_moment = Cumulant::new(vec![centered_product(children.clone())], None).rc();
 
-    let mut sub: Vec<(CircuitRc, f64)> = vec![(centered_moment.clone(), 1.)];
+    let mut sub: Vec<(CircuitRc, f64)> = vec![(centered_moment, 1.)];
 
     for p_indexs in partitions(children.len()) {
         let p = p_indexs
             .iter()
-            .map(|is| {
-                is.iter()
-                    .map(|i| (i.clone(), children[*i].clone()))
-                    .collect_vec()
-            })
+            .map(|is| is.iter().map(|i| (*i, children[*i].clone())).collect_vec())
             .collect_vec();
 
         if p.iter().any(|b| b.len() < 2) || p.len() == 1 {
