@@ -4,15 +4,14 @@ use std::{
 };
 
 use anyhow::Result;
-use circuit_base::{
-    clicolor,
-    print::{color, CliColor, PrintOptions},
-    CircuitNode, CircuitNodeSelfOnlyHash, CircuitRc,
-};
+use circuit_base::{CircuitNode, CircuitNodeSelfOnlyHash, CircuitRc};
+use circuit_print::print::PrintOptions;
 use itertools::Itertools;
 use pyo3::prelude::*;
 use rr_util::{
+    clicolor,
     name::Name,
+    print::{color, CliColor},
     util::{indent, HashBytes},
 };
 use rustc_hash::FxHashMap as HashMap;
@@ -117,14 +116,14 @@ pub fn diff_circuits(
             let mut new_options = options.clone();
             new_options.colorer = Some(PrintOptions::fixed_color(same_color));
             let idstr = color(&format!(" # same {}", id), same_color);
-            let child_printed = new_options.repr(new.clone())?;
+            let child_printed = new_options.repr(new)?;
             let child_printed =
                 once(&(child_printed.lines().next().unwrap().to_owned() + &idstr) as &str)
                     .chain(child_printed.lines().skip(1))
                     .join("\n");
             result.push_str(&indent(child_printed, last_child_stack.len() * 2));
-            result.push_str("\n");
-            seen_diffs.insert(diffkey.clone(), idstr);
+            result.push('\n');
+            seen_diffs.insert(diffkey, idstr);
             return Ok(());
         }
         if compute_self_hash(
@@ -153,8 +152,8 @@ pub fn diff_circuits(
             ));
             let idstr = color(&format!(" # changed {}", id), same_self_color);
             result.push_str(&idstr);
-            seen_diffs.insert(diffkey.clone(), idstr);
-            result.push_str("\n");
+            seen_diffs.insert(diffkey, idstr);
+            result.push('\n');
 
             assert_eq!(new.num_children(), old.num_children());
             for (i, (new_child, old_child)) in zip(new.children(), old.children()).enumerate() {
@@ -184,19 +183,13 @@ pub fn diff_circuits(
         }
         let mut new_options = options.clone();
         new_options.colorer = Some(PrintOptions::fixed_color(new_color));
-        result.push_str(&indent(
-            new_options.repr(new.clone())?,
-            last_child_stack.len() * 2,
-        ));
-        result.push_str("\n");
+        result.push_str(&indent(new_options.repr(new)?, last_child_stack.len() * 2));
+        result.push('\n');
 
         let mut new_options = options.clone();
         new_options.colorer = Some(PrintOptions::fixed_color(removed_color));
-        result.push_str(&indent(
-            new_options.repr(old.clone())?,
-            last_child_stack.len() * 2,
-        ));
-        result.push_str("\n");
+        result.push_str(&indent(new_options.repr(old)?, last_child_stack.len() * 2));
+        result.push('\n');
 
         Ok(())
     }
@@ -220,9 +213,9 @@ pub fn diff_circuits(
         if &result[result.len() - 1..] != "\n" {
             result.push('\n');
         }
-        write!(
+        writeln!(
             result,
-            "{} {} {} {}\n",
+            "{} {} {} {}",
             color(&format!("{}: Same as original", same_color), same_color),
             color(
                 &format!("{}: Children changed", same_self_color),
