@@ -45,7 +45,7 @@ pub fn distribute(
                 return Ok(einsum.rc());
             } else {
                 bail!(DistributeError::OperandIdxTooLarge {
-                    einsum: einsum.clone(),
+                    einsum,
                     operand_idx,
                 })
             };
@@ -67,7 +67,7 @@ pub fn distribute(
             |i, new_einsum| distribute_rec(new_einsum, traversal_per_child[i].clone()),
         );
         if distributed.is_err() && allow_partial_distribute {
-            return Ok(einsum.clone().rc());
+            return Ok(einsum.rc());
         }
         Ok(distributed?.rc())
     }
