@@ -1,12 +1,12 @@
-use anyhow::{bail, Context, Result};
+use anyhow::{anyhow, bail, Context, Result};
 use circuit_base::{generalfunction::*, Array, GeneralFunctionSpec};
 use pyo3::{prelude::*, types::PyTuple};
 use rand::{self, Rng};
 use rr_util::{
     py_types::{assert_tensors_close, ExtraPySelfOps, Tensor},
-    tensor_util::Shape,
+    shape::{shape_is_known, shape_join_eq, Shape, Size},
 };
-use rustc_hash::FxHashSet as HashSet;
+use smallvec::ToSmallVec;
 
 /// I now think this maybe should have been written in python.
 /// Not to bad to port I guess...
@@ -57,6 +57,14 @@ impl Default for GeneralFunctionSpecTester {
     }
 }
 
+fn gen_size_in(rng: &mut impl Rng, start: usize, end: usize) -> Size {
+    if rng.gen_bool(0.01) {
+        Size::NONE
+    } else {
+        Size::known(rng.gen_range(start..end))
+    }
+}
+
 #[pymethods]
 impl GeneralFunctionSpecTester {
     #[new]
@@ -158,26 +166,17 @@ impl GeneralFunctionSpecTester {
             }
         }
 
-        let all_batch_shapes: HashSet<&[usize]> = std::iter::once(&shape[..current_num_batch_dims])
-            .chain(
-                shapes
-                    .iter()
-                    .zip(&input_batchability)
-                    .filter_map(|(s, is_batch)| is_batch.then(|| &s[..current_num_batch_dims])),
-            )
-            .collect();
-        if all_batch_shapes.len() != 1 {
-            bail!(
-                "inputs and output have non-matching 'batch' shapes, all_batch_shapes={:?}",
-                all_batch_shapes
-            );
-        }
+        shapes.iter().zip(&input_batchability).filter_map(|(s, is_batch)| is_batch.then(|| &s[..current_num_batch_dims]))
+            .try_fold(shape[..current_num_batch_dims].to_smallvec(), |a, b| {
+                shape_join_eq(&a, b, ||
+                anyhow!("inputs and output have non-matching 'batch' shapes, a={a:?} b={b:?} input_shapes={shapes:?} output_shape={shape:?}"))
+            })?;
 
         let mut rng = rand::thread_rng();
 
         let mut run_sample = |num_batch_dims, random_inputs: bool| {
             let batch_shape: Shape = (0..num_batch_dims)
-                .map(|_| rng.gen_range(self.start_shape_num..self.end_shape_num))
+                .map(|_| gen_size_in(&mut rng, self.start_shape_num, self.end_shape_num))
                 .collect();
 
             let new_shapes: Vec<Shape> = shapes
@@ -224,7 +223,7 @@ impl GeneralFunctionSpecTester {
             );
             }
 
-            if &new_info.input_batchability != &input_batchability {
+            if new_info.input_batchability != input_batchability {
                 bail!(
                     "{}changed input_batchability when only batching was changed\n{}\n{}",
                     prefix,
@@ -257,7 +256,7 @@ impl GeneralFunctionSpecTester {
                 shapes
                     .iter()
                     .chain(std::iter::once(shape))
-                    .map(|x| x.iter().product::<usize>())
+                    .map(|x| x.iter().map(|x| x.t().unwrap_or(1)).product::<usize>())
                     .sum::<usize>()
                     < self.randn_size_cap
             };
@@ -266,12 +265,13 @@ impl GeneralFunctionSpecTester {
                 && get_count_find(&shapes, &shape)
                 && get_count_find(&new_shapes, &new_info.shape)
                 && num_batch_dims < current_num_batch_dims
+                && shapes.iter().all(|x| shape_is_known(x))
             // only run random on < orig
             {
                 // TODO: swap to f64 as needed!
                 let tensors: Vec<_> = shapes
                     .iter()
-                    .map(|shape| Array::randn(shape.clone()).value)
+                    .map(|shape| Array::randn(shape.clone()).unwrap().value)
                     .collect();
 
                 let out_tensor = spec
@@ -290,7 +290,7 @@ impl GeneralFunctionSpecTester {
                     );
                 }
                 let dims_to_remove = current_num_batch_dims - num_batch_dims;
-                if shape[..dims_to_remove].iter().any(|x| *x == 0) {
+                if shape[..dims_to_remove].iter().any(|x| x.is(0)) {
                     return Ok(());
                 }
                 let idxs: Py<PyTuple> = Python::with_gil(|py| {
@@ -298,7 +298,7 @@ impl GeneralFunctionSpecTester {
                         py,
                         shape[..dims_to_remove]
                             .iter()
-                            .map(|s| rng.gen_range(0..*s).into_py(py)),
+                            .map(|s| s.map(|x| rng.gen_range(0..x)).into_py(py)),
                     )
                     .into()
                 });
@@ -359,7 +359,9 @@ impl GeneralFunctionSpecTester {
                     .map(|_| {
                         let ndim = rng.gen_range(self.start_ndim..self.end_ndim);
                         (0..ndim)
-                            .map(|_| rng.gen_range(self.start_shape_num..self.end_shape_num))
+                            .map(|_| {
+                                gen_size_in(&mut rng, self.start_shape_num, self.end_shape_num)
+                            })
                             .collect()
                     })
                     .collect();
@@ -495,7 +497,9 @@ fn test_index_general_function() -> Result<()> {
                 .map(|min_dims| {
                     let ndim = rng.gen_range(tester.start_ndim.max(min_dims)..tester.end_ndim);
                     (0..ndim)
-                        .map(|_| rng.gen_range(tester.start_shape_num..tester.end_shape_num))
+                        .map(|_| {
+                            gen_size_in(&mut rng, tester.start_shape_num, tester.end_shape_num)
+                        })
                         .collect()
                 })
                 .collect::<Vec<_>>()
