@@ -6,7 +6,6 @@ use std::{
 
 use anyhow::{anyhow, bail, Context, Result};
 use circuit_base::{
-    expand_node::{expand_node, replace_expand_map, MapReplaceExpander, ReplaceMapRc},
     module::{are_args_used, is_intersecting_free_syms, SymbolSetRc},
     CircuitNode, CircuitRc, Module, ModuleArgSpec, ModuleSpec, Rearrange, Symbol,
 };
@@ -16,6 +15,7 @@ use circuit_rewrites::{
         elim_no_input_module, extract_rewrite_raw, module_remove_unused_inputs, module_strip_args,
     },
 };
+use expand_node::{expand_node, replace_expand_map, MapReplaceExpander, ReplaceMapRc};
 use get_update_node::{AnyFound, IterativeMatcher, IterativeMatcherRc, MatcherData};
 use itertools::{izip, multiunzip, Itertools};
 use macro_rules_attribute::apply;
@@ -31,7 +31,8 @@ use rr_util::{
     fn_struct, impl_eq_by_big_hash,
     name::Name,
     pycall, python_error_exception,
-    rearrange_spec::{OpSize, RearrangeSpec, RearrangeSpecError},
+    rearrange_spec::{RearrangeSpec, RearrangeSpecError},
+    shape::Size,
     sv, tu8v, unwrap,
     util::{DimNumMaker, HashBytes},
     IndexSet,
@@ -220,7 +221,7 @@ pub fn fuse_modules_impl(
 ) -> Result<Vec<(CircuitRc, ModuleArgSpec)>> {
     // reverse to get inner shapes inside of outer
     let batch_shapes_per_mod: Vec<_> = orig_modules
-        .into_iter()
+        .iter()
         .map(|m| m.aligned_batch_shape())
         .collect();
     let cumulative_batch_shapes = batch_shapes_per_mod
@@ -236,14 +237,14 @@ pub fn fuse_modules_impl(
         .rev(); // reverse again
 
     let out = new_nodes
-        .into_iter()
+        .iter()
         .zip(cumulative_batch_shapes)
         .zip(orig_modules)
         .zip(batch_shapes_per_mod)
         .map(|(((items, cum_batch_shape), m), overall_mod_batch_shape)| {
             (*items)
-                .to_owned()
-                .into_iter()
+                .iter()
+                .cloned()
                 .zip(m.args())
                 .zip(m.spec.arg_specs.clone())
                 .zip(m.spec.batch_shapes(m.args_slice()))
@@ -251,9 +252,10 @@ pub fn fuse_modules_impl(
                     let missing_ndims = overall_mod_batch_shape.len() - orig_batch_shape.len();
                     let non_batch = orig_node.ndim() - orig_batch_shape.len();
                     let extra_dims = node.ndim() - orig_node.ndim();
-                    let any_batching = extra_dims > 0 || orig_batch_shape.len() > 0;
+                    let any_batching = extra_dims > 0 || !orig_batch_shape.is_empty();
                     let needs_pad_orig = missing_ndims > 0 && extra_dims > 0;
-                    let new_node = if (any_batching && cum_batch_shape.len() > 0) || needs_pad_orig
+                    let new_node = if (any_batching && !cum_batch_shape.is_empty())
+                        || needs_pad_orig
                     {
                         let mut maker = DimNumMaker::default();
                         let same_outer = maker.next_range(extra_dims);
@@ -273,12 +275,12 @@ pub fn fuse_modules_impl(
                             .context("too many dims in fuse module for flatten"));
                         }
 
-                        let mut sizes = sv![OpSize::NONE; maker.running];
+                        let mut sizes = sv![Size::NONE; maker.running];
                         for i in pad_for_orig.clone() {
-                            sizes[i] = Some(overall_mod_batch_shape[i - pad_for_orig.start]).into();
+                            sizes[i] = overall_mod_batch_shape[i - pad_for_orig.start];
                         }
                         for i in cum.clone() {
-                            sizes[i] = Some(cum_batch_shape[i - cum.start]).into();
+                            sizes[i] = cum_batch_shape[i - cum.start];
                         }
 
                         let spec = RearrangeSpec::new(
@@ -292,15 +294,14 @@ pub fn fuse_modules_impl(
                                 .chain(pad_for_orig)
                                 .chain(same_orig_batch)
                                 .chain(cum)
-                                .chain(same_non_batch.clone())
+                                .chain(same_non_batch)
                                 .map(|i| tu8v![i as u8])
                                 .collect(),
                             sizes,
-                        )
-                        .unwrap();
+                        )?;
 
                         let rep_name = node.info().name.map(|x| format!("{} rep_fuse", x).into());
-                        Rearrange::nrc(node, spec, rep_name)
+                        Rearrange::try_new(node, spec, rep_name)?.rc()
                     } else {
                         node
                     };
@@ -345,7 +346,8 @@ impl NestedModuleItems {
             assert_eq!(self_m.num_args(), base.len());
         }
 
-        let flat = fuse_modules_impl(&modules, &new_nodes_flat)?;
+        let flat = fuse_modules_impl(&modules, &new_nodes_flat)
+            .with_context(|| format!("fusing modules {modules:#?}"))?;
         let flat_sym_to_arg = flat
             .iter()
             .map(|(circ, arg_spec)| (arg_spec.symbol.clone(), circ.clone()))
@@ -643,15 +645,21 @@ impl ModulePusher {
         skip_module: IterativeMatcherRc,
         mode: PushDownMode,
     ) -> Result<CircOrGet> {
-        self.push_down_modules_rec(
-            circuit,
-            get,
-            skip_module,
-            mode,
-            &Default::default(),
-            &Default::default(),
-        )
-        .map(|x| x.0)
+        let out = self
+            .push_down_modules_rec(
+                circuit.clone(),
+                get,
+                skip_module,
+                mode,
+                &Default::default(),
+                &Default::default(),
+            )
+            .map(|x| x.0)
+            .context("pushing down modules")?;
+        if let CircOrGet::Circ(ref c) = out {
+            assert_eq!(c.shape(), circuit.shape());
+        }
+        Ok(out)
     }
 
     // we could do a generic function on the bools, but I assume this has better compile time
@@ -698,7 +706,7 @@ impl ModulePusher {
 
         if is_get && found_get {
             let out = get_fin()?;
-            out_set.insert(out.clone());
+            out_set.insert(out);
         }
 
         let found_override = found_get && mode == PushDownMode::PushOverride;
@@ -842,7 +850,7 @@ impl ModulePusher {
                 m.spec.circuit.clone(),
                 spec_circuit_get,
                 spec_circuit_skip_module,
-                &items.push(m.clone(), new_nodes_flat.clone(), new_nodes_base)?,
+                &items.push(m.clone(), new_nodes_flat, new_nodes_base)?,
                 extra_replacements,
             )?;
             if mode == PushDownMode::PushOverride {
@@ -859,6 +867,20 @@ impl ModulePusher {
                 (out, Default::default(), false)
             } else {
                 prior_inner_circuits.insert(m.spec.circuit.clone());
+                let out = if mode == PushDownMode::Circ {
+                    let out = out.unwrap();
+                    let m_shape = m.shape();
+                    if m_shape.len() > out.ndim() {
+                        Some(out.repeat_to(
+                            m_shape.clone(),
+                            m.info().name.map(|x| format!("{} rep_batch", x).into()),
+                        )?)
+                    } else {
+                        Some(out)
+                    }
+                } else {
+                    out
+                };
                 (out, prior_inner_circuits, termed_prior_iter)
             }
         } else {
@@ -972,7 +994,7 @@ impl ModulePusher {
                             .chain(arg_spec.symbol.shape())
                             .copied()
                             .collect(),
-                        arg_spec.symbol.uuid.clone(),
+                        arg_spec.symbol.uuid,
                         arg_spec.symbol.info().name,
                     );
 
@@ -996,7 +1018,7 @@ impl ModulePusher {
                 // if we have extra dims, we have to pull the symbols and resub to sync up batching
                 let new_flat = update_flat(items.flat.clone(), rep_expand_extra_dims);
                 let new_finished_circuit = self.replace_expander.replace_expand_with_map(
-                    circuit.clone(),
+                    circuit,
                     &ReplaceMapRc::new(updated_extra_replacements),
                 )?;
                 (new_finished_circuit, new_flat)
@@ -1034,7 +1056,7 @@ impl ModulePusher {
                     })
                     .collect();
                 let new_finished_circuit = self.replace_expander.replace_expand_with_map(
-                    circuit.clone(),
+                    circuit,
                     &ReplaceMapRc::new(updated_extra_replacements),
                 )?;
                 (new_finished_circuit, rev_nested)
@@ -1079,11 +1101,7 @@ impl ModulePusher {
                 Some(overall_mod_count),
             )
             .context("namer failed in push down module")?;
-        Ok(Module::new(
-            nodes,
-            ModuleSpec { circuit, arg_specs },
-            name.clone(),
-        ))
+        Ok(Module::new(nodes, ModuleSpec { circuit, arg_specs }, name))
     }
 
     fn build_module_flat(
@@ -1112,7 +1130,7 @@ impl ModulePusher {
         modules: Vec<Module>,
     ) -> Result<CircuitRc> {
         if modules.is_empty() {
-            return Ok(base_circuit.clone());
+            return Ok(base_circuit);
         }
         let cum_rev_mods = modules
             .clone()
@@ -1206,7 +1224,7 @@ impl ModuleConstructUpdateBindings {
         &mut self,
         m: Module,
         count_mods: usize,
-        outer_updated_syms: IndexSet<Symbol>,
+        outer_updated_syms: IndexSet<CircuitRc>,
         applied_modules: &[Module],
         overall_count: usize,
     ) -> Result<(CircuitRc, CircuitRc, Vec<Module>)> {
@@ -1252,7 +1270,7 @@ impl ModuleConstructUpdateBindings {
             })
             .collect::<Result<Vec<_>>>()?
             .into_iter()
-            .filter_map(|x| x)
+            .flatten()
             .unzip();
 
         let (new_spec_circuit, base_circuit, prior_modules) = if count_mods > 1 && !self.is_flatten
@@ -1262,13 +1280,13 @@ impl ModuleConstructUpdateBindings {
                 count_mods - 1,
                 outer_updated_syms
                     .into_iter()
-                    .chain(arg_specs.iter().map(|x| x.symbol.clone()))
+                    .chain(arg_specs.iter().map(|x| x.symbol.crc()))
                     .collect(),
                 &applied_modules[..applied_modules.len() - 1],
                 overall_count,
             )?
         } else {
-            (m.spec.circuit.clone(), m.spec.circuit.clone(), Vec::new())
+            (m.spec.circuit.clone(), m.spec.circuit, Vec::new())
         };
 
         let new_modules = if self.is_flatten {
@@ -1295,7 +1313,7 @@ impl ModuleConstructUpdateBindings {
         )
         .context("module construction failed in update bindings after updating nodes")?;
 
-        if m_out.aligned_batch_shape().len() > 0 {
+        if !m_out.aligned_batch_shape().is_empty() {
             bail!("batching update bindings case not supported, consider doing some conforming (TODO: better err)");
         }
 
@@ -1439,14 +1457,14 @@ impl SymbolExtractor {
                             circ| {
             match out_syms.entry(arg_spec.symbol.clone()) {
                 Entry::Occupied(entry) => {
-                    if &entry.get().0 != &circ {
+                    if entry.get().0 != circ {
                         bail!(ExtractSymbolsError::BoundInputInconsistent {
                             symbol: arg_spec.symbol,
                             old_bound: entry.get().0.clone(),
                             new_bound: circ
                         })
                     }
-                    if &entry.get().1 != &arg_spec {
+                    if entry.get().1 != arg_spec {
                         bail!(ExtractSymbolsError::ArgSpecInconsistent {
                             old_arg_spec: entry.get().1.clone(),
                             new_arg_spec: arg_spec
@@ -1496,7 +1514,7 @@ impl SymbolExtractor {
                         return Some(Err(ExtractSymbolsError::BatchedInput {
                             node_ndim: node.ndim(),
                             symbol_ndim: arg_spec.symbol.ndim(),
-                            symbol: arg_spec.symbol.clone(),
+                            symbol: arg_spec.symbol,
                             node,
                         }
                         .into()));
@@ -1737,7 +1755,7 @@ pub fn pull_out_of_modules(
         module_construct_callback,
         circuit_to_arg_spec,
     )?;
-    if b.len() != 0 {
+    if !b.is_empty() {
         bail!("pull_out_of_modules: still pulling circuits before returning (call pull_out_of_modules_extract to get results, or use a matcher that deposits them earlier) {:?}", b)
     }
     Ok(a)
@@ -1790,15 +1808,15 @@ pub fn pull_out_of_modules_get(
                 new_matcher2[0].clone(),
             )?;
 
-            let vars: IndexSet<Symbol> =
-                m.spec.arg_specs.iter().map(|s| s.symbol.clone()).collect();
+            let vars: IndexSet<CircuitRc> =
+                m.spec.arg_specs.iter().map(|s| s.symbol.crc()).collect();
 
             let (rs_with_fvs, rs_no_fvs): (Vec<_>, _) = rs
                 .iter()
                 .cloned()
                 .partition(|(x, _)| is_intersecting_free_syms(x, &vars));
 
-            let m_new = if rs_with_fvs.len() > 0 {
+            let m_new = if !rs_with_fvs.is_empty() {
                 let rs_fvs_new = itertools::process_results(
                     rs_with_fvs.iter().map(|(c, y)| -> Result<_> {
                         let x = replace_expand_map(
@@ -1858,14 +1876,13 @@ pub fn pull_out_of_modules_get(
                 circ.clone()
             };
 
-            let (_, _, new_matcher2_again) = matcher_new_binds
-                .clone()
-                .match_iterate_continue(m_new.clone())?;
+            let (_, _, new_matcher2_again) =
+                matcher_new_binds.match_iterate_continue(m_new.clone())?;
 
             let mut out = rs_no_fvs;
             let circ_out = m_new.map_children_enumerate(|i, c| {
                 Ok(if i == 0 {
-                    if rs_with_fvs.len() == 0 {
+                    if rs_with_fvs.is_empty() {
                         spec_circ_2.clone()
                     } else {
                         c
@@ -1886,7 +1903,7 @@ pub fn pull_out_of_modules_get(
                 })
             })?;
 
-            let circ_out = if rs_with_fvs.len() > 0 && let Some(cb) = &module_construct_callback {
+            let circ_out = if !rs_with_fvs.is_empty() && let Some(cb) = &module_construct_callback {
                 pycall!(cb, (circ_out, circ,), anyhow)?
             } else {
                 circ_out
@@ -1905,9 +1922,9 @@ pub fn pull_out_of_modules_get(
         }
     }
 
-    Ok(recurse(
+    recurse(
         circuit,
         matcher.clone(),
         matcher_new_binds.unwrap_or(matcher),
-    )?)
+    )
 }
