@@ -151,13 +151,13 @@ pub enum NestSpec<I: NestInfo> {
 }
 
 macro_rules! nest_spec_for {
-    ($name:ident {
+    ($name:ident [$($deriv:ident ,)*] {
         $(
             $field:ident : $field_ty:ty,
         )*
     }) => {
     paste::paste!{
-    #[derive(Clone, Debug)]
+    #[derive(Clone, Debug, $($deriv)*)]
     pub struct [<Nest $name sInfo>] {
         $(
             pub $field: $field_ty,
@@ -279,34 +279,19 @@ macro_rules! nest_spec_for {
     };
 }
 
-nest_spec_for!(Einsum {
+nest_spec_for!(Einsum [Default,] {
     name: Option<rr_util::name::Name>,
     out_axes_perm: Option<rr_util::util::EinsumAxes>,
     shrink_out_axes: bool,
 });
-nest_spec_for!(Add {
+nest_spec_for!(Add [Default,] {
     name: Option<rr_util::name::Name>,
 });
 
-impl Default for NestEinsumsInfo {
-    fn default() -> Self {
-        Self {
-            name: None,
-            out_axes_perm: None,
-            shrink_out_axes: false,
-        }
-    }
-}
-impl Default for NestAddsInfo {
-    fn default() -> Self {
-        Self { name: None }
-    }
-}
-
 impl NestInfo for NestEinsumsInfo {
     type Full = NestEinsumsFullInfo;
     fn to_full(self, orig_full: Option<&Self::Full>) -> Self::Full {
-        let out_axes = orig_full.map(|x| x.out_axes.clone()).flatten();
+        let out_axes = orig_full.and_then(|x| x.out_axes.clone());
         Self::Full {
             name: self.name,
             out_axes,
@@ -448,7 +433,7 @@ impl NestFullInfo for NestEinsumsFullInfo {
 
         let out_circ = Einsum::nrc(new_args, einsum.out_axes.clone(), einsum.info().name);
 
-        finish_with_extra_and_circ(&rec_on_children, &extra, out_circ)
+        finish_with_extra_and_circ(rec_on_children, &extra, out_circ)
     }
 
     fn nest_flat_strict_impl(
@@ -490,7 +475,7 @@ impl NestFullInfo for NestAddsFullInfo {
             .collect();
         let out_circ = Add::nrc(new_operands, add.info().name);
 
-        finish_with_extra_and_circ(&rec_on_children, &vec![(); rec_on_children.len()], out_circ)
+        finish_with_extra_and_circ(rec_on_children, &vec![(); rec_on_children.len()], out_circ)
     }
 
     fn nest_flat_strict_impl(
@@ -520,7 +505,7 @@ impl NestIdxsInfo<NestEinsumsFullInfo> {
         Self {
             idxs: new_idxs,
             info: NestEinsumsFullInfo::new(
-                self.info.name.clone(),
+                self.info.name,
                 self.info
                     .out_axes
                     .as_ref()
@@ -537,14 +522,14 @@ impl<I: NestFullInfo> NestIdxsItem<I> {
     fn all_indices(&self) -> Vec<usize> {
         match self {
             Self::Single(x) => vec![*x],
-            Self::Many(x) => x.into_iter().flat_map(|x| x.idxs.all_indices()).collect(),
+            Self::Many(x) => x.iter().flat_map(|x| x.idxs.all_indices()).collect(),
         }
     }
     fn add(&self, n: usize) -> Self {
         match self {
             Self::Single(x) => Self::Single(x + n),
             Self::Many(x) => Self::Many(
-                x.into_iter()
+                x.iter()
                     .map(|x| NestIdxsInfo {
                         idxs: x.idxs.add(n),
                         info: x.info.clone(),
@@ -576,7 +561,7 @@ impl<I: NestFullInfo> NestIdxsItem<I> {
                                 }
                                 (start.min(new_start), end.max(new_end)) // this fold is a bit silly, we could just get first + last
                             } else {
-                                return (start, end);
+                                (start, end)
                             }
                         });
                 if start == usize::MAX {
@@ -608,7 +593,7 @@ pub struct NestEnumerationItem<I: NestFullInfo>(NestIdxsItem<I>);
 impl<I: NestFullInfo> NestEnumerationItem<I> {
     pub fn new(idxs: NestIdxsItem<I>) -> Self {
         let all_indices = idxs.all_indices();
-        assert!(&all_indices == &(0..all_indices.len()).collect::<Vec<usize>>());
+        assert!(all_indices == (0..all_indices.len()).collect::<Vec<usize>>());
         Self(idxs)
     }
 
@@ -627,7 +612,7 @@ pub fn check_permutation_rest(
     let perm_set: HashSet<_> = perm.iter().cloned().collect();
     if perm.len() != perm_set.len() {
         bail!(PermError::IntsNotUnique {
-            ints: perm.iter().cloned().collect()
+            ints: perm.to_vec()
         })
     }
     let count_set = (0..count).collect::<HashSet<_>>();
@@ -639,7 +624,7 @@ pub fn check_permutation_rest(
         })
     }
     let rest: HashSet<usize> = count_set.difference(&perm_set).cloned().collect();
-    if !allow_rest && rest.len() != 0 {
+    if !allow_rest && !rest.is_empty() {
         bail!(NestError::PermutationMissesIdxsAndNoRestInSpec { missed_idxs: rest })
     }
     Ok(rest)
@@ -736,7 +721,7 @@ impl<I: NestInfo> NestSpec<I> {
                             .iter()
                             .map(|x| circuits[*x].clone())
                             .collect(),
-                        intersection: intersection,
+                        intersection,
                     });
                 }
                 if v_set.intersection(&all_ints_set).count() > 0 {
@@ -749,7 +734,7 @@ impl<I: NestInfo> NestSpec<I> {
                             .iter()
                             .map(|x| circuits[*x].clone())
                             .collect(),
-                        intersection: intersection,
+                        intersection,
                     });
                 }
             }
@@ -772,7 +757,7 @@ impl<I: NestInfo> NestSpec<I> {
             .convert_to_named_idxs_impl(
                 mapping,
                 &all_matchers
-                    .into_iter()
+                    .iter()
                     .zip(ints_per_matcher)
                     .map(|(a, b)| (a, &b[..]))
                     .collect(),
@@ -835,10 +820,8 @@ impl<I: NestInfo> NestSpec<I> {
             None
         };
         match self {
-            Self::Info(NestSpecInfo { info, .. }) => info.clone().to_full(orig_info).clone(),
-            Self::Sub(_) => orig_info
-                .map(|x| x.clone())
-                .unwrap_or_else(Default::default),
+            Self::Info(NestSpecInfo { info, .. }) => info.clone().to_full(orig_info),
+            Self::Sub(_) => orig_info.cloned().unwrap_or_default(),
         }
     }
 
@@ -878,8 +861,7 @@ impl<I: NestInfo> NestSpec<I> {
             Ok(ints)
         };
 
-        let find_rest =
-            |ints: &[usize]| Self::find_rest(ints.into_iter().cloned().collect(), mapping);
+        let find_rest = |ints: &[usize]| Self::find_rest(ints.iter().cloned().collect(), mapping);
 
         let handle_many = |all_ints: Vec<_>, items, flat| {
             assert!(is_unique(&all_ints));
@@ -966,7 +948,7 @@ pub fn finish_with_extra_and_circ<I: NestFullInfo>(
     assert_eq!(rec_on_children.len(), extra_items.len());
     let mut running_count = 0;
     let (out, all_mappings): (_, Vec<_>) = rec_on_children
-        .into_iter()
+        .iter()
         .zip(extra_items)
         .map(|((circ, maybe_items, mapping), extra)| {
             let (new_item, additional_count) = match maybe_items {
@@ -1090,7 +1072,7 @@ where
         .zip(updated.per_child_with_term(num_children))
         .zip(updated_current_matcher_to_pair)
         .map(|((c, new_matcher), updated_matchers_to_pair)| {
-            flatten_impl(c, new_matcher.clone(), updated_matchers_to_pair)
+            flatten_impl(c, new_matcher, updated_matchers_to_pair)
         })
         .collect::<Result<Vec<_>>>()?;
     let x = x
@@ -1217,7 +1199,7 @@ pub fn nest_flat_einsum_strict_rec(
                     if let Some(perm) = spec.info.out_axes_perm {
                         if perm.len() != sub_out_axes.len() {
                             bail!(NestError::PermHasWrongLen {
-                                perm: perm.clone(),
+                                perm,
                                 expected_len: sub_out_axes.len()
                             })
                         }
