@@ -2,5 +2,5 @@
 import einops
 
 import interp.circuit.circuit_compiler.util as circ_compiler_util
-import interp.circuit.interop_rust.interop_rust as interop
+import rust_circuit.interop_rust as interop
 from interp.circuit import computational_node, constant
