@@ -1,4 +1,5 @@
 use std::{
+    fmt::{self, Debug},
     iter::zip,
     num::NonZeroUsize,
     ops::{Deref, DerefMut},
@@ -27,10 +28,10 @@ impl TinyVecU8 {
     unsafe fn raw(&self) -> &[u8; 8] {
         std::mem::transmute::<&NonZeroUsize, &[u8; 8]>(&self.0)
     }
-    unsafe fn pointer_mut(&mut self) -> &mut Box<Vec<u8>> {
-        std::mem::transmute::<&mut NonZeroUsize, &mut Box<Vec<u8>>>(&mut self.0)
+    unsafe fn pointer_mut(&mut self) -> &mut Vec<u8> {
+        &mut *(std::mem::transmute::<&mut NonZeroUsize, &mut Box<Vec<u8>>>(&mut self.0))
     }
-    unsafe fn pointer(&self) -> &Box<Vec<u8>> {
+    unsafe fn pointer(&self) -> &Vec<u8> {
         std::mem::transmute::<&NonZeroUsize, &Box<Vec<u8>>>(&self.0)
     }
     #[inline]
@@ -112,7 +113,7 @@ impl TinyVecU8 {
                 result.raw_mut()[1..1 + values.len()].copy_from_slice(values);
             }
         }
-        return result;
+        result
     }
     pub fn unique(&self) -> Self {
         self.iter().unique().cloned().collect()
@@ -147,7 +148,7 @@ impl Drop for TinyVecU8 {
     fn drop(&mut self) {
         if !self.is_inline() {
             // println!("dropping");
-            drop(unsafe { self.pointer() });
+            drop(unsafe { Box::from_raw(self.0.get() as *mut Vec<u8>) });
         }
     }
 }
@@ -244,7 +245,7 @@ impl U8Set {
         zip(self.0, other.0).all(|(sub, sup)| sub & !sup == 0) // could use simd
     }
     pub fn is_superset(&self, other: &U8Set) -> bool {
-        other.is_subset(&self)
+        other.is_subset(self)
     }
     pub fn union(&self, other: &U8Set) -> U8Set {
         Self([
@@ -384,3 +385,114 @@ fn test_tiny_u8() {
         println!("{}", i);
     }
 }
+
+/// OpSize is a memory optimization of Option<usize> that stores a 63-bit int and one "is this none" bit
+#[derive(Copy, Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
+pub struct OpSize(pub u64);
+impl OpSize {
+    pub const SHIFT: usize = 63;
+    pub const NONE: OpSize = OpSize(1 << Self::SHIFT);
+    // sometimes into/from can't do type interference, so we have aliases
+    pub fn t(self) -> Option<usize> {
+        self.into()
+    }
+    pub fn f(val: Option<usize>) -> Self {
+        val.into()
+    }
+    pub fn some(x: usize) -> Self {
+        const _: () = assert!(std::mem::size_of::<usize>() <= std::mem::size_of::<u64>());
+        assert!(Self::is_valid_some(x));
+        OpSize(x as u64)
+    }
+    #[inline(always)]
+    pub fn is_valid_some(x: usize) -> bool {
+        x >> Self::SHIFT == 0
+    }
+    pub fn is_some(&self) -> bool {
+        self.0 >> Self::SHIFT == 0
+    }
+    pub fn is_none(&self) -> bool {
+        self.0 >> Self::SHIFT != 0
+    }
+    pub fn unwrap(self) -> usize {
+        assert!(self.is_some());
+        self.0 as usize
+    }
+    pub fn unwrap_or(self, x: usize) -> usize {
+        if self.is_none() {
+            x
+        } else {
+            self.unwrap()
+        }
+    }
+    pub fn and_then<F>(self, f: F) -> Self
+    where
+        F: FnOnce(usize) -> OpSize,
+    {
+        if self.is_some() {
+            f(self.unwrap())
+        } else {
+            self
+        }
+    }
+    pub fn map<F>(self, f: F) -> Self
+    where
+        F: FnOnce(usize) -> usize,
+    {
+        if self.is_some() {
+            Self::some(f(self.unwrap()))
+        } else {
+            self
+        }
+    }
+    pub fn map_or<F, T>(self, default: T, f: F) -> T
+    where
+        F: FnOnce(usize) -> T,
+    {
+        if self.is_some() {
+            f(self.unwrap())
+        } else {
+            default
+        }
+    }
+    pub fn it(&self) -> u64 {
+        self.0
+    }
+}
+
+impl Debug for OpSize {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        self.t().fmt(f)
+    }
+}
+
+impl From<Option<usize>> for OpSize {
+    fn from(x: Option<usize>) -> Self {
+        match x {
+            None => Self::NONE,
+            Some(value) => OpSize::some(value),
+        }
+    }
+}
+
+impl From<OpSize> for Option<usize> {
+    fn from(x: OpSize) -> Self {
+        if x.0 >> OpSize::SHIFT == 1 {
+            None
+        } else {
+            Some(x.0 as usize)
+        }
+    }
+}
+
+impl<'source> FromPyObject<'source> for OpSize {
+    fn extract(obj: &'source PyAny) -> PyResult<Self> {
+        Ok(OpSize::f(obj.extract()?))
+    }
+}
+
+impl IntoPy<PyObject> for OpSize {
+    fn into_py(self, py: Python<'_>) -> PyObject {
+        self.t().into_py(py)
+    }
+}
