@@ -1,16 +1,71 @@
 use pyo3::{
-    types::{PyModule, PyType},
-    Py, PyResult, PyTypeInfo, Python,
+    exceptions::PyException, pyclass, pymethods, types::PyModule, Py, PyErr, PyObject, PyResult,
+    Python,
 };
 
-pub fn py_get_type<T: PyTypeInfo>() -> Py<PyType> {
-    Python::with_gil(|py| T::type_object(py).into())
+// TODO: remove when/if we drop 3.11 support
+#[pyclass(extends=PyException)]
+pub struct ExceptionWithRustContext {
+    #[pyo3(get)]
+    exc: PyObject,
+    context: anyhow::Error,
 }
 
-pub trait HasPythonException {
-    fn get_py_err(&self, string: String) -> pyo3::PyErr;
-    fn register(py: Python<'_>, m: &PyModule) -> PyResult<()>;
-    fn print_stub(py: Python<'_>) -> PyResult<String>;
+#[pymethods]
+impl ExceptionWithRustContext {
+    fn __str__(&self) -> PyResult<String> {
+        Python::with_gil(|_py| {
+            let ctx = &self.context;
+            Ok(format!("{ctx:?}\n(this is an ExceptionWithRustContext, access e.exc or upgrade to 3.11 to get the underlying exception)\n"))
+        })
+    }
+}
+
+pub struct ErrorInfo {
+    pub try_to_py_err: fn(&(dyn std::error::Error + 'static), s: &str) -> Option<PyErr>,
+    pub register: fn(Python<'_>, &PyModule) -> PyResult<()>,
+    pub print_stub: fn(Python<'_>) -> PyResult<String>,
+}
+inventory::collect!(ErrorInfo);
+
+pub fn register_exceptions(py: Python<'_>, m: &PyModule) -> PyResult<()> {
+    for x in inventory::iter::<ErrorInfo> {
+        (x.register)(py, m)?;
+    }
+    Ok(())
+}
+
+pub fn anyhow_to_py_err(err: anyhow::Error) -> PyErr {
+    let root_cause = err.root_cause();
+    let s = format!("{:?}", err);
+    for x in inventory::iter::<ErrorInfo> {
+        if let Some(r) = (x.try_to_py_err)(root_cause, &s) {
+            return r;
+        }
+    }
+
+    pyo3::anyhow::anyhow_to_py_err(err, |py, e, ctx| {
+        PyErr::from_value(
+            Py::new(
+                py,
+                ExceptionWithRustContext {
+                    exc: e.into_value(py).into(),
+                    context: ctx,
+                },
+            )
+            .unwrap()
+            .as_ref(py),
+        )
+    })
+}
+
+pub fn print_exception_stubs(py: Python<'_>) -> PyResult<String> {
+    itertools::process_results(
+        (inventory::iter::<ErrorInfo>)
+            .into_iter()
+            .map(|x| (x.print_stub)(py)),
+        |it| it.intersperse("\n".to_string()).collect(),
+    )
 }
 
 #[macro_export]
@@ -71,7 +126,7 @@ macro_rules! python_error_exception {
                         Python::with_gil(|py| {
                             let dict = pyo3::types::PyDict::new(py);
                             $(
-                                dict.set_item(stringify!($arg), self.$arg.clone().into_py(py)).unwrap();
+                                dict.set_item::<&str, PyObject>(stringify!($arg), self.$arg.clone().into_py(py)).unwrap();
                             )*
                             dict.into()
                         })
@@ -134,7 +189,7 @@ macro_rules! python_error_exception {
             create_exception, types::PyModule, PyErr, PyResult, PyTypeInfo, Python,
         };
 
-        use $crate::errors_util::HasPythonException;
+        use $crate::errors_util::ErrorInfo;
 
         create_exception!(
             rust_circuit,
@@ -150,19 +205,8 @@ macro_rules! python_error_exception {
             );
         )*
 
-
-        impl HasPythonException for super::$name {
-            fn get_py_err(&self, err_msg_string: String) -> pyo3::PyErr {
-                use super::$name::*;
-                match self {
-                    $(
-                        $err_name { $($arg,)* } => {
-                            PyErr::new::<$sub_excep_name, _>($sub_excep_info_name { err_msg_string, $($arg : $arg.clone(),)* })
-                        },
-                    )*
-                }
-            }
-            fn register(py: Python<'_>, m: &PyModule) -> PyResult<()> {
+        paste::paste!{
+            fn [<register_ $name:snake>](py: Python<'_>, m: &PyModule) -> PyResult<()> {
                 m.add(
                     stringify!($excep_name),
                     py.get_type::<$excep_name>(),
@@ -177,7 +221,8 @@ macro_rules! python_error_exception {
 
                 Ok(())
             }
-            fn print_stub(py : Python<'_>) -> PyResult<String> {
+
+            fn [<print_stub_ $name:snake>](py : Python<'_>) -> PyResult<String> {
                 let out = [
                     format!("class {}({}): ...", $excep_name::NAME, <$base_ty>::type_object(py).name()?),
                     $(
@@ -186,6 +231,27 @@ macro_rules! python_error_exception {
                 ].join("\n");
                 Ok(out)
             }
+
+            fn [<get_py_err_ $name:snake>](x: &super::$name, err_msg_string: String) -> pyo3::PyErr {
+                use super::$name::*;
+                match x {
+                    $(
+                        $err_name { $($arg,)* } => {
+                            PyErr::new::<$sub_excep_name, _>($sub_excep_info_name { err_msg_string, $($arg : $arg.clone(),)* })
+                        },
+                    )*
+                }
+            }
+
+            inventory::submit! {
+                ErrorInfo {
+                    try_to_py_err: |err, s| {
+                        err.downcast_ref::<super::$name>().map(|x| [<get_py_err_ $name:snake>](x, s.to_owned()))
+                    },
+                    register: [<register_ $name:snake>],
+                    print_stub: [<print_stub_ $name:snake>],
+                }
+            }
         }
     }
 }
