@@ -66,7 +66,7 @@ impl<K: Eq + Hash + Default + Clone + Debug, V: Debug + Clone> LruCache<K, V> {
         cache
     }
     pub fn drop_last(&mut self) {
-        if self.map.len() > 0 {
+        if !self.map.is_empty() {
             unsafe {
                 let entry = (*self.tail).prev;
                 (*(*entry).prev).next = self.tail;
@@ -158,21 +158,17 @@ impl TensorCacheRrfs {
             .unwrap_or_else(|| {
                 let tensor = get_tensor_prefix(&prefix)?;
                 let tensor: Tensor = Python::with_gil(|py| {
-                    Ok::<Tensor, PyErr>(
-                        tensor
-                            .tensor()
-                            .getattr(py, "to")?
-                            .call(
-                                py,
-                                (),
-                                Some(
-                                    &[("device".to_owned(), self.device.clone())].into_py_dict(py),
-                                ),
-                            )?
-                            .extract(py)?,
-                    )
+                    tensor
+                        .tensor()
+                        .getattr(py, "to")?
+                        .call(
+                            py,
+                            (),
+                            Some([("device".to_owned(), self.device.clone())].into_py_dict(py)),
+                        )?
+                        .extract(py)
                 })?;
-                let size = tensor.shape().iter().cloned().product::<usize>()
+                let size = tensor.shape().iter().map(|x| x.unwrap()).product::<usize>()
                     * tensor.device_dtype().size();
                 if size > self.size_cutoff {
                     self.large.insert(prefix, tensor.clone(), size)
@@ -187,7 +183,7 @@ impl TensorCacheRrfs {
         self.small
             .get(&prefix)
             .or_else(|| self.large.get(&prefix))
-            .map(|z| z.clone())
+            .cloned()
     }
 }
 
