@@ -43,27 +43,27 @@ impl Ord for Name {
     }
 }
 
-impl Into<Name> for &str {
-    fn into(self) -> Name {
-        Name::new(self)
+impl From<&str> for Name {
+    fn from(val: &str) -> Self {
+        Name::new(val)
     }
 }
-impl Into<Name> for String {
-    fn into(self) -> Name {
-        Name::new(&self)
+impl From<String> for Name {
+    fn from(val: String) -> Self {
+        Name::new(&val)
     }
 }
 
-impl Into<&'static str> for Name {
-    fn into(self) -> &'static str {
-        self.str()
+impl From<Name> for &'static str {
+    fn from(val: Name) -> Self {
+        val.str()
     }
 }
 
 // Not recommended! just use &'static str!
-impl Into<String> for Name {
-    fn into(self) -> String {
-        self.str().to_owned()
+impl From<Name> for String {
+    fn from(val: Name) -> Self {
+        val.str().to_owned()
     }
 }
 
@@ -104,6 +104,26 @@ pub struct NameInterner {
     to_name: HashMap<&'static str, Name>,
 }
 
+// newtype over Name with fast Ord
+#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
+pub struct NameOrdById(pub Name);
+
+impl PartialOrd for NameOrdById {
+    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
+        Some(self.cmp(other))
+    }
+}
+impl Ord for NameOrdById {
+    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
+        self.0 .0.cmp(&other.0 .0)
+    }
+}
+impl std::fmt::Display for NameOrdById {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        self.0.fmt(f)
+    }
+}
+
 impl NameInterner {
     pub fn num_interned_strings() -> usize {
         NAME_INTERNER.lock().unwrap().names.len()
