@@ -5,7 +5,7 @@ use std::{
     sync::Mutex,
 };
 
-use anyhow::{bail, Result};
+use anyhow::{anyhow, bail, Result};
 use num_bigint::BigUint;
 use once_cell::sync::Lazy;
 /// This file optimizes tensor contraction order, like Python's opt_einsum
@@ -24,6 +24,7 @@ use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 
 use crate::{
     name::Name,
+    shape::{shape_try_as_known, Size},
     tensor_util::ParseError,
     union_find::UnionFind,
     util::{dict_to_list, EinsumAxes, NamedAxes, ALPHABET, ALPHABET_INV},
@@ -198,8 +199,8 @@ fn contraction_ssa_ids_to_recycled(
     for c in contractions.iter() {
         result.push(c.iter().map(|x| map[*x] as usize).collect());
         for i in c.iter() {
-            for j in *i..map.len() {
-                map[j] -= 1;
+            for j in map.iter_mut().skip(*i) {
+                *j -= 1;
             }
         }
     }
@@ -214,8 +215,9 @@ pub struct EinsumSpec {
     #[pyo3(get)]
     pub output_ints: Vec<usize>,
     #[pyo3(get)]
-    pub int_sizes: Vec<usize>,
+    pub int_sizes: Vec<Size>,
 }
+
 #[pymethods]
 impl EinsumSpec {
     #[new]
@@ -224,8 +226,8 @@ impl EinsumSpec {
         out_ints: Vec<usize>,
         int_sizes: &PyDict,
     ) -> Result<Self> {
-        let sizes_dict: HashMap<usize, usize> = int_sizes.extract()?;
-        let int_sizes = dict_to_list(&sizes_dict, None);
+        let sizes_dict: HashMap<usize, Size> = int_sizes.extract()?;
+        let int_sizes = dict_to_list(&sizes_dict, None, Size::known(0));
         let result = EinsumSpec {
             input_ints: operand_ints,
             output_ints: out_ints,
@@ -245,10 +247,10 @@ impl EinsumSpec {
         let all_output_in_input = self.output_ints.iter().all(|x| ints_in_input.contains(x));
         let all_ints_have_sizes = ints_in_input
             .iter()
-            .all(|x| (*x < self.int_sizes.len()) && self.int_sizes[*x] > 0);
+            .all(|x| (*x < self.int_sizes.len()) && self.int_sizes[*x].t().unwrap_or(1) > 0);
         all_output_in_input && all_ints_have_sizes
     }
-    pub fn shapes(&self) -> (Vec<Vec<usize>>, Vec<usize>) {
+    pub fn shapes(&self) -> (Vec<Vec<Size>>, Vec<Size>) {
         (
             self.input_ints
                 .iter()
@@ -343,7 +345,7 @@ impl EinsumSpec {
         let out_words = self
             .output_ints
             .iter()
-            .map(|a| (axis_to_name.get(a).unwrap().string().to_owned())) // output ints included in input ints
+            .map(|a| (axis_to_name.get(a).unwrap().string())) // output ints included in input ints
             .collect::<Vec<String>>()
             .join(" ");
         Some(format!("fancy: {operand_words} -> {out_words}"))
@@ -370,8 +372,13 @@ impl EinsumSpec {
         if self.input_ints.is_empty() {
             return BigUint::from(0usize);
         }
+        let int_sizes = self
+            .int_sizes
+            .iter()
+            .map(|x| x.t().unwrap_or(1))
+            .collect::<Vec<_>>();
         let used_ints = self.ints_in_input();
-        let used_int_sizes = used_ints.iter().map(|x| self.int_sizes[*x]);
+        let used_int_sizes = used_ints.iter().map(|x| int_sizes[*x]);
         let used_int_sizes_prod: BigUint = used_int_sizes
             .map(BigUint::from)
             .fold(BigUint::from(1usize), |a, b| a * b);
@@ -379,13 +386,13 @@ impl EinsumSpec {
         let output_size: BigUint = self
             .output_ints
             .iter()
-            .map(|x| BigUint::from(self.int_sizes[*x]))
+            .map(|x| BigUint::from(int_sizes[*x]))
             .fold(BigUint::from(1usize), |a, b| a * b);
 
         let ints_in_output = HashSet::from_iter(self.output_ints.iter().copied());
         let non_output_size: BigUint = used_ints
             .difference(&ints_in_output)
-            .map(|x| BigUint::from(self.int_sizes[*x]))
+            .map(|x| BigUint::from(int_sizes[*x]))
             .fold(BigUint::from(1usize), |a, b| a * b);
         let sums = output_size * (non_output_size - 1usize);
         muls + sums
@@ -464,13 +471,15 @@ impl EinsumSpec {
         }
     }
 
-    pub fn optimize_pre_contract_subsets(
+    fn optimize_pre_contract_subsets(
         &self,
         dont_contract_subsets_into: Option<Vec<usize>>,
-    ) -> (Vec<Vec<usize>>, Vec<usize>) {
+    ) -> Result<(Vec<Vec<usize>>, Vec<usize>)> {
         let dont = dont_contract_subsets_into.unwrap_or_default();
         let mut result: Vec<Vec<usize>> = vec![];
         let mut stack: Vec<usize> = self.input_ints.iter().map(ints_to_bitmap).collect();
+        let int_sizes = shape_try_as_known(&self.int_sizes)
+            .ok_or(anyhow!("can't optimize einsum with unknown sizes!"))?;
         loop {
             let mut res = vec![0, 0];
             let mut big_sup: usize = 0;
@@ -480,12 +489,12 @@ impl EinsumSpec {
                     if !dont.contains(&j)
                         && i != j
                         && sup | sub == *sup
-                        && bitmap_and_sizes_to_size(*sup, &self.int_sizes) <= candidate_cost
+                        && bitmap_and_sizes_to_size(*sup, int_sizes) <= candidate_cost
                     {
                         res[0] = min(i, j);
                         res[1] = max(i, j);
                         big_sup = *sup;
-                        candidate_cost = bitmap_and_sizes_to_size(*sup, &self.int_sizes);
+                        candidate_cost = bitmap_and_sizes_to_size(*sup, int_sizes);
                     }
                 }
             }
@@ -500,7 +509,7 @@ impl EinsumSpec {
         }
         result.push((0..stack.len()).collect());
 
-        (result, stack)
+        Ok((result, stack))
     }
 
     /// based on opt_einsum's dynamic programming algorithm https://github.com/dgasmith/opt_einsum/blob/master/opt_einsum/paths.py#L1124
@@ -518,6 +527,8 @@ impl EinsumSpec {
             return Ok(vec![vec![0]]);
         }
 
+        let int_sizes = shape_try_as_known(&self.int_sizes)
+            .ok_or(anyhow!("can't optimize einsum with unknown sizes!"))?;
         let tensor_ints: Vec<Vec<usize>> = self.input_ints.clone();
         let subgraphs =
             get_disconnected_sugraphs(&tensor_ints, &get_int_to_tensor_appearance(&tensor_ints));
@@ -532,7 +543,7 @@ impl EinsumSpec {
             let sg_bmp: usize = ints_to_bitmap(sg);
             let mut cost_cap: usize = max(
                 2,
-                bitmap_and_sizes_to_size(sg_bmp & out_bmp, &self.int_sizes) * 64,
+                bitmap_and_sizes_to_size(sg_bmp & out_bmp, int_sizes) * 64,
             );
             let cost_cap_increment: usize = 8;
             assert!(
@@ -557,7 +568,6 @@ impl EinsumSpec {
                         }),
                     )
                 })
-                .into_iter()
                 .collect();
             while dp.last().unwrap().is_empty() {
                 for n in 2..dp[1].len() + 1 {
@@ -575,10 +585,7 @@ impl EinsumSpec {
                                         .cost
                                         .checked_add(dp_entry_2.cost)
                                         .unwrap_or(usize::MAX)
-                                        .checked_add(bitmap_and_sizes_to_size(
-                                            union,
-                                            &self.int_sizes,
-                                        ))
+                                        .checked_add(bitmap_and_sizes_to_size(union, int_sizes))
                                         .unwrap_or(usize::MAX);
                                     let new_operands = s1 | s2;
                                     let op_set_already = dp_here_vec[0].contains_key(&new_operands);
@@ -613,7 +620,7 @@ impl EinsumSpec {
                                             &tensor_bmps_used,
                                             out_bmp,
                                         );
-                                        if bitmap_and_sizes_to_size(new_indices, &self.int_sizes)
+                                        if bitmap_and_sizes_to_size(new_indices, int_sizes)
                                             < mem_cap
                                             && (op_set_already
                                                 || dp_here_vec[0].len() < hash_map_cap)
@@ -655,7 +662,7 @@ impl EinsumSpec {
             &tensor_bmps_used,
             all_operands_bmp,
             out_bmp,
-            &self.int_sizes,
+            int_sizes,
         );
         // unsafe{breakpoint()}
         let ssa_result = contraction_to_ssa_ids(&overall_contraction, tensor_ints.len());
@@ -680,7 +687,7 @@ impl EinsumSpec {
             bail!("rust einsum optimization only supports einsum ints up to 63")
         }
         let (mut result, new_input_ints) =
-            self.optimize_pre_contract_subsets(dont_contract_subsets_into);
+            self.optimize_pre_contract_subsets(dont_contract_subsets_into)?;
         // this is ugly, todo fix
         if new_input_ints.len() == 1 && result.len() > 1 {
             result.pop();
@@ -708,7 +715,8 @@ impl EinsumSpec {
 // non python methods
 impl EinsumSpec {
     pub fn int_map(&self, ints_map: &HashMap<usize, usize>) -> EinsumSpec {
-        let mut new_int_sizes: Vec<usize> = vec![0; *ints_map.values().max().unwrap_or(&0) + 1];
+        let mut new_int_sizes: Vec<Size> =
+            vec![Size::known(0); *ints_map.values().max().unwrap_or(&0) + 1];
         for (k, v) in ints_map {
             new_int_sizes[*v] = self.int_sizes[*k];
         }
@@ -731,6 +739,17 @@ impl EinsumSpec {
             int_sizes: new_int_sizes,
         };
     }
+    pub fn from_known(
+        input_ints: Vec<Vec<usize>>,
+        output_ints: Vec<usize>,
+        int_sizes: Vec<usize>,
+    ) -> Self {
+        return Self {
+            input_ints,
+            output_ints,
+            int_sizes: int_sizes.iter().map(|x| Size::known(*x)).collect(),
+        };
+    }
 }
 
 static OPT_EINSUM_CACHE: Lazy<
@@ -747,6 +766,7 @@ static OPT_EINSUM_CACHE: Lazy<
         >,
     >,
 > = Lazy::new(|| Mutex::new(HashMap::default()));
+
 #[pyfunction]
 pub fn optimize_einsum_spec_cached(
     spec: EinsumSpec,
@@ -757,9 +777,9 @@ pub fn optimize_einsum_spec_cached(
 ) -> Result<Vec<Vec<usize>>> {
     let key = (
         spec.clone(),
-        check_outer.clone(),
-        mem_cap.clone(),
-        hash_map_cap.clone(),
+        check_outer,
+        mem_cap,
+        hash_map_cap,
         dont_contract_subsets_into.clone(),
     );
     if let Some(r) = OPT_EINSUM_CACHE.lock().unwrap().get(&key) {
