@@ -1,6 +1,6 @@
 use std::{fmt, ops::Deref};
 
-use anyhow::{Context, Result};
+use anyhow::{bail, Context, Result};
 use macro_rules_attribute::apply;
 use pyo3::{
     once_cell::GILLazyPy,
@@ -11,7 +11,8 @@ use pyo3::{
 use rustc_hash::FxHashMap as HashMap;
 
 use crate::{
-    tensor_util::{Shape, TorchDevice, TorchDeviceDtype, TorchDtype},
+    shape::{shape_into_known, shape_is_known, Shape},
+    tensor_util::{TorchDevice, TorchDeviceDtype, TorchDtype},
     util::{EinsumAxes, HashBytes, HashBytesToPy},
 };
 pub struct PyUtils {
@@ -39,6 +40,7 @@ pub struct PyUtils {
     pub builtins: Py<PyModule>,
     pub gc: Py<PyModule>,
     pub optimizing_symbolic_size_warning: PyObject,
+    pub optimizing_modules_warning: PyObject,
     pub maybe_dtype_to_maybe_string: PyObject,
     pub random_i64: PyObject,
     pub index_union_rebase: PyObject,
@@ -99,6 +101,7 @@ pub static PY_UTILS: GILLazyPy<PyUtils> = GILLazyPy::new_py(|py| {
         builtins: builtins.into(),
         gc: PyModule::import(py, "gc").unwrap().into(),
         optimizing_symbolic_size_warning: get("OptimizingSymbolicSizeWarning"),
+        optimizing_modules_warning: get("OptimizingModulesWarning"),
         maybe_dtype_to_maybe_string: get("maybe_dtype_to_maybe_string"),
         random_i64: get("random_i64"),
         index_union_rebase: get("index_union_rebase"),
@@ -131,7 +134,7 @@ macro_rules! make_py_func {
     } => {
         paste::paste!{
             $(#[$m])*
-            $vi fn [<$name _py>]<'py>($py : Python<'py>, $($arg_name : $arg_ty,)*) -> $ret_ty {
+            $vi fn [<$name _py>]($py : Python<'_>, $($arg_name : $arg_ty,)*) -> $ret_ty {
                 $($tt)*
             }
             $(#[$m])* // TODO: maybe shouldn't apply to both??
@@ -158,9 +161,16 @@ pub fn assert_tensors_close(a: Tensor, b: Tensor) -> Result<()> {
 #[apply(make_py_func)]
 #[py_ident(py)]
 pub fn scalar_to_tensor(v: f64, shape: Shape, device_dtype: TorchDeviceDtype) -> Result<Tensor> {
+    if !shape_is_known(&shape) {
+        bail!("scalar_to_tensor of shape with unknown sizes {shape:?}")
+    }
     PY_UTILS
         .scalar_to_tensor
-        .call(py, (v, shape, device_dtype), None)
+        .call(
+            py,
+            (v, PyShape(shape_into_known(shape)), device_dtype),
+            None,
+        )
         .context("scalar to tensor")?
         .extract::<Tensor>(py)
         .context("scalar to tensor extract")
@@ -169,9 +179,16 @@ pub fn scalar_to_tensor(v: f64, shape: Shape, device_dtype: TorchDeviceDtype) ->
 #[apply(make_py_func)]
 #[py_ident(py)]
 pub fn i64_to_tensor(v: i64, shape: Shape, device_dtype: TorchDeviceDtype) -> Result<Tensor> {
+    if !shape_is_known(&shape) {
+        bail!("i64_to_tensor of shape with unknown sizes {shape:?}")
+    }
     PY_UTILS
         .scalar_to_tensor
-        .call(py, (v, shape, device_dtype), None)
+        .call(
+            py,
+            (v, PyShape(shape_into_known(shape)), device_dtype),
+            None,
+        )
         .context("scalar to tensor")?
         .extract::<Tensor>(py)
         .context("scalar to tensor extract")
@@ -216,10 +233,17 @@ pub fn einops_repeat(
 
 #[apply(make_py_func)]
 #[py_ident(py)]
-pub fn un_flat_concat(tensor: &Tensor, split_shapes: Vec<Shape>) -> PyResult<Vec<Tensor>> {
+pub fn un_flat_concat(tensor: &Tensor, split_shapes: Vec<Vec<usize>>) -> PyResult<Vec<Tensor>> {
     PY_UTILS
         .un_flat_concat
-        .call(py, (tensor, split_shapes), None)
+        .call(
+            py,
+            (
+                tensor,
+                split_shapes.into_iter().map(PyShape).collect::<Vec<_>>(),
+            ),
+            None,
+        )
         .context("un flat concat")?
         .extract(py)
 }
@@ -379,7 +403,7 @@ pub struct Tensor {
     tensor: PyObject,
     pub shape: Shape, /* cache shape so doesn't have to be recomputed on reconstruct etc (not uber efficient I think) */
     hash: Option<HashBytes>,
-    device_dtype: TorchDeviceDtype,
+    pub device_dtype: TorchDeviceDtype,
 }
 
 impl fmt::Debug for Tensor {
@@ -454,7 +478,7 @@ impl Tensor {
     }
 
     pub fn hash_base16(&self) -> Option<String> {
-        self.hash.as_ref().map(|x| base16::encode_lower(x))
+        self.hash.as_ref().map(base16::encode_lower)
     }
 
     pub fn flatten(&self) -> Tensor {
@@ -500,6 +524,16 @@ impl Tensor {
             })
         }
     }
+
+    pub fn max_usize(&self) -> Result<usize> {
+        Python::with_gil(|py| {
+            Ok(self
+                .tensor
+                .call_method1(py, "max", ())?
+                .call_method1(py, "item", ())?
+                .extract(py)?)
+        })
+    }
 }
 
 #[pyfunction]
@@ -507,8 +541,8 @@ pub fn hash_tensor(t: Tensor) -> Result<HashBytesToPy> {
     Ok((*t.hashed()?.hash().unwrap()).into())
 }
 
-#[derive(FromPyObject)]
-pub struct PyShape(pub Shape);
+#[derive(FromPyObject, Clone, Debug)]
+pub struct PyShape(pub Vec<usize>);
 
 impl IntoPy<PyObject> for PyShape {
     fn into_py(self, py: Python<'_>) -> PyObject {
@@ -516,6 +550,15 @@ impl IntoPy<PyObject> for PyShape {
     }
 }
 
+#[derive(FromPyObject, Clone, Debug)]
+pub struct PySymShape(pub Shape);
+
+impl IntoPy<PyObject> for PySymShape {
+    fn into_py(self, py: Python<'_>) -> PyObject {
+        PyTuple::new(py, self.0).into_py(py)
+    }
+}
+
 #[derive(FromPyObject)]
 pub struct PyEinsumAxes(pub EinsumAxes);
 
