@@ -16,16 +16,18 @@ use smallvec::SmallVec as Sv;
 use thiserror::Error;
 use uuid::uuid;
 
+#[cfg(test)]
+use crate::shape::shape_from_known;
 use crate::{
-    compact_data::{TinyVecU8, U8Set},
+    compact_data::TinyVecU8,
     make_single_many,
     name::Name,
     py_types::{einops_repeat, PyOpAtAxes, Tensor, PY_CIRCUIT_ITEMS, PY_UTILS},
-    python_error_exception, sv,
-    symbolic_size::{SymbolicSizeProduct, SIZE_PROD_MATCH},
+    python_error_exception,
+    shape::{Shape, Size, SIZE_MATCH},
+    sv,
     tensor_util::{
-        check_canon_axes, check_canon_idxs, PyParseError, Shape, TensorAxisIndex, TensorIndex,
-        TorchDevice,
+        check_canon_axes, check_canon_idxs, PyParseError, TensorAxisIndex, TensorIndex, TorchDevice,
     },
     tu8v,
     util::{
@@ -33,91 +35,6 @@ use crate::{
         NamedAxes, ALPHABET,
     },
 };
-/// OpSize is a memory optimization of Option<usize> that stores a 63-bit int and one "is this none" bit
-#[derive(Copy, Clone, PartialEq, Eq)]
-pub struct OpSize(pub u64);
-pub type OpShape = Sv<[OpSize; 8]>;
-
-impl OpSize {
-    pub const SHIFT: usize = 63;
-    pub const NONE: OpSize = OpSize(1 << Self::SHIFT);
-    // sometimes into/from can't do type interference, so we have aliases
-    fn t(self) -> Option<usize> {
-        self.into()
-    }
-
-    fn f(val: Option<usize>) -> Self {
-        val.into()
-    }
-    pub fn is_some(&self) -> bool {
-        self.0 >> Self::SHIFT == 0
-    }
-    pub fn is_none(&self) -> bool {
-        self.0 >> Self::SHIFT != 0
-    }
-    pub fn unwrap(self) -> usize {
-        assert!(self.is_some());
-        self.0 as usize
-    }
-    pub fn unwrap_or(self, x: usize) -> usize {
-        if self.is_none() {
-            x
-        } else {
-            self.unwrap()
-        }
-    }
-}
-
-impl Debug for OpSize {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        self.t().fmt(f)
-    }
-}
-
-impl From<Option<usize>> for OpSize {
-    fn from(x: Option<usize>) -> Self {
-        assert!(std::mem::size_of::<usize>() <= std::mem::size_of::<u64>()); // optimized out I assume
-        match x {
-            None => Self::NONE,
-            Some(value) => {
-                assert_eq!(value >> Self::SHIFT, 0);
-                OpSize(value as u64)
-            }
-        }
-    }
-}
-
-impl From<usize> for OpSize {
-    fn from(x: usize) -> Self {
-        Self::from(Some(x))
-    }
-}
-
-impl From<OpSize> for Option<usize> {
-    fn from(x: OpSize) -> Self {
-        if x.0 >> OpSize::SHIFT == 1 {
-            None
-        } else {
-            Some(x.0 as usize)
-        }
-    }
-}
-
-impl<'source> FromPyObject<'source> for OpSize {
-    fn extract(obj: &'source PyAny) -> PyResult<Self> {
-        Ok(OpSize::f(obj.extract()?))
-    }
-}
-
-impl IntoPy<PyObject> for OpSize {
-    fn into_py(self, py: Python<'_>) -> PyObject {
-        self.t().into_py(py)
-    }
-}
-
-pub fn shape_to_op_shape(shape: &Shape) -> OpShape {
-    shape.iter().map(|x| OpSize(*x as u64)).collect()
-}
 
 // RInnerInts distribution is heavy tailed - almost everything's 1, but there might be up to 10.
 // we use multiple of 8 bc they're u8 and we'd just pad if we used less
@@ -132,7 +49,7 @@ pub struct RearrangeSpec {
     #[pyo3(get)]
     pub output_ints: RInts,
     #[pyo3(get)]
-    pub int_sizes: OpShape,
+    pub int_sizes: Shape,
 }
 
 make_single_many!(PyCountsAtAxes, usize, Vec);
@@ -144,7 +61,7 @@ pub fn check_permutation<T: Eq + hash::Hash + Clone + Into<usize>>(perm: &[T]) -
             ints: perm.iter().cloned().map(Into::into).collect()
         })
     }
-    if &perm_set != &(0..perm.len()).collect::<HashSet<_>>() {
+    if perm_set != (0..perm.len()).collect::<HashSet<_>>() {
         bail!(PermError::NotContiguousInts {
             ints: perm.iter().cloned().map(Into::into).collect(),
             count: perm.len(),
@@ -163,7 +80,7 @@ pub fn check_canon_axes_as(rank: usize, axes: &[i64]) -> Result<RInnerInts> {
 impl fmt::Display for RearrangeSpec {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         // plausible that default display shouldn't show_size_if_present, but debug should. (idk)
-        f.write_str(&self.to_string_impl(true, true, true))
+        f.write_str(&self.to_string_impl(true, true))
     }
 }
 impl fmt::Debug for RearrangeSpec {
@@ -180,7 +97,7 @@ impl FromStr for RearrangeSpec {
             Regex::new(&format!(
                 r"({p})|({a}\w*(?::{p})?)|\(\s*((?:({a}\w*(?::{p})?|{p})\s*)*)\)",
                 a = ALPHA_NON_NUMERIC,
-                p = *SIZE_PROD_MATCH
+                p = *SIZE_MATCH
             ))
             .unwrap()
         });
@@ -194,7 +111,7 @@ impl FromStr for RearrangeSpec {
                 });
             };
 
-        let mut sizes: OpShape = sv![];
+        let mut sizes: Shape = sv![];
         let mut string_to_int: HashMap<String, u8> = HashMap::default();
         let new_int = |s: usize| {
             assert!(s > 0);
@@ -206,14 +123,12 @@ impl FromStr for RearrangeSpec {
             Ok((s - 1) as u8)
         };
         let mut f = |s: &str| -> Result<RInts> {
-            let handle_int = |num, sizes: &mut OpShape| {
-                sizes.push(Some(num).into());
+            let handle_size = |num: Size, sizes: &mut Shape| {
+                sizes.push(num);
                 new_int(sizes.len())
             };
-            let parse_num = |s: &str| {
-                SymbolicSizeProduct::parse_to_usize(s).context("rearrange num didn't parse")
-            };
-            let mut handle_string_nosize = |s: String, sizes: &mut OpShape| {
+            let parse_num = |s: &str| Ok(Size::from_str(s)?);
+            let mut handle_string_nosize = |s: String, sizes: &mut Shape| {
                 if let Some(old) = string_to_int.get(&s) {
                     Ok(*old)
                 } else {
@@ -223,11 +138,11 @@ impl FromStr for RearrangeSpec {
                     Ok(new_int)
                 }
             };
-            let mut handle_string = |string: String, sizes: &mut OpShape| -> Result<u8> {
+            let mut handle_string = |string: String, sizes: &mut Shape| -> Result<u8> {
                 if let [string, num] = string.split(':').collect::<Vec<_>>()[..] {
                     let num = parse_num(num)?;
                     let result = handle_string_nosize(string.to_owned(), sizes)?;
-                    sizes[result as usize] = Some(num).into();
+                    sizes[result as usize] = num;
                     Ok(result)
                 } else {
                     handle_string_nosize(string, sizes)
@@ -238,7 +153,7 @@ impl FromStr for RearrangeSpec {
                 .map(|tl| {
                     if let Some(c) = tl.get(1) {
                         let num = parse_num(c.as_str())?;
-                        Ok(tu8v![handle_int(num, &mut sizes)?])
+                        Ok(tu8v![handle_size(num, &mut sizes)?])
                     } else if let Some(c) = tl.get(2) {
                         Ok(tu8v![handle_string(c.as_str().to_owned(), &mut sizes)?])
                     } else if let Some(c) = tl.get(3) {
@@ -247,7 +162,7 @@ impl FromStr for RearrangeSpec {
                             .filter(|z| !z.is_empty())
                             .map(|s| {
                                 parse_num(s)
-                                    .and_then(|z| handle_int(z, &mut sizes))
+                                    .and_then(|z| handle_size(z, &mut sizes))
                                     .or_else(|_e| handle_string(s.to_owned(), &mut sizes))
                             })
                             .collect::<Result<_>>()
@@ -261,7 +176,7 @@ impl FromStr for RearrangeSpec {
         };
 
         Self::new(f(input)?, f(output)?, sizes)
-            .context("failed to construct rearrange spec after parsing")
+            .with_context(|| format!("failed to construct rearrange spec after parsing {string}"))
     }
 }
 
@@ -281,7 +196,7 @@ impl FromStr for RearrangeSpec {
 #[pymethods]
 impl RearrangeSpec {
     #[new]
-    pub fn new(input_ints: RInts, output_ints: RInts, int_sizes: OpShape) -> Result<Self> {
+    pub fn new(input_ints: RInts, output_ints: RInts, int_sizes: Shape) -> Result<Self> {
         let out = RearrangeSpec {
             input_ints,
             output_ints,
@@ -306,7 +221,7 @@ impl RearrangeSpec {
         Self::new(
             (0..rank).map(|i| tu8v![i]).collect(),
             sv![(0..rank).collect()],
-            sv![OpSize::from(None); rank as usize],
+            sv![Size::NONE; rank as usize],
         )
         .unwrap()
     }
@@ -318,7 +233,7 @@ impl RearrangeSpec {
         Ok(Self::new(
             sv![(0..rank as u8).collect()],
             (0..rank).map(|i| tu8v![i as u8]).collect(),
-            shape_to_op_shape(&shape),
+            shape,
         )
         .unwrap())
     }
@@ -340,9 +255,9 @@ impl RearrangeSpec {
             .chain(axis as u8..(ndim - 1) as u8)
             .map(|i| tu8v![i])
             .collect();
-        let op_shape: OpShape = iter::repeat(OpSize::NONE)
+        let op_shape: Shape = iter::repeat(Size::NONE)
             .take(ndim - 1)
-            .chain(shape_to_op_shape(&shape))
+            .chain(shape)
             .collect();
 
         assert_eq!(op_shape.len(), out_ints.len());
@@ -356,7 +271,7 @@ impl RearrangeSpec {
         Self::new(
             (0..rank).map(|i| tu8v![i]).collect(),
             (0..rank).map(|i| tu8v![i]).collect(),
-            sv![OpSize::from(None); rank as usize],
+            sv![Size::NONE; rank as usize],
         )
         .unwrap()
     }
@@ -390,9 +305,8 @@ impl RearrangeSpec {
                 .map(|i| tu8v![i as u8])
                 .collect(),
             new_shape
-                .iter()
-                .map(|i| OpSize(*i as u64))
-                .chain(std::iter::repeat(OpSize::NONE).take(old_rank))
+                .into_iter()
+                .chain(std::iter::repeat(Size::NONE).take(old_rank))
                 .collect(),
         )
         .expect("Internal error - this should always work");
@@ -434,7 +348,7 @@ impl RearrangeSpec {
         let out = Self::new(
             (0..permutation.len() as u8).map(|i| tu8v![i]).collect(),
             permutation.iter().map(|i| tu8v![*i as u8]).collect(),
-            sv![OpSize::NONE;permutation.len()],
+            sv![Size::NONE;permutation.len()],
         )
         .unwrap();
 
@@ -449,7 +363,7 @@ impl RearrangeSpec {
         let out = Self::new(
             (0..count as u8).map(|i| tu8v![i]).collect(),
             permute_combine,
-            sv![OpSize::NONE; count],
+            sv![Size::NONE; count],
         )
         .unwrap();
         Ok(out)
@@ -511,30 +425,18 @@ impl RearrangeSpec {
     }
 
     pub fn to_einops_string(&self) -> String {
-        self.to_string_impl(false, false, false)
+        self.to_string_impl(false, false)
     }
 
     #[pyo3(name = "to_string", signature = (
         size_on_letter = true,
-        use_symbolic = true,
         show_size_if_present = true
     ))]
-    pub fn to_string_impl(
-        &self,
-        size_on_letter: bool,
-        use_symbolic: bool,
-        show_size_if_present: bool,
-    ) -> String {
+    pub fn to_string_impl(&self, size_on_letter: bool, show_size_if_present: bool) -> String {
         let ints_needing_sizes = (!show_size_if_present).then(|| self.ints_needing_sizes());
         let ints_needing_sizes = ints_needing_sizes.as_ref();
 
-        let int_to_str = |i: usize| {
-            if use_symbolic {
-                SymbolicSizeProduct::from(i).to_string()
-            } else {
-                i.to_string()
-            }
-        };
+        let int_to_str = |i: usize| i.to_string();
 
         let get_str = |ints: &RInts, ints_in_other: &HashSet<AxisInt>| {
             ints.iter()
@@ -550,13 +452,13 @@ impl RearrangeSpec {
                                     format!(
                                         "{}:{}",
                                         ALPHABET[*i as usize].to_owned(),
-                                        int_to_str(self.int_sizes[*i as usize].unwrap() as usize)
+                                        int_to_str(self.int_sizes[*i as usize].unwrap())
                                     )
                                 } else {
                                     ALPHABET[*i as usize].to_owned()
                                 }
                             } else {
-                                int_to_str(self.int_sizes[*i as usize].unwrap() as usize)
+                                int_to_str(self.int_sizes[*i as usize].unwrap())
                             }
                         })
                         .collect::<Vec<String>>();
@@ -653,7 +555,7 @@ impl RearrangeSpec {
             self.int_sizes
                 .iter()
                 .cloned()
-                .chain(vec![OpSize::f(None); batch_rank].into_iter())
+                .chain(vec![Size::NONE; batch_rank].into_iter())
                 .collect(),
         )
         .unwrap()
@@ -672,7 +574,7 @@ impl RearrangeSpec {
     /// apply rearrange to tensor
     pub fn apply(&self, tensor: Tensor) -> Result<Tensor> {
         let (string, letter_sizes) = self
-            .conform_to_input_shape(&tensor.shape())
+            .conform_to_input_shape(tensor.shape())
             .context("conform to input shape failed in apply")?
             .to_einops_string_and_letter_sizes();
         einops_repeat(&tensor, string, letter_sizes)
@@ -681,37 +583,14 @@ impl RearrangeSpec {
         format!("{:?}", self)
     }
     pub fn shapes(&self) -> Result<(Shape, Shape)> {
-        let used_ints: U8Set = self
-            .input_ints
-            .iter()
-            .chain(&self.output_ints)
-            .flatten()
-            .cloned()
-            .collect();
-        if self
-            .int_sizes
-            .iter()
-            .enumerate()
-            .any(|(i, x)| x.is_none() && used_ints.contains(i as u8))
-        {
-            bail!(RearrangeSpecError::HasWildcardSizes { spec: self.clone() });
-        }
         Ok((
             self.input_ints
                 .iter()
-                .map(|x| {
-                    x.iter()
-                        .map(|y| self.int_sizes[*y as usize].0 as usize)
-                        .product()
-                })
+                .map(|x| x.iter().map(|y| self.int_sizes[*y as usize]).product())
                 .collect(),
             self.output_ints
                 .iter()
-                .map(|x| {
-                    x.iter()
-                        .map(|y| self.int_sizes[*y as usize].0 as usize)
-                        .product()
-                })
+                .map(|x| x.iter().map(|y| self.int_sizes[*y as usize]).product())
                 .collect(),
         ))
     }
@@ -732,38 +611,48 @@ impl RearrangeSpec {
     /// # Examples
     /// 1. if special_case_ones, remove all indices of size 1
     /// ```
-    /// # use rr_util::{sv,tu8v, rearrange_spec::{OpSize, RearrangeSpec}};
+    /// # use rr_util::{sv,tu8v, rearrange_spec::RearrangeSpec, shape::Size};
     /// let spec = RearrangeSpec::new(
     ///     sv![tu8v![0, 1], tu8v![2, 3]],
     ///     sv![tu8v![3, 2, 1, 0]],
-    ///     sv![OpSize(3), OpSize(1), OpSize(1), OpSize(1)],
+    ///     sv![
+    ///         Size::known(3),
+    ///         Size::known(1),
+    ///         Size::known(1),
+    ///         Size::known(1)
+    ///     ],
     /// )
     /// .unwrap();
     /// let canonicalized_spec =
-    ///     RearrangeSpec::new(sv![tu8v![0], tu8v![]], sv![tu8v![0]], sv![OpSize(3)]).unwrap();
+    ///     RearrangeSpec::new(sv![tu8v![0], tu8v![]], sv![tu8v![0]], sv![Size(3)]).unwrap();
     /// assert_eq!(spec.canonicalize(true), canonicalized_spec);
     /// ```
     /// 2. Merge all sequences of indices that always appear together in the same order inside parentheses
     /// 3. Renumber the int indices sequentially based on when they first appear, inputs first then outputs
     /// ```
-    /// # use rr_util::{sv,tu8v, rearrange_spec::{OpSize, RearrangeSpec}};
+    /// # use rr_util::{sv,tu8v, rearrange_spec::RearrangeSpec, shape::Size};
     /// let spec = RearrangeSpec::new(
     ///     sv![tu8v![0], tu8v![1, 2], tu8v![3, 4, 5]],
     ///     sv![tu8v![1, 2], tu8v![5], tu8v![3, 4], tu8v![0]],
     ///     sv![
-    ///         OpSize(2),
-    ///         OpSize(2),
-    ///         OpSize(2),
-    ///         OpSize(2),
-    ///         OpSize(2),
-    ///         OpSize(2),
+    ///         Size::known(2),
+    ///         Size::known(2),
+    ///         Size::known(2),
+    ///         Size::known(2),
+    ///         Size::known(2),
+    ///         Size::known(2),
     ///     ],
     /// )
     /// .unwrap();
     /// let canonicalized_spec = RearrangeSpec::new(
     ///     sv![tu8v![0], tu8v![1], tu8v![2, 3]],
     ///     sv![tu8v![1], tu8v![3], tu8v![2], tu8v![0]],
-    ///     sv![OpSize(2), OpSize(4), OpSize(4), OpSize(2)],
+    ///     sv![
+    ///         Size::known(2),
+    ///         Size::known(4),
+    ///         Size::known(4),
+    ///         Size::known(2)
+    ///     ],
     /// )
     /// .unwrap();
     /// assert_eq!(spec.canonicalize(true), canonicalized_spec);
@@ -845,7 +734,7 @@ impl RearrangeSpec {
         }
 
         let mut map: HashMap<AxisInt, AxisInt> = HashMap::default();
-        let mut new_int_sizes: OpShape = sv![];
+        let mut new_int_sizes: Shape = sv![];
         let mut already_accounted_shape = vec![false; int_sizes.len()];
 
         let mut merge_tuples_and_renumber_fn = |ints: RInts| {
@@ -864,10 +753,9 @@ impl RearrangeSpec {
                             if let Some(old_size) = int_sizes[i as usize] {
                                 let new_int_sizes_idx =
                                     *map.get(&find(i, &mut prev_idx)).unwrap() as usize;
-                                new_int_sizes[new_int_sizes_idx] = Some(
-                                    new_int_sizes[new_int_sizes_idx].unwrap() as usize * old_size,
-                                )
-                                .into();
+                                new_int_sizes[new_int_sizes_idx] =
+                                    Some(new_int_sizes[new_int_sizes_idx].unwrap() * old_size)
+                                        .into();
                             }
                             already_accounted_shape[i as usize] = true;
                         }
@@ -935,7 +823,7 @@ impl RearrangeSpec {
                 .int_sizes
                 .iter()
                 .cloned()
-                .chain(vec![OpSize(1); (next_int - first_int_to_add) as usize])
+                .chain(vec![Size::known(1); (next_int - first_int_to_add) as usize])
                 .collect(),
         };
         if !allow_rust_invalid {
@@ -974,11 +862,6 @@ impl RearrangeSpec {
     }
 }
 
-pub enum ExpandToSpecOrShape {
-    Spec(RearrangeSpec),
-    SetShape(OpShape),
-}
-
 impl RearrangeSpec {
     pub fn ints_in_inp_it(&self) -> impl Iterator<Item = AxisInt> + '_ {
         self.input_ints.iter().flatten().copied()
@@ -1001,24 +884,13 @@ impl RearrangeSpec {
         Ok(Self::ident(rank as u8))
     }
 
-    pub fn new_canon(input_ints: RInts, output_ints: RInts, int_sizes: OpShape) -> Self {
+    pub fn new_canon(input_ints: RInts, output_ints: RInts, int_sizes: Shape) -> Self {
         RearrangeSpec::new(input_ints, output_ints, int_sizes)
             .unwrap()
             .canonicalize(true)
     }
 
     pub fn expand_to(&self, shape: &Shape) -> Result<RearrangeSpec> {
-        match self.expand_to_spec_or_shape(shape, true)? {
-            ExpandToSpecOrShape::Spec(spec) => Ok(spec),
-            ExpandToSpecOrShape::SetShape(_) => unreachable!(),
-        }
-    }
-
-    pub fn expand_to_spec_or_shape(
-        &self,
-        shape: &Shape,
-        error_on_ambiguous: bool,
-    ) -> Result<ExpandToSpecOrShape> {
         if shape.len() != self.input_ints.len() {
             bail!(RearrangeSpecError::NotConformable {
                 spec: self.clone(),
@@ -1026,14 +898,11 @@ impl RearrangeSpec {
             });
         }
 
-        let mut out_shape: OpShape = sv![OpSize::NONE; shape.len()];
         let mut int_sizes = self.int_sizes.clone();
-        for (dim, (inner_ints, &expand_size, out_size)) in
-            izip!(&self.input_ints, shape, &mut out_shape).enumerate()
-        {
+        for (dim, (inner_ints, &expand_size)) in izip!(&self.input_ints, shape).enumerate() {
             if inner_ints
                 .iter()
-                .any(|x| self.int_sizes[*x as usize].is_none())
+                .any(|x| !self.int_sizes[*x as usize].is_known())
             {
                 // none size, so this is fine
                 continue;
@@ -1042,33 +911,33 @@ impl RearrangeSpec {
                 .iter()
                 .map(|x| self.int_sizes[*x as usize].unwrap())
                 .product::<usize>();
-            if current_size == expand_size {
+            if expand_size.is(current_size) {
                 continue;
             }
             if inner_ints.len() == 1 {
                 // change to wild card if single non-matching int
-                int_sizes[inner_ints[0] as usize] = OpSize::NONE;
+                int_sizes[inner_ints[0] as usize] = Size::NONE;
                 continue;
             }
-            let inner_int_sizes: Vec<_> = inner_ints
-                .iter()
-                .map(|x| self.int_sizes[*x as usize].unwrap())
-                .collect();
-            let divides_size: Vec<_> = inner_int_sizes
-                .iter()
-                .map(|&x| x != 0 && expand_size % x == 0)
-                .collect();
-            let num_failing_to_divide = divides_size
-                .iter()
-                .map(|divides| (!divides) as usize)
-                .sum::<usize>();
-            if num_failing_to_divide != 1 {
-                if error_on_ambiguous {
+            if let Some(expand_size) = expand_size.t() {
+                let inner_int_sizes: Vec<_> = inner_ints
+                    .iter()
+                    .map(|x| self.int_sizes[*x as usize].unwrap())
+                    .collect();
+                let divides_size: Vec<_> = inner_int_sizes
+                    .iter()
+                    .map(|&x| x != 0 && expand_size % x == 0)
+                    .collect();
+                let num_failing_to_divide = divides_size
+                    .iter()
+                    .map(|divides| (!divides) as usize)
+                    .sum::<usize>();
+                if num_failing_to_divide != 1 {
                     bail!(RearrangeSpecError::AmbiguousExpand {
                         expand_size,
                         current_size,
                         inner_ints: inner_ints.clone(),
-                        inner_int_sizes: inner_int_sizes.clone(),
+                        inner_int_sizes,
                         divides_size,
                         num_failing_to_divide,
                         dim,
@@ -1076,25 +945,17 @@ impl RearrangeSpec {
                         spec: self.clone()
                     })
                 }
-                *out_size = Some(current_size).into();
-                continue;
+                // if we have exactly 1 symbolic idx, set this size to None
+                let non_dividing_idx = divides_size
+                    .iter()
+                    .enumerate()
+                    .find_map(|(i, b)| (!b).then_some(i))
+                    .unwrap();
+                int_sizes[inner_ints[non_dividing_idx] as usize] = Size::NONE;
             }
-
-            // if we have exactly 1 symbolic idx, set this size to None
-            let non_dividing_idx = divides_size
-                .iter()
-                .enumerate()
-                .find_map(|(i, b)| (!b).then_some(i))
-                .unwrap();
-            int_sizes[inner_ints[non_dividing_idx] as usize] = OpSize::NONE;
-        }
-        if out_shape.iter().any(|x| x.is_some()) {
-            return Ok(ExpandToSpecOrShape::SetShape(out_shape));
         }
 
-        Ok(ExpandToSpecOrShape::Spec(
-            Self::new(self.input_ints.clone(), self.output_ints.clone(), int_sizes).unwrap(),
-        ))
+        Ok(Self::new(self.input_ints.clone(), self.output_ints.clone(), int_sizes).unwrap())
     }
 
     pub fn conform_to_input_shape(&self, shape: &Shape) -> Result<RearrangeSpec> {
@@ -1106,32 +967,34 @@ impl RearrangeSpec {
         }
         let mut int_sizes = self.int_sizes.clone();
         for (ints, &l) in zip(&self.input_ints, shape) {
-            let none_indices: Vec<usize> = ints
-                .iter()
-                .enumerate()
-                .filter(|(_i, x)| self.int_sizes[**x as usize].is_none())
-                .map(|(i, _x)| i)
-                .collect();
-            let non_wildcard_size: usize = ints
-                .iter()
-                .map(|x| self.int_sizes[*x as usize].unwrap_or(1))
-                .product();
-            let has_wildcard = none_indices.len() > 0;
-            if (!has_wildcard && l != non_wildcard_size)
-                || (has_wildcard && l % non_wildcard_size != 0)
-            {
-                bail!(RearrangeSpecError::NotConformable {
-                    spec: self.clone(),
-                    shape: shape.clone(),
-                });
-            }
+            if let Some(l) = l.t() {
+                let none_indices: Vec<usize> = ints
+                    .iter()
+                    .enumerate()
+                    .filter(|(_i, x)| !self.int_sizes[**x as usize].is_known())
+                    .map(|(i, _x)| i)
+                    .collect();
+                let non_wildcard_size: usize = ints
+                    .iter()
+                    .map(|x| self.int_sizes[*x as usize].unwrap_or(1))
+                    .product();
+                let has_wildcard = !none_indices.is_empty();
+                if (!has_wildcard && l != non_wildcard_size)
+                    || (has_wildcard && l % non_wildcard_size != 0)
+                {
+                    bail!(RearrangeSpecError::NotConformable {
+                        spec: self.clone(),
+                        shape: shape.clone(),
+                    });
+                }
 
-            if has_wildcard {
-                if none_indices.len() > 1 {
-                    bail!(RearrangeSpecError::TooManyWildcardSizes { spec: self.clone() });
-                } else if none_indices.len() == 1 {
-                    int_sizes[ints[none_indices[0]] as usize] =
-                        OpSize((l / non_wildcard_size) as u64);
+                if has_wildcard {
+                    if none_indices.len() > 1 {
+                        bail!(RearrangeSpecError::TooManyWildcardSizes { spec: self.clone() });
+                    } else if none_indices.len() == 1 {
+                        int_sizes[ints[none_indices[0]] as usize] =
+                            Size::known(l / non_wildcard_size);
+                    }
                 }
             }
         }
@@ -1147,32 +1010,34 @@ impl RearrangeSpec {
         }
         let mut int_sizes = self.int_sizes.clone();
         for (ints, &l) in zip(&self.output_ints, shape) {
-            let none_indices: Vec<usize> = ints
-                .iter()
-                .enumerate()
-                .filter(|(_i, x)| self.int_sizes[**x as usize].is_none())
-                .map(|(i, _x)| i)
-                .collect();
-            let non_wildcard_size: usize = ints
-                .iter()
-                .map(|x| self.int_sizes[*x as usize].unwrap_or(1))
-                .product();
-            let has_wildcard = none_indices.len() > 0;
-            if (!has_wildcard && l != non_wildcard_size)
-                || (has_wildcard && l % non_wildcard_size != 0)
-            {
-                bail!(RearrangeSpecError::NotConformable {
-                    spec: self.clone(),
-                    shape: shape.clone(),
-                });
-            }
+            if let Some(l) = l.t() {
+                let none_indices: Vec<usize> = ints
+                    .iter()
+                    .enumerate()
+                    .filter(|(_i, x)| !self.int_sizes[**x as usize].is_known())
+                    .map(|(i, _x)| i)
+                    .collect();
+                let non_wildcard_size: usize = ints
+                    .iter()
+                    .map(|x| self.int_sizes[*x as usize].unwrap_or(1))
+                    .product();
+                let has_wildcard = !none_indices.is_empty();
+                if (!has_wildcard && l != non_wildcard_size)
+                    || (has_wildcard && l % non_wildcard_size != 0)
+                {
+                    bail!(RearrangeSpecError::NotConformable {
+                        spec: self.clone(),
+                        shape: shape.clone(),
+                    });
+                }
 
-            if has_wildcard {
-                if none_indices.len() > 1 {
-                    bail!(RearrangeSpecError::TooManyWildcardSizes { spec: self.clone() });
-                } else if none_indices.len() == 1 {
-                    int_sizes[ints[none_indices[0]] as usize] =
-                        OpSize((l / non_wildcard_size) as u64);
+                if has_wildcard {
+                    if none_indices.len() > 1 {
+                        bail!(RearrangeSpecError::TooManyWildcardSizes { spec: self.clone() });
+                    } else if none_indices.len() == 1 {
+                        int_sizes[ints[none_indices[0]] as usize] =
+                            Size::some(l / non_wildcard_size);
+                    }
                 }
             }
         }
@@ -1189,14 +1054,14 @@ impl RearrangeSpec {
         if !is_unique(&ints) {
             bail!(RearrangeSpecError::IntsNotUnique {
                 is_input: true,
-                ints: ints,
+                ints,
             })
         }
         let ints: Vec<_> = self.ints_in_out_it().collect();
         if !is_unique(&ints) {
             bail!(RearrangeSpecError::IntsNotUnique {
                 is_input: false,
-                ints: ints,
+                ints,
             })
         }
 
@@ -1292,7 +1157,7 @@ impl RearrangeSpec {
                 .map(|i| tu8v![i as u8])
                 .collect(),
             (0..output_shape.len()).map(|i| tu8v![i as u8]).collect(),
-            shape_to_op_shape(output_shape),
+            output_shape.clone(),
         )
         .unwrap()
     }
@@ -1307,7 +1172,10 @@ impl RearrangeSpec {
         RearrangeSpec::new(
             input_ints.iter().map(|i| tu8v![*i]).collect(),
             output_ints.iter().map(|i| tu8v![*i]).collect(),
-            shape_to_op_shape(&dict_to_list(&int_sizes_usize, None).into_iter().collect()),
+            dict_to_list(&int_sizes_usize, None, 0)
+                .into_iter()
+                .map(Size::some)
+                .collect(),
         )
     }
 
@@ -1324,7 +1192,7 @@ impl RearrangeSpec {
         hasher.update(&uuid!("ebd66768-bb20-43a4-8f5a-fd51e76a0333").into_bytes()); // delimit with uuid
         for size in self.int_sizes.iter() {
             hasher.update(&uuid!("ebd66768-bb20-43a4-8f5a-fd51e76a0333").into_bytes()); // delimit with uuid
-            hasher.update(&size.0.to_le_bytes());
+            hasher.update(&size.to_le_bytes());
         }
         hasher.finalize().into()
     }
@@ -1338,23 +1206,23 @@ impl RearrangeSpec {
     ///
     /// # Examples
     /// ```
-    /// # use rr_util::{sv,tu8v, rearrange_spec::{OpSize, RearrangeSpec}};
+    /// # use rr_util::{sv,tu8v, rearrange_spec::RearrangeSpec, shape::Size};
     /// let inner = RearrangeSpec::new(
     ///     sv![tu8v![0], tu8v![1]],
     ///     sv![tu8v![0, 1]],
-    ///     sv![OpSize(2), OpSize(3)],
+    ///     sv![Size::known(2), Size::known(3)],
     /// )
     /// .unwrap();
     /// let outer = RearrangeSpec::new(
     ///     sv![tu8v![0]],
     ///     sv![tu8v![0], tu8v![1]],
-    ///     sv![OpSize(6), OpSize(4)],
+    ///     sv![Size::known(6), Size::known(4)],
     /// )
     /// .unwrap();
     /// let fused = RearrangeSpec::new(
     ///     sv![tu8v![0], tu8v![1]],
     ///     sv![tu8v![0, 1], tu8v![2]],
-    ///     sv![OpSize(2), OpSize(3), OpSize(4)],
+    ///     sv![Size::known(2), Size::known(3), Size::known(4)],
     /// )
     /// .unwrap();
     /// assert_eq!(RearrangeSpec::fuse(&inner, &outer).unwrap(), fused);
@@ -1483,10 +1351,7 @@ impl RearrangeSpec {
         Ok(Self::new(
             input_ints,
             output_ints,
-            outer_sizes
-                .iter()
-                .map(|&x| OpSize::from(Some(x as usize)))
-                .collect(),
+            outer_sizes.iter().map(|&x| Size::some(x)).collect(),
         )
         .unwrap()
         .canonicalize(true))
@@ -1564,10 +1429,10 @@ impl RearrangeSpec {
         Ok(Self::new(
             input_ints.into_iter().map(|i| tu8v![i as u8]).collect(),
             output_ints.into_iter().map(|i| tu8v![i as u8]).collect(),
-            std::iter::repeat(OpSize::NONE)
+            std::iter::repeat(Size::NONE)
                 .take(orig_ndim)
                 .map(Ok)
-                .chain(sorted_counts.iter().map(|x| Ok(OpSize((*x).try_into()?))))
+                .chain(sorted_counts.iter().map(|x| Ok(Size::some(*x))))
                 .collect::<Result<_>>()?,
         )
         .unwrap())
@@ -1635,7 +1500,7 @@ impl RearrangeSpec {
                 .int_sizes
                 .iter()
                 .enumerate()
-                .map(|(i, l)| int_indices[&(i as u8)].axis_len(l.unwrap()).into())
+                .map(|(i, l)| Some(int_indices[&(i as u8)].axis_len(l.unwrap())).into())
                 .collect(),
             input_ints: self
                 .input_ints
@@ -1689,60 +1554,38 @@ impl RearrangeSpec {
 
 #[test]
 fn test_expand_to() {
+    let s = shape_from_known;
+
     let spec = RearrangeSpec::from_str("a->a").unwrap();
-    assert_eq!(spec.expand_to(&[8].into_iter().collect()).unwrap(), spec);
-    assert_eq!(spec.expand_to(&[2].into_iter().collect()).unwrap(), spec);
+    assert_eq!(spec.expand_to(&s(&[8])).unwrap(), spec);
+    assert_eq!(spec.expand_to(&s(&[2])).unwrap(), spec);
 
     let spec_sized = RearrangeSpec::from_str("a:5->a").unwrap();
-    assert_eq!(
-        spec_sized.expand_to(&[5].into_iter().collect()).unwrap(),
-        spec_sized
-    );
-    assert_eq!(
-        spec_sized.expand_to(&[2].into_iter().collect()).unwrap(),
-        spec
-    );
-    assert_eq!(
-        spec_sized.expand_to(&[6].into_iter().collect()).unwrap(),
-        spec
-    );
+    assert_eq!(spec_sized.expand_to(&s(&[5])).unwrap(), spec_sized);
+    assert_eq!(spec_sized.expand_to(&s(&[2])).unwrap(), spec);
+    assert_eq!(spec_sized.expand_to(&s(&[6])).unwrap(), spec);
 
     let spec = RearrangeSpec::from_str("(a:5) b (k:2 y:7)-> (k a b) y").unwrap();
+    assert_eq!(spec.expand_to(&s(&[5, 3, 14])).unwrap(), spec);
     assert_eq!(
-        spec.expand_to(&[5, 3, 14].into_iter().collect()).unwrap(),
-        spec
-    );
-    assert_eq!(
-        spec.expand_to(&[2, 3, 14].into_iter().collect()).unwrap(),
+        spec.expand_to(&s(&[2, 3, 14])).unwrap(),
         RearrangeSpec::from_str("a b (k:2 y:7)-> (k a b) y").unwrap()
     );
     assert_eq!(
-        spec.expand_to(&[7, 3, 14].into_iter().collect()).unwrap(),
+        spec.expand_to(&s(&[7, 3, 14])).unwrap(),
         RearrangeSpec::from_str("a b (k:2 y:7)-> (k a b) y").unwrap()
     );
     assert_eq!(
-        spec.expand_to(&[7, 3, 21].into_iter().collect()).unwrap(),
+        spec.expand_to(&s(&[7, 3, 21])).unwrap(),
         RearrangeSpec::from_str("a b (k y:7)-> (k a b) y").unwrap()
     );
     assert_eq!(
-        spec.expand_to(&[7, 3, 18].into_iter().collect()).unwrap(),
+        spec.expand_to(&s(&[7, 3, 18])).unwrap(),
         RearrangeSpec::from_str("a b (k:2 y)-> (k a b) y").unwrap()
     );
-    assert!(spec.expand_to(&[5, 3, 5].into_iter().collect()).is_err());
-    assert!(spec
-        .expand_to(&[2, 5, 3, 14].into_iter().collect())
-        .is_err());
-    assert!(spec.expand_to(&[5, 3, 28].into_iter().collect()).is_err());
-}
-
-fn fancy_size_format<'a, T: IntoIterator<Item = &'a usize>>(x: T) -> String {
-    format!(
-        "[{}]",
-        x.into_iter()
-            .map(|s| SymbolicSizeProduct::from(*s).to_string())
-            .collect::<Vec<_>>()
-            .join(",")
-    )
+    assert!(spec.expand_to(&s(&[5, 3, 5])).is_err());
+    assert!(spec.expand_to(&s(&[2, 5, 3, 14])).is_err());
+    assert!(spec.expand_to(&s(&[5, 3, 28])).is_err());
 }
 
 #[apply(python_error_exception)]
@@ -1772,17 +1615,15 @@ pub enum RearrangeSpecError {
     },
 
     #[error(
-        "expanding to expand_size={fancy_size} (!= current_size={fancy_current_size}) {} {}\n{s}\n{c}\n({e_name})",
+        "expanding to expand_size={expand_size} (!= current_size={current_size}) {} {}\n{s}\n{c}\n({e_name})",
         "was ambiguous for",
         format!(
-            "inner_ints={:?}, inner_int_sizes={}, divides_size={:?}, num_failing_to_divide={}!=1",
+            "inner_ints={:?}, inner_int_sizes={:?}, divides_size={:?}, num_failing_to_divide={}!=1",
             inner_ints,
-            fancy_size_format(inner_int_sizes.iter()),
+            inner_int_sizes,
             divides_size,
             num_failing_to_divide
         ),
-        fancy_size = SymbolicSizeProduct::from(*expand_size).to_string(),
-        fancy_current_size =  SymbolicSizeProduct::from(*current_size).to_string(),
         s = concat!(
             "ambiguous expansion happens if:\n",
             "  - there aren't any wildcards\n",
@@ -1794,7 +1635,7 @@ pub enum RearrangeSpecError {
         c = format!(
             "dim={}, shape={:?}, spec={}",
             dim,
-            fancy_size_format(shape.iter()),
+            shape,
             spec
         )
     )]
@@ -1837,13 +1678,13 @@ pub enum RearrangeSpecError {
     #[error("int which is only in output doesn't have size, only_in_output={only_in_output:?}, corresponding_sizes={corresponding_sizes:?}, int_sizes={int_sizes:?} ({e_name})")]
     IntOnlyInOutputWithoutSize {
         only_in_output: Vec<AxisInt>,
-        corresponding_sizes: Vec<OpSize>,
-        int_sizes: OpShape,
+        corresponding_sizes: Vec<Size>,
+        int_sizes: Shape,
     },
 
     // TODO: improve error message to highlight failure
     #[error("ints={ints:?} int_sizes={int_sizes:?} ({e_name})")]
-    InputAxisHasMultipleWildCards { ints: RInts, int_sizes: OpShape },
+    InputAxisHasMultipleWildCards { ints: RInts, int_sizes: Shape },
 
     // TODO: improve error message to highlight failure
     #[error("len_shape={len_shape} > 255 ({e_name})")]
