@@ -5,7 +5,7 @@ use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use smallvec::{Array, SmallVec as Sv};
 use uuid::Uuid;
 
-use crate::{opt_einsum::EinsumSpec, rearrange_spec::OpSize};
+use crate::{compact_data::OpSize, opt_einsum::EinsumSpec, shape::Size};
 
 pub trait RustRepr {
     /// Serialize objects to Rust expression source code
@@ -92,6 +92,15 @@ impl<T: RustRepr> RustRepr for Vec<T> {
         format!("vec![{}]", strings.join(","))
     }
 }
+impl RustRepr for Size {
+    fn repr(&self) -> String {
+        if !self.is_known() {
+            "Size::NONE".to_owned()
+        } else {
+            format!("Size::known({})", self.unwrap())
+        }
+    }
+}
 impl<T: Array> RustRepr for Sv<T>
 where
     T::Item: RustRepr,
@@ -154,7 +163,7 @@ pub fn rustfmt_string(string: &str) -> String {
     let templated = format!("fn main(){{\n{}\n}}", string);
     let tempfilename = "/tmp/rustfmt_string_temp";
     let mut tempfile = File::create(tempfilename).unwrap();
-    tempfile.write(templated.as_bytes()).unwrap();
+    tempfile.write_all(templated.as_bytes()).unwrap();
     let tempfile = File::open(tempfilename).unwrap();
     String::from_utf8(
         std::process::Command::new("rustfmt")
