@@ -22,7 +22,7 @@ use crate::{
 
 static TENSOR_CACHE_DIR: Lazy<String> = Lazy::new(|| {
     std::env::var("TENSORS_CACHE_DIR")
-        .unwrap_or(std::env::var("HOME").unwrap() + &"/tensors_by_hash_cache")
+        .unwrap_or(std::env::var("HOME").unwrap() + "/tensors_by_hash_cache")
 });
 static UNSYNCED_LIST: Lazy<String> = Lazy::new(|| TENSOR_CACHE_DIR.clone() + "/.unsynced_list");
 
@@ -58,13 +58,7 @@ pub fn split_prefix(prefix: &str) -> (Vec<String>, String) {
 
 pub fn write_tensor_to_dir_tree(dir: &str, tensor: Tensor, force: bool) -> Result<bool> {
     let hash_parts = get_tensor_hash_parts(&tensor);
-    let dir = dir.to_owned()
-        + "/"
-        + &hash_parts[..SECTIONS.len() - 1]
-            .iter()
-            .cloned()
-            .collect::<Vec<_>>()
-            .join("/");
+    let dir = dir.to_owned() + "/" + &hash_parts[..SECTIONS.len() - 1].to_vec().join("/");
     fs::create_dir_all(&dir)?;
     let filename = dir + "/" + hash_parts.last().unwrap();
     if !force && fs::metadata(&filename).is_ok() {
@@ -93,10 +87,10 @@ pub fn get_only_file_in_nested_dirs(dir: &str) -> Result<String> {
             if files.len() > 1 {
                 bail!(anyhow!("prefix dir contains multiple things"));
             }
-            if files.len() == 0 {
+            if files.is_empty() {
                 bail!(anyhow!("dir has no files"));
             }
-            dir = dir + "/" + &files[0].as_ref().unwrap().file_name().to_str().unwrap();
+            dir = dir + "/" + files[0].as_ref().unwrap().file_name().to_str().unwrap();
         }
     }
 }
@@ -106,7 +100,7 @@ pub fn get_tensor_dir_tree(dir: &str, hash_prefix: &str) -> Result<Tensor> {
 }
 
 pub fn get_filename_dir_tree(dir: &str, hash_prefix: &str) -> Result<(String, String)> {
-    let (wholes, part) = split_prefix(&hash_prefix);
+    let (wholes, part) = split_prefix(hash_prefix);
     let look_dir = dir.to_owned() + "/" + &wholes.join("/");
     let files = fs::read_dir(&look_dir)
         .with_context(|| format!("tensors dir doesnt exist {}", &look_dir))?;
@@ -120,7 +114,7 @@ pub fn get_filename_dir_tree(dir: &str, hash_prefix: &str) -> Result<(String, St
         1 => {
             let longpath = look_dir
                 + "/"
-                + &matching_entries[0]
+                + matching_entries[0]
                     .as_ref()
                     .unwrap()
                     .file_name()
@@ -137,7 +131,7 @@ pub fn get_filename_dir_tree(dir: &str, hash_prefix: &str) -> Result<(String, St
             } else {
                 get_only_file_in_nested_dirs(&longpath)
             }
-            .map(|file| (file.clone(), file[dir.len()..].replace("/", "")))
+            .map(|file| (file.clone(), file[dir.len()..].replace('/', "")))
         }
         _ => {
             bail!(anyhow!("Found multiple tensors matching hash prefix"))
@@ -187,9 +181,9 @@ pub fn register_tensor_unsynced(hash: &str) -> Result<()> {
 pub fn save_tensor(tensor: Tensor, force: bool) -> Result<bool> {
     fs::create_dir_all(&*TENSOR_CACHE_DIR)?;
 
-    let result = write_tensor_to_dir_tree(&*TENSOR_CACHE_DIR, tensor.clone(), force);
+    let result = write_tensor_to_dir_tree(&TENSOR_CACHE_DIR, tensor.clone(), force);
     if force{
-        write_tensor_to_dir_tree(&get_rrfs_tensor_db_dir(), tensor.clone(), force)?;
+        write_tensor_to_dir_tree(&get_rrfs_tensor_db_dir(), tensor, force)?;
     }else if let Ok(did_write) = result && did_write{
         register_tensor_unsynced(&tensor.hashed()?.hash_base16().unwrap())?;
     }
@@ -262,14 +256,14 @@ pub fn sync_specific_tensors(tensor_hash_prefixes: Vec<String>, parallelism: usi
         let tx = tx.clone();
         pool.execute(move || {
             tx.send((move || {
-                let local_filename = get_filename_dir_tree(&*TENSOR_CACHE_DIR, &hash_prefix_str)?.0;
+                let local_filename = get_filename_dir_tree(&TENSOR_CACHE_DIR, &hash_prefix_str)?.0;
                 let remote_filename =
                     local_filename.replace(&*TENSOR_CACHE_DIR, &get_rrfs_tensor_db_dir());
-                let remote_filename_dir: Vec<_> = remote_filename.split("/").collect();
+                let remote_filename_dir: Vec<_> = remote_filename.split('/').collect();
                 let remote_filename_dir = remote_filename_dir
                     [..remote_filename_dir.len().checked_add_signed(-1).unwrap()]
                     .join("/");
-                fs::create_dir_all(&remote_filename_dir).context("failed to create dir in tree")?;
+                fs::create_dir_all(remote_filename_dir).context("failed to create dir in tree")?;
                 copy_without_setting_permissions(&local_filename, &remote_filename)
                     .context("failed to copy")?;
                 eprintln!("wrote {}", hash_prefix_str);
@@ -304,10 +298,10 @@ pub fn sync_specific_tensors(tensor_hash_prefixes: Vec<String>, parallelism: usi
 
 #[pyfunction]
 pub fn get_tensor_prefix(prefix: &str) -> Result<Tensor> {
-    get_tensor_dir_tree(&*TENSOR_CACHE_DIR, &prefix).or_else(|_e| {
-        let result = get_tensor_dir_tree(&get_rrfs_tensor_db_dir(), &prefix);
+    get_tensor_dir_tree(&TENSOR_CACHE_DIR, prefix).or_else(|_e| {
+        let result = get_tensor_dir_tree(&get_rrfs_tensor_db_dir(), prefix);
         if let Ok(t) = &result {
-            write_tensor_to_dir_tree(&*TENSOR_CACHE_DIR, t.clone(), false)?;
+            write_tensor_to_dir_tree(&TENSOR_CACHE_DIR, t.clone(), false)?;
         }
         result
     })
@@ -315,17 +309,17 @@ pub fn get_tensor_prefix(prefix: &str) -> Result<Tensor> {
 
 #[pyfunction]
 pub fn ensure_prefix_local(prefix: &str) -> Result<()> {
-    get_filename_dir_tree(&*TENSOR_CACHE_DIR, &prefix)
+    get_filename_dir_tree(&TENSOR_CACHE_DIR, prefix)
         .map(|_| ())
         .or_else(|_e| {
             let remote_filename = get_filename_dir_tree(&get_rrfs_tensor_db_dir(), prefix)?.0;
             let local_filename =
-                remote_filename.replace(&get_rrfs_tensor_db_dir(), &*TENSOR_CACHE_DIR);
-            let local_filename_dir: Vec<_> = local_filename.split("/").collect();
+                remote_filename.replace(&get_rrfs_tensor_db_dir(), &TENSOR_CACHE_DIR);
+            let local_filename_dir: Vec<_> = local_filename.split('/').collect();
             let local_filename_dir = local_filename_dir
                 [..local_filename_dir.len().checked_add_signed(-1).unwrap()]
                 .join("/");
-            fs::create_dir_all(&local_filename_dir).context("failed to create dir in tree")?;
+            fs::create_dir_all(local_filename_dir).context("failed to create dir in tree")?;
             copy_without_setting_permissions(&remote_filename, &local_filename)
                 .context("failed to copy")?;
             Ok(())
@@ -334,7 +328,7 @@ pub fn ensure_prefix_local(prefix: &str) -> Result<()> {
 
 #[pyfunction]
 pub fn migrate_tensors_from_old_dir(dir: &str) -> Result<()> {
-    let files: Vec<String> = fs::read_dir(&dir)
+    let files: Vec<String> = fs::read_dir(dir)
         .unwrap()
         .map(|x| x.unwrap().file_name().to_str().unwrap().to_owned())
         .collect();
