@@ -1,5 +1,4 @@
 use std::{
-    borrow::Borrow,
     cmp::{max, min},
     fmt::{self, Display},
     iter::zip,
@@ -8,6 +7,7 @@ use std::{
 };
 
 use anyhow::{anyhow, bail, Context, Error, Result};
+use itertools::izip;
 use macro_rules_attribute::apply;
 use pyo3::{
     exceptions,
@@ -20,21 +20,19 @@ use pyo3::{
     AsPyPointer, FromPyObject, IntoPy, PyAny, PyErr, PyObject, PyResult, Python, ToPyObject,
 };
 use rustc_hash::FxHashMap as HashMap;
-use smallvec::SmallVec as Sv;
 use thiserror::Error;
 use uuid::uuid;
 
-use super::py_types::{ExtraPySelfOps, PyShape};
+use super::py_types::ExtraPySelfOps;
 use crate::{
     atr, filter_by_variant,
     lru_cache::TensorCacheRrfs,
     py_types::{use_rust_comp, Tensor, PY_UTILS},
-    pycall, python_error_exception, sv,
-    symbolic_size::SymbolicSizeProduct,
+    pycall, python_error_exception,
+    shape::{Shape, Size},
     tensor_db::{get_tensor_prefix, save_tensor},
     util::HashBytes,
 };
-pub type Shape = Sv<[usize; 4]>; // TODO: maybe this should be struct which preserves invariant that dim shapes >= 0?
 
 /// casts tensors whith None device/dtype to specified device/dtype
 pub fn cast_tensors(tensors: &[Tensor], out_device_dtype: TorchDeviceDtype) -> Vec<Tensor> {
@@ -73,24 +71,16 @@ pub fn upcast_tensor_device_dtypes(tensors: &[Tensor]) -> Vec<Tensor> {
 }
 
 fn upcast_devices_if_needed(devices: &[TorchDevice]) -> Option<TorchDevice> {
-    if devices.len() == 0 || devices[1..].iter().all(|z| z == &devices[0]) {
+    if devices.is_empty() || devices[1..].iter().all(|z| z == &devices[0]) {
         return None;
     }
-    devices
-        .iter()
-        .cloned()
-        .filter(|z| *z != TorchDevice::Cpu)
-        .next()
+    devices.iter().cloned().find(|z| *z != TorchDevice::Cpu)
 }
 fn upcast_dtypes_if_needed(dtypes: &[TorchDtype]) -> Option<TorchDtype> {
-    if dtypes.len() == 0 || dtypes[1..].iter().all(|z| z == &dtypes[0]) {
+    if dtypes.is_empty() || dtypes[1..].iter().all(|z| z == &dtypes[0]) {
         return None;
     }
-    dtypes
-        .iter()
-        .cloned()
-        .filter(|z| *z != TorchDtype::float64)
-        .next()
+    dtypes.iter().cloned().find(|z| *z != TorchDtype::float64)
 }
 
 #[pyfunction]
@@ -124,135 +114,6 @@ pub fn upcast_tensor_devices(tensors: &[Tensor]) -> Vec<Tensor> {
     tensors.to_owned()
 }
 
-pub fn eq_or_symbolic(x: &[usize], y: &[usize]) -> bool {
-    x.len() == y.len()
-        && x.iter().zip(y).all(|(a, b)| {
-            SymbolicSizeProduct::has_symbolic(*a) || SymbolicSizeProduct::has_symbolic(*b) || a == b
-        })
-}
-
-#[crate::pyo3::pyfunction]
-#[pyo3(name = "broadcast_shapes", signature = (*shapes))]
-pub fn broadcast_shapes_py(shapes: Vec<Shape>) -> Result<PyShape> {
-    broadcast_shapes(&shapes).map(PyShape)
-}
-
-/// broadcast, but don't broadcast on 1s
-pub fn right_align_shapes(shapes: &[impl Borrow<[usize]>]) -> Result<Shape> {
-    broadcast_shapes_impl(shapes, false)
-}
-
-pub fn broadcast_shapes(shapes: &[impl Borrow<[usize]>]) -> Result<Shape> {
-    broadcast_shapes_impl(shapes, true)
-}
-
-pub fn broadcast_shapes_impl(
-    shapes: &[impl Borrow<[usize]>],
-    special_case_ones: bool,
-) -> Result<Shape> {
-    if shapes.is_empty() {
-        return Ok(sv![]);
-    }
-    if shapes.len() == 1 {
-        return Ok(shapes[0].borrow().iter().cloned().collect());
-    }
-    let ranks: Vec<usize> = shapes.iter().map(|x| x.borrow().len()).collect();
-    let out_rank = *ranks.iter().max().unwrap();
-    let mut existing = vec![None; out_rank];
-    for axis_from_end in 1..out_rank + 1 {
-        for (i, shape) in shapes.iter().enumerate() {
-            let axis_for_shape = ranks[i].checked_sub(axis_from_end);
-            match axis_for_shape {
-                None => {}
-                Some(j) => {
-                    let axis_len = shape.borrow()[j];
-                    let cur = existing[out_rank - axis_from_end];
-                    let success = cur
-                        .map(|x| {
-                            x == axis_len
-                                || (special_case_ones && x == 1 && axis_len != 0)
-                                || (special_case_ones && axis_len == 1 && x != 0)
-                        })
-                        .unwrap_or(true);
-                    if !success {
-                        bail!(MiscInputError::NotBroadcastable {
-                            shapes: shapes
-                                .iter()
-                                .map(|x| x.borrow().iter().cloned().collect())
-                                .collect(),
-                        });
-                    }
-                    existing[out_rank - axis_from_end] =
-                        Some(cur.map(|x| max(x, axis_len)).unwrap_or(axis_len));
-                }
-            }
-        }
-    }
-    Ok(existing.into_iter().map(|x| x.unwrap()).collect())
-}
-
-#[test]
-fn test_broadcast_shapes() {
-    let run = |x: &[Shape]| broadcast_shapes(x);
-    let s = |x: Shape| x;
-    assert_eq!(run(&[]).unwrap(), Shape::new());
-    assert_eq!(run(&[sv![], sv![]]).unwrap(), s(sv![]));
-    assert_eq!(run(&[sv![], sv![]]).unwrap(), s(sv![]));
-    assert_eq!(run(&[sv![1], sv![]]).unwrap(), s(sv![1]));
-    assert_eq!(run(&[sv![], sv![1]]).unwrap(), s(sv![1]));
-    assert_eq!(run(&[sv![3], sv![]]).unwrap(), s(sv![3]));
-    assert_eq!(run(&[sv![], sv![3]]).unwrap(), s(sv![3]));
-    assert_eq!(run(&[sv![1], sv![1]]).unwrap(), s(sv![1]));
-    assert_eq!(run(&[sv![3], sv![3]]).unwrap(), s(sv![3]));
-    assert_eq!(run(&[sv![1], sv![4, 3]]).unwrap(), s(sv![4, 3]));
-    assert_eq!(run(&[sv![4, 3], sv![4, 1]]).unwrap(), s(sv![4, 3]));
-    assert_eq!(run(&[sv![4, 3], sv![1, 3]]).unwrap(), s(sv![4, 3]));
-    assert_eq!(run(&[sv![4, 1], sv![4, 3]]).unwrap(), s(sv![4, 3]));
-    assert_eq!(run(&[sv![1, 3], sv![4, 3]]).unwrap(), s(sv![4, 3]));
-    assert_eq!(
-        run(&[sv![1, 3], sv![4, 3], sv![5, 4, 1]]).unwrap(),
-        s(sv![5, 4, 3])
-    );
-    assert_eq!(
-        run(&[sv![3, 4, 5], sv![3, 4, 5], sv![3, 4, 5]]).unwrap(),
-        s(sv![3, 4, 5])
-    );
-
-    assert!(run(&[sv![5], sv![6]]).is_err());
-    assert!(run(&[sv![3, 5], sv![5, 1]]).is_err());
-    assert!(run(&[sv![5, 0], sv![5, 1]]).is_err());
-    assert!(run(&[sv![5, 2], sv![5, 1]]).is_ok());
-    assert!(run(&[sv![1, 2, 2], sv![1, 2], sv![3, 2, 2, 2, 172112]]).is_err());
-
-    let run_align = |x: &[Shape]| right_align_shapes(x);
-
-    assert_eq!(run_align(&[]).unwrap(), Shape::new());
-    assert_eq!(run_align(&[sv![], sv![]]).unwrap(), s(sv![]));
-    assert_eq!(run_align(&[sv![], sv![]]).unwrap(), s(sv![]));
-    assert_eq!(run_align(&[sv![1], sv![]]).unwrap(), s(sv![1]));
-    assert_eq!(run_align(&[sv![], sv![1]]).unwrap(), s(sv![1]));
-    assert_eq!(run_align(&[sv![3], sv![]]).unwrap(), s(sv![3]));
-    assert_eq!(run_align(&[sv![], sv![3]]).unwrap(), s(sv![3]));
-    assert_eq!(run_align(&[sv![1], sv![1]]).unwrap(), s(sv![1]));
-    assert_eq!(run_align(&[sv![3], sv![3]]).unwrap(), s(sv![3]));
-    assert!(run_align(&[sv![1], sv![4, 3]]).is_err());
-    assert!(run_align(&[sv![4, 3], sv![4, 1]]).is_err());
-    assert!(run_align(&[sv![4, 3], sv![1, 3]]).is_err());
-    assert!(run_align(&[sv![4, 1], sv![4, 3]]).is_err());
-    assert!(run_align(&[sv![1, 3], sv![4, 3]]).is_err());
-    assert!(run_align(&[sv![1, 3], sv![4, 3], sv![5, 4, 1]]).is_err());
-    assert_eq!(
-        run_align(&[sv![3, 4, 5], sv![3, 4, 5], sv![3, 4, 5]]).unwrap(),
-        s(sv![3, 4, 5])
-    );
-
-    assert!(run_align(&[sv![5], sv![6]]).is_err());
-    assert!(run_align(&[sv![3, 5], sv![5, 1]]).is_err());
-    assert!(run_align(&[sv![5, 0], sv![5, 1]]).is_err());
-    assert!(run_align(&[sv![5, 2], sv![5, 1]]).is_err());
-    assert!(run_align(&[sv![1, 2, 2], sv![1, 2], sv![3, 2, 2, 2, 172112]]).is_err());
-}
-
 #[apply(python_error_exception)]
 #[base_error_name(MiscInput)]
 #[base_exception(PyValueError)]
@@ -300,7 +161,7 @@ pub enum IndexError {
         shape: Shape,
         at: usize,
         axis: TensorAxisIndex,
-        l: usize,
+        l: Size,
     },
 
     #[error("index rank too high: {index_rank} vs {node_rank} ({e_name})")]
@@ -344,16 +205,17 @@ impl Slice {
         if x < 0 {
             l.saturating_sub((-x) as usize)
         } else {
-            l.min(x as usize)
+            // don't min with l to match Index bounds checking
+            x as usize
         }
     }
 
     pub fn start_u(self, l: usize) -> usize {
-        self.start.map(|x| Self::to_unsigned_loc(x, l)).unwrap_or(0)
+        self.start.map_or(0, |x| Self::to_unsigned_loc(x, l))
     }
 
     pub fn stop_u(self, l: usize) -> usize {
-        self.stop.map(|x| Self::to_unsigned_loc(x, l)).unwrap_or(l)
+        self.stop.map_or(l, |x| Self::to_unsigned_loc(x, l))
     }
 
     pub fn to_uslice(self, l: usize) -> USlice {
@@ -363,8 +225,16 @@ impl Slice {
         }
     }
 
-    fn size(self, l: usize) -> usize {
-        self.stop_u(l).saturating_sub(self.start_u(l))
+    fn size(self, l: Size) -> Size {
+        match l.t() {
+            None => match (self.start, self.stop) {
+                (Some(x), Some(y)) if x >= 0 && y >= 0 => {
+                    Size::known(y.checked_sub(x).unwrap().try_into().unwrap())
+                }
+                _ => Size::NONE,
+            },
+            Some(l) => Size::known(self.stop_u(l).saturating_sub(self.start_u(l))),
+        }
     }
 
     fn canonicalize(self, l: usize) -> Self {
@@ -389,8 +259,8 @@ impl Slice {
         }
     }
 
-    pub fn is_identity(self, l: usize) -> bool {
-        self.start_u(l) == 0 && self.stop_u(l) == l
+    pub fn is_identity(self, l: Size) -> bool {
+        (self.start.is_none() && self.stop.is_none()) || (l.is_known() && self.size(l) == l)
     }
 }
 
@@ -400,7 +270,7 @@ impl<'source> FromPyObject<'source> for Slice {
 
         // could also use never type if that was supported : /
         let step: Option<isize> = py_slice.getattr("step").unwrap().extract()?;
-        if step != None {
+        if step.is_some() {
             return Err(PyErr::new::<exceptions::PyValueError, _>(
                 "step must be None!",
             ));
@@ -577,6 +447,7 @@ impl TensorAxisIndex {
         stop: None,
     });
 
+    // TODO should be cached
     pub fn tensor_canon(tensor: &IndexTensor) -> Self {
         pycall!(PY_UTILS.canon_index_tensor, (tensor,))
     }
@@ -588,7 +459,7 @@ impl TensorAxisIndex {
             .collect::<Option<Vec<_>>>()
         {
             containings.sort_by(|a, b| a.start.cmp(&b.start));
-            let mut result = containings[0].clone();
+            let mut result = containings[0];
             for c in &containings[1..] {
                 if c.start > result.stop {
                     return None;
@@ -615,9 +486,10 @@ impl TensorAxisIndex {
                     .collect(),
             )
         } else {
+            // TODO probably should cache?
             pycall!(
                 PY_UTILS.index_union_rebase,
-                (indices.clone(), l, device_dtype.device.clone())
+                (indices.clone(), l, device_dtype.device)
             )
         }
     }
@@ -649,11 +521,12 @@ impl TensorAxisIndex {
         self.offset(-(uslice.start as i64))
     }
 
-    pub fn canonicalize(&self, l: usize) -> Self {
+    pub fn canonicalize(&self, l: &Size) -> Self {
         match self {
-            TensorAxisIndex::Single(i) => TensorAxisIndex::Single((*i + l as i64) % l as i64),
+            TensorAxisIndex::Single(i) if let Some(l) = l.t() => TensorAxisIndex::Single((*i + l as i64) % l as i64),
+            TensorAxisIndex::Slice(sl) if let Some(l) = l.t() => TensorAxisIndex::Slice(sl.canonicalize(l)),
             TensorAxisIndex::Tensor(t) => TensorAxisIndex::tensor_canon(t),
-            TensorAxisIndex::Slice(sl) => TensorAxisIndex::Slice(sl.canonicalize(l)),
+            _ => self.clone()
         }
     }
 
@@ -664,7 +537,7 @@ impl TensorAxisIndex {
         })
     }
 
-    pub fn is_identity(&self, l: usize) -> bool {
+    pub fn is_identity(&self, l: Size) -> bool {
         match self {
             TensorAxisIndex::Slice(s) => s.is_identity(l),
             _ => false, // dont bother to check tensor, if you want that canon first
@@ -711,7 +584,7 @@ impl TensorAxisIndex {
         match self {
             TensorAxisIndex::Single(_) => 0,
             TensorAxisIndex::Slice(s) => s.to_uslice(l).length(),
-            TensorAxisIndex::Tensor(t) => t.0.shape[0],
+            TensorAxisIndex::Tensor(t) => t.0.shape[0].unwrap(),
         }
     }
 }
@@ -787,18 +660,32 @@ fn sat_as_expected() {
 }
 
 impl TensorIndex {
+    pub fn expand_to_shape(&self, old_shape: &[Size], new_shape: &[Size]) -> Self {
+        TensorIndex(
+            izip!(&self.0, new_shape, old_shape)
+                .map(|(x, news, olds)| {
+                    if !x.is_identity(*news) && x.is_identity(*olds) {
+                        TensorAxisIndex::IDENT
+                    } else {
+                        x.clone()
+                    }
+                })
+                .collect(),
+        )
+    }
+
     pub fn apply_to_shape(&self, shape: &Shape) -> Shape {
         zip(&self.0, shape)
-            .filter_map(|(idx, &l)| match idx {
+            .filter_map(|(idx, l)| match idx {
                 TensorAxisIndex::Single(_i) => None,
                 TensorAxisIndex::Tensor(t) => {
                     if t.shape().is_empty() {
                         None
                     } else {
-                        Some(t.shape()[0])
+                        Some(Size::known(t.shape()[0].unwrap()))
                     }
                 }
-                TensorAxisIndex::Slice(sl) => Some(sl.size(l)),
+                TensorAxisIndex::Slice(sl) => Some(sl.size(*l)),
             })
             .collect()
     }
@@ -807,13 +694,13 @@ impl TensorIndex {
         TensorIndex(
             zip(&self.0, shape)
                 .map(|(ax, l)| match ax {
-                    TensorAxisIndex::Slice(slice) => TensorAxisIndex::Slice(Slice {
-                        start: if slice.start_u(*l) == 0 {
+                    TensorAxisIndex::Slice(slice) if let Some(l) = l.t() => TensorAxisIndex::Slice(Slice {
+                        start: if slice.start_u(l) == 0 {
                             None
                         } else {
                             slice.start
                         },
-                        stop: if slice.stop_u(*l) == *l {
+                        stop: if slice.stop_u(l) == l {
                             None
                         } else {
                             slice.stop
@@ -825,7 +712,7 @@ impl TensorIndex {
         )
     }
 
-    pub fn is_identity(&self, shape: &[usize]) -> bool {
+    pub fn is_identity(&self, shape: &[Size]) -> bool {
         zip(&self.0, shape).all(|(idx, &l)| idx.is_identity(l))
     }
 
@@ -850,21 +737,23 @@ impl TensorIndex {
             Ok(())
         };
 
-        for (at, (idx, &l)) in zip(&self.0, shape).enumerate() {
-            match idx {
-                &TensorAxisIndex::Single(i) => check(at, Some(i), l, false)?,
-                &TensorAxisIndex::Slice(Slice { start, stop }) => {
-                    check(at, start, l, true)?;
-                    check(at, stop, l, true)?;
+        for (at, (idx, l)) in zip(&self.0, shape).enumerate() {
+            if let Some(l) = l.t() {
+                match *idx {
+                    TensorAxisIndex::Single(i) => check(at, Some(i), l, false)?,
+                    TensorAxisIndex::Slice(Slice { start, stop }) => {
+                        check(at, start, l, true)?;
+                        check(at, stop, l, true)?;
 
-                    let mod_l_idx = |x| if x < 0 { l as i64 + x } else { x };
-                    let start_u: i64 = start.map_or(0, mod_l_idx);
-                    let stop_u: i64 = stop.map_or(l as i64, mod_l_idx);
-                    if start_u > stop_u {
-                        get_err(at)?;
+                        let mod_l_idx = |x| if x < 0 { l as i64 + x } else { x };
+                        let start_u: i64 = start.map_or(0, mod_l_idx);
+                        let stop_u: i64 = stop.map_or(l as i64, mod_l_idx);
+                        if start_u > stop_u {
+                            get_err(at)?;
+                        }
                     }
+                    _ => (),
                 }
-                _ => (),
             }
         }
 
@@ -875,7 +764,7 @@ impl TensorIndex {
         self.validate(shape).expect("invalid tensor index");
         TensorIndex(
             zip(&self.0, shape)
-                .map(|(idx, &l)| idx.canonicalize(l))
+                .map(|(idx, l)| idx.canonicalize(l))
                 .collect(),
         )
     }
@@ -928,7 +817,7 @@ impl TensorIndex {
                 }
                 TensorAxisIndex::Tensor(t) => {
                     hasher.update(t.hashed().unwrap().hash().unwrap());
-                } // dont bother to check tensor, if you want that canon first
+                }
                 TensorAxisIndex::Slice(sl) => sl.update_hash(&mut hasher),
             }
         }
@@ -992,7 +881,7 @@ impl TensorIndex {
                         return Ok(TensorAxisIndex::Single(int));
                     }
                     if let Some(no_t) = x.strip_prefix('t') {
-                        let no_t_pre_bracket = no_t.split("[").next().unwrap().trim();
+                        let no_t_pre_bracket = no_t.split('[').next().unwrap().trim();
                         return tensor_cache
                             .as_mut()
                             .map(|tc| tc.get_tensor(no_t_pre_bracket.to_owned()))
@@ -1305,7 +1194,7 @@ impl TorchDtype {
         Python::with_gil(|py| {
             if !x
                 .as_ref(py)
-                .is_instance(&PY_UTILS.torch.getattr(py, "dtype")?.as_ref(py))?
+                .is_instance(PY_UTILS.torch.getattr(py, "dtype")?.as_ref(py))?
             {
                 bail!("not torch dtype")
             }
@@ -1363,7 +1252,7 @@ impl TorchDevice {
         Python::with_gil(|py| {
             if !x
                 .as_ref(py)
-                .is_instance(&PY_UTILS.torch.getattr(py, "device")?.as_ref(py))?
+                .is_instance(PY_UTILS.torch.getattr(py, "device")?.as_ref(py))?
             {
                 bail!("not torch device")
             }
@@ -1407,7 +1296,7 @@ impl TorchDeviceDtype {
         Python::with_gil(|py| {
             PY_UTILS
                 .cast_tensor
-                .call(py, (tensor, self.clone()), None)
+                .call(py, (tensor, *self), None)
                 .unwrap()
                 .extract(py)
                 .unwrap()
@@ -1415,7 +1304,7 @@ impl TorchDeviceDtype {
     }
 
     pub fn op(&self) -> TorchDeviceDtypeOp {
-        self.clone().into()
+        (*self).into()
     }
 
     pub fn __repr__(&self) -> String {
@@ -1476,8 +1365,8 @@ impl TorchDeviceDtypeOp {
             device: match (&self.device, &other.device) {
                 (Some(l), Some(r)) if l != r => {
                     bail!(MiscInputError::ChildrenMultipleDevices {
-                        a: self.device.clone(),
-                        b: other.device.clone(),
+                        a: self.device,
+                        b: other.device,
                     });
                 }
                 (l, r) => l.as_ref().or(r.as_ref()).cloned(),
@@ -1485,8 +1374,8 @@ impl TorchDeviceDtypeOp {
             dtype: match (&self.dtype, &other.dtype) {
                 (Some(l), Some(r)) if l != r => {
                     bail!(MiscInputError::ChildrenMultipleDtypes {
-                        a: self.dtype.clone(),
-                        b: other.dtype.clone(),
+                        a: self.dtype,
+                        b: other.dtype,
                     });
                 }
                 (l, r) => l.as_ref().or(r.as_ref()).cloned(),
@@ -1497,26 +1386,26 @@ impl TorchDeviceDtypeOp {
 
     pub fn unwrap_or_defaults(self) -> TorchDeviceDtype {
         TorchDeviceDtype {
-            dtype: self.dtype.unwrap_or_else(|| TorchDtype::float64),
-            device: self.device.unwrap_or_else(|| TorchDevice::Cpu),
+            dtype: self.dtype.unwrap_or(TorchDtype::float64),
+            device: self.device.unwrap_or(TorchDevice::Cpu),
         }
     }
 
     pub fn size(&self) -> usize {
-        self.clone().unwrap_or_defaults().size()
+        (*self).unwrap_or_defaults().size()
     }
     pub fn overwrite(&self, concrete: &TorchDeviceDtype) -> TorchDeviceDtype {
-        let mut result = concrete.clone();
+        let mut result = *concrete;
         if let Some(x) = &self.device {
-            result.device = x.clone();
+            result.device = *x;
         }
         if let Some(x) = &self.dtype {
-            result.dtype = x.clone();
+            result.dtype = *x;
         }
         result
     }
     pub fn override_other(&self, other: TorchDeviceDtypeOp) -> TorchDeviceDtypeOp {
-        let mut result = other.clone();
+        let mut result = other;
         if let Some(x) = &self.device {
             result.device = Some(*x);
         }
@@ -1602,17 +1491,17 @@ fn test_parse_numeric() {
     assert!(parse_numeric::<f64>("_.38748_3741").is_err());
     assert!(parse_numeric::<f64>("_1.1").is_err());
 
-    assert_eq!(parse_numeric::<usize>("38748_3741").unwrap(), 38748_3741);
+    assert_eq!(parse_numeric::<usize>("38748_3741").unwrap(), 387_483_741);
     assert_eq!(
         parse_numeric::<usize>("38748_____3741").unwrap(),
-        38748_____3741
+        387_483_741
     );
     assert_eq!(
         parse_numeric::<usize>("3_____8748_3741").unwrap(),
-        3_____8748_3741
+        387_483_741
     );
     assert_eq!(parse_numeric::<f64>("3.38473").unwrap(), 3.38473);
     assert_eq!(parse_numeric::<f64>("3.").unwrap(), 3.);
-    assert_eq!(parse_numeric::<f64>("3____.3").unwrap(), 3____.3);
-    assert_eq!(parse_numeric::<f64>("3.____3__").unwrap(), 3____.3__);
+    assert_eq!(parse_numeric::<f64>("3____.3").unwrap(), 3.3);
+    assert_eq!(parse_numeric::<f64>("3.____3__").unwrap(), 3.3);
 }
