@@ -132,11 +132,11 @@ pub fn inverse_permutation(perm: &Vec<usize>) -> Vec<usize> {
 }
 
 /// element at k in vec is v, max
-pub fn dict_to_list(dict: &HashMap<usize, usize>, max: Option<usize>) -> Vec<usize> {
+pub fn dict_to_list<T: Clone>(dict: &HashMap<usize, T>, max: Option<usize>, default: T) -> Vec<T> {
     let max = max.unwrap_or(*dict.keys().max().unwrap_or(&0));
-    let mut result = vec![0; max + 1];
+    let mut result = vec![default; max + 1];
     for (k, v) in dict.iter() {
-        result[*k] = *v;
+        result[*k] = v.clone();
     }
     result
 }
@@ -361,12 +361,12 @@ pub fn test_outer_product() {
 }
 
 // note this adds an element at the end so you can always look up arr[i+1] to get end
-pub fn cumsum<T: std::ops::AddAssign + Default + Copy>(col: &[T]) -> Vec<T> {
+pub fn cumsum<T: std::ops::Add<Output = T> + Default + Copy>(col: &[T]) -> Vec<T> {
     col.iter()
         .chain(std::iter::once(&Default::default()))
         .scan(Default::default(), |state: &mut T, el| {
             let old_state = *state;
-            *state += *el;
+            *state = *state + *el;
             Some(old_state)
         })
         .collect()
@@ -466,7 +466,7 @@ impl<T> EmptySingleMany<T> {
         if x.is_empty() {
             Self::Empty
         } else if x.len() == 1 {
-            Self::Single(x.pop().unwrap().into())
+            Self::Single(x.pop().unwrap())
         } else {
             Self::Many(x)
         }
@@ -481,7 +481,7 @@ impl<T> From<Vec<T>> for EmptySingleMany<T> {
 
 pub fn split_first_take<T: Default>(x: &mut [T]) -> Option<(T, impl Iterator<Item = T> + '_)> {
     x.split_first_mut()
-        .map(|(l, r)| (mem::take(l), r.into_iter().map(mem::take)))
+        .map(|(l, r)| (mem::take(l), r.iter_mut().map(mem::take)))
 }
 
 pub fn flip_op_result<T, E>(x: Option<Result<T, E>>) -> Result<Option<T>, E> {
@@ -748,16 +748,11 @@ pub fn transpose<T>(v: Vec<Vec<T>>, inner_len: usize) -> Vec<Vec<T>> {
         .collect()
 }
 
+#[derive(Default)]
 pub struct DimNumMaker {
     pub running: usize,
 }
 
-impl Default for DimNumMaker {
-    fn default() -> Self {
-        Self { running: 0 }
-    }
-}
-
 impl DimNumMaker {
     pub fn next(&mut self) -> usize {
         let out = self.running;
