@@ -1,7 +1,6 @@
 from __future__ import annotations
 
 import functools
-import hashlib
 import itertools
 import math
 import uuid
@@ -25,12 +24,10 @@ from ._rust import (
     Matcher,
     MatcherIn,
     ModuleArgSpec,
-    PrintOptions,
     PushDownIndexEinsumNoopError,
     Rearrange,
     RearrangeSpec,
     Scalar,
-    SetSymbolicShape,
     Symbol,
     TorchAxisIndex,
     add_elim_zeros,
@@ -60,7 +57,7 @@ from ._rust import (
     rearrange_fuse,
     traverse_until_depth,
 )
-from .py_utils import I, make_index_at
+from .py_utils import I, S, make_index_at
 
 T = TypeVar("T")
 
@@ -1071,7 +1068,7 @@ def split_einsum_concat(
         before = sum(node.children[j].shape[axis] for j in range(0, g[0]))
         end = before + sum(node.children[j].shape[axis] for j in range(g[0], g[-1] + 1))
 
-        new_slice.append(I[before:end])
+        new_slice.append(S[before:end])
 
     def run_on_index(new_idx: Index, _, name: str):
         if not push_down_through_concat:
@@ -1669,22 +1666,19 @@ def push_down_indexes_through_modules(
     syms = set(get_free_symbols(circ))
     no_parent_fvs = Matcher(lambda c: syms.isdisjoint(set(get_free_symbols(c))))
 
-    def is_sym(y):
-        return y.is_symbol() or (isinstance(y, SetSymbolicShape) and is_sym(y.children[0]))
-
-    def rename_idx_set_shape(y) -> Circuit:
-        if isinstance(y, SetSymbolicShape) or y.is_index():
-            y = y.map_child(0, rename_idx_set_shape)
+    def rename_idx(y) -> Circuit:
+        if y.is_index():
+            y = y.map_child(0, rename_idx)
             return y.rename(y.get_autoname())
         return y
 
     def update_bind(y):
-        y = rename_idx_set_shape(y)
+        y = rename_idx(y)
         if y.is_index():
             return push_down(y.cast_index())
         return y
 
-    m = Matcher(Index) & (lambda y: is_sym(y.children[0])) & no_parent_fvs
+    m = Matcher(Index) & (lambda y: y.children[0].is_symbol()) & no_parent_fvs
 
     def construct(m, old):
         m2 = module_remove_unused_inputs(
