@@ -262,7 +262,7 @@ def chain_excluding(parent: IterativeMatcher, child: IterativeMatcherIn, term_ea
     """Matches `child` from `parent`, excluding any intermediate nodes matched by `term_early_at`."""
     if term_early_at is False:
         return parent.chain(child)
-    return restrict(parent.chain(IterativeMatcher(child)), term_early_at=term_early_at)
+    return parent.chain(restrict(child, term_early_at=term_early_at))
 
 
 @attrs.define(frozen=True)
