@@ -46,7 +46,7 @@ sym.uuid
 
 # symbols aren't explicitly computable
 assert sym.is_explicitly_computable is False
-with pytest.raises(rc.TensorEvalNotExplicitlyComputableError) as exc:
+with pytest.raises(rc.SchedulingEvaluateCalledWithSymsError) as exc:
     sym.evaluate()
 print(exc.exconly())
 
@@ -180,7 +180,6 @@ print(excinfo.exconly())
 
 # %%
 
-s0, s1, s2, *_ = rc.symbolic_sizes()
 sym_size_mlp = P(
     f"""
 'bilinear_mlp' Einsum h,h,oh->o
@@ -188,13 +187,12 @@ sym_size_mlp = P(
     'bilinear_mlp.fold_pre' Rearrange (a:2 b) -> a:2 b
       'bilinear_mlp.pre' Einsum i,hi->h
                             # symbolic sizes can be indicated with <NUM>s
-        'bilinear_mlp.input' [0s] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52
+        'bilinear_mlp.input' [?] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52
                                  # we can also indicate products (including products of symbolic sizes)
-        'bilinear_mlp.w.proj_in' [2*1s, 0s] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
+        'bilinear_mlp.w.proj_in' [?, ?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
   'bilinear_mlp.pre1' Index [1,:]
     'bilinear_mlp.fold_pre'
-                                 # you can also use `rc.symbolic_sizes()` which returns the full list
-  'bilinear_mlp.w.proj_out' [2s, {s1}] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
+  'bilinear_mlp.w.proj_out' [?, ?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
 """
 )
 
@@ -208,14 +206,6 @@ sym_size = rc.Expander(
 )(sym_size_mlp, fancy_validate=True)
 printer.print(sym_size)
 
-# note the 'SetSymbolicShape' node. this is used to introduce contraints on symbolic sizes
-# (you shouldn't typically need to do anything by hand with these nodes, they should just be automatically managed in expand)
-
-# %%
-
-# you can access a list of constraints if you so wish!
-print(sym_size.symbolic_size_constraints)
-
 # %%
 
 # quick test of printing/parsing
@@ -226,11 +216,10 @@ parsed = P(
     'bilinear_mlp.fold_pre' Rearrange c d e (a:2 b) -> c d e a:2 b
       'bilinear_mlp.pre' Einsum jkli,hi->jklh
         'rep_input' [3,5,7,17] Scalar 0.7
-        'bilinear_mlp.w.proj_in set_shape' [2*1s,17] SetSymbolicShape
-          'bilinear_mlp.w.proj_in' [2*1s,0s] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
+        'bilinear_mlp.w.proj_in' [?,?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
   'bilinear_mlp.pre1' Index [:,:,:,1,:]
     'bilinear_mlp.fold_pre'
-  'bilinear_mlp.w.proj_out' [2s,1s] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
+  'bilinear_mlp.w.proj_out' [?,?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
 """
 )
 assert rc.deep_normalize(parsed) == rc.deep_normalize(sym_size)
@@ -261,9 +250,6 @@ printer.print(sym_size_in_out)
 assert sym_size_in_out.is_explicitly_computable
 assert mlp_expanded == sym_size_in_out
 
-# now note that the 'SetSymbolicShape' is gone and we're explicitly computable.
-# We've resolved all of the shapes!
-
 # %%
 
 # we'll get an error if we make shapes inconsistent (in this case an einsum error)
@@ -287,16 +273,11 @@ scl0, scl1, v, new0, new1 = P.parse_circuits(
   2 [7] Scalar 1.5
   3
   4 [3] Scalar 1.7
-'new0' [0s] Symbol
-'new1' [0s] Symbol
+'new0' [?] Symbol
+'new1' [?] Symbol
 """
 )
 
-with pytest.raises(rc.SymbolicSizeSetFailedToSatisfyContraintsError) as exc2:
-    # expand node just expands 1 level on replaced inputs (like expander)
-    rc.expand_node(v, [scl0, new0, scl1, new1])
-print(exc2.exconly())
-
 # %% [markdown]
 
 # # Modules
@@ -326,9 +307,9 @@ print(exc2.exconly())
 
 input_sym, proj_in_sym, proj_out_sym, mlp = P.parse_circuits(
     """
-'bilinear_mlp.input' [0s] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52
-'bilinear_mlp.w.proj_in' [2*1s, 0s] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
-'bilinear_mlp.w.proj_out' [2s, 1s] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
+'bilinear_mlp.input' [?] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52
+'bilinear_mlp.w.proj_in' [?, ?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
+'bilinear_mlp.w.proj_out' [?, ?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
 'bilinear_mlp' Einsum h,h,oh->o
   'bilinear_mlp.pre0' Index [0,:]
     'bilinear_mlp.fold_pre' Rearrange (a:2 b) -> a:2 b
@@ -785,11 +766,11 @@ printer.print(mlp_2_layer.substitute(name_prefix="outer."))
 s = """
 'a.norm' Module
   'ln' Add
-    'ln.w.bias' [0s] Symbol 621c7792-0177-45ab-87c5-7ff1c3bec487
+    'ln.w.bias' [?] Symbol 621c7792-0177-45ab-87c5-7ff1c3bec487
     'ln.y_scaled' Einsum h,h->h
       'ln.y' Einsum h,->h
         'ln.mean_subbed' Add
-          'ln.input' [0s] Symbol 981b4d2a-711b-4a9d-a11c-d859c311e80c
+          'ln.input' [?] Symbol 981b4d2a-711b-4a9d-a11c-d859c311e80c
           'ln.neg_mean' Einsum h,z,->z
             'ln.input'
             'ln.neg' [1] Scalar -1
@@ -803,10 +784,10 @@ s = """
               'ln.mean_subbed'
               'ln.mean_subbed'
               'ln.c.recip_hidden_size'
-      'ln.w.scale' [0s] Symbol 0fa341c3-34b3-4699-847f-08674808b28a
-  'a.norm.input' [1s,0s] Symbol 6a622698-fd68-4d25-aeee-e8d38e68049e ! 'ln.input'
-  'a.ln.w.bias' [0s] Symbol 2c737289-2702-404c-a22e-ad37c2652620 ! 'ln.w.bias'
-  'a.ln.w.scale' [0s] Symbol c564149d-e226-4e3d-8e47-7d6e2ceea99e ! 'ln.w.scale'
+      'ln.w.scale' [?] Symbol 0fa341c3-34b3-4699-847f-08674808b28a
+  'a.norm.input' [?,?] Symbol 6a622698-fd68-4d25-aeee-e8d38e68049e ! 'ln.input'
+  'a.ln.w.bias' [?] Symbol 2c737289-2702-404c-a22e-ad37c2652620 ! 'ln.w.bias'
+  'a.ln.w.scale' [?] Symbol c564149d-e226-4e3d-8e47-7d6e2ceea99e ! 'ln.w.scale'
 """
 a_norm = P(s)
 # (see below for extract symbols explanation)
@@ -815,11 +796,11 @@ extracted_sub = extracted_a_norm.substitute(name_prefix="w.a.")
 expected_extracted = """
 'a.norm' Module
   'w.a.ln' Add
-    'a.ln.w.bias' [0s] Symbol 2c737289-2702-404c-a22e-ad37c2652620
+    'a.ln.w.bias' [?] Symbol 2c737289-2702-404c-a22e-ad37c2652620
     'w.a.ln.y_scaled' Einsum h,h->h
       'ln.y' Einsum h,->h
         'ln.mean_subbed' Add
-          'ln.input' [0s] Symbol 981b4d2a-711b-4a9d-a11c-d859c311e80c
+          'ln.input' [?] Symbol 981b4d2a-711b-4a9d-a11c-d859c311e80c
           'ln.neg_mean' Einsum h,z,->z
             'ln.input'
             'ln.neg' [1] Scalar -1
@@ -833,8 +814,8 @@ expected_extracted = """
               'ln.mean_subbed'
               'ln.mean_subbed'
               'ln.c.recip_hidden_size'
-      'a.ln.w.scale' [0s] Symbol c564149d-e226-4e3d-8e47-7d6e2ceea99e
-  'a.norm.input' [1s,0s] Symbol 6a622698-fd68-4d25-aeee-e8d38e68049e ! 'ln.input'
+      'a.ln.w.scale' [?] Symbol c564149d-e226-4e3d-8e47-7d6e2ceea99e
+  'a.norm.input' [?,?] Symbol 6a622698-fd68-4d25-aeee-e8d38e68049e ! 'ln.input'
 """
 assert extracted_sub == P(expected_extracted)
 
@@ -942,8 +923,8 @@ print(exc8.exconly())
 # (which aren't removed by conform_all_modules), then you'll get a warning
 s_with_sym = """
 0 Einsum ij,jk->ik
-  1 [1s,2s] Symbol 6594dafb-81f8-49a3-8381-6ffc4746b474
-  2 [2s,3s] Symbol bc646cfe-a299-4580-9a54-d8d5001a566b
+  1 [?,?] Symbol 6594dafb-81f8-49a3-8381-6ffc4746b474
+  2 [?,?] Symbol bc646cfe-a299-4580-9a54-d8d5001a566b
 """
 
 with pytest.warns(rc.OptimizingSymbolicSizeWarning) as w:
