@@ -144,6 +144,8 @@ SplitConfigDict = dict[int, list[tuple[rc.TorchAxisIndex, str]]]
 SplitConfigSetDict = Union[set[int], SplitConfigDict]
 SplitConfig = Union[Literal["full"], SplitConfigSetDict]
 
+SplitPosConfig = dict[int, list[tuple[Union[slice, int], str]]]
+
 
 def split_by_head(c: rc.Circuit, split: SplitConfig = "full"):
     """
@@ -271,9 +273,6 @@ def flatten_res(c: rc.Circuit, flatten_components: bool = False):
     )
 
 
-SplitPosConfig = dict[int, list[tuple[str, Union[slice, int]]]]
-
-
 def configure_transformer(
     c: rc.Circuit,
     to: Optional[To] = To.ATTN_MLP_NORM,
@@ -346,14 +345,12 @@ def configure_transformer(
             def try_(f, x):
                 return f(x) or x
 
+            c = c.update({"b.input idx idx"}, lambda x: op.unwrap(rc.index_fuse(x.cast_index())))
             c = c.update(
-                {"b.input set_shape set_shape idx", "b.input set_shape idx", "b.input idx"},
+                {"b.input idx"},
                 lambda x: try_(rc.index_elim_identity, try_(rc.index_concat_drop_unreached, x)),
             )
 
-    # pull this out into function maybe
-    c = c.update(rc.SetSymbolicShape, lambda x: x.rename(x.get_autoname()))
-
     return c
 
 
@@ -375,12 +372,26 @@ def batch_mul(
     return x.evolve(out_axes=x.out_axes[:insert_at_idx_v] + tuple(einsum_axes) + x.out_axes[insert_at_idx_v:])
 
 
+# po = rc.PrintOptions(
+#     # traversal=rc.restrict(m, True),
+#     colorer=rc.PrintOptions.type_colorer(),
+#     bijection=False,
+#     shape_only_when_necessary=False,arrows=True
+# )
+
+
 def split_by_pos(circ: rc.Circuit, config: SplitPosConfig, len_: int):
+    """
+    requires that weights have been pushed down all the way to heads, sequence length is specified as len_ (sequence_length in configure_transformer), and the model has causal attention (is_causal in configure_transformer)
+
+    {layer: [(slice, name), (int, name2), ...], layer2: ...}: split these layers by position into this list of groups
+    """
+
     def push_down(x):
         return rc.push_down_index(
             x,
             rc.restrict(
-                {rc.SetSymbolicShape, rc.Symbol, rc.Index},
+                {rc.Symbol, rc.Index},
                 True,
             ),
             allow_partial_pushdown=True,
@@ -389,7 +400,6 @@ def split_by_pos(circ: rc.Circuit, config: SplitPosConfig, len_: int):
 
     def update_layer(ca, splits: list[Union[int, slice]], split_names: list[str]):
         mask = ca.get_unique("a.mask")
-        mask = rc.SetSymbolicShape(mask, (len_, len_))
         mask = rc.Concat(
             *(
                 rc.Concat(
@@ -408,12 +418,7 @@ def split_by_pos(circ: rc.Circuit, config: SplitPosConfig, len_: int):
         )
 
         ca = rc.Expander(
-            ("a.mask", lambda _: mask),
-            ("a.v.input", lambda x: rc.SetSymbolicShape(x, (len_, x.shape[1]))),
-            ("a.k.input", lambda x: rc.SetSymbolicShape(x, (len_, x.shape[1]))),
-            ("a.q.input", lambda x: rc.SetSymbolicShape(x, (len_, x.shape[1]))),
-            ("a.input", lambda x: rc.SetSymbolicShape(x, (len_, x.shape[1]))),
-            ("p.input", lambda x: rc.SetSymbolicShape(x, (len_, x.shape[1]))),
+            ("a.mask", lambda _: mask), ("b.input", lambda x: rc.Index(x, (slice(0, len_), (slice(None, None)))))
         )(ca)
         ca = algrw.split_to_concat(
             ca, 0, splits, split_names, run_on_index=lambda x, *_: push_down(x).rename(f"{x.name}")
@@ -422,7 +427,6 @@ def split_by_pos(circ: rc.Circuit, config: SplitPosConfig, len_: int):
         ca = ca.update(
             "a.mask causal", lambda x: op.unwrap(rc.concat_elim_identity(x.cast_concat()))
         )  # left by push_down_index
-        # ca.print(po)
         ca = ca.update("a.attn_scores", rc.deep_pull_concat_messy)
 
         def upd_softmax(x: rc.Circuit) -> rc.Concat:
@@ -458,11 +462,11 @@ def split_by_pos(circ: rc.Circuit, config: SplitPosConfig, len_: int):
             rc.Matcher("a.attn_scores_raw").chain(rc.restrict(rc.Index, start_depth=1, end_depth=2)), push_down
         )
 
-        ca = ca.update(
-            rc.Matcher("a.v").chain(rc.restrict("a.v.input set_shape", end_depth=2))
-            | rc.Matcher("a.k").chain(rc.restrict("a.k.input set_shape", end_depth=2)),
-            lambda a: rc.Index(a, [slice(0, a.shape[0]), slice(None, None)]),
-        )  # fixup last
+        # ca = ca.update(
+        #     rc.Matcher("a.v").chain(rc.restrict("a.v.input set_shape", end_depth=2))
+        #     | rc.Matcher("a.k").chain(rc.restrict("a.k.input set_shape", end_depth=2)),
+        #     lambda a: rc.Index(a, [slice(0, a.shape[0]), slice(None, None)]),
+        # )  # fixup last
 
         def split_idx(ix):
             ix = ix.cast_index()
@@ -470,28 +474,28 @@ def split_by_pos(circ: rc.Circuit, config: SplitPosConfig, len_: int):
                 return ix
             return rc.Index(rc.Index(ix.node, (slice(0, ix.idx[0].stop), slice(None, None))), ix.idx)
 
-        ca = ca.update({"a.q.input set_shape idx", "p.input set_shape idx"}, split_idx)
-
-        ca = ca.update(
-            {"a.norm.input", "p.norm.input"}, lambda x: rc.SetSymbolicShape(x, (len_, x.shape[1]))
-        )  # needed to make push_down_indexes_through_modules work here, otherwise push down index gets stuck on a symbolic shape
+        ca = ca.update({"a.q.input idx", "p.input idx"}, split_idx)
 
         ca = algrw.push_down_indexes_through_modules(
             ca,
             {
-                "a.v.input set_shape idx",
-                "a.k.input set_shape idx",
-                "a.q.input set_shape idx",
-                "p.input set_shape idx",
-                "p.input set_shape",
+                "a.v.input idx",
+                "a.k.input idx",
+                "a.q.input idx",
+                "p.input idx",
+                "p.input",
             },
             push_down=push_down,
         )
         return ca
 
     for (l, c) in config.items():
-        names = [n for n, _ in c]
-        splits = [s for _, s in c]
+        if isinstance(c[0][0], str):
+            raise Exception(
+                "split_by_pos config was changed to match split_by_head, was (name, idxs), now is (idxs, name) at innermost level"
+            )
+        names = [n for _, n in c]
+        splits = [s for s, _ in c]
         circ = circ.update(
             rc.IterativeMatcher(f"b{l}.call").spec_circuit_matcher(), lambda x: update_layer(x, splits, names)
         )
