@@ -4,7 +4,7 @@ import functools
 import json
 import random
 from copy import copy
-from typing import Any, Callable, Dict, List, Literal, Optional, TypeVar, Union, cast
+from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, TypeVar, Union, cast
 from uuid import UUID, uuid4
 
 import cattrs
@@ -28,14 +28,12 @@ from ._rust import (
     Parser,
     PrintOptions,
     Scalar,
-    Shape,
     Symbol,
     TorchDeviceDtypeOp,
     deep_module_remove_unused_inputs,
     get_free_symbols,
     module_new_bind,
     simp,
-    symbolic_sizes,
 )
 from .py_utils import I
 
@@ -48,24 +46,6 @@ P = Parser()
 # weights are 'x.w.y' ('w' isn't a placeholder). Constants are 'x.c.y'.
 # then, callers can do 'z.x' or 'x.call' (or similar)
 
-(
-    HIDDEN,
-    # OUT_HIDDEN,
-    SEQ,
-    SEQ_Q,
-    SEQ_K,
-    HEADS,
-    MLP_PROJ,
-    BILINEAR_MLP_PROJ,
-    QK_DIM,
-    V_DIM,
-    LOG_LIKELYHOOD_CLASSES,
-    VOCAB_SIZE,
-    HALF_ROTARY_DIM,  # 1/4 of QK_DIM
-    *_,
-) = symbolic_sizes()
-OUT_HIDDEN = HIDDEN  # fix out hidden = hidden as needed (https://github.com/redwoodresearch/unity/issues/1718, https://github.com/redwoodresearch/unity/issues/1716)
-
 
 def to_name_dict(*circs: Circuit):
     out = {c.name: c for c in circs}
@@ -112,7 +92,7 @@ def zero_input(circs: Dict[str, Circuit], to_zero: str):
     return to_name_dict(
         *(
             # we could make this faster if we needed
-            deep_module_remove_unused_inputs(simp(c.update(to_zero, lambda x: Scalar(0.0, shape=x.shape)))).update(
+            deep_module_remove_unused_inputs(simp(c.update(to_zero, lambda _: Scalar(0.0, shape=())))).update(
                 True, lambda x: x.rename(x.name.removesuffix(" rem_unused"))
             )
             for s, c in circs.items()
@@ -124,7 +104,7 @@ def zero_input(circs: Dict[str, Circuit], to_zero: str):
 layernorm_str = f"""
 'ln.y' Einsum h,->h
   'ln.mean_subbed' Add
-    'ln.input' [{HIDDEN}] Symbol 981b4d2a-711b-4a9d-a11c-d859c311e80c
+    'ln.input' [?] Symbol 981b4d2a-711b-4a9d-a11c-d859c311e80c
     'ln.neg_mean' Einsum h,z,->z # z is padding dim for add
       'ln.input'
       'ln.neg' [1] Scalar -1
@@ -142,10 +122,10 @@ layernorm_str = f"""
 'ln.input'
 
 'ln' Add
-  'ln.w.bias' [{HIDDEN}] Symbol 621c7792-0177-45ab-87c5-7ff1c3bec487
+  'ln.w.bias' [?] Symbol 621c7792-0177-45ab-87c5-7ff1c3bec487
   'ln.y_scaled' Einsum h,h->h
     'ln.y'
-    'ln.w.scale' [{HIDDEN}] Symbol 0fa341c3-34b3-4699-847f-08674808b28a
+    'ln.w.scale' [?] Symbol 0fa341c3-34b3-4699-847f-08674808b28a
 
 'ln.w.bias'
 'ln.w.scale'
@@ -157,14 +137,14 @@ layernorm = ModuleLibraryRet.mk(layernorm_circs, "ln")
 batchnorm_str = f"""
 'bn.y' Einsum h,h->h
   'bn.mean_subbed' Add
-    'bn.input' [{HIDDEN}] Symbol cfe0aa25-1214-4d3e-ad53-417fedbb9b7b
+    'bn.input' [?] Symbol cfe0aa25-1214-4d3e-ad53-417fedbb9b7b
     'bn.neg_mean' Einsum h,->h
-      'bn.mean' [{HIDDEN}] Symbol 217403c9-03b6-4444-8f72-de03b95229d0
+      'bn.mean' [?] Symbol 217403c9-03b6-4444-8f72-de03b95229d0
       'bn.neg' [] Scalar -1
   'bn.rsqrt' GeneralFunction rsqrt
     'bn.var_p_eps' Add
       'bn.c.eps' [] Scalar 1e-5
-      'bn.var' [{HIDDEN}] Symbol c4e41ca7-e3ab-429e-81b6-25bb14423fb4
+      'bn.var' [?] Symbol c4e41ca7-e3ab-429e-81b6-25bb14423fb4
 
 'bn.input'
 'bn.mean'
@@ -172,10 +152,10 @@ batchnorm_str = f"""
 
 # same as ln above, we could dedudup if we wanted
 'bn' Add
-  'bn.w.bias' [{HIDDEN}] Symbol 3e861c08-9a74-4348-87df-3752a8557aea
+  'bn.w.bias' [?] Symbol 3e861c08-9a74-4348-87df-3752a8557aea
   'bn.y_scaled' Einsum h,h->h
     'bn.y'
-    'bn.w.scale' [{HIDDEN}] Symbol ccbac7e0-c931-422b-9724-1e24ab6e9c63
+    'bn.w.scale' [?] Symbol ccbac7e0-c931-422b-9724-1e24ab6e9c63
 
 'bn.w.bias'
 'bn.w.scale'
@@ -202,23 +182,23 @@ raw_attention_str = f"""
   'a.comb_v' Einsum qk,kV->qV
     'a.attn_probs' GeneralFunction softmax
       'a.attn_scores' Add
-        'a.attn_scores_raw' [{SEQ_Q}, {SEQ_K}] Einsum qc,kc,,qk->qk
-          'a.q_p_bias' [{SEQ_Q}, {QK_DIM}] Add
+        'a.attn_scores_raw' [?, ?] Einsum qc,kc,,qk->qk
+          'a.q_p_bias' [?, ?] Add
             'a.q' Einsum qd,cd->qc
-              'a.q.input' [{SEQ_Q}, {HIDDEN}] Symbol 4f80d1a1-86a4-4e44-94f7-909ec7089061
-              'a.w.q_h' [{QK_DIM}, {HIDDEN}] Symbol 665efa60-d86c-40d5-92b2-b96d11686a8b
-            'a.w.q_bias_h' [{QK_DIM}] Symbol 7d531f53-6cce-4bf3-82db-5fe8b2eef974
-          'a.k_p_bias' [{SEQ_K}, {QK_DIM}] Add
+              'a.q.input' [?, ?] Symbol 4f80d1a1-86a4-4e44-94f7-909ec7089061
+              'a.w.q_h' [?, ?] Symbol 665efa60-d86c-40d5-92b2-b96d11686a8b
+            'a.w.q_bias_h' [?] Symbol 7d531f53-6cce-4bf3-82db-5fe8b2eef974
+          'a.k_p_bias' [?, ?] Add
             'a.k' Einsum kd,cd->kc
-              'a.k.input' [{SEQ_K}, {HIDDEN}] Symbol 664bddee-28ca-47e7-9fb7-9a718de06619
-              'a.w.k_h' [{QK_DIM}, {HIDDEN}] Symbol 41177709-446d-4588-b9e5-c2bbf59d53a0
-            'a.w.k_bias_h' [{QK_DIM}] Symbol a891aae4-3c24-4713-afc7-8b954c6fc1b5
+              'a.k.input' [?, ?] Symbol 664bddee-28ca-47e7-9fb7-9a718de06619
+              'a.w.k_h' [?, ?] Symbol 41177709-446d-4588-b9e5-c2bbf59d53a0
+            'a.w.k_bias_h' [?] Symbol a891aae4-3c24-4713-afc7-8b954c6fc1b5
           'a.c.div_head_size' GeneralFunction rsqrt
             'a.c.head_size' GeneralFunction last_dim_size
-              'a.c.bias_for_head_size' [{QK_DIM}] Einsum jk->j # size on this line is just asserted
-                'a.w.k_h' [{QK_DIM}, {HIDDEN}]
+              'a.c.bias_for_head_size' [?] Einsum jk->j # size on this line is just asserted
+                'a.w.k_h' [?, ?]
           # mask is true at where positions *are* allowed. (This differs from old pos_mask code)
-          'a.mask' [{SEQ_Q}, {SEQ_K}] Symbol ccfe5bc9-b402-42dd-a5e1-191e6fb7c268
+          'a.mask' [?, ?] Symbol ccfe5bc9-b402-42dd-a5e1-191e6fb7c268
         'a.score_neg_inf_bias' Einsum qk,->qk
           'a.not_mask' Module
             'not_mask'
@@ -226,10 +206,10 @@ raw_attention_str = f"""
           'a.neg_inf' [] Scalar -10_000.0
     'a.v_p_bias' Add
       'a.v' Einsum kd,Vd->kV
-        'a.v.input' [{SEQ_K}, {HIDDEN}] Symbol 8fd4c632-7f28-49ee-84cc-3dde997e0693
-        'a.w.v_h' [{V_DIM}, {HIDDEN}] Symbol 79b6ebff-f9d0-411a-bcdc-530cc13e1524
-      'a.w.v_bias_h' [{V_DIM}] Symbol dfb2c4ec-9378-40ee-a360-7d58f2b96954
-  'a.w.o_h' [{OUT_HIDDEN}, {V_DIM}] Symbol 11a116cb-2168-4725-a06f-1b61a8ca6797
+        'a.v.input' [?, ?] Symbol 8fd4c632-7f28-49ee-84cc-3dde997e0693
+        'a.w.v_h' [?, ?] Symbol 79b6ebff-f9d0-411a-bcdc-530cc13e1524
+      'a.w.v_bias_h' [?] Symbol dfb2c4ec-9378-40ee-a360-7d58f2b96954
+  'a.w.o_h' [?, ?] Symbol 11a116cb-2168-4725-a06f-1b61a8ca6797
 
 'a.q.input'
 'a.k.input'
@@ -247,8 +227,8 @@ raw_attention_str = f"""
 'a.head.on_inp' Module
   'a.head'
   'a.qk_input' Add ! 'a.q.input'
-    'a.input' [{SEQ}, {HIDDEN}] Symbol f9eabd07-e2ab-4ed4-8b4a-c9c039d61835
-    'a.pos_input' [{SEQ}, {HIDDEN}] Symbol eab8313e-d910-4174-8dbe-c612954eec34
+    'a.input' [?, ?] Symbol f9eabd07-e2ab-4ed4-8b4a-c9c039d61835
+    'a.pos_input' [?, ?] Symbol eab8313e-d910-4174-8dbe-c612954eec34
   'a.qk_input' ! 'a.k.input'
   'a.input' ! 'a.v.input'
 
@@ -260,14 +240,14 @@ raw_attention_str = f"""
     # batch over head
     'a.heads' Module
       'a.head.on_inp'
-      'a.w.q' [{HEADS}, {QK_DIM}, {HIDDEN}] Symbol cf8f9c58-1875-45b0-8007-66b66b8a405a ! 'a.w.q_h'
-      'a.w.q_bias' [{HEADS}, {QK_DIM}] Symbol e896a43e-ba35-43d5-97e7-e189caf3278b ! 'a.w.q_bias_h'
-      'a.w.k' [{HEADS}, {QK_DIM}, {HIDDEN}] Symbol c342f1cd-71e6-4848-86b4-1c3ffdd46753 ! 'a.w.k_h'
-      'a.w.k_bias' [{HEADS}, {QK_DIM}] Symbol 13465b29-c28a-4d55-8a27-22c26cc01c69 ! 'a.w.k_bias_h'
-      'a.w.v' [{HEADS}, {V_DIM}, {HIDDEN}] Symbol a90db69c-f4ad-47d6-8b76-faa4b107dacd ! 'a.w.v_h'
-      'a.w.v_bias' [{HEADS}, {V_DIM}] Symbol ea64ee26-2438-4aa9-b777-0ef4f5f70e74 ! 'a.w.v_bias_h'
-      'a.w.o' [{HEADS}, {OUT_HIDDEN}, {V_DIM}] Symbol 24e0e5cf-8b68-4bf2-b198-17237523b237 ! 'a.w.o_h'
-  'a.w.o_bias' [{OUT_HIDDEN}] Symbol 61d54061-63e2-4a5d-9c7a-15c1ecc53b95
+      'a.w.q' [?, ?, ?] Symbol cf8f9c58-1875-45b0-8007-66b66b8a405a ! 'a.w.q_h'
+      'a.w.q_bias' [?, ?] Symbol e896a43e-ba35-43d5-97e7-e189caf3278b ! 'a.w.q_bias_h'
+      'a.w.k' [?, ?, ?] Symbol c342f1cd-71e6-4848-86b4-1c3ffdd46753 ! 'a.w.k_h'
+      'a.w.k_bias' [?, ?] Symbol 13465b29-c28a-4d55-8a27-22c26cc01c69 ! 'a.w.k_bias_h'
+      'a.w.v' [?, ?, ?] Symbol a90db69c-f4ad-47d6-8b76-faa4b107dacd ! 'a.w.v_h'
+      'a.w.v_bias' [?, ?] Symbol ea64ee26-2438-4aa9-b777-0ef4f5f70e74 ! 'a.w.v_bias_h'
+      'a.w.o' [?, ?, ?] Symbol 24e0e5cf-8b68-4bf2-b198-17237523b237 ! 'a.w.o_h'
+  'a.w.o_bias' [?] Symbol 61d54061-63e2-4a5d-9c7a-15c1ecc53b95
 
 'a.w.q'
 'a.w.q_bias'
@@ -281,18 +261,16 @@ raw_attention_str = f"""
 
 # We split into two pieces so we can apply rotary positional embeddings to only the first.
 # Also, we use HALF_ROTARY_DIM to enforce that the rotary dim is even.
-'qk_with_rot' [{SEQ}, {QK_DIM}] SetSymbolicShape
-  'qk_with_rot.re' [{SEQ}, 2*2*{HALF_ROTARY_DIM}] Rearrange q s:2 e:2 rh -> q (s:2 e:2 rh)
-    'qk_with_rot.cat' [{SEQ}, 2, 2, {HALF_ROTARY_DIM}] Concat 1
-      'qk_with_rot.q_rot' Rearrange q e rh -> q 1 e rh
-        'qk_rot' [{SEQ}, 2, {HALF_ROTARY_DIM}] GeneralFunction apply_rotary_pos_emb at rust_circuit.generalfuncs.rotary_pos_emb:ApplyRotaryPosEmb
-          'qk_rot.inp' Index [:, 0]
-            'qk_split' [{SEQ}, 2, 2, {HALF_ROTARY_DIM}] Rearrange q (s:2 e:2 rh) -> q s:2 e:2 rh
-              'qk_split.set' [{SEQ}, 2*2*{HALF_ROTARY_DIM}] SetSymbolicShape
-                # q or k
-                'qk' [{SEQ}, {QK_DIM}] Symbol 442c7d04-1948-4904-b96a-5c0acf22f19c
-      'qk_with_rot.q_pass' Index [:, 1:]
-        'qk_split'
+'qk_with_rot' [?, ?] Rearrange q s:2 e:2 rh -> q (s:2 e:2 rh)
+  'qk_with_rot.cat' [?, 2, 2, ?] Concat 1
+    'qk_with_rot.q_rot' Rearrange q e rh -> q 1 e rh
+      'qk_rot' [?, 2, ?] GeneralFunction apply_rotary_pos_emb at rust_circuit.generalfuncs.rotary_pos_emb:ApplyRotaryPosEmb
+        'qk_rot.inp' Index [:, 0]
+          'qk_split' [?, 2, 2, ?] Rearrange q (s:2 e:2 rh) -> q s:2 e:2 rh
+            # q or k
+            'qk' [?, ?] Symbol 442c7d04-1948-4904-b96a-5c0acf22f19c
+    'qk_with_rot.q_pass' Index [:, 1:]
+      'qk_split'
 
 # Most of the below is copied from a.head; could factor into a module but backcompat is annoying
 
@@ -354,7 +332,7 @@ def get_attention(bias: bool = False, pos: bool = False, rope: bool = False):
     return ModuleLibraryRet.mk(circs, circ)
 
 
-m_input = Symbol((HIDDEN,), UUID("13ec4cdc-5f25-4969-870b-0bfa2300187b"), name="m.input")
+m_input = Symbol((None,), UUID("13ec4cdc-5f25-4969-870b-0bfa2300187b"), name="m.input")
 m_base_circs: Dict[str, Circuit] = {"m.input": m_input}
 
 
@@ -366,12 +344,12 @@ raw_bilinear_mlp_str = f"""
         'm.pre_p_bias' Add
           'm.pre' Einsum i,hi->h
             'm.input'
-            'm.w.proj_in' [{BILINEAR_MLP_PROJ*2}, {HIDDEN}] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
-          'm.w.in_bias' [{BILINEAR_MLP_PROJ*2}] Symbol 886b5425-cd19-4db4-871a-c46cb1a23114
+            'm.w.proj_in' [?, ?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
+          'm.w.in_bias' [?] Symbol 886b5425-cd19-4db4-871a-c46cb1a23114
     'm.pre1' Index [1,:]
       'm.fold_pre'
-    'm.w.proj_out' [{OUT_HIDDEN}, {BILINEAR_MLP_PROJ}] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
-  'm.w.out_bias' [{OUT_HIDDEN}] Symbol 7efddf2e-20af-492c-a94e-1d19468f333f
+    'm.w.proj_out' [?, ?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
+  'm.w.out_bias' [?] Symbol 7efddf2e-20af-492c-a94e-1d19468f333f
 
 'm.w.proj_in'
 'm.w.in_bias'
@@ -401,10 +379,10 @@ def get_pointwise_mlp(function_str: str = "gelu", output_bias: bool = False):
           'm.pre' Add
             'm.pre_mul' Einsum i,hi->h
               'm.input'
-              'm.w.proj_in' [{MLP_PROJ}, {HIDDEN}] Symbol 5217f963-0cdb-460e-bb1f-f82f7fbb3cd9
-            'm.w.in_bias' [{MLP_PROJ}] Symbol c870ec00-8c6f-4080-907c-703ea85dde48
-        'm.w.proj_out' [{OUT_HIDDEN}, {MLP_PROJ}] Symbol fdefa9af-a7d6-4a38-a7ed-5ce816c6efe7
-      'm.w.out_bias' [{OUT_HIDDEN}] Symbol 113c30ba-fa88-4f34-a301-ea7912e03064
+              'm.w.proj_in' [?, ?] Symbol 5217f963-0cdb-460e-bb1f-f82f7fbb3cd9
+            'm.w.in_bias' [?] Symbol c870ec00-8c6f-4080-907c-703ea85dde48
+        'm.w.proj_out' [?, ?] Symbol fdefa9af-a7d6-4a38-a7ed-5ce816c6efe7
+      'm.w.out_bias' [?] Symbol 113c30ba-fa88-4f34-a301-ea7912e03064
 
     'm.input'
     'm.w.proj_in'
@@ -426,16 +404,16 @@ def get_norm_bind(prefix: str, uuids: List[UUID], norm_type: Literal["ln", "bn"]
     bind_str = f"""
     '{p}.norm' Module
       '{nt}'
-      '{p}.norm.input' [{SEQ}, {HIDDEN}] Symbol {uuids.pop(4)} ! '{nt}.input'"""
+      '{p}.norm.input' [?, ?] Symbol {uuids.pop(4)} ! '{nt}.input'"""
 
     if norm_type == "bn":
         bind_str += f"""
-      '{p}.{nt}.mean' [{HIDDEN}] Symbol {uuids.pop(3)} ! '{nt}.mean'
-      '{p}.{nt}.var' [{HIDDEN}] Symbol {uuids.pop(2)} ! '{nt}.var'"""
+      '{p}.{nt}.mean' [?] Symbol {uuids.pop(3)} ! '{nt}.mean'
+      '{p}.{nt}.var' [?] Symbol {uuids.pop(2)} ! '{nt}.var'"""
 
     bind_str += f"""
-      '{p}.{nt}.w.bias' [{HIDDEN}] Symbol {uuids.pop(1)} ! '{nt}.w.bias'
-      '{p}.{nt}.w.scale' [{HIDDEN}] Symbol {uuids.pop(0)} ! '{nt}.w.scale'
+      '{p}.{nt}.w.bias' [?] Symbol {uuids.pop(1)} ! '{nt}.w.bias'
+      '{p}.{nt}.w.scale' [?] Symbol {uuids.pop(0)} ! '{nt}.w.scale'
 
     '{p}.norm.input'"""
 
@@ -501,7 +479,7 @@ def get_norm_call(prefix: str, circs: Dict[str, Circuit], mod: Optional[str] = N
     return ModuleLibraryRet.mk(circs, f"{prefix}.norm_call")
 
 
-b_input = Symbol((SEQ, HIDDEN), UUID("5837c4fd-f5ac-4bff-8456-abf3e95bcf36"), name="b.input")
+b_input = Symbol((None, None), UUID("5837c4fd-f5ac-4bff-8456-abf3e95bcf36"), name="b.input")
 
 
 def rename_circs_to_keys(d: Dict[str, Circuit], rename_suffix: str = ""):
@@ -571,7 +549,6 @@ class TransformerBlockParams:
 
         attn_only_str = f"""
         'b.resid.a' Add
-          # 'b.a.set' [{SEQ}, {HIDDEN}] SetSymbolicShape # not currently needed as HIDDEN == OUT_HIDDEN
           'b.a' Module
             '{attn_with_norm.name}'
             'b.input' ! 'a{'.norm' if self.norm_type is not None else ''}.input'
@@ -590,7 +567,7 @@ class TransformerBlockParams:
           'b.add' Add
             'b.a' Module
               '{attn.name}'
-              'p.input' [{SEQ}, {HIDDEN}] Symbol e76c1fd7-a230-4295-9199-17dc8dad79d0 ! 'a.input'
+              'p.input' [?, ?] Symbol e76c1fd7-a230-4295-9199-17dc8dad79d0 ! 'a.input'
             'b.m' Module
               '{mlp.name}'
               'p.input' ! 'm.input'
@@ -612,7 +589,6 @@ class TransformerBlockParams:
             mlp_input = cur_resid_name
             block_str = f"""
             'b.resid.m' Add
-              # 'b.m.set' [{SEQ}, {HIDDEN}] SetSymbolicShape # not currently needed as HIDDEN == OUT_HIDDEN
               'b.m' Module
                 '{mlp_with_norm.name}'
                 '{mlp_input}' ! 'm{'.norm' if self.norm_type is not None else ''}.input'
@@ -738,7 +714,7 @@ class TransformerBlockParams:
 log_likelyhood_str = f"""
 'll' GeneralFunction gen_index_at_0_batch_x_c
   'log_probs' GeneralFunction log_softmax
-    'll.input' [{LOG_LIKELYHOOD_CLASSES}] Symbol b9d111e5-b793-4f63-84a0-2f8590b9f39c
+    'll.input' [?] Symbol b9d111e5-b793-4f63-84a0-2f8590b9f39c
   'll.label' [] Symbol 74c503b9-fe2c-4f8b-9350-7351a292c351
 
 'll.input'
@@ -772,12 +748,11 @@ class TransformerParams:
         # keeps right ordering
         circs = dict(x for b in blocks for x in b.circs.items())
 
-        out: Circuit = Symbol((SEQ, HIDDEN), UUID("ece2bb5d-c6d6-4b7a-93a5-32b5ac264888"), "t.input")
+        out: Circuit = Symbol((None, None), UUID("ece2bb5d-c6d6-4b7a-93a5-32b5ac264888"), "t.input")
         assert "t.input" not in circs
         circs["t.input"] = out
         for i, block in enumerate(blocks):
             out = module_new_bind(block.body, ("b.input", out), name=f"b{i}.call")
-            assert out.shape == (SEQ, HIDDEN)
             circs[out.name] = out
 
         if self.norm_type is not None and self.use_norm_output:
@@ -793,10 +768,10 @@ class TransformerParams:
         logits_str = f"""
         't.logits' Einsum sh,vh->sv
           '{out.name}'
-          't.w.unembed' [{VOCAB_SIZE}, {HIDDEN}] Symbol 85d5a05a-ef9e-4910-a967-3f27951f67cf
+          't.w.unembed' [?, ?] Symbol 85d5a05a-ef9e-4910-a967-3f27951f67cf
         't.logits_p_bias' Add
           't.logits'
-          't.w.unembed_bias' [{VOCAB_SIZE}] Symbol 62d9f91a-5df2-4eb8-af75-5bdfb15eb974
+          't.w.unembed_bias' [?] Symbol 62d9f91a-5df2-4eb8-af75-5bdfb15eb974
 
         't.w.unembed'
         't.w.unembed_bias'
@@ -848,7 +823,7 @@ class TransformerParams:
         num_heads: int = 7,
         seq_len: int = 9,
         vocab_size: int = 11,
-        batch_shape: Shape = (13, 2),
+        batch_shape: Tuple[int, ...] = (13, 2),
         device_dtype: TorchDeviceDtypeOp = TorchDeviceDtypeOp(),
     ):
         """be warned this is not random init, it's GARBAGE, it won't train or behave properly at all"""
@@ -906,7 +881,7 @@ class TransformerInfo:
         inp_mask_has_q: bool = False,
         prefix: str = "t",
     ):
-        seq_len = inp_tok_embeds.shape[-2]
+        seq_len: int = cast(int, inp_tok_embeds.shape[-2])
 
         if pos_embed_weights is None:
             pos_embeds = None
