@@ -7,7 +7,7 @@ from typing import Any, Callable, Dict, Optional
 import rust_circuit as rc
 from rust_circuit import Circuit, PrintOptions
 
-ui_default_hidden_matcher = rc.Matcher(rc.SetSymbolicShape, rc.Scalar, rc.Tag)
+ui_default_hidden_matcher = rc.Matcher(rc.Scalar, rc.Tag)
 ui_no_ln_matcher = rc.Matcher("ln", rc.Matcher.regex(r"\.*ln\.*"))
 
 
@@ -38,10 +38,8 @@ def circuit_to_json(circ: Circuit) -> Any:
             "childIndices": childIndices,
             "meta": {
                 "name": circ.op_name,
-                "shape": [
-                    {"otherFactor": x.other_factor, "symbolicSizes": x.symbolic_sizes} for x in circ.symbolic_shape()
-                ],
-                "numel": rc.oom_fmt(math.prod(circ.shape)),
+                "shape": [None if x == rc.Size.NONE else x for x in circ.shape],
+                "numel": rc.oom_fmt(circ.numel),
                 "kind": circ.__class__.__name__,
                 "extra": extra_info,
             },
