@@ -69,6 +69,7 @@ from rust_circuit.algebric_rewrite import (
 from .util import dedup_with_order, st_Circuit
 
 P = Parser()
+po = rc.PrintOptions(tensor_index_literal=True, bijection=False)
 
 
 def is_zero(c: Circuit):
@@ -177,8 +178,8 @@ def check_transform_equality(
         trans_eval = trans
         circuit_eval = circuit
 
-    hypothesis.note(circuit_eval.repr())
-    hypothesis.note(trans_eval.repr())
+    hypothesis.note(circuit_eval.repr(po))
+    hypothesis.note(trans_eval.repr(po))
     actual, expected = (trans_eval.evaluate(), circuit_eval.evaluate())
     target_diff(actual.type(expected.dtype), expected)  # , evaluator=evaluator)
     return trans
