@@ -2,7 +2,7 @@ import uuid
 
 import pytest
 
-from rust_circuit import Add, Array, Cumulant, Einsum, Symbol, circuit_is_leaf, circuit_is_var, print_circuit_type_check
+from rust_circuit import Add, Array, Cumulant, Einsum, Symbol, print_circuit_type_check
 
 
 def test_circuit_type():
@@ -16,8 +16,8 @@ def test_circuit_type():
         print_circuit_type_check(int)  # type:ignore
     with pytest.raises(TypeError):
         print_circuit_type_check(None)  # type:ignore
-    assert circuit_is_leaf(Symbol([2, 3], uuid.uuid4(), name="hi"))
-    assert not circuit_is_var(Symbol([2, 3], uuid.uuid4(), name="hi"))
+    assert Symbol([2, 3], uuid.uuid4(), name="hi").is_leaf()
+    assert not Symbol([2, 3], uuid.uuid4(), name="hi").is_var()
 
 
 if __name__ == "__main__":
