@@ -36,7 +36,6 @@ from rust_circuit import (
     Index,
     IterativeMatcher,
     Matcher,
-    MiscInputNotBroadcastableError,
     Module,
     ModuleArgSpec,
     ModulePusher,
@@ -48,6 +47,7 @@ from rust_circuit import (
     Sampler,
     Scalar,
     Shape,
+    ShapeNotBroadcastableError,
     Symbol,
     TorchDeviceDtypeOp,
     all_children,
@@ -91,18 +91,25 @@ def test_subst_all_modules(circ):
     assert_close(circ.evaluate(), circ1.evaluate())
 
 
-@pytest.mark.xfail
+@hypothesis.given(get_c_st(max_growth_steps=10))
+def test_conform_all_modules(circ):
+    circ1 = rc.conform_all_modules(circ)
+    assert not any("internal_expand" in x.name for x in all_children(circ1))
+    assert_close(circ.evaluate(), circ1.evaluate())
+
+
 @hypothesis.given(
-    get_c_st(probs_default=CP.kw(all=1, Module=20, Symbol=20, Cumulant=0), from_other=False, max_growth_steps=20),
+    get_c_st(probs_default=CP.kw(all=1, Module=20, Symbol=20, Cumulant=0), from_other=False, max_growth_steps=30),
     st.data(),
 )
 @mark_not_interesting_if(rc.PushDownModulePushingPastModuleWhichOverridesSymError)
+@mark_not_interesting_if(rc.ConstructModuleTriedToBatchUnbatchableInputError)
 def test_push_down_modules(circ, d):
     skips = d.draw(st_c.st_subsets([x.name for x in all_children(circ) if x.is_module()]))
     cb = d.draw(st.builds(rc.ModulePusher.remove_and_elim_callback, st.booleans(), st.booleans(), st.booleans()))
     pu = rc.ModulePusher(flatten_modules=d.draw(st.booleans()), module_construct_callback=cb)
     note(circ.repr(po))
-    circ1 = pu.push_down_modules(circ, traversal=rc.traversal, skip_module=skips)
+    circ1 = pu.push_down_modules(circ, traversal=rc.new_traversal(), skip_module=skips)
     note(circ1.repr(po))
     assert_close(circ.evaluate(), circ1.evaluate())
 
@@ -148,6 +155,47 @@ def test_upd_nested_id(circ, d):
     assert_close(circ.evaluate(), circ1.evaluate())
 
 
+def dedup_pairwise(l, f):
+    out = []
+    for x in l:
+        if not any(f(x, y) for y in out):
+            out.append(x)
+    return out
+
+
+@hypothesis.given(
+    get_c_st(from_other=False, max_growth_steps=10, probs_per_depth=[CP.kw(all=0, branches=1, Cumulant=0)]).filter(
+        lambda c: c.num_children > 0
+    ),
+    st.data(),
+)
+@mark_not_interesting_if(rc.RearrangeSpecNotConformableError)
+@mark_not_interesting_if(rc.ConstructModuleTriedToBatchUnbatchableInputError)
+def test_expander_prop(circ, d):
+    children = list(rc.all_children(circ))
+    children.sort()
+    cs = d.draw(st.lists(st.sampled_from(children), unique=True, min_size=1, max_size=2))
+    cs = dedup_pairwise(cs, lambda x, y: (x in rc.all_children(y) or (y in rc.all_children(x))))
+    b = d.draw(st.integers(1, 5))
+    # print("\ncirc:")
+    # circ.print()
+    # print("circs:")
+    # for c in cs:
+    #     c.print()
+    new_cs = [d.draw(get_c_st(shape=(b,) + c.shape, from_other=False, max_growth_steps=4)) for c in cs]
+    note(circ.repr(po))
+    expanded = rc.Expander(*[(c, lambda _, new_c=new_c: new_c) for (c, new_c) in zip(cs, new_cs)])(circ)
+    note(expanded.repr(po))
+    d = {c: new_c for (c, new_c) in zip(cs, new_cs)}
+
+    def upd(x, i):
+        return x.update(set(cs), lambda a: rc.Index(d[a], (i,) + ((slice(None, None),) * a.ndim)))
+
+    concatted = rc.Concat(*[upd(circ, i).unsqueeze(0) for i in range(b)], axis=0)
+    note(concatted.repr(po))
+    assert_close(expanded.evaluate(), concatted.evaluate())
+
+
 def test_expand_node():
     circ: Circuit = Rearrange(
         Symbol.new_with_none_uuid((2, 3), "0"), RearrangeSpec([[0], [1]], [[1], [0]], [None, None])
@@ -186,7 +234,7 @@ def test_expand_node():
     print(expanded)
 
     with pytest.raises(ExpandFixedIndexError):
-        circ = Index(Symbol.new_with_none_uuid((2, 3), "0"), I[0:2, 0])
+        circ = Index(Symbol.new_with_none_uuid((2, 3), "0"), I[0:1, 0])
         expanded = expand_node(circ, [Array.randn(5, 3)])
         print(expanded)
 
@@ -205,7 +253,7 @@ def test_expand_node():
     print(expanded)
 
     circ = Add(Symbol.new_with_none_uuid((2, 3, 1), "0"), Symbol.new_with_none_uuid((2, 1, 5), "1"))
-    with pytest.raises(MiscInputNotBroadcastableError):
+    with pytest.raises(ShapeNotBroadcastableError):
         expand_node(
             circ, [Symbol.new_with_none_uuid((13, 7, 3, 1), "0"), Symbol.new_with_none_uuid((1, 7, 1, 11), "1")]
         )
@@ -232,13 +280,13 @@ def test_expand_node():
     print(expanded)
 
     circ = Add(Symbol.new_with_none_uuid((), "0"), Symbol.new_with_none_uuid((2, 1, 5), "1"))
-    with pytest.raises(MiscInputNotBroadcastableError):
+    with pytest.raises(ShapeNotBroadcastableError):
         expand_node(
             circ, [Symbol.new_with_none_uuid((7, 3, 1), "0"), Symbol.new_with_none_uuid((1, 3, 11, 9, 4, 5), "1")]
         )
 
     circ = Add(Symbol.new_with_none_uuid((), "0"), Symbol.new_with_none_uuid((2, 1, 5), "1"))
-    with pytest.raises(MiscInputNotBroadcastableError):
+    with pytest.raises(ShapeNotBroadcastableError):
         expand_node(
             circ, [Symbol.new_with_none_uuid((1, 3, 1), "0"), Symbol.new_with_none_uuid((1, 3, 11, 9, 4, 5), "1")]
         )
@@ -310,7 +358,7 @@ def test_expand_node():
         )
         print(expanded)
 
-    with pytest.raises(MiscInputNotBroadcastableError):
+    with pytest.raises(ShapeNotBroadcastableError):
         circ = Einsum.from_einsum_string(
             "ab,bc->ac", Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((3, 5), "1")
         )
@@ -322,7 +370,7 @@ def test_expand_node():
         circ = Concat(Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((4, 3), "1"), axis=0)
         expanded = expand_node(circ, [Symbol.new_with_none_uuid((3, 3), "0"), Symbol.new_with_none_uuid((4, 3), "1")])
 
-    with pytest.raises(MiscInputNotBroadcastableError):
+    with pytest.raises(ShapeNotBroadcastableError):
         circ = Concat(Symbol.new_with_none_uuid((2, 3), "0"), Symbol.new_with_none_uuid((3, 3), "1"), axis=0)
         expanded = expand_node(
             circ, [Symbol.new_with_none_uuid((2, 2, 3), "0"), Symbol.new_with_none_uuid((3, 3, 3), "1")]
@@ -388,14 +436,14 @@ def test_expand_node():
     assert expanded.shape == (7, 11, 4, 3, 2)
 
     s = """
-    'rep_fst' [2s, 3s] Symbol 0bac3288-00e7-41bf-bd59-ac04de67ba47
-    'rep_snd' [4s, 5s] Symbol d5f29a2c-5927-4237-a7ea-7fe7ec2c3be6
+    'rep_fst' [?, ?] Symbol 0bac3288-00e7-41bf-bd59-ac04de67ba47
+    'rep_snd' [?, ?] Symbol d5f29a2c-5927-4237-a7ea-7fe7ec2c3be6
     'a' Module
       'add' Add
         'arg0' [] Symbol 7d2c48d4-6110-4759-be98-de15bfb20af2
         'arg1' [] Symbol 86b517ef-b83c-4ea0-9568-f89bb2106db9
-      'fst' [0s, 1s] Symbol c41b44e7-4f04-4124-bf02-34e0133c7c19 ! 'arg0'
-      'snd' [0s, 1s] Symbol 823c2e81-99a2-46b3-bec1-329109e06e20 ! 'arg1'
+      'fst' [?, ?] Symbol c41b44e7-4f04-4124-bf02-34e0133c7c19 ! 'arg0'
+      'snd' [?, ?] Symbol 823c2e81-99a2-46b3-bec1-329109e06e20 ! 'arg1'
     """
     rep_fst, rep_snd, c = Parser().parse_circuits(s)
     c = c.cast_module()
@@ -404,8 +452,8 @@ def test_expand_node():
     assert out.shape[1] in [rep_fst.shape[1], rep_snd.shape[1]]
 
     s = """
-    'rep_fst' [3s] Symbol 0bac3288-00e7-41bf-bd59-ac04de67ba47
-    'rep_snd' [8, 5s] Symbol d5f29a2c-5927-4237-a7ea-7fe7ec2c3be6
+    'rep_fst' [?] Symbol 0bac3288-00e7-41bf-bd59-ac04de67ba47
+    'rep_snd' [8, ?] Symbol d5f29a2c-5927-4237-a7ea-7fe7ec2c3be6
     'a' Module
       'add' Add
         'arg0' [] Symbol 7d2c48d4-6110-4759-be98-de15bfb20af2
@@ -420,14 +468,14 @@ def test_expand_node():
     assert out.shape[1] in [rep_fst.shape[0], rep_snd.shape[1]]
 
     s = """
-    'rep_fst' [7, 9, 3s] Symbol 0bac3288-00e7-41bf-bd59-ac04de67ba47
-    'rep_snd' [8, 5s] Symbol d5f29a2c-5927-4237-a7ea-7fe7ec2c3be6
+    'rep_fst' [7, 9, ?] Symbol 0bac3288-00e7-41bf-bd59-ac04de67ba47
+    'rep_snd' [8, ?] Symbol d5f29a2c-5927-4237-a7ea-7fe7ec2c3be6
     'a' Module
       'add' Add
-        'arg0' [0s] Symbol 7d2c48d4-6110-4759-be98-de15bfb20af2
-        'arg1' [0s] Symbol 86b517ef-b83c-4ea0-9568-f89bb2106db9
-      'fst' [7s] Symbol c41b44e7-4f04-4124-bf02-34e0133c7c19 ! 'arg0'
-      'snd' [12s, 7s] Symbol 823c2e81-99a2-46b3-bec1-329109e06e20 ! 'arg1'
+        'arg0' [?] Symbol 7d2c48d4-6110-4759-be98-de15bfb20af2
+        'arg1' [?] Symbol 86b517ef-b83c-4ea0-9568-f89bb2106db9
+      'fst' [?] Symbol c41b44e7-4f04-4124-bf02-34e0133c7c19 ! 'arg0'
+      'snd' [?, ?] Symbol 823c2e81-99a2-46b3-bec1-329109e06e20 ! 'arg1'
     'b' Module
       'a'
       'rep_fst' ! 'fst'
@@ -442,14 +490,14 @@ def test_expand_node():
 
     for sym_on_top in [False, True]:
         s = f"""
-        'rep_fst' [12, {"7s" if sym_on_top else 8}, 9, 7] Symbol 0bac3288-00e7-41bf-bd59-ac04de67ba47
-        'rep_snd' [13, 12, 9, 10, {8 if sym_on_top else "7s"}, 7] Symbol d5f29a2c-5927-4237-a7ea-7fe7ec2c3be6
+        'rep_fst' [12, {"?" if sym_on_top else 8}, 9, 7] Symbol 0bac3288-00e7-41bf-bd59-ac04de67ba47
+        'rep_snd' [13, 12, 9, 10, {8 if sym_on_top else "?"}, 7] Symbol d5f29a2c-5927-4237-a7ea-7fe7ec2c3be6
         'a' Module
           'add' Add
-            'arg0' [1s, 0s] Symbol 7d2c48d4-6110-4759-be98-de15bfb20af2
-            'arg1' [0s] Symbol 86b517ef-b83c-4ea0-9568-f89bb2106db9
-          'fst' [21s, 24s, 20s] Symbol c41b44e7-4f04-4124-bf02-34e0133c7c19 ! 'arg0'
-          'snd' [23s, 22s, 21s, 20s] Symbol 823c2e81-99a2-46b3-bec1-329109e06e20 ! 'arg1'
+            'arg0' [?, ?] Symbol 7d2c48d4-6110-4759-be98-de15bfb20af2
+            'arg1' [?] Symbol 86b517ef-b83c-4ea0-9568-f89bb2106db9
+          'fst' [?, ?, ?] Symbol c41b44e7-4f04-4124-bf02-34e0133c7c19 ! 'arg0'
+          'snd' [?, ?, ?, ?] Symbol 823c2e81-99a2-46b3-bec1-329109e06e20 ! 'arg1'
         'b' Module
           'a'
           'rep_fst' ! 'fst'
@@ -466,10 +514,10 @@ def test_expand_node():
     'rep_snd' [13, 12,  10, 102, 7] Symbol d5f29a2c-5927-4237-a7ea-7fe7ec2c3be6
     'a' Module
       'add' Add
-        'arg0' [0s] Symbol 7d2c48d4-6110-4759-be98-de15bfb20af2
-        'arg1' [0s] Symbol 86b517ef-b83c-4ea0-9568-f89bb2106db9
-      'fst' [     21s, 20s] Symbol c41b44e7-4f04-4124-bf02-34e0133c7c19 ! 'arg0'
-      'snd' [22s, 21s, 20s] Symbol 823c2e81-99a2-46b3-bec1-329109e06e20 ! 'arg1'
+        'arg0' [?] Symbol 7d2c48d4-6110-4759-be98-de15bfb20af2
+        'arg1' [?] Symbol 86b517ef-b83c-4ea0-9568-f89bb2106db9
+      'fst' [   ?, ?] Symbol c41b44e7-4f04-4124-bf02-34e0133c7c19 ! 'arg0'
+      'snd' [?, ?, ?] Symbol 823c2e81-99a2-46b3-bec1-329109e06e20 ! 'arg1'
     'b' Module
       'a'
       'rep_fst' ! 'fst'
@@ -589,23 +637,6 @@ def test_bind():
     )
 
 
-def test_symbolic_size_constraints():
-    s0, s1, *_ = symbolic_sizes()
-
-    circ = Add(
-        Symbol.new_with_random_uuid((s0, s1), name="first"), Symbol.new_with_random_uuid((s0, s1), name="second")
-    )
-    assert circ.symbolic_size_constraints == set()
-
-    circ_replaced = Expander(
-        ("first", lambda _: Scalar(0.5, (s0, 3), "first_input")),
-    )(circ, fancy_validate=True)
-    assert len(circ_replaced.symbolic_size_constraints) == 1
-    constraint = circ_replaced.symbolic_size_constraints.pop()
-    assert constraint.l.other_factor == 1 and constraint.l.symbolic_sizes == [1]
-    assert constraint.r.other_factor == 3 and constraint.r.symbolic_sizes == []
-
-
 def test_replace_deep():
     replace_expand_bottom_up_dict(
         Einsum.from_einsum_string(
@@ -737,12 +768,6 @@ def test_full_transformer():
     raw_test_full_transformer(Params("bn", False, False, False))
 
 
-def test_symbolic_prime():
-    s = symbolic_sizes()[0]
-    for i in range(2, s):
-        assert s % i != 0
-
-
 def test_module_conform():
     spec = get_attention().spec
     module = Module(
@@ -813,11 +838,11 @@ def get_complex_module_test(
     b_inner_alpha = "".join(alpha[: len(batch_inner_0)])
     s = f"""
     'inner' Einsum az,bz,cl->bac
-      'inner.sym_arg0' [0s, 4s] Symbol a409cdc3-d962-4df1-8d1b-c5c8d8b50639
+      'inner.sym_arg0' [?, ?] Symbol a409cdc3-d962-4df1-8d1b-c5c8d8b50639
       'inner.concat' Concat 0
-        'inner.sym_arg1' [3, 4s] Symbol 02927290-fc4f-45b7-9822-267918e897d6
-        'outer.sym_arg0' [7, 4s] Symbol 6dcfaf1e-5e35-418a-ad88-3d79ae32096c
-      'outer.sym_arg1' [3s, 5s] Symbol ecf273f5-82c5-4fbf-8097-d9b1a250e035
+        'inner.sym_arg1' [3, ?] Symbol 02927290-fc4f-45b7-9822-267918e897d6
+        'outer.sym_arg0' [7, ?] Symbol 6dcfaf1e-5e35-418a-ad88-3d79ae32096c
+      'outer.sym_arg1' [?, ?] Symbol ecf273f5-82c5-4fbf-8097-d9b1a250e035
 
     'inner.call' Module
       'inner'
@@ -841,12 +866,10 @@ def get_complex_module_test(
     ]
     inner_call = inner_call_v.cast_module()
     outer = outer_v.cast_module()
-    s0 = symbolic_sizes()[0]
-    s3 = symbolic_sizes()[3]
     inner_batch = broadcast_shapes(batch_inner_0, batch_inner_1)
     outer_batch = broadcast_shapes(batch_outer_0, batch_outer_1)
-    assert inner.shape == (10, s0, s3)
-    assert inner_call.shape == (inner_batch + (10, 7, s3))
+    assert inner.shape == (10, rc.Size.NONE, rc.Size.NONE)
+    assert inner_call.shape == (inner_batch + (10, 7, rc.Size.NONE))
     assert outer.shape == (outer_batch + inner_batch + (10, 7, 9))
 
     base_value = outer.evaluate()
