@@ -123,11 +123,11 @@ def test_batching_top():
     'bilinear_mlp.pre0' Index [0,:]
       'bilinear_mlp.fold_pre' Rearrange (a:2 b) -> a:2 b
         'bilinear_mlp.pre' Einsum i,hi->h
-          'bilinear_mlp.input' [0s] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52
-          'bilinear_mlp.w.proj_in' [2*1s,0s] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
+          'bilinear_mlp.input' [?] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52
+          'bilinear_mlp.w.proj_in' [?,?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
     'bilinear_mlp.pre1' Index [1,:]
       'bilinear_mlp.fold_pre'
-    'bilinear_mlp.w.proj_out' [2s,1s] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
+    'bilinear_mlp.w.proj_out' [?,?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
   0 [64,17] Array rand ! 'bilinear_mlp.input'
   'b.proj_in' [22,17] Array rand ! 'bilinear_mlp.w.proj_in' ftt
   'b.proj_out' [13,11] Array rand ! 'bilinear_mlp.w.proj_out' ftt
