@@ -2043,11 +2043,11 @@ mlp_2_layer_str = """
     'bilinear_mlp.pre0' Index [0,:]
       'bilinear_mlp.fold_pre' Rearrange (a:2 b) -> a:2 b
         'bilinear_mlp.pre' Einsum i,hi->h
-          'bilinear_mlp.input' [0s] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52
-          'bilinear_mlp.w.proj_in' [2*1s,0s] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
+          'bilinear_mlp.input' [?] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52
+          'bilinear_mlp.w.proj_in' [?,?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
     'bilinear_mlp.pre1' Index [1,:]
       'bilinear_mlp.fold_pre'
-    'bilinear_mlp.w.proj_out' [2s,1s] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
+    'bilinear_mlp.w.proj_out' [?,?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
   'b.bilinear_mlp' Module ! 'bilinear_mlp.input'
     'bilinear_mlp'
     'b.input' [3,5,7,17] Scalar 7 ! 'bilinear_mlp.input'
@@ -2089,14 +2089,14 @@ def test_print_comments():
         == """
 'outer_b.bilinear_mlp' Module
 ├‣'bilinear_mlp' Einsum h,h,oh->o # Spec
-│ ├‣'bilinear_mlp.pre0' [1s] Index [0,:] # h
+│ ├‣'bilinear_mlp.pre0' [?] Index [0,:] # h
 │ │ └‣'bilinear_mlp.fold_pre' Rearrange (a:2 b) -> a:2 b
 │ │   └‣'bilinear_mlp.pre' Einsum i,hi->h
-│ │     ├‣'bilinear_mlp.input' [0s] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52 # i
-│ │     └‣'bilinear_mlp.w.proj_in' [2*1s,0s] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6 # hi
-│ ├‣'bilinear_mlp.pre1' [1s] Index [1,:] # h
+│ │     ├‣'bilinear_mlp.input' [?] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52 # i
+│ │     └‣'bilinear_mlp.w.proj_in' [?,?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6 # hi
+│ ├‣'bilinear_mlp.pre1' [?] Index [1,:] # h
 │ │ └‣'bilinear_mlp.fold_pre'
-│ └‣'bilinear_mlp.w.proj_out' [2s,1s] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf # oh
+│ └‣'bilinear_mlp.w.proj_out' [?,?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf # oh
 ├‣'b.bilinear_mlp' Module ! 'bilinear_mlp.input'
 │ ├‣'bilinear_mlp' # Spec
 │ ├‣'b.input' [3,5,7,17] Scalar 7 ! 'bilinear_mlp.input'
@@ -2130,14 +2130,14 @@ def test_commenters():
         == """
 'outer_b.bilinear_mlp' Module # line length: 7 # node count: 15
   'bilinear_mlp' Einsum h,h,oh->o # line length: 17 # node count: 8
-    'bilinear_mlp.pre0' [1s] Index [0,:] # line length: 17 # node count: 5
+    'bilinear_mlp.pre0' [?] Index [0,:] # line length: 16 # node count: 5
       'bilinear_mlp.fold_pre' Rearrange (a:2 b) -> a:2 b # line length: 27 # node count: 4
         'bilinear_mlp.pre' Einsum i,hi->h # line length: 15 # node count: 3
-          'bilinear_mlp.input' [0s] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52 # line length: 49 # node count: 1
-          'bilinear_mlp.w.proj_in' [2*1s,0s] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6 # line length: 54 # node count: 1
-    'bilinear_mlp.pre1' [1s] Index [1,:] # line length: 17 # node count: 5
+          'bilinear_mlp.input' [?] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52 # line length: 48 # node count: 1
+          'bilinear_mlp.w.proj_in' [?,?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6 # line length: 50 # node count: 1
+    'bilinear_mlp.pre1' [?] Index [1,:] # line length: 16 # node count: 5
       'bilinear_mlp.fold_pre' # line length: 27 # node count: 4
-    'bilinear_mlp.w.proj_out' [2s,1s] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf # line length: 52 # node count: 1
+    'bilinear_mlp.w.proj_out' [?,?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf # line length: 50 # node count: 1
   'b.bilinear_mlp' Module ! 'bilinear_mlp.input' # line length: 7 # node count: 12
     'bilinear_mlp' # line length: 17 # node count: 8
     'b.input' [3,5,7,17] Scalar 7 ! 'bilinear_mlp.input' # line length: 20 # node count: 1
@@ -2162,13 +2162,13 @@ def test_commenters():
         == """
 outer_b.bilinear_mlp Module # line length: 7 # node count: 15
   bilinear_mlp Einsum h,h,oh->o # line length: 17 # node count: 8
-    bilinear_mlp.pre0 [1s] Index [0,:] ... # line length: 17 # node count: 5
-    bilinear_mlp.pre1 [1s] Index [1,:] ... # line length: 17 # node count: 5
-    bilinear_mlp.w.proj_out [2s,1s] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf # line length: 52 # node count: 1
-  b.bilinear_mlp Module ! bilinear_mlp.input [0s] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52 # line length: 7 # node count: 12
+    bilinear_mlp.pre0 [?] Index [0,:] ... # line length: 16 # node count: 5
+    bilinear_mlp.pre1 [?] Index [1,:] ... # line length: 16 # node count: 5
+    bilinear_mlp.w.proj_out [?,?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf # line length: 50 # node count: 1
+  b.bilinear_mlp Module ! bilinear_mlp.input [?] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52 # line length: 7 # node count: 12
     bilinear_mlp # line length: 17 # node count: 8
     b.input [3,5,7,17] Scalar 7 ! bilinear_mlp.input # line length: 20 # node count: 1
-    b.proj_in [22,17] Scalar 8 ! bilinear_mlp.w.proj_in [2*1s,0s] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6 ftt # line length: 17 # node count: 1
+    b.proj_in [22,17] Scalar 8 ! bilinear_mlp.w.proj_in [?,?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6 ftt # line length: 17 # node count: 1
     b.proj_out [13,11] Scalar 9 ! bilinear_mlp.w.proj_out ftt # line length: 17 # node count: 1
   b_new.proj_in [22,13] Scalar 10 ! bilinear_mlp.w.proj_in ftt # line length: 18 # node count: 1
   b_new.proj_out [20,11] Scalar 11 ! bilinear_mlp.w.proj_out ftt # line length: 18 # node count: 1
