@@ -24,13 +24,14 @@ from interp.circuit.circuit_utils import evaluate_fn
 from interp.circuit.computational_node import Index as pIndex
 from interp.circuit.testing import strategies as st_c
 from interp.circuit.testing.topdown_circuit import CircuitProbs as CP
+from interp.circuit.testing.topdown_circuit import st_Circuit
 from interp.circuit.testing.utils import mark_not_interesting_if
 from interp.tools.indexer import I
 from rust_circuit import Add as rAdd
 from rust_circuit import Array as rArray
 from rust_circuit import Circuit as rCircuit
 from rust_circuit import Concat as rConcat
-from rust_circuit import ConstructArrayHasReservedSymbolicShapeError, DistributeNoopError
+from rust_circuit import DistributeNoopError
 from rust_circuit import Einsum as rEinsum
 from rust_circuit import GeneralFunction as rGeneralFunction
 from rust_circuit import Index as rIndex
@@ -248,8 +249,11 @@ def batch_to_concat_args_st(draw):
 )
 def test_evaluate_batch(args):
     circ, dim, bs = args
+    circ = rc.cast_circuit(circ, TorchDeviceDtypeOp(dtype="float64"))
     hypothesis.assume(not any(x.is_scatter() for x in rc.all_children(circ)))
+    hypothesis.note(circ.repr(rc.PrintOptions(bijection=False)))
     circ1 = batch_to_concat(circ, dim, bs)
+    hypothesis.note(circ1.repr(rc.PrintOptions(bijection=False)))
     assert_close(circ.evaluate(), circ1.evaluate())
 
 
@@ -260,8 +264,8 @@ def test_evaluate_batch(args):
 )
 def test_optimize_batch(args):
     circ, dim, bs = args
+    circ = rc.cast_circuit(circ, TorchDeviceDtypeOp(dtype="float64"))
     hypothesis.assume(not any(x.is_scatter() for x in rc.all_children(circ)))
-    circ = r_cast_circuit(circ, TorchDeviceDtypeOp(dtype="float64"))
     circ1 = batch_to_concat(circ, dim, bs)
     assert_close(circ.evaluate().to(dtype=torch.float64), optimize_and_evaluate(circ1).to(dtype=torch.float64))
 
@@ -296,7 +300,7 @@ def get_rewrite_st(setup):
     shape = st.sampled_from(options["shapes"]) if "shapes" in options else st_np.array_shapes(min_dims=1, max_dims=3)
     circuit_type_settings = {"Add": (1, add_settings)} | options.get("circuit_type_settings", {})
     probs_per_depth = [CP.kw(all=0, **{k: circuit_type_settings.get(k, 1) for k in x}) for x in probs_normalized]
-    return get_c_st(shape, probs_per_depth, max_growth_steps=options.get("max_growth_steps", 20), from_other=False)
+    return get_c_st(shape, probs_per_depth, max_growth_steps=options.get("max_growth_steps", 10), from_other=False)
 
 
 def deep_pull_concat_strip(c: rCircuit) -> rCircuit:
@@ -400,7 +404,7 @@ rewrite_test_setups = [
     (concat_pull_removable_axes, ["Concat", ["Scalar", "Rearrange"]]),
     (einsum_pull_removable_axes, ["Einsum", ["Scalar", "Rearrange"]]),
     (add_make_broadcasts_explicit, ["Add"]),
-    (distribute_all, ["Einsum", "Add"], {"max_growth_steps": 10}),
+    (distribute_all, ["Einsum", "Add"], {"max_growth_steps": 5}),
     (PartialName(distribute, operand_idx=0, do_broadcasts=True), ["Einsum", "Add"]),
     (einsum_of_permute_merge, ["Einsum", "Rearrange"]),  # i really don't know why this needs f64
     (permute_of_einsum_merge, ["Rearrange", "Einsum"]),
@@ -1564,11 +1568,10 @@ def test_simp_nodes_low(circ):
     assert r_count_nodes(simped) / 2 < r_count_nodes(circ)
 
 
-@hypothesis.settings(deadline=None)
-@hypothesis.given(circ=get_c_st())
-@pytest.mark.parametrize("naive_schedule", [False])
-@pytest.mark.xfail
+@hypothesis.given(circ=get_c_st(max_growth_steps=20))
+@pytest.mark.parametrize("naive_schedule", [True, False])
 def test_scheduled_execution(circ, naive_schedule):
+    hypothesis.note(circ.repr(rc.PrintOptions(bijection=False)))
     raw_test_scheduled_execution(circ, naive_schedule)
 
 
@@ -1608,8 +1611,6 @@ def test_scheduled_execution_manual():
         assert_close(normal_eval, s_eval)
 
 
-@pytest.mark.xfail
-@hypothesis.settings(deadline=None)
 @hypothesis.given(circ=get_c_st())
 @mark_not_interesting_if(SchedulingOOMError)
 def test_rust_opt_eval(circ):
@@ -1727,11 +1728,6 @@ def test_adjust_scale():
             assert_close(naive_result, torch.full((), 0.0, dtype=torch.float32))
 
 
-def test_array_symbolic_error():
-    with pytest.raises(ConstructArrayHasReservedSymbolicShapeError):
-        rArray(torch.randn(10_007, 1))
-
-
 def test_eval_many():
     arrconst = rArray(torch.randn(10, 10))
     circs = [rAdd(arrconst, arrconst), arrconst, rEinsum((arrconst, (0, 1)), out_axes=())]
@@ -1810,18 +1806,6 @@ def test_rearrange_from_string():
     assert z.output_ints == [[3, 2], [0], [1]]
     assert z.int_sizes == [17, None, 4, None]
 
-    sym_sizes = symbolic_sizes()
-    s0 = sym_sizes[0]
-    s1 = sym_sizes[1]
-    s6 = sym_sizes[6]
-    s17 = sym_sizes[17]
-    s240 = sym_sizes[240]
-    z = rRearrangeSpec.from_string("(adf:0s other c:0s*1s d:14*2*6s) b2 -> (b2 c) adf other 0s (d 17s) 240s")
-    assert z.to_string() == "(a:0s b c:0s*1s d:28*6s) e -> (e c:0s*1s) a:0s b 0s (d:28*6s 17s) 240s"
-    assert z.input_ints == [[0, 1, 2, 3], [4]]
-    assert z.output_ints == [[4, 2], [0], [1], [5], [3, 6], [7]]
-    assert z.int_sizes == [s0, None, s0 * s1, 28 * s6, None, s0, s17, s240]
-
 
 def test_flatten():
     x: Any = rScalar(0.3)
