@@ -1,6 +1,6 @@
 use std::fs;
 
-use circuit_base::parsing::Parser;
+use circuit_print::parsing::Parser;
 use circuit_rewrites::circuit_optimizer::{
     optimize_circuit, OptimizationContext, OptimizationSettings,
 };
